/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 
/*
** QGL_WIN.C
**
** This file implements the operating system binding of GL to QGL function
** pointers. When doing a port of Quake2 you must implement the following
** two functions:
**
** QGL_Init () - loads libraries, assigns function pointers, etc.
** QGL_Shutdown () - unloads libraries, NULLs function pointers
*/ 
#include <float.h>
#include "gl_local.h"
#include "glw_win.h"

int (WINAPI * qwglChoosePixelFormat) (HDC, CONST PIXELFORMATDESCRIPTOR *);
int (WINAPI * qwglDescribePixelFormat) (HDC, int, UINT, LPPIXELFORMATDESCRIPTOR);
int (WINAPI * qwglGetPixelFormat) (HDC);
BOOL (WINAPI * qwglSetPixelFormat) (HDC, int, CONST PIXELFORMATDESCRIPTOR *);
BOOL (WINAPI * qwglSwapBuffers) (HDC);

BOOL (WINAPI * qwglCopyContext) (HGLRC, HGLRC, UINT);
HGLRC (WINAPI * qwglCreateContext) (HDC);
HGLRC (WINAPI * qwglCreateLayerContext) (HDC, int);
BOOL (WINAPI * qwglDeleteContext) (HGLRC);
HGLRC (WINAPI * qwglGetCurrentContext) (VOID);
HDC (WINAPI * qwglGetCurrentDC) (VOID);
PROC (WINAPI * qwglGetProcAddress) (LPCSTR);
BOOL (WINAPI * qwglMakeCurrent) (HDC, HGLRC);
BOOL (WINAPI * qwglShareLists) (HGLRC, HGLRC);
BOOL (WINAPI * qwglUseFontBitmaps) (HDC, DWORD, DWORD, DWORD);

BOOL (WINAPI * qwglUseFontOutlines) (HDC, DWORD, DWORD, DWORD, FLOAT,
									 FLOAT, int, LPGLYPHMETRICSFLOAT);

BOOL (WINAPI * qwglDescribeLayerPlane) (HDC, int, int, UINT,
										LPLAYERPLANEDESCRIPTOR);
int (WINAPI * qwglSetLayerPaletteEntries) (HDC, int, int, int,
										 CONST COLORREF *);
int (WINAPI * qwglGetLayerPaletteEntries) (HDC, int, int, int,
										 COLORREF *);
BOOL (WINAPI * qwglRealizeLayerPalette) (HDC, int, BOOL);
BOOL (WINAPI * qwglSwapLayerBuffers) (HDC, UINT);

BOOL ( WINAPI * qwglSwapInterval)( int interval );
BOOL ( WINAPI * qwglGetDeviceGammaRampEXT)( unsigned char *, unsigned char *, unsigned char * );
BOOL ( WINAPI * qwglSetDeviceGammaRampEXT)( const unsigned char *, const unsigned char *, const unsigned char * );

PROC  MHGLGetProcAddress (LPCSTR FuncName)
{
	int i;
	char *VendorCode[] = {"", "ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX", NULL};
	PROC thePROC;

	// see is the function supported by the core implementation
	if (thePROC = qwglGetProcAddress (FuncName)) return thePROC;

	// start at 1 cos we've already tested the no vendor code variant
	for (i = 1; ; i++)
	{
		// oh shit...
		if (!VendorCode[i]) break;

		thePROC = qwglGetProcAddress (va ("%s%s", FuncName, VendorCode[i]));

		if (thePROC)
		{
			VID_Printf (PRINT_ALL, "...\"%s\" Unavailable, reverting to \"%s%s\"\n", FuncName, FuncName, VendorCode[i]);
			return thePROC;
		}
	}

	// try falling back on an fp version if the int version is not there
	// (a certain vendor claims to support OpenGL 2.0 but does not have some of the i entry points...)
	if (FuncName[strlen (FuncName) - 1] == 'i' || (FuncName[strlen (FuncName) - 2] == 'i' && FuncName[strlen (FuncName) - 1] == 'v'))
	{
		char NewFuncName[256];

		// take a copy
		strcpy (NewFuncName, FuncName);

		// switch the i to f
		if (FuncName[strlen (FuncName) - 1] == 'i')
			NewFuncName[strlen (NewFuncName) - 1] = 'f';
		else NewFuncName[strlen (NewFuncName) - 2] = 'f';

		// now rerun the test using the f versions...
		for (i = 0; ; i++)
		{
			// oh shit...
			if (!VendorCode[i]) break;

			thePROC = qwglGetProcAddress (va ("%s%s", NewFuncName, VendorCode[i]));

			if (thePROC)
			{
				VID_Printf (PRINT_ALL, "...\"%s\" Unavailable, reverting to \"%s%s\"\n", FuncName, NewFuncName, VendorCode[i]);
				return thePROC;
			}
		}
	}

	VID_Printf (PRINT_ALL, "...\"%s\" not supported by your OpenGL Implementation\n", FuncName);
	return NULL;
}

/*
** QGL_Shutdown
**
** Unloads the specified DLL then nulls out all the proc pointers.
*/ 
void QGL_Shutdown (void) 
{
	if (glw_state.hinstOpenGL) 
	{
		FreeLibrary (glw_state.hinstOpenGL);
		glw_state.hinstOpenGL = NULL;
	}

	glw_state.hinstOpenGL = NULL;

	glAccum = NULL;
	glAlphaFunc = NULL;
	glAreTexturesResident = NULL;
	glArrayElement = NULL;
	glBegin = NULL;
	glBindTexture = NULL;
	glBitmap = NULL;
	glBlendFunc = NULL;
	glCallList = NULL;
	glCallLists = NULL;
	glClear = NULL;
	glClearAccum = NULL;
	glClearColor = NULL;
	glClearDepth = NULL;
	glClearIndex = NULL;
	glClearStencil = NULL;
	glClipPlane = NULL;
	glColor3b = NULL;
	glColor3bv = NULL;
	glColor3d = NULL;
	glColor3dv = NULL;
	glColor3f = NULL;
	glColor3fv = NULL;
	glColor3i = NULL;
	glColor3iv = NULL;
	glColor3s = NULL;
	glColor3sv = NULL;
	glColor3ub = NULL;
	glColor3ubv = NULL;
	glColor3ui = NULL;
	glColor3uiv = NULL;
	glColor3us = NULL;
	glColor3usv = NULL;
	glColor4b = NULL;
	glColor4bv = NULL;
	glColor4d = NULL;
	glColor4dv = NULL;
	glColor4f = NULL;
	glColor4fv = NULL;
	glColor4i = NULL;
	glColor4iv = NULL;
	glColor4s = NULL;
	glColor4sv = NULL;
	glColor4ub = NULL;
	glColor4ubv = NULL;
	glColor4ui = NULL;
	glColor4uiv = NULL;
	glColor4us = NULL;
	glColor4usv = NULL;
	glColorMask = NULL;
	glColorMaterial = NULL;
	glColorPointer = NULL;
	glCopyPixels = NULL;
	glCopyTexImage1D = NULL;
	glCopyTexImage2D = NULL;
	glCopyTexSubImage1D = NULL;
	glCopyTexSubImage2D = NULL;
	glCullFace = NULL;
	glDeleteLists = NULL;
	glDeleteTextures = NULL;
	glDepthFunc = NULL;
	glDepthMask = NULL;
	glDepthRange = NULL;
	glDisable = NULL;
	glDisableClientState = NULL;
	glDrawArrays = NULL;
	glDrawBuffer = NULL;
	glDrawElements = NULL;
	glDrawPixels = NULL;
	glEdgeFlag = NULL;
	glEdgeFlagPointer = NULL;
	glEdgeFlagv = NULL;
	glEnable = NULL;
	glEnableClientState = NULL;
	glEnd = NULL;
	glEndList = NULL;
	glEvalCoord1d = NULL;
	glEvalCoord1dv = NULL;
	glEvalCoord1f = NULL;
	glEvalCoord1fv = NULL;
	glEvalCoord2d = NULL;
	glEvalCoord2dv = NULL;
	glEvalCoord2f = NULL;
	glEvalCoord2fv = NULL;
	glEvalMesh1 = NULL;
	glEvalMesh2 = NULL;
	glEvalPoint1 = NULL;
	glEvalPoint2 = NULL;
	glFeedbackBuffer = NULL;
	glFinish = NULL;
	glFlush = NULL;
	glFogf = NULL;
	glFogfv = NULL;
	glFogi = NULL;
	glFogiv = NULL;
	glFrontFace = NULL;
	glFrustum = NULL;
	glGenLists = NULL;
	glGenTextures = NULL;
	glGetBooleanv = NULL;
	glGetClipPlane = NULL;
	glGetDoublev = NULL;
	glGetError = NULL;
	glGetFloatv = NULL;
	glGetIntegerv = NULL;
	glGetLightfv = NULL;
	glGetLightiv = NULL;
	glGetMapdv = NULL;
	glGetMapfv = NULL;
	glGetMapiv = NULL;
	glGetMaterialfv = NULL;
	glGetMaterialiv = NULL;
	glGetPixelMapfv = NULL;
	glGetPixelMapuiv = NULL;
	glGetPixelMapusv = NULL;
	glGetPointerv = NULL;
	glGetPolygonStipple = NULL;
	glGetString = NULL;
	glGetTexEnvfv = NULL;
	glGetTexEnviv = NULL;
	glGetTexGendv = NULL;
	glGetTexGenfv = NULL;
	glGetTexGeniv = NULL;
	glGetTexImage = NULL;
	glGetTexLevelParameterfv = NULL;
	glGetTexLevelParameteriv = NULL;
	glGetTexParameterfv = NULL;
	glGetTexParameteriv = NULL;
	glHint = NULL;
	glIndexMask = NULL;
	glIndexPointer = NULL;
	glIndexd = NULL;
	glIndexdv = NULL;
	glIndexf = NULL;
	glIndexfv = NULL;
	glIndexi = NULL;
	glIndexiv = NULL;
	glIndexs = NULL;
	glIndexsv = NULL;
	glIndexub = NULL;
	glIndexubv = NULL;
	glInitNames = NULL;
	glInterleavedArrays = NULL;
	glIsEnabled = NULL;
	glIsList = NULL;
	glIsTexture = NULL;
	glLightModelf = NULL;
	glLightModelfv = NULL;
	glLightModeli = NULL;
	glLightModeliv = NULL;
	glLightf = NULL;
	glLightfv = NULL;
	glLighti = NULL;
	glLightiv = NULL;
	glLineStipple = NULL;
	glLineWidth = NULL;
	glListBase = NULL;
	glLoadIdentity = NULL;
	glLoadMatrixd = NULL;
	glLoadMatrixf = NULL;
	glLoadName = NULL;
	glLogicOp = NULL;
	glMap1d = NULL;
	glMap1f = NULL;
	glMap2d = NULL;
	glMap2f = NULL;
	glMapGrid1d = NULL;
	glMapGrid1f = NULL;
	glMapGrid2d = NULL;
	glMapGrid2f = NULL;
	glMaterialf = NULL;
	glMaterialfv = NULL;
	glMateriali = NULL;
	glMaterialiv = NULL;
	glMatrixMode = NULL;
	glMultMatrixd = NULL;
	glMultMatrixf = NULL;
	glNewList = NULL;
	glNormal3b = NULL;
	glNormal3bv = NULL;
	glNormal3d = NULL;
	glNormal3dv = NULL;
	glNormal3f = NULL;
	glNormal3fv = NULL;
	glNormal3i = NULL;
	glNormal3iv = NULL;
	glNormal3s = NULL;
	glNormal3sv = NULL;
	glNormalPointer = NULL;
	glOrtho = NULL;
	glPassThrough = NULL;
	glPixelMapfv = NULL;
	glPixelMapuiv = NULL;
	glPixelMapusv = NULL;
	glPixelStoref = NULL;
	glPixelStorei = NULL;
	glPixelTransferf = NULL;
	glPixelTransferi = NULL;
	glPixelZoom = NULL;
	glPointSize = NULL;
	glPolygonMode = NULL;
	glPolygonOffset = NULL;
	glPolygonStipple = NULL;
	glPopAttrib = NULL;
	glPopClientAttrib = NULL;
	glPopMatrix = NULL;
	glPopName = NULL;
	glPrioritizeTextures = NULL;
	glPushAttrib = NULL;
	glPushClientAttrib = NULL;
	glPushMatrix = NULL;
	glPushName = NULL;
	glRasterPos2d = NULL;
	glRasterPos2dv = NULL;
	glRasterPos2f = NULL;
	glRasterPos2fv = NULL;
	glRasterPos2i = NULL;
	glRasterPos2iv = NULL;
	glRasterPos2s = NULL;
	glRasterPos2sv = NULL;
	glRasterPos3d = NULL;
	glRasterPos3dv = NULL;
	glRasterPos3f = NULL;
	glRasterPos3fv = NULL;
	glRasterPos3i = NULL;
	glRasterPos3iv = NULL;
	glRasterPos3s = NULL;
	glRasterPos3sv = NULL;
	glRasterPos4d = NULL;
	glRasterPos4dv = NULL;
	glRasterPos4f = NULL;
	glRasterPos4fv = NULL;
	glRasterPos4i = NULL;
	glRasterPos4iv = NULL;
	glRasterPos4s = NULL;
	glRasterPos4sv = NULL;
	glReadBuffer = NULL;
	glReadPixels = NULL;
	glRectd = NULL;
	glRectdv = NULL;
	glRectf = NULL;
	glRectfv = NULL;
	glRecti = NULL;
	glRectiv = NULL;
	glRects = NULL;
	glRectsv = NULL;
	glRenderMode = NULL;
	glRotated = NULL;
	glRotatef = NULL;
	glScaled = NULL;
	glScalef = NULL;
	glScissor = NULL;
	glSelectBuffer = NULL;
	glShadeModel = NULL;
	glStencilFunc = NULL;
	glStencilMask = NULL;
	glStencilOp = NULL;
	glTexCoord1d = NULL;
	glTexCoord1dv = NULL;
	glTexCoord1f = NULL;
	glTexCoord1fv = NULL;
	glTexCoord1i = NULL;
	glTexCoord1iv = NULL;
	glTexCoord1s = NULL;
	glTexCoord1sv = NULL;
	glTexCoord2d = NULL;
	glTexCoord2dv = NULL;
	glTexCoord2f = NULL;
	glTexCoord2fv = NULL;
	glTexCoord2i = NULL;
	glTexCoord2iv = NULL;
	glTexCoord2s = NULL;
	glTexCoord2sv = NULL;
	glTexCoord3d = NULL;
	glTexCoord3dv = NULL;
	glTexCoord3f = NULL;
	glTexCoord3fv = NULL;
	glTexCoord3i = NULL;
	glTexCoord3iv = NULL;
	glTexCoord3s = NULL;
	glTexCoord3sv = NULL;
	glTexCoord4d = NULL;
	glTexCoord4dv = NULL;
	glTexCoord4f = NULL;
	glTexCoord4fv = NULL;
	glTexCoord4i = NULL;
	glTexCoord4iv = NULL;
	glTexCoord4s = NULL;
	glTexCoord4sv = NULL;
	glTexCoordPointer = NULL;
	glTexEnvf = NULL;
	glTexEnvfv = NULL;
	glTexEnvi = NULL;
	glTexEnviv = NULL;
	glTexGend = NULL;
	glTexGendv = NULL;
	glTexGenf = NULL;
	glTexGenfv = NULL;
	glTexGeni = NULL;
	glTexGeniv = NULL;
	glTexImage1D = NULL;
	glTexImage2D = NULL;
	glTexParameterf = NULL;
	glTexParameterfv = NULL;
	glTexParameteri = NULL;
	glTexParameteriv = NULL;
	glTexSubImage1D = NULL;
	glTexSubImage2D = NULL;
	glTranslated = NULL;
	glTranslatef = NULL;
	glVertex2d = NULL;
	glVertex2dv = NULL;
	glVertex2f = NULL;
	glVertex2fv = NULL;
	glVertex2i = NULL;
	glVertex2iv = NULL;
	glVertex2s = NULL;
	glVertex2sv = NULL;
	glVertex3d = NULL;
	glVertex3dv = NULL;
	glVertex3f = NULL;
	glVertex3fv = NULL;
	glVertex3i = NULL;
	glVertex3iv = NULL;
	glVertex3s = NULL;
	glVertex3sv = NULL;
	glVertex4d = NULL;
	glVertex4dv = NULL;
	glVertex4f = NULL;
	glVertex4fv = NULL;
	glVertex4i = NULL;
	glVertex4iv = NULL;
	glVertex4s = NULL;
	glVertex4sv = NULL;
	glVertexPointer = NULL;
	glViewport = NULL;

	qwglCopyContext = NULL;
	qwglCreateContext = NULL;
	qwglCreateLayerContext = NULL;
	qwglDeleteContext = NULL;
	qwglDescribeLayerPlane = NULL;
	qwglGetCurrentContext = NULL;
	qwglGetCurrentDC = NULL;
	qwglGetLayerPaletteEntries = NULL;
	qwglGetProcAddress = NULL;
	qwglMakeCurrent = NULL;
	qwglRealizeLayerPalette = NULL;
	qwglSetLayerPaletteEntries = NULL;
	qwglShareLists = NULL;
	qwglSwapLayerBuffers = NULL;
	qwglUseFontBitmaps = NULL;
	qwglUseFontOutlines = NULL;

	qwglChoosePixelFormat = NULL;
	qwglDescribePixelFormat = NULL;
	qwglGetPixelFormat = NULL;
	qwglSetPixelFormat = NULL;
	qwglSwapBuffers = NULL;

	qwglSwapInterval	 = NULL;

	qwglGetDeviceGammaRampEXT = NULL;
	qwglSetDeviceGammaRampEXT = NULL;
}

#	pragma warning (disable : 4113 4133 4047) 
#	define GPA(a) GetProcAddress (glw_state.hinstOpenGL, a) 

/*
** QGL_Init
**
** This is responsible for binding our gl function pointers to 
** the appropriate GL stuff. In Windows this means doing a 
** LoadLibrary and a bunch of calls to GetProcAddress. On other
** operating systems we need to do the right thing, whatever that
** might be.
** 
*/ 
qboolean QGL_Init (const char *dllname) 
{
	// update 3Dfx gamma irrespective of underlying DLL
	{
		char envbuffer[1024];
		float g;

		g = 2.00 * (0.8 - (vid_gamma->value - 0.5)) + 1.0F;
		Com_sprintf (envbuffer, sizeof (envbuffer) , "SSTV2_GAMMA = %f", g);
		_putenv (envbuffer);
		Com_sprintf (envbuffer, sizeof (envbuffer) , "SST_GAMMA = %f", g);
		_putenv (envbuffer);
	}

	if ((glw_state.hinstOpenGL = LoadLibrary (dllname)) == 0) 
	{
		char *buf = NULL;

		FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError () , MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT) , (LPTSTR) &buf, 0, NULL);
		VID_Printf (PRINT_ALL, "%s\n", buf);
		return false;
	}

	gl_config.allow_cds = true;

	glAccum = GPA("glAccum");
	glAlphaFunc = GPA("glAlphaFunc");
	glAreTexturesResident = GPA("glAreTexturesResident");
	glArrayElement = GPA("glArrayElement");
	glBegin = GPA("glBegin");
	glBindTexture = GPA("glBindTexture");
	glBitmap = GPA("glBitmap");
	glBlendFunc = GPA("glBlendFunc");
	glCallList = GPA("glCallList");
	glCallLists = GPA("glCallLists");
	glClear = GPA("glClear");
	glClearAccum = GPA("glClearAccum");
	glClearColor = GPA("glClearColor");
	glClearDepth = GPA("glClearDepth");
	glClearIndex = GPA("glClearIndex");
	glClearStencil = GPA("glClearStencil");
	glClipPlane = GPA("glClipPlane");
	glColor3b = GPA("glColor3b");
	glColor3bv = GPA("glColor3bv");
	glColor3d = GPA("glColor3d");
	glColor3dv = GPA("glColor3dv");
	glColor3f = GPA("glColor3f");
	glColor3fv = GPA("glColor3fv");
	glColor3i = GPA("glColor3i");
	glColor3iv = GPA("glColor3iv");
	glColor3s = GPA("glColor3s");
	glColor3sv = GPA("glColor3sv");
	glColor3ub = GPA("glColor3ub");
	glColor3ubv = GPA("glColor3ubv");
	glColor3ui = GPA("glColor3ui");
	glColor3uiv = GPA("glColor3uiv");
	glColor3us = GPA("glColor3us");
	glColor3usv = GPA("glColor3usv");
	glColor4b = GPA("glColor4b");
	glColor4bv = GPA("glColor4bv");
	glColor4d = GPA("glColor4d");
	glColor4dv = GPA("glColor4dv");
	glColor4f = GPA("glColor4f");
	glColor4fv = GPA("glColor4fv");
	glColor4i = GPA("glColor4i");
	glColor4iv = GPA("glColor4iv");
	glColor4s = GPA("glColor4s");
	glColor4sv = GPA("glColor4sv");
	glColor4ub = GPA("glColor4ub");
	glColor4ubv = GPA("glColor4ubv");
	glColor4ui = GPA("glColor4ui");
	glColor4uiv = GPA("glColor4uiv");
	glColor4us = GPA("glColor4us");
	glColor4usv = GPA("glColor4usv");
	glColorMask = GPA("glColorMask");
	glColorMaterial = GPA("glColorMaterial");
	glColorPointer = GPA("glColorPointer");
	glCopyPixels = GPA("glCopyPixels");
	glCopyTexImage1D = GPA("glCopyTexImage1D");
	glCopyTexImage2D = GPA("glCopyTexImage2D");
	glCopyTexSubImage1D = GPA("glCopyTexSubImage1D");
	glCopyTexSubImage2D = GPA("glCopyTexSubImage2D");
	glCullFace = GPA("glCullFace");
	glDeleteLists = GPA("glDeleteLists");
	glDeleteTextures = GPA("glDeleteTextures");
	glDepthFunc = GPA("glDepthFunc");
	glDepthMask = GPA("glDepthMask");
	glDepthRange = GPA("glDepthRange");
	glDisable = GPA("glDisable");
	glDisableClientState = GPA("glDisableClientState");
	glDrawArrays = GPA("glDrawArrays");
	glDrawBuffer = GPA("glDrawBuffer");
	glDrawElements = GPA("glDrawElements");
	glDrawPixels = GPA("glDrawPixels");
	glEdgeFlag = GPA("glEdgeFlag");
	glEdgeFlagPointer = GPA("glEdgeFlagPointer");
	glEdgeFlagv = GPA("glEdgeFlagv");
	glEnable = GPA("glEnable");
	glEnableClientState = GPA("glEnableClientState");
	glEnd = GPA("glEnd");
	glEndList = GPA("glEndList");
	glEvalCoord1d				 = GPA("glEvalCoord1d");
	glEvalCoord1dv = GPA("glEvalCoord1dv");
	glEvalCoord1f = GPA("glEvalCoord1f");
	glEvalCoord1fv = GPA("glEvalCoord1fv");
	glEvalCoord2d = GPA("glEvalCoord2d");
	glEvalCoord2dv = GPA("glEvalCoord2dv");
	glEvalCoord2f = GPA("glEvalCoord2f");
	glEvalCoord2fv = GPA("glEvalCoord2fv");
	glEvalMesh1 = GPA("glEvalMesh1");
	glEvalMesh2 = GPA("glEvalMesh2");
	glEvalPoint1 = GPA("glEvalPoint1");
	glEvalPoint2 = GPA("glEvalPoint2");
	glFeedbackBuffer = GPA("glFeedbackBuffer");
	glFinish = GPA("glFinish");
	glFlush = GPA("glFlush");
	glFogf = GPA("glFogf");
	glFogfv = GPA("glFogfv");
	glFogi = GPA("glFogi");
	glFogiv = GPA("glFogiv");
	glFrontFace = GPA("glFrontFace");
	glFrustum = GPA("glFrustum");
	glGenLists = GPA("glGenLists");
	glGenTextures = GPA("glGenTextures");
	glGetBooleanv = GPA("glGetBooleanv");
	glGetClipPlane = GPA("glGetClipPlane");
	glGetDoublev = GPA("glGetDoublev");
	glGetError = GPA("glGetError");
	glGetFloatv = GPA("glGetFloatv");
	glGetIntegerv = GPA("glGetIntegerv");
	glGetLightfv = GPA("glGetLightfv");
	glGetLightiv = GPA("glGetLightiv");
	glGetMapdv = GPA("glGetMapdv");
	glGetMapfv = GPA("glGetMapfv");
	glGetMapiv = GPA("glGetMapiv");
	glGetMaterialfv = GPA("glGetMaterialfv");
	glGetMaterialiv = GPA("glGetMaterialiv");
	glGetPixelMapfv = GPA("glGetPixelMapfv");
	glGetPixelMapuiv = GPA("glGetPixelMapuiv");
	glGetPixelMapusv = GPA("glGetPixelMapusv");
	glGetPointerv = GPA("glGetPointerv");
	glGetPolygonStipple = GPA("glGetPolygonStipple");
	glGetString = GPA("glGetString");
	glGetTexEnvfv = GPA("glGetTexEnvfv");
	glGetTexEnviv = GPA("glGetTexEnviv");
	glGetTexGendv = GPA("glGetTexGendv");
	glGetTexGenfv = GPA("glGetTexGenfv");
	glGetTexGeniv = GPA("glGetTexGeniv");
	glGetTexImage = GPA("glGetTexImage");
	glGetTexLevelParameterfv = GPA("glGetLevelParameterfv");
	glGetTexLevelParameteriv = GPA("glGetLevelParameteriv");
	glGetTexParameterfv = GPA("glGetTexParameterfv");
	glGetTexParameteriv = GPA("glGetTexParameteriv");
	glHint = GPA("glHint");
	glIndexMask = GPA("glIndexMask");
	glIndexPointer = GPA("glIndexPointer");
	glIndexd = GPA("glIndexd");
	glIndexdv = GPA("glIndexdv");
	glIndexf = GPA("glIndexf");
	glIndexfv = GPA("glIndexfv");
	glIndexi = GPA("glIndexi");
	glIndexiv = GPA("glIndexiv");
	glIndexs = GPA("glIndexs");
	glIndexsv = GPA("glIndexsv");
	glIndexub = GPA("glIndexub");
	glIndexubv = GPA("glIndexubv");
	glInitNames = GPA("glInitNames");
	glInterleavedArrays = GPA("glInterleavedArrays");
	glIsEnabled = GPA("glIsEnabled");
	glIsList = GPA("glIsList");
	glIsTexture = GPA("glIsTexture");
	glLightModelf = GPA("glLightModelf");
	glLightModelfv = GPA("glLightModelfv");
	glLightModeli = GPA("glLightModeli");
	glLightModeliv = GPA("glLightModeliv");
	glLightf = GPA("glLightf");
	glLightfv = GPA("glLightfv");
	glLighti = GPA("glLighti");
	glLightiv = GPA("glLightiv");
	glLineStipple = GPA("glLineStipple");
	glLineWidth = GPA("glLineWidth");
	glListBase = GPA("glListBase");
	glLoadIdentity = GPA("glLoadIdentity");
	glLoadMatrixd = GPA("glLoadMatrixd");
	glLoadMatrixf = GPA("glLoadMatrixf");
	glLoadName = GPA("glLoadName");
	glLogicOp = GPA("glLogicOp");
	glMap1d = GPA("glMap1d");
	glMap1f = GPA("glMap1f");
	glMap2d = GPA("glMap2d");
	glMap2f = GPA("glMap2f");
	glMapGrid1d = GPA("glMapGrid1d");
	glMapGrid1f = GPA("glMapGrid1f");
	glMapGrid2d = GPA("glMapGrid2d");
	glMapGrid2f = GPA("glMapGrid2f");
	glMaterialf = GPA("glMaterialf");
	glMaterialfv = GPA("glMaterialfv");
	glMateriali = GPA("glMateriali");
	glMaterialiv = GPA("glMaterialiv");
	glMatrixMode = GPA("glMatrixMode");
	glMultMatrixd = GPA("glMultMatrixd");
	glMultMatrixf = GPA("glMultMatrixf");
	glNewList = GPA("glNewList");
	glNormal3b = GPA("glNormal3b");
	glNormal3bv = GPA("glNormal3bv");
	glNormal3d = GPA("glNormal3d");
	glNormal3dv = GPA("glNormal3dv");
	glNormal3f = GPA("glNormal3f");
	glNormal3fv = GPA("glNormal3fv");
	glNormal3i = GPA("glNormal3i");
	glNormal3iv = GPA("glNormal3iv");
	glNormal3s = GPA("glNormal3s");
	glNormal3sv = GPA("glNormal3sv");
	glNormalPointer = GPA("glNormalPointer");
	glOrtho = GPA("glOrtho");
	glPassThrough = GPA("glPassThrough");
	glPixelMapfv = GPA("glPixelMapfv");
	glPixelMapuiv = GPA("glPixelMapuiv");
	glPixelMapusv = GPA("glPixelMapusv");
	glPixelStoref = GPA("glPixelStoref");
	glPixelStorei = GPA("glPixelStorei");
	glPixelTransferf = GPA("glPixelTransferf");
	glPixelTransferi = GPA("glPixelTransferi");
	glPixelZoom = GPA("glPixelZoom");
	glPointSize = GPA("glPointSize");
	glPolygonMode = GPA("glPolygonMode");
	glPolygonOffset = GPA("glPolygonOffset");
	glPolygonStipple = GPA("glPolygonStipple");
	glPopAttrib = GPA("glPopAttrib");
	glPopClientAttrib = GPA("glPopClientAttrib");
	glPopMatrix = GPA("glPopMatrix");
	glPopName = GPA("glPopName");
	glPrioritizeTextures = GPA("glPrioritizeTextures");
	glPushAttrib = GPA("glPushAttrib");
	glPushClientAttrib = GPA("glPushClientAttrib");
	glPushMatrix = GPA("glPushMatrix");
	glPushName = GPA("glPushName");
	glRasterPos2d = GPA("glRasterPos2d");
	glRasterPos2dv = GPA("glRasterPos2dv");
	glRasterPos2f = GPA("glRasterPos2f");
	glRasterPos2fv = GPA("glRasterPos2fv");
	glRasterPos2i = GPA("glRasterPos2i");
	glRasterPos2iv = GPA("glRasterPos2iv");
	glRasterPos2s = GPA("glRasterPos2s");
	glRasterPos2sv = GPA("glRasterPos2sv");
	glRasterPos3d = GPA("glRasterPos3d");
	glRasterPos3dv = GPA("glRasterPos3dv");
	glRasterPos3f = GPA("glRasterPos3f");
	glRasterPos3fv = GPA("glRasterPos3fv");
	glRasterPos3i = GPA("glRasterPos3i");
	glRasterPos3iv = GPA("glRasterPos3iv");
	glRasterPos3s = GPA("glRasterPos3s");
	glRasterPos3sv = GPA("glRasterPos3sv");
	glRasterPos4d = GPA("glRasterPos4d");
	glRasterPos4dv = GPA("glRasterPos4dv");
	glRasterPos4f = GPA("glRasterPos4f");
	glRasterPos4fv = GPA("glRasterPos4fv");
	glRasterPos4i = GPA("glRasterPos4i");
	glRasterPos4iv = GPA("glRasterPos4iv");
	glRasterPos4s = GPA("glRasterPos4s");
	glRasterPos4sv = GPA("glRasterPos4sv");
	glReadBuffer = GPA("glReadBuffer");
	glReadPixels = GPA("glReadPixels");
	glRectd = GPA("glRectd");
	glRectdv = GPA("glRectdv");
	glRectf = GPA("glRectf");
	glRectfv = GPA("glRectfv");
	glRecti = GPA("glRecti");
	glRectiv = GPA("glRectiv");
	glRects = GPA("glRects");
	glRectsv = GPA("glRectsv");
	glRenderMode = GPA("glRenderMode");
	glRotated = GPA("glRotated");
	glRotatef = GPA("glRotatef");
	glScaled = GPA("glScaled");
	glScalef = GPA("glScalef");
	glScissor = GPA("glScissor");
	glSelectBuffer = GPA("glSelectBuffer");
	glShadeModel = GPA("glShadeModel");
	glStencilFunc = GPA("glStencilFunc");
	glStencilMask = GPA("glStencilMask");
	glStencilOp = GPA("glStencilOp");
	glTexCoord1d = GPA("glTexCoord1d");
	glTexCoord1dv = GPA("glTexCoord1dv");
	glTexCoord1f = GPA("glTexCoord1f");
	glTexCoord1fv = GPA("glTexCoord1fv");
	glTexCoord1i = GPA("glTexCoord1i");
	glTexCoord1iv = GPA("glTexCoord1iv");
	glTexCoord1s = GPA("glTexCoord1s");
	glTexCoord1sv = GPA("glTexCoord1sv");
	glTexCoord2d = GPA("glTexCoord2d");
	glTexCoord2dv = GPA("glTexCoord2dv");
	glTexCoord2f = GPA("glTexCoord2f");
	glTexCoord2fv = GPA("glTexCoord2fv");
	glTexCoord2i = GPA("glTexCoord2i");
	glTexCoord2iv = GPA("glTexCoord2iv");
	glTexCoord2s = GPA("glTexCoord2s");
	glTexCoord2sv = GPA("glTexCoord2sv");
	glTexCoord3d = GPA("glTexCoord3d");
	glTexCoord3dv = GPA("glTexCoord3dv");
	glTexCoord3f = GPA("glTexCoord3f");
	glTexCoord3fv = GPA("glTexCoord3fv");
	glTexCoord3i = GPA("glTexCoord3i");
	glTexCoord3iv = GPA("glTexCoord3iv");
	glTexCoord3s = GPA("glTexCoord3s");
	glTexCoord3sv = GPA("glTexCoord3sv");
	glTexCoord4d = GPA("glTexCoord4d");
	glTexCoord4dv = GPA("glTexCoord4dv");
	glTexCoord4f = GPA("glTexCoord4f");
	glTexCoord4fv = GPA("glTexCoord4fv");
	glTexCoord4i = GPA("glTexCoord4i");
	glTexCoord4iv = GPA("glTexCoord4iv");
	glTexCoord4s = GPA("glTexCoord4s");
	glTexCoord4sv = GPA("glTexCoord4sv");
	glTexCoordPointer = GPA("glTexCoordPointer");
	glTexEnvf = GPA("glTexEnvf");
	glTexEnvfv = GPA("glTexEnvfv");
	glTexEnvi = GPA("glTexEnvi");
	glTexEnviv = GPA("glTexEnviv");
	glTexGend = GPA("glTexGend");
	glTexGendv = GPA("glTexGendv");
	glTexGenf = GPA("glTexGenf");
	glTexGenfv = GPA("glTexGenfv");
	glTexGeni = GPA("glTexGeni");
	glTexGeniv = GPA("glTexGeniv");
	glTexImage1D = GPA("glTexImage1D");
	glTexImage2D = GPA("glTexImage2D");
	glTexParameterf = GPA("glTexParameterf");
	glTexParameterfv = GPA("glTexParameterfv");
	glTexParameteri = GPA("glTexParameteri");
	glTexParameteriv = GPA("glTexParameteriv");
	glTexSubImage1D = GPA("glTexSubImage1D");
	glTexSubImage2D = GPA("glTexSubImage2D");
	glTranslated = GPA("glTranslated");
	glTranslatef = GPA("glTranslatef");
	glVertex2d = GPA("glVertex2d");
	glVertex2dv = GPA("glVertex2dv");
	glVertex2f = GPA("glVertex2f");
	glVertex2fv = GPA("glVertex2fv");
	glVertex2i = GPA("glVertex2i");
	glVertex2iv = GPA("glVertex2iv");
	glVertex2s = GPA("glVertex2s");
	glVertex2sv = GPA("glVertex2sv");
	glVertex3d = GPA("glVertex3d");
	glVertex3dv = GPA("glVertex3dv");
	glVertex3f = GPA("glVertex3f");
	glVertex3fv = GPA("glVertex3fv");
	glVertex3i = GPA("glVertex3i");
	glVertex3iv = GPA("glVertex3iv");
	glVertex3s = GPA("glVertex3s");
	glVertex3sv = GPA("glVertex3sv");
	glVertex4d = GPA("glVertex4d");
	glVertex4dv = GPA("glVertex4dv");
	glVertex4f = GPA("glVertex4f");
	glVertex4fv = GPA("glVertex4fv");
	glVertex4i = GPA("glVertex4i");
	glVertex4iv = GPA("glVertex4iv");
	glVertex4s = GPA("glVertex4s");
	glVertex4sv = GPA("glVertex4sv");
	glVertexPointer = GPA("glVertexPointer");
	glViewport = GPA("glViewport");

	qwglCopyContext = GPA("wglCopyContext");
	qwglCreateContext = GPA("wglCreateContext");
	qwglCreateLayerContext = GPA("wglCreateLayerContext");
	qwglDeleteContext = GPA("wglDeleteContext");
	qwglDescribeLayerPlane = GPA("wglDescribeLayerPlane");
	qwglGetCurrentContext = GPA("wglGetCurrentContext");
	qwglGetCurrentDC = GPA("wglGetCurrentDC");
	qwglGetLayerPaletteEntries = GPA("wglGetLayerPaletteEntries");
	qwglGetProcAddress = GPA("wglGetProcAddress");
	qwglMakeCurrent = GPA("wglMakeCurrent");
	qwglRealizeLayerPalette = GPA("wglRealizeLayerPalette");
	qwglSetLayerPaletteEntries = GPA("wglSetLayerPaletteEntries");
	qwglShareLists = GPA("wglShareLists");
	qwglSwapLayerBuffers = GPA("wglSwapLayerBuffers");
	qwglUseFontBitmaps = GPA("wglUseFontBitmapsA");
	qwglUseFontOutlines = GPA("wglUseFontOutlinesA");

	qwglChoosePixelFormat = GPA("wglChoosePixelFormat");
	qwglDescribePixelFormat = GPA("wglDescribePixelFormat");
	qwglGetPixelFormat = GPA("wglGetPixelFormat");
	qwglSetPixelFormat = GPA("wglSetPixelFormat");
	qwglSwapBuffers = GPA("wglSwapBuffers");

	qwglSwapInterval = 0;
	glPointParameterf = 0;
	glPointParameterfv = 0;
	glColorTable = 0;
	glActiveTexture = 0;
	glMultiTexCoord2f = 0;

	return true;
}


#pragma warning (default : 4113 4133 4047) 



