/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// texture management system
// this would probably make a nice C++ class, but I'm a dinosaur, so what the hey?


#include "quakedef.h"
#include "wad.h"
#include "cvar.h"
#include "screen.h"
#include "cmd.h"
#include "render.h"
#include "client.h"
#include "view.h"
#include "gl_model.h"
#include "glquake.h"
#include "glext.h"



// we keep everything except map textures which are potentially replacable
typedef struct
{
	int			texnum;
	char		identifier[64];

	int			width;
	int			height;

	qboolean	mipmap;
	qboolean	alpha;

	unsigned short crc;

	qboolean	current;
} gltexture_t;


// MAX_MAP_TEXTURES is 512, double that cos each texture *could* potentially have a fullbright,
// MAX_MAP_ENTITIES is 1024, all of them aside from 2 (worldspawn and info_player_start)
// *could* be alias models (it's highly unlikely though), each of which could also have a
// fullbright, so add 2048 (to be safer), then add a bit for overhead for the 2D stuff
#define	MAX_GLTEXTURES	3328

gltexture_t	gltextures[MAX_GLTEXTURES];

int			numgltextures;


// true if we're reloading textures
qboolean reload = false;

int detailtexture;

/*
==========================================================================================

					TEXTURE MANAGEMENT

==========================================================================================
*/


void GL_TileTexture (unsigned int *in, unsigned int *out, int inWidth, int inHeight, int tileX, int tileY)
{
	int outPos = 0;
	int i, j, k, l;
	int Row;

	for (l = 0; l < tileY; l++)
	{
		for (k = 0, Row = 0; k < inHeight; k++, Row += inHeight)
		{
			for (j = 0; j < tileX; j++)
			{
				for (i = 0; i < inWidth; i++)
				{
					out[outPos++] = in[i + Row];
				}
			}
		}
	}
}


/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	int		i;
	gltexture_t	*glt;

	for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
	{
		if (!strcmp (identifier, glt->identifier))
			return gltextures[i].texnum;
	}

	return -1;
}


void GL_TexDump_f (void)
{
	int		i;
	gltexture_t	*glt;
	FILE *f;

	f = fopen ("texdump.txt", "w");

	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		fprintf (f, "%5i   %s\n", gltextures[i].texnum, glt->identifier);

	fclose (f);
}


/*
===============
BuildIntensityTable

Got this from Quake II and if you don't like it it's just too bad
===============
*/
// we have a gamma table in Quake 1 already
static byte intensitytable[256];
static byte intensitytable2d[256];

void BuildIntensityTable (void)
{
	int i, j;

	for (i = 0; i < 256; i++)
	{
		// intensity table for world texes
		j = i * 2;
		if (j > 255) j = 255;

		intensitytable[i] = j;

		// intensity table for 2d view texes
		j = (i * 4) / 3;
		if (j > 255) j = 255;

		intensitytable2d[i] = j;
	}
}


/*
===============
GL_LightScaleTexture

Got this from Quake II and if you don't like it it's just too bad
===============
*/
void GL_LightScaleTexture (unsigned *texin, int inwidth, int inheight)
{
	int	i, c;
	byte *p;

	p = (byte *) texin;

	c = inwidth * inheight;

	for (i = 0; i < c; i++, p += 4)
	{
		p[0] = gammatable[intensitytable[p[0]]];
		p[1] = gammatable[intensitytable[p[1]]];
		p[2] = gammatable[intensitytable[p[2]]];
	}
}


/*
===============
GL_LightScale2DTexture

Similar to above but used for non-mipped texes, gives a less extreme scaling
===============
*/
void GL_LightScale2DTexture (unsigned *texin, int inwidth, int inheight)
{
	int	i, c;
	byte *p;

	p = (byte *) texin;

	c = inwidth * inheight;

	for (i = 0; i < c; i++, p += 4)
	{
		p[0] = gammatable[intensitytable2d[p[0]]];
		p[1] = gammatable[intensitytable2d[p[1]]];
		p[2] = gammatable[intensitytable2d[p[2]]];
	}
}


/*
===============
GL_ResampleTexture

Got this from Quake II, based on the assumption that the Quake II version of it *must*
necessarily be better (dunno if it actually is, can't really see much difference)
===============
*/
void GL_ResampleTexture (unsigned *texIn, int inwidth, int inheight, unsigned *texOut, int outwidth, int outheight)
{
	int i, j;
	unsigned *inrow, *inrow2;
	unsigned theFrac, fracstep;
	unsigned p1[1024], p2[1024];
	byte *pix1, *pix2, *pix3, *pix4;

	if (inwidth == outwidth && inheight == outheight)
	{
		// do a straight copy
		for (i = 0; i < inwidth * inheight; i++)
			texOut[i] = texIn[i];

		return;
	}

	fracstep = inwidth * 0x10000 / outwidth;
	theFrac = fracstep >> 2;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	theFrac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	for (i = 0; i < outheight; i++, texOut += outwidth)
	{
		inrow = texIn + inwidth * (int) ((i + 0.25) * inheight / outheight);
		inrow2 = texIn + inwidth * (int) ((i + 0.75) * inheight / outheight);

		// this isn't actually *used* any more
		theFrac = fracstep >> 1;

		for (j = 0; j < outwidth; j++)
		{
			pix1 = (byte *) inrow + p1[j];
			pix2 = (byte *) inrow + p2[j];
			pix3 = (byte *) inrow2 + p1[j];
			pix4 = (byte *) inrow2 + p2[j];

			((byte *) (texOut + j))[0] = (pix1[0] + pix2[0] + pix3[0] + pix4[0]) >> 2;
			((byte *) (texOut + j))[1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1]) >> 2;
			((byte *) (texOut + j))[2] = (pix1[2] + pix2[2] + pix3[2] + pix4[2]) >> 2;
			((byte *) (texOut + j))[3] = (pix1[3] + pix2[3] + pix3[3] + pix4[3]) >> 2;
		}
	}
}


// maximum texture size
extern int max_size;


/*
===============
ColorDiff

Get the difference between 2 colours.  This is accomplished by first converting to
greyscale, then comparing and subtracting.  There are better ways of doing this, but
greyscaling gives the most appropriate visual results for Quake textures.  The
alpha component is generally ignored.
===============
*/
unsigned int ColorDiff (unsigned int p1, unsigned int p2)
{
	byte *rgba1 = (byte *) &p1;
	byte *rgba2 = (byte *) &p2;

	int avg1;
	int avg2;

	// NTSC greyscale conversion standard
	avg1 = ((int) rgba1[0] * 30 + (int) rgba1[1] * 59 + (int) rgba1[2] * 11) / 300;
	avg2 = ((int) rgba2[0] * 30 + (int) rgba2[1] * 59 + (int) rgba2[2] * 11) / 300;

	if (avg1 > avg2)
		return avg1 - avg2;
	else return avg2 - avg1;
}


/*
===============
ColorAvg

Get the average of 2 colours, including alpha component
===============
*/
unsigned int ColorAvg (unsigned int p1, unsigned int p2)
{
	byte *rgba1 = (byte *) &p1;
	byte *rgba2 = (byte *) &p2;
	byte *rgbaavg;
	unsigned int avg;

	rgbaavg = (byte *) &avg;

	rgbaavg[0] = (byte) (((int) rgba1[0] + (int) rgba2[0]) / 2);
	rgbaavg[1] = (byte) (((int) rgba1[1] + (int) rgba2[1]) / 2);
	rgbaavg[2] = (byte) (((int) rgba1[2] + (int) rgba2[2]) / 2);
	rgbaavg[3] = (byte) (((int) rgba1[3] + (int) rgba2[3]) / 2);

	return avg;
}


/*
===============
ChooseLerp

Choose an interpolation function based on the lowest difference (i.e. which colour is
nearest to the original pixel): "colormin" is a confusing name, really...
===============
*/
unsigned int ChooseLerp (unsigned int me, unsigned int row, unsigned int col, unsigned int diag)
{
	unsigned int d1;
	unsigned int d2;
	unsigned int d3;
	unsigned int d4;
	unsigned int colormin;

	d1 = ColorDiff (me, row);
	d2 = ColorDiff (me, col);
	d3 = ColorDiff (me, diag);
	d4 = ColorDiff (row, col);

	colormin = d1;

	if (colormin > d2) colormin = d2;
	if (colormin > d3) colormin = d3;
	if (colormin > d4) colormin = d4;

	if (colormin == d1)
		return ColorAvg (me, row);
	else if (colormin == d2)
		return ColorAvg (me, col);
	else if (colormin == d3)
		return ColorAvg (me, diag);
	else return ColorAvg (me, ColorAvg (row, col));
}


/*
=================================
GL_InterpolatePixels

directional average scaling algorithm from http://www.compuphase.com/graphic/scale2.htm

this is edge-based, so it's very good at bringing out detail in complex textures when viewed
close up.  gives sharper results than bilinear or bicubic, without the "spotty" appearance of
my own algorithm (removed).  it's tolerant of multiple passes so long as MHQuakes internal
1024 x 1024 maximum is adhered to.

the paper at the url above discusses use of bit interleaving to determine colour differences.
i've abandoned that as it gives undesirable results with quake textures (likely the fault of
quake textures rather than anything in the paper).  instead i'm using option 2: converting to
greyscale.  the speed implications mentioned in this paper don't really apply to quake as the 
calculations don't occur at run time and the bottleneck is elsewhere (OpenGL texture uploading).
=================================
*/
void GL_InterpolatePixels (unsigned int *scaled, int w, int h)
{
	int hcount;
	int wcount;

	int x;

	// 2D arrays to make life easier for all of us
	static unsigned int temp1[1024][1024];	// copy of original
	static unsigned int temp2[1024][1024];	// upsampled

	// row up, col left, row down, col right, same row, same col
	int pos[6];

	int avg[4];	// rgba

	// copy the original data into our first 2D array
	for (hcount = 0, x = 0; hcount < h; hcount++)
	{
		for (wcount = 0; wcount < w; wcount++, x++)
		{
			temp1[hcount][wcount] = scaled[x];
		}
	}

	for (hcount = 0; hcount < h; hcount++)
	{
		// determine the positions of the 8 surrounding pixels - part one: rows
		// row up
		pos[0] = hcount - 1;
		if (pos[0] < 0) pos[0] += h;

		// row down
		pos[2] = hcount + 1;
		if (pos[2] >= h) pos[2] -= h;

		// same row
		pos[4] = hcount;

		for (wcount = 0; wcount < w; wcount++)
		{
			// determine the positions of the 8 surrounding pixels - part two: columns
			// col left
			pos[1] = wcount - 1;
			if (pos[1] < 0) pos[1] += w;

			// col right
			pos[3] = wcount + 1;
			if (pos[3] >= w) pos[3] -= w;

			// same col
			pos[5] = wcount;

			// each pixel in temp1 maps to 4 pixels in temp2
			temp2[hcount * 2][wcount * 2] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[0]][pos[5]], temp1[pos[4]][pos[1]], temp1[pos[0]][pos[1]]);
			temp2[hcount * 2][wcount * 2 + 1] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[0]][pos[5]], temp1[pos[4]][pos[3]], temp1[pos[0]][pos[3]]);
			temp2[hcount * 2 + 1][wcount * 2] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[2]][pos[5]], temp1[pos[4]][pos[1]], temp1[pos[2]][pos[1]]);
			temp2[hcount * 2 + 1][wcount * 2 + 1] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[2]][pos[5]], temp1[pos[4]][pos[3]], temp1[pos[2]][pos[3]]);
		}
	}

	// copy temp2 back to scaled
	for (hcount = 0, x = 0; hcount < h * 2; hcount++)
	{
		for (wcount = 0; wcount < w * 2; wcount++, x++)
		{
			scaled[x] = temp2[hcount][wcount];
		}
	}
}


/*
===============
GL_Upload32

We make sure in advance that all 2D images coming in here are padded with alpha to make them
a power of 2 and their s and t coords modified accordingly (only if they aren't already a
power of 2) - 2D images are less tolerant of resampling, etc, than the 3D stuff is.

Anything that is not a power of 2 is resampled *up* to the next power of 2, then capped at
either 1024 * 1024 or the max texture size, whichever is lower.  LINEAR_MIPMAP_LINEAR filtering
is always used so as to avoid abrubt transitions between miplevels in the world.  A GL_RGBA 
internal format is preferred for everything - using the full 32 bits should always be faster 
than having to unpack texture data before use.  Even the lowliest budget "price of a pack of 
fags" card has 32MB these days so memory is not a consideration.

Mipmapped textures are lightscaled in advance of uploading to increase the contrast range
and thereby give the illusion that our crappy 8-bit sources are something a little better!!!
===============
*/
qboolean lerptex;

void GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	int w;
	int h;
	static unsigned int scaled[1024 * 1024];

	// get the power of 2 equal to or above the texture size
	for (w = 1; w < width; w *= 2);
	for (h = 1; h < height; h *= 2);

	// cap the size
	if (w > 1024) w = 1024;
	if (h > 1024) h = 1024;

	// cap at max texture size
	if (w > max_size) w = max_size;
	if (h > max_size) h = max_size;

	// either scale it to the new size or do a straight copy
	GL_ResampleTexture (data, width, height, scaled, w, h);

	// don't interpolate cos it's too big
	if (w > 512 || h > 512) lerptex = false;

	if (lerptex)
	{
		GL_InterpolatePixels (scaled, w, h);

		w *= 2;
		h *= 2;

		lerptex = false;
	}

	if (mipmap)
	{
		GL_LightScaleTexture (scaled, w, h);

		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  w, 
					  h, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  scaled);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);
	}
	else
	{
		GL_LightScale2DTexture (scaled, w, h);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  w, 
					  h, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  scaled);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}


/*
===============
GL_Upload8
===============
*/
void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	static unsigned trans[1024 * 1024];
	int i;
	int s;
	int p;

	s = width * height;
	alpha = false;

	for (i = 0; i < s; i++)
	{
		p = data[i];

		if (p == 255)
		{
			alpha = true;
			trans[i] = 0;
		}
		else trans[i] = d_8to24table[p];
	}

	GL_Upload32 (trans, width, height, mipmap, alpha);
}


int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha)
{
	int	i;
	gltexture_t	*glt;
	unsigned short crc;

	crc = CRC_Block (data, (width * height));

	// see if the texture is allready present
	if (identifier[0])
	{
		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		{
			if (!strcmp (identifier, glt->identifier))
			{
				// the data doesn't match so reload it
				if (crc != glt->crc) goto CacheMismatch;

				// if we're reloading, reload it
				if (reload) goto CacheMismatch;

				// mark it as current
				gltextures[i].current = true;

				return gltextures[i].texnum;
			}
		}
	}

	glt = &gltextures[numgltextures];
	numgltextures++;

	strcpy (glt->identifier, identifier);
	glt->texnum = texture_extension_number++;

CacheMismatch:;
	glt->width = width;
	glt->height = height;
	glt->mipmap = mipmap;

	// mark it as current
	gltextures[numgltextures - 1].current = true;

	// determine if we need/want to interpolate it
	if (!mipmap)
	{
		// don't interpolate 2D stuff
		lerptex = false;
	}
	else if (identifier[0] == 'f' &&
		identifier[1] == 'b' &&
		identifier[2] == 'r' &&
		identifier[3] == '_' &&
		identifier[4] == 'p' &&
		identifier[5] == 'r' &&
		identifier[6] == 'o' &&
		identifier[7] == 'g' &&
		identifier[8] == 's' &&
		identifier[9] == '/')
	{
		// don't interpolate skins cos they're generally larger and it takes too long
		// (makes minimal difference anyway)
		lerptex = false;
	}
	else if (identifier[0] == 'f' &&
		identifier[1] == 'b' &&
		identifier[2] == 'r' &&
		identifier[3] == '_')
	{
		lerptex = true;
	}
	else if (alpha)
	{
		// don't interpolate if it has alpha cos it screws up the borders
		lerptex = false;
	}
	else if (identifier[0] == 'p' &&
		identifier[1] == 'r' &&
		identifier[2] == 'o' &&
		identifier[3] == 'g' &&
		identifier[4] == 's' &&
		identifier[5] == '/')
	{
		// don't interpolate skins cos they're generally larger and it takes too long
		// (makes minimal difference anyway)
		lerptex = false;
	}
	else if (identifier[0] == '*')
	{
		// don't interpolate liquid texes
		lerptex = false;
	}
	else
	{
		// interpolate everything else
		lerptex = true;
	}

	glBindTexture (GL_TEXTURE_2D, glt->texnum);

	GL_Upload8 (data, width, height, mipmap, alpha);

	return glt->texnum;
}


/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (qpic_t *pic, char *picname)
{
	return GL_LoadTexture (picname, pic->width, pic->height, pic->data, false, true);
}


/*
================
FindFullbrightTexture

Determine if a given texture has fullbright components
================
*/
qboolean FindFullbrightTexture (byte *pixels, int num_pix)
{
    int i;

    for (i = 0; i < num_pix; i++)
        if (pixels[i] > 223 && pixels[i] != 255)
            return true;

    return false;
}


/*
================
ConvertPixels

Converts a textures non-fullbright components to a specified alpha colour
================
*/
void ConvertPixels (byte *pixels, int num_pixels, int alphaColour)
{
    int i;

    for (i = 0; i < num_pixels; i++)
        if (pixels[i] < 224)
            pixels[i] = alphaColour;
}



/*
================
RemoveFullBrights

Removes all fullbright components from a texture
================
*/
void RemoveFullBrights (byte *pixels, int num_pixels, int alphaColour)
{
    int i;

    for (i = 0; i < num_pixels; i++)
        if (pixels[i] > 223 && pixels[i] != 255)
            pixels[i] = alphaColour;
}



/*
================
AverageColour

Determine the average colour (expressed in RGB rather than palletted) of all colours in
a texture
================
*/
float avg[3];

void AverageColour (byte *pixels, int num_pixels)
{
    int i;
	unsigned rgba;
	int count;

	avg[0] = avg[1] = avg[2] = 0;

    for (i = 0, count = 0; i < num_pixels; i++)
	{
		// ignore alpha
		if (pixels[i] == 255) continue;

		rgba = d_8to24table[pixels[i]];
		count++;

		avg[0] += ((byte *) &rgba)[0];
		avg[1] += ((byte *) &rgba)[1];
		avg[2] += ((byte *) &rgba)[2];
	}

	if (!count) return;

	avg[0] /= count;
	avg[1] /= count;
	avg[2] /= count;
}


/*
================
GL_LoadMapTexture

Given a pre-filled in texture_t this will do the GL stuff on it
================
*/
void GL_LoadMapTexture (texture_t *tx)
{
	char fbr_mask_name[64];
	static byte temp[1024 * 1024];
	int pixels = tx->width * tx->height;

	tx->gl_texturenum = GL_LoadTexture (tx->name,
										tx->width,
										tx->height,
										(byte *) (tx+1), 
										true,
										false);

	// check for fullbright pixels in the texture - but do not check liquid textures,
	// since they will be rendered fullbright anyway...
	if ((tx->name[0] != '*') && (FindFullbrightTexture ((byte *) (tx + 1), pixels)))
	{
		// first off, let's be nice and non-destructive to the original pixels
		memcpy ((byte *) temp, (byte *) (tx + 1), pixels);

		// convert any non-fullbright pixel to fully transparent
		ConvertPixels (temp, pixels, 255);

		// derive a new name for the texture to prevent mismatches
		sprintf (fbr_mask_name, "fbr_%s", tx->name);

		// and now load it - we need alpha for the non-fullbright components
		tx->fullbright = GL_LoadTexture (fbr_mask_name,
										 tx->width, 
										 tx->height, 
										 temp, 
										 true, 
										 true);
	}
	else tx->fullbright = 0;

	// if it's a liquid texture we find an underwater fog colour for it
	if (tx->name[0] == '*')
	{
		AverageColour ((byte *) (tx + 1), pixels);

		tx->uwater_fog[0] = avg[0] / 256;
		tx->uwater_fog[1] = avg[1] / 256;
		tx->uwater_fog[2] = avg[2] / 256;
		tx->uwater_fog[3] = 0.2;
	}
}


// oh what a steal!!!
void fractalnoise (byte *noise, int size, int startgrid)
{
	int x, y, g, g2, amplitude, min, max, size1 = size - 1, sizepower, gridpower;
	int *noisebuf;

#define n(x, y) noisebuf[((y) & size1) * size + ((x) & size1)]

	for (sizepower = 0; (1 << sizepower) < size; sizepower++);

	if (size != (1 << sizepower))
		Sys_Error ("fractalnoise: size must be power of 2\n");

	for (gridpower = 0; (1 << gridpower) < startgrid; gridpower++);

	if (startgrid != (1 << gridpower))
		Sys_Error ("fractalnoise: grid must be power of 2\n");

	startgrid = bound (0, startgrid, size);

	amplitude = 0xFFFF; // this gets halved before use

	noisebuf = malloc (size * size * sizeof (int));
	memset (noisebuf, 0, size * size * sizeof (int));

	for (g2 = startgrid; g2; g2 >>= 1)
	{
		// brownian motion (at every smaller level there is random behavior)
		amplitude >>= 1;

		for (y = 0; y < size; y += g2)
		{
			for (x = 0; x < size; x += g2)
			{
				n(x, y) += (rand () & amplitude);
			}
		}

		g = g2 >> 1;

		if (g)
		{
			// subdivide, diamond-square algorithm (really this has little to do with squares)
			// diamond
			for (y = 0; y < size; y += g2)
			{
				for (x = 0; x < size; x += g2)
				{
					n(x + g, y + g) = (n(x, y) + n(x + g2, y) + n(x, y + g2) + n(x + g2, y + g2)) >> 2;
				}
			}

			// square
			for (y = 0; y < size; y += g2)
			{
				for (x = 0; x < size; x += g2)
				{
					n(x + g, y) = (n(x, y) + n(x + g2, y) + n(x + g, y - g) + n(x + g, y + g)) >> 2;
					n(x, y + g) = (n(x, y) + n(x, y + g2) + n(x - g, y + g) + n(x + g,y + g)) >> 2;
				}
			}
		}
	}

	// find range of noise values
	min = max = 0;

	for (y = 0; y < size; y++)
	{
		for (x = 0; x < size; x++)
		{
			if (n(x, y) < min) min = n(x, y);
			if (n(x, y) > max) max = n(x, y);
		}
	}

	max -= min;
	max++;

	// normalize noise and copy to output
	for (y = 0;y < size;y++)
	{
		for (x = 0; x < size; x++)
		{
			*noise++ = (byte) (((n(x, y) - min) * 256) / max);
		}
	}

	free (noisebuf);
#undef n
}


void LoadDetailTexture (int detailwidth, int detailheight)
{
	int i, x, y, light;
	float vc[3], vx[3], vy[3], vn[3], lightdir[3];
	byte data[256][256][4], noise[256][256];

	detailtexture = texture_extension_number;
	texture_extension_number++;

	lightdir[0] = 0.5;
	lightdir[1] = 1;
	lightdir[2] = -0.25;

	VectorNormalize (lightdir);

	fractalnoise (&noise[0][0], 256, 256 >> 4);

	for (y = 0; y < detailheight; y++)
	{
		for (x = 0; x < detailwidth; x++)
		{
			vc[0] = x;
			vc[1] = y;

			vc[2] = noise[y][x] * (1.0f / 32.0f);

			vx[0] = x + 1;
			vx[1] = y;

			vx[2] = noise[y][(x + 1) % 256] * (1.0f / 32.0f);

			vy[0] = x;
			vy[1] = y + 1;

			vy[2] = noise[(y + 1) % 256][x] * (1.0f / 32.0f);

			VectorSubtract (vx, vc, vx);
			VectorSubtract (vy, vc, vy);
			CrossProduct (vx, vy, vn);
			VectorNormalize (vn);

			light = 128 - DotProduct (vn, lightdir) * 128;
			light = bound (0, light, 255);

			data[y][x][0] = data[y][x][1] = data[y][x][2] = light;
			data[y][x][3] = 255;
		}
	}

	// make it into a texture
	glBindTexture (GL_TEXTURE_2D, detailtexture);

	// mipmap this, but don't send it through GL_Upload32 or it will be lightscaled!
	gluBuild2DMipmaps (GL_TEXTURE_2D, 
					   GL_RGBA, 
					   detailwidth, 
					   detailheight, 
					   GL_RGBA, 
					   GL_UNSIGNED_BYTE, 
					   data);

	// always use LINEAR_MIPMAP_LINEAR
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


/****************************************/


// after changing resolution, any texture which is current will go
// corrupt, so reload them all
void GL_ReloadLightmap (int lightmapnum);

void GL_ReloadTextures (void)
{
	// tell GL_LoadTexture that we're reloading
	reload = true;

	// do the console stuff first
	GL_LoadCharset ();

	// the MH logo - VERY important...
	MHLoadLogo ();

	// particles
	R_InitParticleTexture ();

	// map textures - if we have a map loaded, that is...!!!
	if (r_worldentity.model && cl.worldmodel)
	{
		int i;
		int x;
		model_t *mod;

		for (x = 1; x < MAX_MODELS; x++)
		{
			mod = cl.model_precache[x];

			if (!mod) continue;

			switch (mod->type)
			{
			case mod_brush:
				// these will be loaded with the world
				if (mod->name[0] == '*') break;

				for (i = 0; i < mod->numtextures; i++)
				{
					if (mod->textures[i])
					{
						GL_LoadMapTexture (mod->textures[i]);
					}
				}

				break;

			case mod_alias:
				break;

			case mod_sprite:
				break;

			default:
				break;
			}
		}

		// lightmaps - the data is kept in main memory so we just reload from that
		for (i = 0; i < 64; i++)
			GL_ReloadLightmap (i);
	}

	// OK, we're done reloading now
	reload = false;
}


/*
===============
GL_TexmanInit

Initialize the texture management system
===============
*/
void GL_TexmanInit (void)
{
	memset (gltextures, 0, sizeof (gltextures));

	// set up the intensity table for texture lightscaling
	BuildIntensityTable ();

	LoadDetailTexture (256, 256);
}


/*
===============
GL_TexManMapLoadFinish

Does some housekeeping after loading the map
===============
*/
void GL_TexManMapLoadFinish (void)
{
}
