/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 

#include "quakedef.h"


qboolean FrameHasLava;
extern msurface_t *genericchain;
float hhaze_pmatrix[16];
float hhaze_mvmatrix[16];
int hhaze_viewport[4];

#define LAVA_RENDER_SCALE 72.0f

qboolean warpkill;


static void transform_point (float out[4], const float m[16], const float in[4])
{
#define MATRIX(row,col)  m[col * 4 + row]
	out[0] = MATRIX(0, 0) * in[0] + MATRIX(0, 1) * in[1] + MATRIX(0, 2) * in[2] + MATRIX(0, 3) * in[3];
	out[1] = MATRIX(1, 0) * in[0] + MATRIX(1, 1) * in[1] + MATRIX(1, 2) * in[2] + MATRIX(1, 3) * in[3];
	out[2] = MATRIX(2, 0) * in[0] + MATRIX(2, 1) * in[1] + MATRIX(2, 2) * in[2] + MATRIX(2, 3) * in[3];
	out[3] = MATRIX(3, 0) * in[0] + MATRIX(3, 1) * in[1] + MATRIX(3, 2) * in[2] + MATRIX(3, 3) * in[3];
#undef MATRIX
}


// version of gluProject from the mesa library converted to use vec3_t's as it's in and out
void mhqgluProject (vec3_t vin, const float modelMatrix[16], const float projMatrix[16], const int viewport[4], vec3_t vout)
{
	float in[4];
	float out[4];

	// save to a local cos we will be modifying it here
	in[0] = vin[0];
	in[1] = vin[1];
	in[2] = vin[2];
	in[3] = 1.0;

	transform_point (out, modelMatrix, in);
	transform_point (in, projMatrix, out);

	// could not project
	if (in[3] == 0.0) return;

	in[0] /= in[3];
	in[1] /= in[3];
	in[2] /= in[3];

	// transform to viewport coords and store back to vout
	vout[0] = viewport[0] + (1 + in[0]) * viewport[2] / 2;
	vout[1] = viewport[1] + (1 + in[1]) * viewport[3] / 2;
	vout[2] = (1 + in[2]) / 2;
}


void R_ProjectSurfaceVerts (void)
{
	glpoly_t *p;
	msurface_t *surf;

	for (surf = genericchain; surf; surf = surf->genericchain)
	{
		// do each poly
		for (p = surf->polys; p; p = p->next)
		{
			// project poly to window coords
			mhqgluProject (p->midpoint, hhaze_mvmatrix, hhaze_pmatrix, hhaze_viewport, p->wincoords);
		}
	}
}


void R_HazeMaskSurfacesOnDrawList (vec3_t p1, vec3_t p2, vec3_t p3, vec3_t p4)
{
	vec3_t screenmins = {999999, 999999, 999999};
	vec3_t screenmaxs = {-999999, -999999, -999999};
	vec3_t screenpoints[4];
	msurface_t *surf;

	// convert vertex points to screen coords.  z should be the same in all cases
	mhqgluProject (p1, hhaze_mvmatrix, hhaze_pmatrix, hhaze_viewport, screenpoints[0]);
	mhqgluProject (p2, hhaze_mvmatrix, hhaze_pmatrix, hhaze_viewport, screenpoints[1]);
	mhqgluProject (p3, hhaze_mvmatrix, hhaze_pmatrix, hhaze_viewport, screenpoints[2]);
	mhqgluProject (p4, hhaze_mvmatrix, hhaze_pmatrix, hhaze_viewport, screenpoints[3]);

	// get the screen mins and maxs
	if (screenpoints[0][0] < screenmins[0]) screenmins[0] = screenpoints[0][0];
	if (screenpoints[1][0] < screenmins[0]) screenmins[0] = screenpoints[1][0];
	if (screenpoints[2][0] < screenmins[0]) screenmins[0] = screenpoints[2][0];
	if (screenpoints[3][0] < screenmins[0]) screenmins[0] = screenpoints[3][0];

	if (screenpoints[0][1] < screenmins[1]) screenmins[1] = screenpoints[0][1];
	if (screenpoints[1][1] < screenmins[1]) screenmins[1] = screenpoints[1][1];
	if (screenpoints[2][1] < screenmins[1]) screenmins[1] = screenpoints[2][1];
	if (screenpoints[3][1] < screenmins[1]) screenmins[1] = screenpoints[3][1];

	if (screenpoints[0][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[0][0];
	if (screenpoints[1][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[1][0];
	if (screenpoints[2][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[2][0];
	if (screenpoints[3][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[3][0];

	if (screenpoints[0][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[0][1];
	if (screenpoints[1][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[1][1];
	if (screenpoints[2][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[2][1];
	if (screenpoints[3][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[3][1];

	// z should always be the same because these points will always face the player
	screenmins[2] = screenmaxs[2] = screenpoints[0][2];

	// re-use the detail texture chain for these as it saves on unnecessary nested loops and omits surfs we
	// don't need to check
	for (surf = genericchain; surf; surf = surf->genericchain)
	{
		// already masked
		if (surf->polys->hazemask) continue;

		// surf is in front of the haze.  z should be the same in both screenmins and screenmaxs cases
		// so we only need to check out one of them.  these surfs will only have 1 poly so don't bother walking the chain
		if (surf->polys[0].wincoords[2] < screenmins[2]) continue;

		// surf is behind the haze in the z order - now see is it directly behind
		if (surf->polys[0].wincoords[0] < screenmins[0]) continue;		// too far left
		if (surf->polys[0].wincoords[1] < screenmins[1]) continue;		// too far down
		if (surf->polys[0].wincoords[0] > screenmaxs[0]) continue;		// too far right
		if (surf->polys[0].wincoords[1] > screenmaxs[1]) continue;		// too far up

		// surf is directly behind the haze so mark it
		surf->polys->hazemask = true;
	}

	// check has every surf been warped.
	// assume that they have unless we prove otherwise
	warpkill = true;

	// another stroll through the surfs - ho hum.  shouldn't hurt too much this time though...
	// the data will be in the cpu cache and it's cheaper than a loada unnecessary recalcs
	for (surf = genericchain; surf; surf = surf->genericchain)
	{
		// already masked
		if (surf->polys->hazemask) continue;

		// not every surf in the view has been warped if we get this far any time
		warpkill = false;

		// don;t need to check any more
		return;
	}
}


void R_SetupHazeMaskForLavaSurf (msurface_t *surf, vec3_t up, vec3_t right)
{
	glpoly_t *p = surf->polys;
	vec3_t p1;
	vec3_t p2;
	vec3_t p3;
	vec3_t p4;
	int i;
	float offset;
	float scale;

	// a warped surf is a chain of polys so batch up this part of the render per surf rather than per poly
	for (; p; p = p->next)
	{
		scale = (p->midpoint[0] - r_origin[0]) * vpn[0] + (p->midpoint[1] - r_origin[1]) * vpn[1] + (p->midpoint[2] - r_origin[2]) * vpn[2];

		// too near the screen or behind the current viewpoint
		if (scale < 50) continue;

		// do 2 quads rather than one tall one to retain vertical scale when looking up/down
		for (i = 0; i < 2; i++)
		{
			offset = 0.5f + (float) i; // * 0.666f;

			// calculate vertex points for a virtual particle
			p1[0] = p->midpoint[0] - (right[0] + up[0]) * 0.5f;
			p1[1] = p->midpoint[1] - (right[1] + up[1]) * 0.5f;
			p1[2] = p->midpoint[2] - (right[2] + up[2]) * 0.5f + (LAVA_RENDER_SCALE * offset);

			p2[0] = p1[0] + up[0];
			p2[1] = p1[1] + up[1];
			p2[2] = p1[2] + up[2];

			p3[0] = p2[0] + right[0];
			p3[1] = p2[1] + right[1];
			p3[2] = p2[2] + right[2];

			p4[0] = p1[0] + right[0];
			p4[1] = p1[1] + right[1];
			p4[2] = p1[2] + right[2];

			// now check for surface masking
			R_HazeMaskSurfacesOnDrawList (p1, p2, p3, p4);
		}
	}
}


void R_SetupHazeMasks (void)
{
	int i;
	msurface_t *surf;
	vec3_t up, right;

	up[0] = vup[0] * LAVA_RENDER_SCALE;
	up[1] = vup[1] * LAVA_RENDER_SCALE;
	up[2] = vup[2] * LAVA_RENDER_SCALE;

	right[0] = vright[0] * LAVA_RENDER_SCALE;
	right[1] = vright[1] * LAVA_RENDER_SCALE;
	right[2] = vright[2] * LAVA_RENDER_SCALE;

	warpkill = false;

	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		// not in the pvs for this viewleaf (this is only ever true if the surf
		// is also lava, so we don't need to bother checking that condition also)
		if (surf->lavavisframe != r_visframecount) continue;

		// create the haze masks
		R_SetupHazeMaskForLavaSurf (surf, up, right);

		// every surf in the view has been warped
		if (warpkill) return;
	}
}


// fade the surface warp in and out to prevent popping
void R_SurfaceHazeWarp (void)
{
	msurface_t *surf;
	float hazetime = r_frametime * 2.0f;

	// only do this is we have lava visible
	if (FrameHasLava)
	{
		// save out the modelview, projection and viewport
		glGetFloatv (GL_MODELVIEW_MATRIX, hhaze_mvmatrix);
		glGetFloatv (GL_PROJECTION_MATRIX, hhaze_pmatrix);
		glGetIntegerv (GL_VIEWPORT, hhaze_viewport);

		// get the midpoint of each surf on the drawlist in window coords
		R_ProjectSurfaceVerts ();

		// mark those surfs which are masked by haze
		R_SetupHazeMasks ();
	}

	// now set up the haze factors for both surfs that have haze and surfs that haven't.  we do surfs that haven't because
	// one of them might have been hazed the previous frame, and we want to fade out smoothly rather than abruptly going
	// back to un-hazed (which creates an ugly "popping" appearance).  we also do this if we don't have lava visible, as
	// we might have had lava visible the previous frame.
	for (surf = genericchain; surf; surf = surf->genericchain)
	{
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (surf->flags & SURF_UNDERWATER)
		{
			// constant warp with no fading
			surf->polys->warpfade = 0.5f;
			continue;
		}

		if (!surf->polys) continue;

		if (surf->polys->hazemask)
		{
			// fade the warp up if the surf is hazemasked, upper bound it at 1
			surf->polys->warpfade += hazetime;
			if (surf->polys->warpfade >= 0.5f) surf->polys->warpfade = 0.5f;
		}
		else
		{
			// fade the warp down
			surf->polys->warpfade -= hazetime;

			// if the warp is still fading down hazemask it, otherwise lower bound it at 0
			if (surf->polys->warpfade > 0)
				surf->polys->hazemask = true;
			else surf->polys->warpfade = 0;
		}
	}
}


/*
===================
R_MarkLava

Finds all surfs in the current PVS that are lava, totally ignoring frustum culling
===================
*/
void R_MarkLava (void)
{
	int i;
	int j;
	mleaf_t *leaf;
	char *texname;

	FrameHasLava = false;

	for (i = 0, leaf = cl.worldmodel->leafs; i < cl.worldmodel->numleafs; i++, leaf++)
	{
		// only interested in empty here
		if (leaf->contents != CONTENTS_EMPTY) continue;

		for (j = 0; j < leaf->nummarksurfaces; j++)
		{
			leaf->firstmarksurface[j]->lavavisframe = -1;

			if (!(leaf->firstmarksurface[j]->flags & SURF_DRAWTURB)) continue;

			// keep it clean
			texname = leaf->firstmarksurface[j]->texinfo->texture->name;

			if (leaf->firstmarksurface[j]->flags & SURF_DRAWLAVA)
			{
				leaf->firstmarksurface[j]->lavavisframe = r_visframecount;

				FrameHasLava = true;
			}
		}
	}
}

