/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#include "quakedef.h"

// includes for image libs
#include "jpeglib.h"
#include "png.h"

// intel integrated graphics work reasonably well at opengl 1.4 level, but require some hacks to maintain performance
extern qboolean IntelIntegrated;
extern qboolean IntelPicmip;

// true if we're loading water, false otherwise.  only ever set to true through the GL_LoadTexture entrypoint,
// set back to false immediately when done to prevent false positives.
qboolean watertexture;

int mapshottexturenum = 0;

void MD5_Checksum (unsigned char *data, int dataLen, unsigned char *checksum);
char MD5_Compare (unsigned char *digest1, unsigned char *digest2);
void MD5_Copy (unsigned char *srcdigest, unsigned char *dstdigest);

// for testing 'n' shit
void WriteTGA (char *name, int width, int height, byte *data, int bpp)
{
	int i;
	byte buffer[18];
	FILE *f;
	char newname[MAX_OSPATH];

	// remove invalid characters from name
	for (i = 0; ; i++)
	{
		if (!name[i]) break;

		// invalid windows filename characters
		if (name[i] == '\\') name[i] = '_';
		if (name[i] == '/') name[i] = '_';
		if (name[i] == ':') name[i] = '_';
		if (name[i] == '*') name[i] = '_';
		if (name[i] == '?') name[i] = '_';
		if (name[i] == '"') name[i] = '_';
		if (name[i] == '<') name[i] = '_';
		if (name[i] == '>') name[i] = '_';
		if (name[i] == '|') name[i] = '_';
	}

	// construct the new name
	sprintf (newname, "%s.tga", name);

	// open it.
	f = fopen (newname, "wb");

	// oops!  we likely don't have write access to the root Quake folder!
	// like, we *could* be out of disk space, but in *2007*?  get real.
	if (!f) return;

	// construct the header
	memset (buffer, 0, 18);
	buffer[2] = 2;		// uncompressed type
	buffer[12] = width & 255;
	buffer[13] = width >> 8;
	buffer[14] = height & 255;
	buffer[15] = height >> 8;
	buffer[16] = 32;	// pixel size
	buffer[17] = 32;	// flipped

	// write out the header
	fwrite (buffer, 18, 1, f);

	// write out the data
	for (i = 0; i < (width * height); i++)
	{
		if (bpp == 8)
		{
			// swap to BGR
			fputc (((byte *) &d_8to24table[data[i]])[2], f);
			fputc (((byte *) &d_8to24table[data[i]])[1], f);
			fputc (((byte *) &d_8to24table[data[i]])[0], f);
			fputc (((byte *) &d_8to24table[data[i]])[3], f);
		}
		else if (bpp == 24)
		{
			// swap to BGR
			fputc (data[i * 3 + 2], f);
			fputc (data[i * 3 + 1], f);
			fputc (data[i * 3 + 0], f);
			fputc (255, f);
		}
		else if (bpp == 32)
		{
			// swap to BGR
			fputc (data[i * 4 + 2], f);
			fputc (data[i * 4 + 1], f);
			fputc (data[i * 4 + 0], f);
			fputc (data[i * 4 + 3], f);
		}
		else
		{
			// invalid bpp, so just write random garbage
			// in an infinite universe, it might make a pretty picture someday.
			fputc (rand () % 256, f);
			fputc (rand () % 256, f);
			fputc (rand () % 256, f);
			fputc (rand () % 256, f);
		}
	}

	// done!
	fclose (f);
}


char ThePaths[4][64];

char *SearchPaths[5] = {NULL, NULL, NULL, NULL, NULL};

void BuildTextureSearchPaths (char *p1, char *p2, char *p3, char *p4)
{
	// clear down all search paths
	SearchPaths[0] = SearchPaths[1] = SearchPaths[2] = SearchPaths[4] = SearchPaths[5] = NULL;

	if (p1)
	{
		strcpy (ThePaths[0], p1);
		SearchPaths[0] = ThePaths[0];
	}
	else return;

	if (p2)
	{
		strcpy (ThePaths[1], p2);
		SearchPaths[1] = ThePaths[1];
	}
	else return;

	if (p3)
	{
		strcpy (ThePaths[2], p3);
		SearchPaths[2] = ThePaths[2];
	}
	else return;

	if (p4)
	{
		strcpy (ThePaths[3], p4);
		SearchPaths[3] = ThePaths[3];
	}
	else return;
}


void GL_AllocTexture (int *texnum)
{
	glGenTextures (1, texnum);
}


gltexture_t	*gltextures = NULL;


int		gl_solid_format = 3;
int		gl_alpha_format = 4;


int		texels;


#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;

qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	*gl;
	char newname[MAX_OSPATH];

	p = W_GetLumpName (name);
	gl = (glpic_t *) p->data;

	sprintf (newname, "wad/%s", name);

	BuildTextureSearchPaths ("textures/", NULL, NULL, NULL);

	gl->gltexture = GL_LoadTexture (newname, p->width, p->height, p->data, TEX_INCACHE);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		*gl;

	for (pic = menu_cachepics, i = 0; i < menu_numcachepics; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS) Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");

	menu_numcachepics++;
	strcpy (pic->name, path);

	// load the pic from disk
	dat = (qpic_t *)COM_LoadTempFile (path);

	if (!dat) Sys_Error ("Draw_CachePic: failed to load %s", path);
	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
	{
		byte alpha_colour;

		memcpy (menuplyr_pixels, dat->data, dat->width * dat->height);

		// fix up the alpha - assume that the correct alpha colour is the first entry
		alpha_colour = menuplyr_pixels[0];

		for (i = 0; i < dat->width * dat->height; i++)
		{
			if (menuplyr_pixels[i] == alpha_colour)
			{
				// correct alpha colour
				menuplyr_pixels[i] = 255;
			}
		}
	}

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	gl = (glpic_t *) pic->pic.data;
	BuildTextureSearchPaths ("textures/", NULL, NULL, NULL);
	gl->gltexture = GL_LoadTexture (path, dat->width, dat->height, dat->data, TEX_INCACHE);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return &pic->pic;
}


/*
=======================================================================================================================

				TEXTURE LOADING

=======================================================================================================================
*/


/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	gltexture_t	*glt;

	for (glt = gltextures; glt; glt = glt->next)
	{
		if (!strcmp (identifier, glt->identifier))
			return glt->texnum;
	}

	return -1;
}


/*
================
GL_ResampleTexture
================
*/
unsigned scaled[MAX_QUAKE_TEXTURE_SIZE * MAX_QUAKE_TEXTURE_SIZE];

typedef struct resamplepoint_s
{
	int p1;
	int p2;
	int p1percent;
	int p2percent;
} resamplepoint_t;


resamplepoint_t resamplew[MAX_QUAKE_TEXTURE_SIZE];
resamplepoint_t resampleh[MAX_QUAKE_TEXTURE_SIZE];

void MakeResamplePoint (resamplepoint_t *resamp, int inscale, int outscale, int outmult)
{
	int i;

	// set up the resample points across
	for (i = 0; i < outscale; i++)
	{
		// calculate the equalivalent point in the "in" buffer
		float fp = (float) (inscale - 1) * ((float) i / (float) (outscale - 1));

		resamp[i].p1 = (int) floor (fp);
		resamp[i].p2 = (int) ceil (fp);

		// the decimal part is the percentage of p2 used
		resamp[i].p2percent = (int) ((fp - (float) resamp[i].p1) * 100.0f);

		// the rest is the percentage of p2 used
		resamp[i].p1percent = 100 - resamp[i].p2percent;

		resamp[i].p1 *= outmult;
		resamp[i].p2 *= outmult;
	}
}


void GL_ResampleTexture (byte *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight, int bpp)
{
	int h;
	int w;
	int outpoint;
	byte *indata[4];
	byte *outdata;
	int inrgba[4];
	int inpercents[4];
	int resampmult;
	int pix;

	if (bpp == 8)
		resampmult = 1;
	else resampmult = 4;

	// set up our resample points in 2 lookup tables for added speed
	MakeResamplePoint (resamplew, inwidth, outwidth, 1 * resampmult);
	MakeResamplePoint (resampleh, inheight, outheight, inwidth * resampmult);

	for (h = 0, outpoint = 0; h < outheight; h++)
	{
		for (w = 0; w < outwidth; w++, outpoint++)
		{
			// get a valid pointer to the outdata
			outdata = (byte *) &out[outpoint];

			// set up the percents
			inpercents[0] = (resamplew[w].p1percent + resampleh[h].p1percent);
			inpercents[1] = (resamplew[w].p2percent + resampleh[h].p1percent);
			inpercents[2] = (resamplew[w].p1percent + resampleh[h].p2percent);
			inpercents[3] = (resamplew[w].p2percent + resampleh[h].p2percent);

			// initially 0
			inrgba[0] = inrgba[1] = inrgba[2] = inrgba[3] = 0;

			// set up the data points
			if (bpp == 8)
			{
				// in-place upsampling and resampling
				indata[0] = (byte *) &d_8to24table[in[resamplew[w].p1 + resampleh[h].p1]];
				indata[1] = (byte *) &d_8to24table[in[resamplew[w].p2 + resampleh[h].p1]];
				indata[2] = (byte *) &d_8to24table[in[resamplew[w].p1 + resampleh[h].p2]];
				indata[3] = (byte *) &d_8to24table[in[resamplew[w].p2 + resampleh[h].p2]];
			}
			else
			{
				// just resample
				indata[0] = &in[resamplew[w].p1 + resampleh[h].p1];
				indata[1] = &in[resamplew[w].p2 + resampleh[h].p1];
				indata[2] = &in[resamplew[w].p1 + resampleh[h].p2];
				indata[3] = &in[resamplew[w].p2 + resampleh[h].p2];
			}

			// accumulate the data for each point
			for (pix = 0; pix < 4; pix++)
			{
				// accumulate as ints to avoid byte data type wrapping
				inrgba[pix] += (int) indata[0][pix] * inpercents[0];
				inrgba[pix] += (int) indata[1][pix] * inpercents[1];
				inrgba[pix] += (int) indata[2][pix] * inpercents[2];
				inrgba[pix] += (int) indata[3][pix] * inpercents[3];

				// copy across
				outdata[pix] = (byte) (inrgba[pix] / 400);
			}
		}
	}
}


void GL_HalveTextureW (int width, int height, unsigned *data)
{
	int		i, size;
	byte	*out, *in;

	out = in = (byte *) data;
	size = (width * height) >> 1;

	for (i = 0; i < size; i++, out += 4, in += 8)
	{
		out[0] = (in[0] + in[4]) >> 1;
		out[1] = (in[1] + in[5]) >> 1;
		out[2] = (in[2] + in[6]) >> 1;
		out[3] = (in[3] + in[7]) >> 1;
	}
}


void GL_HalveTextureH (int width, int height, unsigned *data)
{
	int		i, j;
	byte	*out, *in;

	out = in = (byte *) data;
	height >>= 1;
	width <<= 2;

	for (i = 0; i < height; i++, in += width)
	{
		for (j = 0; j < width; j += 4, out += 4, in += 4)
		{
			out[0] = (in[0] + in[width + 0]) >> 1;
			out[1] = (in[1] + in[width + 1]) >> 1;
			out[2] = (in[2] + in[width + 2]) >> 1;
			out[3] = (in[3] + in[width + 3]) >> 1;
		}
	}
}


/*
===============
GL_UploadTexture
===============
*/
// sometimes we need to know the size textures were uploaded at
int upload_width;
int upload_height;

// water fog colour
vec3_t waterfog;


byte DigitalVibrance[256] = {255};


void InitVibranceTable (void)
{
	int i;
	int v;

	for (i = 0; i < 256; i++)
	{
		v = i * i;
		v /= 255;
		v += i;

		DigitalVibrance[i] = (v > 255) ? 255 : v;
	}
}


void GL_AlphaMaskTexture (int w, int h, unsigned *data)
{
	int i;
	byte *rgba = (byte *) data;

	for (i = 0; i < (w * h); i++)
	{
		// set alpha to red
		rgba[3] = rgba[1];

		// set r/g/b to white
		rgba[0] = rgba[1] = rgba[2] = 255;

		rgba += 4;
	}
}

glpic_t Font8x8[64];

void GL_MakeFont8x8 (int width, int height, unsigned *data)
{
	int w, h;
	int w1, h1;
	int w2, h2;
	int tile_width;
	int tile_height;
	int picnum;
	int minw, maxw;
	qboolean valid_spot;

	if (width % 8) Sys_Error ("GL_MakeFont8x8: width % 8");
	if (height % 8) Sys_Error ("GL_MakeFont8x8: height % 8");

	tile_width = width / 8;
	tile_height = height / 8;

	// set correct alpha colour
	for (h = 0; h < (height * width); h++)
	{
		// magenta is the standard for alpha with bmps
		if (((byte *) &data[h])[0] == 255 && ((byte *) &data[h])[1] == 0 && ((byte *) &data[h])[2] == 255) data[h] = 0;
	}

	for (h = 0, picnum = 0; h < height; h += tile_height)
	{
		for (w = 0; w < width; w += tile_width, picnum++)
		{
			// t co-ords (may need to invert for bottom-left paradigm :P)
			Font8x8[picnum].tl = (float) h / (float) height;
			Font8x8[picnum].th = (float) (h + tile_height) / (float) height;

			// get leftmost w
			for (h1 = h, minw = width + 1, maxw = -1; h1 < (h + tile_height); h1++)
			{
				qboolean got_minw = false;
				qboolean got_maxw = false;

				h2 = h1 * width;

				for (w1 = w + 1, w2 = (w + tile_width - 1); w1 < w2; w1++, w2--)
				{
					// initially valid
					valid_spot = true;

					// now check it
					if (((byte *) &data[h2 + w1])[3] == 0) valid_spot = false;

					// check validity and set
					if (valid_spot && w1 < minw)
					{
						minw = w1;
						got_minw = true;
					}

					// initially valid
					valid_spot = true;

					// now check it
					if (((byte *) &data[h2 + w2])[3] == 0) valid_spot = false;

					// check validity and set
					if (valid_spot && w2 > maxw)
					{
						maxw = w2;
						got_maxw = true;
					}

					if (got_minw && got_maxw) break;
				}
			}

			// check min and max validity.  if they're bad, just set to the bounding square
			if (maxw < minw || minw >= width || maxw < 0)
			{
				minw = w;
				maxw = w + tile_width;
			}

			// set s co-ords based on leftmost and rightmost w - take one pixel on either side to keep it bilerp happy
			Font8x8[picnum].sl = (float) (minw - 1) / (float) width;
			Font8x8[picnum].sh = (float) (maxw + 1) / (float) width;
		}
	}
}


void GL_UploadTexture (unsigned *data, int width, int height, int flags)
{
	int i;
	unsigned *upload = NULL;
	int scaled_width, scaled_height;

	// get power of 2 equal or above the size
	for (scaled_width = 1; scaled_width < width; scaled_width *= 2);
	for (scaled_height = 1; scaled_height < height; scaled_height *= 2);

	// clamp at internal limit - this is important, as some cards can support larger textures but be limited in memory,
	// so we need to constrain the size a bit more than what the card tells us we can
	if (scaled_width > gl_max_size) scaled_width = gl_max_size;
	if (scaled_height > gl_max_size) scaled_height = gl_max_size;

	texels += scaled_width * scaled_height;

	// what to do next depends on our bpp and whether or not we need to rescale
	if ((flags & TEX_32BPP) && scaled_width == width && scaled_height == height)
	{
		// just copy it straight in
		// ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++
		// ++ the external texture loaders depend on scaled never being used here as a quirk to save on memory allocations.     ++
		// ++ if this code is ever changed so that scaled is used, they will need to be modified to reflect that.               ++
		// ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++ IMPORTANT ++
		upload = data;
		// data will always == scaled here...
	}
	else if (!(flags & TEX_32BPP) && scaled_width == width && scaled_height == height)
	{
		// convert data to bytes
		byte *data8 = (byte *) data;

		// expand 8bpp to 32 bpp
		for (i = 0; i < width * height; i++)
			scaled[i] = d_8to24table[data8[i]];

		// hipnotic alpha hack
		if (flags & TEX_ALPHAHACK)
		{
			int x, y;

			for (i = 0, x = 0; x < width; x++)
			{
				for (y = 0; y < height; y++, i++)
				{
					qboolean hackpos = false;

					// 4 panels
					if (x > 2 && y > 2 && x < 29 && y < 29) hackpos = true;
					if (x > 34 && y > 2 && x < 61 && y < 29) hackpos = true;
					if (x > 2 && y > 34 && x < 29 && y < 61) hackpos = true;
					if (x > 34 && y > 34 && x < 61 && y < 61) hackpos = true;

					if (hackpos)
					{
						data8 = (byte *) &scaled[i];

						// window
						data8[0] = data8[1] = data8[2] = 120 + (rand () % 16);
						data8[3] = 192;
					}
				}
			}
		}

		// set upload pointer to the scaled buffer
		upload = scaled;
	}
	else
	{
		// resample it
		GL_ResampleTexture ((byte *) data, width, height, scaled, scaled_width, scaled_height, (flags & TEX_32BPP) ? 32 : 8);

		// set upload pointer to the scaled buffer
		upload = scaled;
	}

	// paranoia!
	if (!upload) Sys_Error ("GL_UploadTexture: upload buffer not set");

	if ((flags & TEX_32BPP) && (flags & TEX_MIPMAP) && IntelPicmip)
	{
		// halve the size (in place) to preserve texture memory
		GL_HalveTextureW (scaled_width, scaled_height, upload);
		scaled_width /= 2;

		GL_HalveTextureH (scaled_width, scaled_height, upload);
		scaled_height /= 2;
	}

	if (DigitalVibrance[0] > 0) InitVibranceTable ();

#if 0
	for (i = scaled_width * scaled_height - 1; i >= 0; i--)
	{
		byte *rgba = (byte *) &upload[i];

		rgba[0] = DigitalVibrance[rgba[0]];
		rgba[1] = DigitalVibrance[rgba[1]];
		rgba[2] = DigitalVibrance[rgba[2]];
	}
#endif

	// sometimes we need to know the size textures were uploaded at
	upload_width = scaled_width;
	upload_height = scaled_height;

	if (watertexture)
	{
		waterfog[0] = waterfog[1] = waterfog[2] = 0;

		// read in the average upload colour into waterfog
		for (i = 0; i < upload_width * upload_height; i++)
		{
			byte *rgba = (byte *) &upload[i];

			waterfog[0] += (float) rgba[0] / 255.0f;
			waterfog[1] += (float) rgba[1] / 255.0f;
			waterfog[2] += (float) rgba[2] / 255.0f;
		}

		waterfog[0] /= (float) (upload_width * upload_height);
		waterfog[1] /= (float) (upload_width * upload_height);
		waterfog[2] /= (float) (upload_width * upload_height);
	}

	// additional texture processing
	if (flags & TEX_ALPHAMASK) GL_AlphaMaskTexture (upload_width, upload_height, upload);
	if (flags & TEX_FONT8X8) GL_MakeFont8x8 (upload_width, upload_height, upload);

	// set mipmapping
	glHint (GL_GENERATE_MIPMAP_HINT, GL_NICEST);
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, (flags & TEX_MIPMAP) ? GL_TRUE : GL_FALSE);

	// upload to OpenGL
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, upload);

	// apply proper filtering
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (flags & TEX_MIPMAP) ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// clear upload pointer
	upload = NULL;
}


qboolean GL_LoadPNG (char *uploadname, int flags)
{
	unsigned char header[8], **rowpointers;
	png_structp	png;
	png_infop	pnginfo;
	int		y, width, height, bitdepth, colortype;
	int		interlace, compression, filter, bytesperpixel;
	unsigned long	rowbytes;
	int w, h;
	FILE *f;
	byte *texdata;
	unsigned *buffer = NULL;

	COM_FOpenFile (uploadname, &f);

	if (!f) return false;

	fread (header, 1, 8, f);

	if (png_sig_cmp (header, 0, 8))
	{
		fclose (f);
		return false;
	}

	if (!(png = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)))
	{
		fclose (f);
		return false;
	}

	if (!(pnginfo = png_create_info_struct (png)))
	{
		png_destroy_read_struct (&png, &pnginfo, NULL);
		fclose (f);
		return false;
	}

	if (setjmp (png->jmpbuf))
	{
		png_destroy_read_struct (&png, &pnginfo, NULL);
		fclose (f);
		return false;
	}

	png_init_io (png, f);
	png_set_sig_bytes (png, 8);
	png_read_info (png, pnginfo);
	png_get_IHDR (png, pnginfo, (png_uint_32 *) &width, (png_uint_32 *) &height, &bitdepth, &colortype, &interlace, &compression, &filter);

	if (colortype == PNG_COLOR_TYPE_PALETTE)
	{
		png_set_palette_to_rgb (png);
		png_set_filler (png, 255, PNG_FILLER_AFTER);			
	}

	if (colortype == PNG_COLOR_TYPE_GRAY && bitdepth < 8)
		png_set_gray_1_2_4_to_8 (png);
	
	if (png_get_valid(png, pnginfo, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha (png);

	if (colortype == PNG_COLOR_TYPE_GRAY || colortype == PNG_COLOR_TYPE_GRAY_ALPHA)
	{
		png_set_gray_to_rgb (png);
		png_set_filler (png, 255, PNG_FILLER_AFTER);			
	}

	if (colortype != PNG_COLOR_TYPE_RGBA)
		png_set_filler (png, 255, PNG_FILLER_AFTER);
	
	if (bitdepth < 8)
		png_set_expand (png);
	else if (bitdepth == 16)
		png_set_strip_16 (png);

	png_read_update_info (png, pnginfo);
	rowbytes = png_get_rowbytes (png, pnginfo);
	bytesperpixel = png_get_channels (png, pnginfo);
	bitdepth = png_get_bit_depth (png, pnginfo);

	// fill in properties
	// just using width and height are *not* good here...!
	// rowbytes / 4 should also be good for use here, but we'll do it the official way
	w = png_get_image_width (png, pnginfo);
	h = png_get_image_height (png, pnginfo);

	if (bitdepth != 8 || bytesperpixel != 4)
	{	
		png_destroy_read_struct (&png, &pnginfo, NULL);
		fclose (f);
		return false;
	}

	// initialize the return data
	if (LoadExternalIntoScaled (w, h))
	{
		// take it direct into scaled
		texdata = (byte *) scaled;
	}
	else
	{
		// allocate a temp buffer for it
		buffer = (unsigned *) malloc ((w * h * 4));
		texdata = (byte *) buffer;
	}

	// set up the row pointers
	rowpointers = (unsigned char **) malloc (height * 4);

	// and for each row
	for (y = 0; y < height; y++)
		rowpointers[y] = texdata + y * rowbytes;

	// read it all in
	png_read_image (png, rowpointers);
	png_read_end (png, NULL);

	png_destroy_read_struct (&png, &pnginfo, NULL);

	fclose (f);

	// upload
	flags |= TEX_32BPP;
	GL_UploadTexture ((unsigned *) texdata, w, h, flags);

	if (buffer) free (buffer);
	free (rowpointers);

	return true;
}


typedef struct TGAHeader_s
{
	unsigned char id_length;
	unsigned char colormap_type;
	unsigned char image_type;
	unsigned short colormap_index;
	unsigned short colormap_length;
	unsigned char colormap_size;
	unsigned short x_origin;
	unsigned short y_origin;
	unsigned short width;
	unsigned short height;
	unsigned char pixel_size;
	unsigned char attributes;
} TGAHeader_t;


int TGAfgetLittleShort (FILE *f)
{
	unsigned char b1 = fgetc (f);
	unsigned char b2 = fgetc (f);

	return (short) (b1 + b2 * 256);
}


qboolean GL_LoadTGA (char *uploadname, int flags)
{
	TGAHeader_t targa_header;
	int columns;
	int rows;
	int numPixels;
	unsigned char *pixbuf;
	int row, column;
	int flipper;
	FILE *f;
	byte *texdata;
	unsigned *buffer = NULL;

	COM_FOpenFile (uploadname, &f);

	if (!f) return false;

	// read the header
	targa_header.id_length = fgetc (f);
	targa_header.colormap_type = fgetc (f);
	targa_header.image_type = fgetc (f);

	targa_header.colormap_index = TGAfgetLittleShort (f);
	targa_header.colormap_length = TGAfgetLittleShort (f);
	targa_header.colormap_size = fgetc (f);
	targa_header.x_origin = TGAfgetLittleShort (f);
	targa_header.y_origin = TGAfgetLittleShort (f);
	targa_header.width = TGAfgetLittleShort (f);
	targa_header.height = TGAfgetLittleShort (f);
	targa_header.pixel_size = fgetc (f);
	targa_header.attributes = fgetc (f);

	// they call him flipper, flipper, faster than lightning, no-one you see, is smarter than he.
	flipper = !((targa_header.attributes & 0x20) >> 5);

	// validate
	if (targa_header.image_type != 2 && targa_header.image_type != 10)
	{
		fclose (f);
		return false;
	}

	if (targa_header.colormap_type != 0 || (targa_header.pixel_size != 32 && targa_header.pixel_size != 24))
	{
		fclose (f);
		return false;
	}

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	// initialize the return data
	if (LoadExternalIntoScaled (targa_header.width, targa_header.height))
	{
		// take it direct into scaled
		texdata = (byte *) scaled;
	}
	else
	{
		// allocate a temp buffer for it
		buffer = (unsigned *) malloc ((numPixels * 4));
		texdata = (byte *) buffer;
	}

	// skip TARGA image comment
	if (targa_header.id_length != 0) fseek (f, targa_header.id_length, SEEK_CUR);

	if (targa_header.image_type == 2)
	{
		// Uncompressed, RGB images
		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = texdata + row * columns * 4;
			else pixbuf = texdata + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns; column++)
			{
				unsigned char blue = fgetc (f);
				unsigned char green = fgetc (f);
				unsigned char red = fgetc (f);
				unsigned char alphabyte = 255;

				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;

				switch (targa_header.pixel_size)
				{
				case 24:
					*pixbuf++ = 255;
					break;

				case 32:
					alphabyte = fgetc (f);
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10)
	{
		// Runlength encoded RGB images
		unsigned char red;
		unsigned char green;
		unsigned char blue;
		unsigned char alphabyte = 255;
		unsigned char packetHeader;
		unsigned char packetSize;
		unsigned char j;

		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = texdata + row * columns * 4;
			else pixbuf = texdata + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns;)
			{
				packetHeader = fgetc (f);
				packetSize = 1 + (packetHeader & 0x7f);

				if (packetHeader & 0x80)
				{
					// run-length packet
					blue = fgetc (f);
					green = fgetc (f);
					red = fgetc (f);

					switch (targa_header.pixel_size)
					{
					case 24:
						alphabyte = 255;
						break;

					case 32:
						alphabyte = fgetc (f);
						break;
					}
	
					for (j = 0; j < packetSize; j++)
					{
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;

						column++;

						if (column == columns)
						{
							// run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = texdata + row * columns * 4;
							else pixbuf = texdata + ((row - 1) - row) * columns * 4;
						}
					}
				}
				else
				{
					// non run-length packet
					for (j = 0; j < packetSize; j++)
					{
						blue = fgetc (f);
						green = fgetc (f);
						red = fgetc (f);

						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;

						switch (targa_header.pixel_size)
						{
						case 24:
							*pixbuf++ = 255;
							break;

						case 32:
							alphabyte = fgetc (f);
							*pixbuf++ = alphabyte;
							break;
						}

						column++;

						if (column == columns)
						{
							// pixel packet run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = texdata + row * columns * 4;
							else pixbuf = texdata + ((row - 1) - row) * columns * 4;
						}
					}
				}
			}

			breakOut:;
		}
	}

	// upload
	flags |= TEX_32BPP;
	GL_UploadTexture ((unsigned *) texdata, targa_header.width, targa_header.height, flags);

	if (buffer) free (buffer);

	fclose (f);

	// success
	return true;
}


#ifndef BI_RGB
#define BI_RGB 0L
#endif

#ifndef BI_RLE8
#define BI_RLE8 1L
#endif

qboolean GL_LoadBMP (char *uploadname, int flags)
{
	int i, j;
	int run;
	int runlength;
	unsigned char bmIn;
	BITMAPINFO bmInfo;
	BITMAPFILEHEADER bmHeader;
	RGBQUAD bmPalette[256];
	unsigned char *bmData;
	int pad;
	FILE *f;
	byte *texdata;
	unsigned *buffer = NULL;
	LONG TheHeight;

	COM_FOpenFile (uploadname, &f);

	if (!f) return false;

	// read the file header
	fread (&bmHeader, 1, sizeof (bmHeader), f);

	// validate the type
	if (bmHeader.bfType != 0x4D42)
	{
		fclose (f);
		return false;
	}

	// read the BITMAPINFO
	fread (&bmInfo, 1, sizeof (bmInfo), f);

	if (bmInfo.bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo.bmiHeader.biCompression == BI_RLE8 && bmInfo.bmiHeader.biBitCount == 8))
		{
			fclose (f);
			return false;
		}
	}

	// support 4, 8 and 24 BPP modes
	if (bmInfo.bmiHeader.biBitCount != 4 && bmInfo.bmiHeader.biBitCount != 8 && bmInfo.bmiHeader.biBitCount != 24)
	{
		fclose (f);
		return false;
	}

	TheHeight = abs (bmInfo.bmiHeader.biHeight);

	// initialize the return data
	if (LoadExternalIntoScaled (bmInfo.bmiHeader.biWidth, TheHeight))
	{
		// take it direct into scaled
		texdata = (byte *) scaled;
	}
	else
	{
		// allocate a temp buffer for it
		buffer = (unsigned *) malloc ((bmInfo.bmiHeader.biWidth * TheHeight * 4));
		texdata = (byte *) buffer;
	}

	bmData = texdata;

	if (bmInfo.bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo.bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo.bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo.bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo.bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo.bmiHeader.biClrUsed; i++)
		{
			fread (&bmPalette[i], 1, sizeof (RGBQUAD), f);
		}
	}

	// each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
	// out how much padding we need (this may not work with RLE images)
	for (i = 0; ; i++)
	{
		pad = i * 4;
		if (pad >= (bmInfo.bmiHeader.biWidth * 3)) break;
	}

	pad -= bmInfo.bmiHeader.biWidth * 3;

	// now seek to the proper offset of the data
	fseek (f, bmHeader.bfOffBits, SEEK_SET);

	// load the data
	for (i = 0; i < bmInfo.bmiHeader.biWidth * TheHeight; i++)
	{
		if (bmInfo.bmiHeader.biBitCount == 4)
		{
			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[bmIn % 16].rgbRed;
			bmData[5] = bmPalette[bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo.bmiHeader.biBitCount == 8)
		{
			if (bmInfo.bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = fgetc (f);

				if (bmIn == 0)
				{
					bmIn = fgetc (f);
					i++;

					switch (bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = fgetc (f);
						i++;

						bmData += bmIn * 4;

						bmIn = fgetc (f);
						i++;

						bmData += bmIn * bmInfo.bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = fgetc (f);
							i++;

							bmData[0] = bmPalette[bmIn].rgbRed;
							bmData[1] = bmPalette[bmIn].rgbGreen;
							bmData[2] = bmPalette[bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = fgetc (f);
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = bmIn;

					bmIn = fgetc (f);
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[bmIn].rgbRed;
						bmData[1] = bmPalette[bmIn].rgbGreen;
						bmData[2] = bmPalette[bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn].rgbRed;
			bmData[1] = bmPalette[bmIn].rgbGreen;
			bmData[2] = bmPalette[bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else
		{
			// 24 BPP
			bmData[2] = fgetc (f);
			bmData[1] = fgetc (f);
			bmData[0] = fgetc (f);
			bmData[3] = 255;

			bmData += 4;
		}

		// read the padding
		for (j = 0; j < pad; j++) fgetc (f);
	}

bmDone:;

	// check orientation
	if (bmInfo.bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < TheHeight / 2; i++)
		{
			for (j = 0; j < bmInfo.bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo.bmiHeader.biWidth) + j;
				unsigned int dstpos = (((TheHeight - 1) - i) * bmInfo.bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) texdata)[dstpos];
				((unsigned int *) texdata)[dstpos] = ((unsigned int *) texdata)[srcpos];
				((unsigned int *) texdata)[srcpos] = tmp;
			}
		}
	}

	// upload
	flags |= TEX_32BPP;
	GL_UploadTexture ((unsigned *) texdata, bmInfo.bmiHeader.biHeight, TheHeight, flags);

	if (buffer) free (buffer);

	fclose (f);

	return true;
}


qboolean GL_LoadJPG (char *uploadname, int flags)
{
	struct jpeg_decompress_struct cinfo;
	JDIMENSION num_scanlines;
	JSAMPARRAY in;
	struct jpeg_error_mgr jerr;
	int numPixels;
	int row_stride;
	unsigned char *out;
	byte *texdata;
	unsigned *buffer = NULL;
	int count;
	int i;
	int r, g, b;
	FILE *f;

	COM_FOpenFile (uploadname, &f);

	if (!f) return false;

	// set up the decompression.
	cinfo.err = jpeg_std_error (&jerr);
	jpeg_create_decompress (&cinfo);

	// inititalize the source
	jpeg_stdio_src (&cinfo, f);

	// initialize decompression
	(void) jpeg_read_header (&cinfo, TRUE);
	(void) jpeg_start_decompress (&cinfo);

	numPixels = cinfo.image_width * cinfo.image_height;

	// initialize the input buffer - we'll use the in-built memory management routines in the
	// JPEG library because it will automatically free the used memory for us when we destroy
	// the decompression structure.  cool.
	row_stride = cinfo.output_width * cinfo.output_components;
	in = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	// bit of error checking
	if (cinfo.output_components != 3) goto error;
	if ((numPixels * 4) != ((row_stride * cinfo.output_height) + numPixels)) goto error;

	// read the jpeg
	count = 0;

	// initialize the return data
	if (LoadExternalIntoScaled (cinfo.image_width, cinfo.image_height))
	{
		// take it direct into scaled
		texdata = (byte *) scaled;
	}
	else
	{
		// allocate a temp buffer for it
		buffer = (unsigned *) malloc ((numPixels * 4));
		texdata = (byte *) buffer;
	}

	while (cinfo.output_scanline < cinfo.output_height) 
	{
		num_scanlines = jpeg_read_scanlines(&cinfo, in, 1);
		out = in[0];

		for (i = 0; i < row_stride;)
		{
			r = texdata[count++] = out[i++];
			g = texdata[count++] = out[i++];
			b = texdata[count++] = out[i++];

			// alpha is always 255 with JPEGs
			texdata[count++] = 255;
		}
	}

	// finish decompression and destroy the jpeg
	(void) jpeg_finish_decompress (&cinfo);

	// upload
	flags |= TEX_32BPP;
	GL_UploadTexture ((unsigned *) texdata, cinfo.image_width, cinfo.image_height, flags);

	if (buffer) free (buffer);

	fclose (f);

	return true;

error:;
	// have to do this rather than a simple "return false" cos we need to destroy the reader too!!!
	jpeg_destroy_decompress (&cinfo);
	fclose (f);
	return false;
}


typedef struct pcx_s
{
	char manufacturer;
	char version;
	char encoding;
	char bits_per_pixel;

	unsigned short xmin;
	unsigned short ymin;
	unsigned short xmax;
	unsigned short ymax;

	unsigned short hres;
	unsigned short vres;

	unsigned char palette[48];

	char reserved;
	char color_planes;

	unsigned short bytes_per_line;
	unsigned short palette_type;

	char filler[58];

	unsigned int data;			// unbounded
} pcx_t;


qboolean GL_LoadPCX (char *uploadname, int flags)
{
	pcx_t *pcx, pcxbuf;
	unsigned char palette[768];
	unsigned char *pix;
	int x, y;
	int dataByte, runLength;
	int count;
	FILE *f;
	byte *texdata;
	unsigned *buffer = NULL;

	COM_FOpenFile (uploadname, &f);

	if (!f) return false;

	// parse the PCX file
	fread (&pcxbuf, 1, sizeof (pcxbuf), f);

	pcx = &pcxbuf;

	if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 1024 || pcx->ymax >= 1024)
	{
		fclose (f);
		return false;
	}

	// seek to palette
	fseek (f, -768, SEEK_END);
	fread (palette, 1, 768, f);

	fseek (f, sizeof (pcxbuf) - 4, SEEK_SET);

	count = (pcx->xmax + 1) * (pcx->ymax + 1);

	// initialize the return data
	if (LoadExternalIntoScaled ((pcx->xmax + 1), (pcx->ymax + 1)))
	{
		// take it direct into scaled
		texdata = (byte *) scaled;
	}
	else
	{
		// allocate a temp buffer for it
		buffer = (unsigned *) malloc ((count * 4));
		texdata = (byte *) buffer;
	}

	for (y = 0; y <= pcx->ymax; y++)
	{
		pix = texdata + 4 * y * (pcx->xmax + 1);

		for (x = 0; x <= pcx->xmax;)
		{
			dataByte = fgetc (f);

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = fgetc (f);
			}
			else runLength = 1;

			while (runLength-- > 0)
			{
				pix[0] = palette[dataByte * 3];
				pix[1] = palette[dataByte * 3 + 1];
				pix[2] = palette[dataByte * 3 + 2];

				// PCX files never have alpha
				pix[3] = 255;
				pix += 4;
				x++;
			}
		}
	}

	// upload
	flags |= TEX_32BPP;
	GL_UploadTexture ((unsigned *) texdata, (pcx->xmax + 1), (pcx->ymax + 1), flags);

	if (buffer) free (buffer);

	fclose (f);

	return true;
}


int R_LoadMapShot (char *name)
{
	if (!mapshottexturenum)
		GL_AllocTexture (&mapshottexturenum);

	glBindTexture (GL_TEXTURE_2D, mapshottexturenum);

	// failed to load
	if (!GL_LoadTGA (name, TEX_BASIC))
		return 0;

	return mapshottexturenum;
}


typedef struct imageformats_s
{
	char *extension;
	qboolean (*Texloader) (char *, int);
} imageformats_t;

imageformats_t imageformats[] =
{
	// image loaders. jpg is twice cos the jpeg extension is also valid
	{".png", GL_LoadPNG},
	{".tga", GL_LoadTGA},
	{".bmp", GL_LoadBMP},
	{".jpg", GL_LoadJPG},
	{".jpeg", GL_LoadJPG},
	{".pcx", GL_LoadPCX},

	// always end the list with this
	{NULL, NULL},
};


qboolean LoadExternalTexture (char *basename, int flags)
{
	int i, j, x;
	char uploadname[MAX_OSPATH];

	// disable external textures on intel integrated
	if (IntelIntegrated) return false;

	for (j = 0; ; j++)
	{
		// no more search paths
		if (!SearchPaths[j]) break;

		for (i = 0; ; i++)
		{
			// no more extensions
			if (!imageformats[i].extension) break;

			// build the name
			if (flags & TEX_LUMA)
				sprintf (uploadname, "%s%s_luma%s", SearchPaths[j], basename, imageformats[i].extension);
			else sprintf (uploadname, "%s%s%s", SearchPaths[j], basename, imageformats[i].extension);

			// cleanse the name
			for (x = 0; ; x++)
			{
				if (uploadname[x] == 0) break;

				if (uploadname[x] == '*') uploadname[x] = '#';
			}

			// attempt a load, and return true if we succeed
			if (imageformats[i].Texloader (uploadname, flags)) return true;
		}
	}

	return false;
}


/*
================
GL_LoadTexture
================
*/
gltexture_t	*GL_LoadTexture (char *identifier, int width, int height, byte *data, int flags)
{
	int i;
	gltexture_t	*glt;
	gltexture_t	*Free = NULL;
	byte md5digest[16];
	byte ShotgunShells[] = {202, 6, 69, 163, 17, 112, 190, 234, 102, 56, 225, 242, 212, 175, 27, 187};

	// make an md5 of the texture data
	MD5_Checksum (data, (width * height * ((flags & TEX_32BPP) ? 4 : 1)), md5digest);

	// fix white line at base of shotgun shells box
	// one of id's artists needs a good slap here...
	// done before the cache check to prevent reloads of this texture...
	if (MD5_Compare (md5digest, ShotgunShells))
		Q_memcpy (data, data + 32 * 31, 32);

	// see if the texture is already present
	for (glt = gltextures; glt; glt = glt->next)
	{
		if (glt->texnum == 0)
		{
			// mark as a free texture
			Free = glt;

			// don;t check
			continue;
		}

		// don't do a cache check
		if (flags & TEX_NOCACHE) continue;

		if (MD5_Compare (md5digest, glt->md5digest))
		{
			// set the cache mode
			glt->cache = (flags & TEX_INCACHE) ? -1 : 0;

			// return the found texture
			return glt;
		}
	}

	// check for water now
	if (identifier[0] == '*')
		watertexture = true;
	else watertexture = false;

	if (Free)
	{
		// use this free slot for our new texture
		glt = Free;
	}
	else
	{
		// create a new texture object
		glt = (gltexture_t *) malloc (sizeof (gltexture_t));

		// link it in
		glt->next = gltextures;
		gltextures = glt;
	}

	// copy in the data
	sprintf (glt->identifier, "%s", identifier);
	GL_AllocTexture (&glt->texnum);

	glt->width = width;
	glt->height = height;

	glt->flags = flags;

	// remove the extension part from the name
	for (i = 0; ; i++)
	{
		if (glt->identifier[i] == 0) break;

		if (glt->identifier[i] == '.')
		{
			strcpy (&glt->identifier[i], &glt->identifier[i + 4]);

			break;
		}
	}

	// set the cache mode
	glt->cache = (glt->flags & TEX_INCACHE) ? -1 : 0;

	// store the md5
	MD5_Copy (md5digest, glt->md5digest);

	// if it's an mdl, we need to hack the alpha to ensure that it's correct.
	// this test will also catch sprites, but it's non-destructive for them so cool.
	if ((glt->flags & TEX_MDL) && !(glt->flags & TEX_32BPP))
	{
		// assume that the first pixel is the correct alpha colour
		byte alphapix = data[0];
		qboolean isVLight = !(strncmp (identifier, "progs/v_light", 13));

		for (i = 0; i < (width * height); i++)
		{
			// standard alpha
			if (data[i] == alphapix) data[i] = 255;

			// lightning gun hack
			if (isVLight && data[i] == 246) data[i] = 255;
		}

		// WriteTGA (identifier, width, height, data, 8);
	}

	if (hipnotic && (glt->flags & TEX_BSP) && !strncmp (identifier, "sfloor4_3", 9) && width == 64 && height == 64)
		glt->flags |= TEX_ALPHAHACK;

	// upload it
	glBindTexture (GL_TEXTURE_2D, glt->texnum);

	// only send the base data through the loader if we fail to find an external
	// if it's already 32 bit don't bother with the external check
	if (glt->flags & TEX_32BPP)
		GL_UploadTexture ((unsigned *) data, width, height, glt->flags);
	else if (!LoadExternalTexture (glt->identifier, glt->flags))
		GL_UploadTexture ((unsigned *) data, width, height, glt->flags);

	// store out the correct sizes (we don't bother doing this for lumas)
	glt->width = upload_width;
	glt->height = upload_height;

	// no luma by default
	glt->lumatexnum = 0;

	if (watertexture)
	{
		// save out the fog
		glt->colour[0] = waterfog[0];
		glt->colour[1] = waterfog[1];
		glt->colour[2] = waterfog[2];

		// reset to false
		watertexture = false;

		// liquids never have lumas, so no need to check
		return glt;
	}

	// only put lumas on 8 bit mipmapped textures
	if ((glt->flags & TEX_MIPMAP) && !(flags & TEX_32BPP))
	{
		qboolean hasLuma = false;

		// assume that we have a luma always
		GL_AllocTexture (&glt->lumatexnum);
		glBindTexture (GL_TEXTURE_2D, glt->lumatexnum);

		// always look for a luma irrespective of whether the original has one
		if (!LoadExternalTexture (glt->identifier, (glt->flags | TEX_LUMA)) && !(flags & TEX_32BPP))
		{
			// hunt around for a luma
			for (i = 0; i < (width * height); i++)
			{
				// we need to go through all the data so that we can properly set non-luma to black
				if (data[i] > 223 && data[i] != 255)
					hasLuma = true;
				else data[i] = 0;
			}

			// did we find a luma?
			if (hasLuma)
			{
				// yes - upload it now
				GL_UploadTexture ((unsigned *) data, width, height, (glt->flags | TEX_LUMA));
			}
			else
			{
				// remove the luma texture object
				glDeleteTextures (1, &glt->lumatexnum);
				glt->lumatexnum = 0;
			}
		}
	}

	if (glt->lumatexnum) glt->flags |= TEX_LUMA;

	// return the texture object we just used
	return glt;
}


void GL_DeleteTextures (void)
{
	gltexture_t *glt;

	for (glt = gltextures; glt; glt = glt->next)
	{
		if (glt->cache < 0)
		{
			// always keep these textures
			continue;
		}

		// keep unused textures for a few maps in case they're needed
		// smaller textures are kept longer than bigger ones
		if (glt->cache < 5 && glt->width <= 512 && glt->height <= 512)
		{
			// currently in use - increment cache to signify it's been checked
			glt->cache++;
			continue;
		}

		// bigger textures are deleted quicker to save on texture memory
		// these are mainly "special" textures, like muh_bah, which are one-offs
		// we still keep them for two maps just in case... but no more...
		if (glt->cache < 2)
		{
			// currently in use - increment cache to signify it's been checked
			glt->cache++;
			continue;
		}

		// this texture can now be removed
		glDeleteTextures (1, &glt->texnum);
		glDeleteTextures (1, &glt->lumatexnum);

		// set texnum to 0
		glt->texnum = 0;
		glt->lumatexnum = 0;

		// set cache to -1 so it won't be attempted to be deleted
		glt->cache = -1;

		// set identifier
		glt->identifier[0] = 0;

		// no flags
		glt->flags = 0;
	}
}


qboolean LoadExternalIntoScaled (int w, int h)
{
	int s;

	// check size
	if (w > 1024) return false;
	if (h > 1024) return false;

	// check power of 2 in width
	for (s = 1; s < w; s *= 2);

	// width not a power of 2
	if (s != w) return false;

	// check power of 2 in height
	for (s = 1; s < h; s *= 2);

	// height not a power of 2
	if (s != h) return false;

	// ok to load into scaled
	return true;
}

/*
================
GL_LoadBitmapResourceTexture

Loads a BMP resource texture

only bitmaps can be loaded from resources, cos versions of Windows prior to XP don't support other
image formats in resources.  (Should we really worry about this anymore?)
================
*/
void GL_LoadBitmapResourceTexture (gltexture_t *glt, int ResourceID, int flags)
{
	int i, j;
	int run;
	int runlength;
	unsigned char *bmIn;
	BITMAPINFO *bmInfo;
	RGBQUAD bmPalette[256];
	RGBQUAD *PaletteIn;
	unsigned char *buffer = NULL;
	unsigned char *bmData;
	unsigned char *bmRaw;
	int pad;
	HINSTANCE hInstance;
	HANDLE hBitmap;

	// get a handle to the MHQuake executable
	hInstance = GetModuleHandle (NULL);

	// load the resource
	hBitmap = LoadResource (hInstance, FindResource (hInstance, MAKEINTRESOURCE (ResourceID), RT_BITMAP));

	// lock it
	bmRaw = (byte *) LockResource (hBitmap);

	// get the bm info header
	// read the header
	bmInfo = (BITMAPINFO *) bmRaw;

	// skip past the header (for 24 BPP RGB images the pixel data immediately follows, otherwise a palette is next)
	bmRaw += sizeof (BITMAPINFO);

	if (bmInfo->bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo->bmiHeader.biCompression == BI_RLE8 && bmInfo->bmiHeader.biBitCount == 8))
			goto BadBMP;
	}

	// support 4, 8 and 24 BPP modes
	if (bmInfo->bmiHeader.biBitCount != 4 && 
		bmInfo->bmiHeader.biBitCount != 8 && 
		bmInfo->bmiHeader.biBitCount != 24) goto BadBMP;

	// if it's a power of 2 and less than 1024 * 1024 we can load it directly into scaled, otherwise we must allocate a buffer
	if (LoadExternalIntoScaled (bmInfo->bmiHeader.biWidth, abs (bmInfo->bmiHeader.biHeight)))
	{
		// this is safe as the texture loader never uses scaled for 32 bit textures where these conditions are met
		bmData = (byte *) scaled;
	}
	else
	{
		buffer = (byte *) malloc (bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight) * 4);
		bmData = buffer;
	}

	if (bmInfo->bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo->bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo->bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo->bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo->bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo->bmiHeader.biClrUsed; i++)
		{
			PaletteIn = (RGBQUAD *) bmRaw;

			bmPalette[i].rgbBlue = PaletteIn->rgbBlue;
			bmPalette[i].rgbGreen = PaletteIn->rgbGreen;
			bmPalette[i].rgbRed = PaletteIn->rgbRed;
			bmPalette[i].rgbReserved = PaletteIn->rgbReserved;

			bmRaw += sizeof (RGBQUAD);
		}
	}

	// each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
	// out how much padding we need (this may not work with RLE images)
	for (i = 0; ; i++)
	{
		pad = i * 4;
		if (pad >= (bmInfo->bmiHeader.biWidth * 3)) break;
	}

	pad -= bmInfo->bmiHeader.biWidth * 3;

	// load the data
	for (i = 0; i < bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight); i++)
	{
		if (bmInfo->bmiHeader.biBitCount == 4)
		{
			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[*bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[*bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[*bmIn % 16].rgbRed;
			bmData[5] = bmPalette[*bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[*bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo->bmiHeader.biBitCount == 8)
		{
			if (bmInfo->bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = bmRaw++;

				if (*bmIn == 0)
				{
					bmIn = bmRaw++;
					i++;

					switch (*bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * 4;

						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * bmInfo->bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = *bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = bmRaw++;
							i++;

							bmData[0] = bmPalette[*bmIn].rgbRed;
							bmData[1] = bmPalette[*bmIn].rgbGreen;
							bmData[2] = bmPalette[*bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = bmRaw++;
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = *bmIn;

					bmIn = bmRaw++;
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[*bmIn].rgbRed;
						bmData[1] = bmPalette[*bmIn].rgbGreen;
						bmData[2] = bmPalette[*bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn].rgbRed;
			bmData[1] = bmPalette[*bmIn].rgbGreen;
			bmData[2] = bmPalette[*bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else
		{
			// 24 BPP
			bmData[2] = *bmRaw++;
			bmData[1] = *bmRaw++;
			bmData[0] = *bmRaw++;
			bmData[3] = 255;

			bmData += 4;
		}

		// read the padding
		for (j = 0; j < pad; j++) bmRaw++;
	}

bmDone:;

	// revert the bmData pointer
	if (buffer)
		bmData = buffer;
	else bmData = (unsigned char *) scaled;

	// check orientation
	if (bmInfo->bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < abs (bmInfo->bmiHeader.biHeight) / 2; i++)
		{
			for (j = 0; j < bmInfo->bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo->bmiHeader.biWidth) + j;
				unsigned int dstpos = (((abs (bmInfo->bmiHeader.biHeight) - 1) - i) * bmInfo->bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) bmData)[dstpos];
				((unsigned int *) bmData)[dstpos] = ((unsigned int *) bmData)[srcpos];
				((unsigned int *) bmData)[srcpos] = tmp;
			}
		}
	}

	// fill in texture properties
	sprintf (glt->identifier, "****Resource_%i", ResourceID);
	glt->cache = -1;
	GL_AllocTexture (&glt->texnum);
	glt->height = abs (bmInfo->bmiHeader.biHeight);
	glt->width = bmInfo->bmiHeader.biWidth;
	glt->next = NULL;
	glt->lumatexnum = 0;

	flags |= TEX_32BPP;

	// now upload it
	glBindTexture (GL_TEXTURE_2D, glt->texnum);
	GL_UploadTexture ((unsigned *) bmData, glt->width, glt->height, flags);

	// release buffer
	if (buffer) free (buffer);

	return;

BadBMP:;
	// we can't test a return condition so we just crap out instead.  as these textures are part of the application, this
	// is fair to do, as it indicates a corrupted application.
	Sys_Error ("Failed to Load Bitmap Resource %i", ResourceID);

	return;
}


extern texture_t *texturelist;

void R_UpdateWarpTextures (void)
{
	texture_t *tex;
	qboolean ClearState = false;
	float x, y, x2, warptess;

	// range 3 to 64
	float r_waterquality = 8.0f;

	if (cl.paused) return;

	warptess = 128.0 / r_waterquality;

	for (tex = texturelist; tex; tex = tex->texturelist)
	{
		// not liquid
		if (!(tex->name[0] == '*' || tex->name[0] == '#')) continue;

		if (!ClearState)
		{
			ClearState = true;
		}

		glBindTexture (GL_TEXTURE_2D, tex->gltexture->texnum);

		for (x = 0.0; x < 128.0; x = x2)
		{
			x2 = x + warptess;

//			glBegin (GL_TRIANGLE_STRIP);

			for (y = 0.0; y < 128.01; y += warptess) // .01 for rounding errors
			{
//				glTexCoord2f (WARPCALC (x,y), WARPCALC (y,x));
//				glVertex2f (x, y);
//				glTexCoord2f (WARPCALC (x2,y), WARPCALC (y,x2));
//				glVertex2f (x2, y);
			}

//			glEnd();
		}

		//copy to texture
//		GL_Bind (tx->warpimage);
//		glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, glx, gly+glheight-gl_warpimagesize, gl_warpimagesize, gl_warpimagesize);
	}

	if (ClearState)
	{
	}
}

