/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include "quakedef.h"
#include "winquake.h"
#include <commctrl.h>

#define WARP_WIDTH		320
#define WARP_HEIGHT		200
#define BASEWIDTH		320
#define BASEHEIGHT		200

#define MODE_WINDOWED			0
#define NO_MODE					(MODE_WINDOWED - 1)
#define MODE_FULLSCREEN_DEFAULT	(MODE_WINDOWED + 1)

typedef struct gl_vmode_s
{
	modestate_t	type;
	int			width;
	int			height;
	int			modenum;
	int			dib;
	int			fullscreen;
	int			bpp;
	int			halfscreen;
	char		modedesc[17];

	struct gl_vmode_s *next;
} gl_vmode_t;


const char *gl_vendor;
const char *gl_renderer;
const char *gl_version;
const char *gl_extensions;

qboolean		DDActive;
qboolean		scr_skipupdate;

static gl_vmode_t *gl_modelist = NULL;
static gl_vmode_t *gl_windowed_mode = NULL;
static gl_vmode_t *gl_current_mode = NULL;


static int		nummodes;

static DEVMODE	gdevmode;
static qboolean	vid_initialized = false;
static qboolean	windowed, leavecurrentmode;
static qboolean vid_canalttab = false;
static qboolean vid_wassuspended = false;
static int		windowed_mouse;
extern qboolean	mouseactive;  // from in_win.c
static HICON	hIcon;

int			DIBWidth, DIBHeight;
RECT		WindowRect;
DWORD		WindowStyle, ExWindowStyle;

HWND	mainwindow, dibwindow;

int			vid_modenum = NO_MODE;
int			vid_realmode;
int			vid_default = MODE_WINDOWED;
static int	windowed_default;
unsigned char	vid_curpal[256*3];
static qboolean fullsbardraw = false;

static float vid_gamma = 1.0;

HGLRC	baseRC;
HDC		maindc;

HWND WINAPI InitializeWindow (HINSTANCE hInstance, int nCmdShow);

viddef_t	vid;				// global video state

unsigned	d_8to24table[256];

modestate_t	modestate = MS_UNINIT;

void VID_MenuDraw (void);
void VID_MenuKey (int key);

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void AppActivate(BOOL fActive, BOOL minimize);
void ClearAllStates (void);
void VID_UpdateWindowStatus (void);
void GL_Init (void);

PROC glArrayElementEXT;
PROC glColorPointerEXT;
PROC glTexCoordPointerEXT;
PROC glVertexPointerEXT;

typedef void (APIENTRY *lp3DFXFUNC) (int, int, int, int, int, const void*);
lp3DFXFUNC glColorTableEXT;
qboolean is8bit = false;
qboolean isPermedia = false;

//====================================

cvar_t		vid_mode = {"vid_mode","0", false};

// Note that 0 is MODE_WINDOWED
cvar_t		_vid_default_mode = {"_vid_default_mode","0", true};

// Note that 3 is MODE_FULLSCREEN_DEFAULT
cvar_t		_vid_default_mode_win = {"_vid_default_mode_win","3", true};

cvar_t		vid_wait = {"vid_wait","0"};
cvar_t		vid_nopageflip = {"vid_nopageflip","0", true};
cvar_t		_vid_wait_override = {"_vid_wait_override", "0", true};
cvar_t		vid_config_x = {"vid_config_x","800", true};
cvar_t		vid_config_y = {"vid_config_y","600", true};
cvar_t		vid_stretch_by_2 = {"vid_stretch_by_2","1", true};
cvar_t		_windowed_mouse = {"_windowed_mouse","1", true};

int			window_center_x, window_center_y, window_x, window_y, window_width, window_height;
RECT		window_rect;


void CenterWindow (HWND hWndCenter, int width, int height, BOOL lefttopjustify)
{
    int     CenterX, CenterY;

	CenterX = (GetSystemMetrics(SM_CXSCREEN) - width) / 2;
	CenterY = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;
	if (CenterX > CenterY*2)
		CenterX >>= 1;	// dual screens
	CenterX = (CenterX < 0) ? 0: CenterX;
	CenterY = (CenterY < 0) ? 0: CenterY;
	SetWindowPos (hWndCenter, NULL, CenterX, CenterY, 0, 0,
			SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

qboolean VID_SetWindowedMode (gl_vmode_t *mode)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	lastmodestate = modestate;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = mode->width;
	WindowRect.bottom = mode->height;

	DIBWidth = mode->width;
	DIBHeight = mode->height;

	WindowStyle = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx (&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx
	(
		ExWindowStyle,
		"WinQuake",
		"GLQuake",
		WindowStyle,
		rect.left, rect.top,
		width,
		height,
		NULL,
		NULL,
		global_hInstance,
		NULL
	);

	if (!dibwindow) Sys_Error ("Couldn't create DIB window");

	// Center and show the DIB window
	CenterWindow(dibwindow, WindowRect.right - WindowRect.left, WindowRect.bottom - WindowRect.top, false);

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	modestate = MS_WINDOWED;

	// because we have set the background brush for the window to NULL
	// (to avoid flickering when re-sizing the window on the desktop),
	// we clear the window to black when created, otherwise it will be
	// empty while Quake starts up.
	hdc = GetDC (dibwindow);
	PatBlt (hdc, 0, 0, WindowRect.right, WindowRect.bottom, BLACKNESS);
	ReleaseDC (dibwindow, hdc);

	if (vid.conheight > mode->height) vid.conheight = mode->height;
	if (vid.conwidth > mode->width) vid.conwidth = mode->width;

	vid.width = mode->width;
	vid.height = mode->height;

	vid.numpages = 2;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM) TRUE, (LPARAM) hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM) FALSE, (LPARAM) hIcon);

	return true;
}


qboolean VID_SetFullDIBMode (gl_vmode_t *mode)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	// switch resolution if required
	if (!leavecurrentmode)
	{
		gdevmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		gdevmode.dmBitsPerPel = mode->bpp;
		gdevmode.dmPelsWidth = mode->width << mode->halfscreen;
		gdevmode.dmPelsHeight = mode->height;
		gdevmode.dmSize = sizeof (gdevmode);

		if (ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) Sys_Error ("Couldn't set fullscreen DIB mode");
	}

	lastmodestate = modestate;
	modestate = MS_FULLDIB;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = mode->width;
	WindowRect.bottom = mode->height;

	DIBWidth = mode->width;
	DIBHeight = mode->height;

	WindowStyle = WS_POPUP;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx (&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx
	(
		ExWindowStyle,
		"WinQuake",
		"GLQuake",
		WindowStyle,
		rect.left, rect.top,
		width,
		height,
		NULL,
		NULL,
		global_hInstance,
		NULL
	);

	if (!dibwindow) Sys_Error ("Couldn't create DIB window");

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	// Because we have set the background brush for the window to NULL
	// (to avoid flickering when re-sizing the window on the desktop), we
	// clear the window to black when created, otherwise it will be
	// empty while Quake starts up.
	hdc = GetDC (dibwindow);
	PatBlt (hdc, 0, 0, WindowRect.right, WindowRect.bottom, BLACKNESS);
	ReleaseDC (dibwindow, hdc);

	if (vid.conheight > mode->height) vid.conheight = mode->height;
	if (vid.conwidth > mode->width) vid.conwidth = mode->width;

	vid.width = mode->width;
	vid.height = mode->height;

	vid.numpages = 2;

	// needed because we're not getting WM_MOVE messages fullscreen on NT
	window_x = 0;
	window_y = 0;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM) TRUE, (LPARAM) hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM) FALSE, (LPARAM) hIcon);

	return true;
}


int VID_SetMode (gl_vmode_t *mode, unsigned char *palette)
{
	int				original_mode, temp;
	qboolean		stat = false;
    MSG				msg;

	// so Con_Printfs don't mess us up by forcing vid and snd updates
	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;

	CDAudio_Pause ();

	if (vid_modenum == NO_MODE)
		original_mode = windowed_default;
	else
		original_mode = vid_modenum;

	// Set either the fullscreen or windowed mode
	if (mode->type == MS_WINDOWED)
	{
		if (_windowed_mouse.value && key_dest == key_game)
		{
			stat = VID_SetWindowedMode (mode);
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
		else
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			stat = VID_SetWindowedMode (mode);
		}
	}
	else if (mode->type == MS_FULLDIB)
	{
		stat = VID_SetFullDIBMode (mode);
		IN_ActivateMouse ();
		IN_HideMouse ();
	}
	else Sys_Error ("VID_SetMode: Bad mode type in gl_modelist");

	window_width = DIBWidth;
	window_height = DIBHeight;
	VID_UpdateWindowStatus ();

	CDAudio_Resume ();
	scr_disabled_for_loading = temp;

	if (!stat) Sys_Error ("Couldn't set video mode");

	// now we try to make sure we get the focus on the mode switch, because
	// sometimes in some systems we don't.  We grab the foreground, then
	// finish setting up, pump all our messages, and sleep for a little while
	// to let messages finish bouncing around the system, then we put
	// ourselves at the top of the z order, then grab the foreground again,
	// Who knows if it helps, but it probably doesn't hurt
	SetForegroundWindow (mainwindow);
	VID_SetPalette (palette);
	//vid_modenum = modenum;
	Cvar_SetValue ("vid_mode", (float) vid_modenum);

	while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
	{
      	TranslateMessage (&msg);
      	DispatchMessage (&msg);
	}

	Sleep (100);

	SetWindowPos (mainwindow, HWND_TOP, 0, 0, 0, 0, SWP_DRAWFRAME | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW | SWP_NOCOPYBITS);

	SetForegroundWindow (mainwindow);

	// fix the leftover Alt from any Alt-Tab or the like that switched us away
	ClearAllStates ();

	if (!msg_suppress_1) Con_SafePrintf ("Video mode %s initialized.\n", mode->modedesc);

	VID_SetPalette (palette);

	vid.recalc_refdef = 1;

	return true;
}


/*
================
VID_UpdateWindowStatus
================
*/
void VID_UpdateWindowStatus (void)
{

	window_rect.left = window_x;
	window_rect.top = window_y;
	window_rect.right = window_x + window_width;
	window_rect.bottom = window_y + window_height;
	window_center_x = (window_rect.left + window_rect.right) / 2;
	window_center_y = (window_rect.top + window_rect.bottom) / 2;

	IN_UpdateClipCursor ();
}


//====================================

// wglGetProcAddress seems implemented in some weird fashion guaranteed to generate shitloads of warnings.
// it works OK, just that the compiler is paranoid about the code at W4.
#pragma warning (disable: 4113)

// common vendor codes
char *VendorCode[] = {"ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX", NULL};

/*
================
MHGLGetProcAddress

Implements a work-around for earlier OpenGL versions, by checking for vendor-specific entrypoints that
support the required functions.
================
*/
PROC MHGLGetProcAddress (char *funcname, qboolean crash, qboolean *ver)
{
	int i;
	static PROC thePROC;

	// see is the function supported by the core implementation
	if (thePROC = wglGetProcAddress (funcname))
	{
		*ver = (*ver && true);
		return thePROC;
	}

	for (i = 0; ; i++)
	{
		if (!VendorCode[i]) break;

		thePROC = wglGetProcAddress (va ("%s%s", funcname, VendorCode[i]));

		if (thePROC)
		{
			Con_Printf ("%s Unavailable, reverting to %s%s\n", funcname, funcname, VendorCode[i]);
			*ver = (*ver && true);
			return thePROC;
		}
	}

	// oh dear.
	if (crash) Sys_Error ("Core OpenGL function %s Unavailable!!!", funcname);

	*ver = (*ver && false);
	return NULL;
}


// internal capabilities
int gl_num_tmus;
qboolean OpenGL12;
qboolean OpenGL12Img;
qboolean OpenGL13;
qboolean OpenGL14;

void GL_GetFunctionPointers (void)
{
	// initially all true
	OpenGL12 = true;
	OpenGL12Img = true;
	OpenGL13 = true;
	OpenGL14 = true;

	// opengl 1.2 and imaging subset prototypes
	glBlendColor = MHGLGetProcAddress ("glBlendColor", true, &OpenGL12);
	glBlendEquation = MHGLGetProcAddress ("glBlendEquation", true, &OpenGL12);
	glDrawRangeElements = MHGLGetProcAddress ("glDrawRangeElements", true, &OpenGL12);
	glColorTable = MHGLGetProcAddress ("glColorTable", false, &OpenGL12Img);
	glColorTableParameterfv = MHGLGetProcAddress ("glColorTableParameterfv", false, &OpenGL12Img);
	glColorTableParameteriv = MHGLGetProcAddress ("glColorTableParameteriv", false, &OpenGL12Img);
	glCopyColorTable = MHGLGetProcAddress ("glCopyColorTable", false, &OpenGL12Img);
	glGetColorTable = MHGLGetProcAddress ("glGetColorTable", false, &OpenGL12Img);
	glGetColorTableParameterfv = MHGLGetProcAddress ("glGetColorTableParameterfv", false, &OpenGL12Img);
	glGetColorTableParameteriv = MHGLGetProcAddress ("glGetColorTableParameteriv", false, &OpenGL12Img);
	glColorSubTable = MHGLGetProcAddress ("glColorSubTable", false, &OpenGL12Img);
	glCopyColorSubTable = MHGLGetProcAddress ("glCopyColorSubTable", false, &OpenGL12Img);
	glConvolutionFilter1D = MHGLGetProcAddress ("glConvolutionFilter1D", false, &OpenGL12Img);
	glConvolutionFilter2D = MHGLGetProcAddress ("glConvolutionFilter2D", false, &OpenGL12Img);
	glConvolutionParameterf = MHGLGetProcAddress ("glConvolutionParameterf", false, &OpenGL12Img);
	glConvolutionParameterfv = MHGLGetProcAddress ("glConvolutionParameterfv", false, &OpenGL12Img);
	glConvolutionParameteri = MHGLGetProcAddress ("glConvolutionParameteri", false, &OpenGL12Img);
	glConvolutionParameteriv = MHGLGetProcAddress ("glConvolutionParameteriv", false, &OpenGL12Img);
	glCopyConvolutionFilter1D = MHGLGetProcAddress ("glCopyConvolutionFilter1D", false, &OpenGL12Img);
	glCopyConvolutionFilter2D = MHGLGetProcAddress ("glCopyConvolutionFilter2D", false, &OpenGL12Img);
	glGetConvolutionFilter = MHGLGetProcAddress ("glGetConvolutionFilter", false, &OpenGL12Img);
	glGetConvolutionParameterfv = MHGLGetProcAddress ("glGetConvolutionParameterfv", false, &OpenGL12Img);
	glGetConvolutionParameteriv = MHGLGetProcAddress ("glGetConvolutionParameteriv", false, &OpenGL12Img);
	glGetSeparableFilter = MHGLGetProcAddress ("glGetSeparableFilter", false, &OpenGL12Img);
	glSeparableFilter2D = MHGLGetProcAddress ("glSeparableFilter2D", false, &OpenGL12Img);
	glGetHistogram = MHGLGetProcAddress ("glGetHistogram", false, &OpenGL12Img);
	glGetHistogramParameterfv = MHGLGetProcAddress ("glGetHistogramParameterfv", false, &OpenGL12Img);
	glGetHistogramParameteriv = MHGLGetProcAddress ("glGetHistogramParameteriv", false, &OpenGL12Img);
	glGetMinmax = MHGLGetProcAddress ("glGetMinmax", false, &OpenGL12Img);
	glGetMinmaxParameterfv = MHGLGetProcAddress ("glGetMinmaxParameterfv", false, &OpenGL12Img);
	glGetMinmaxParameteriv = MHGLGetProcAddress ("glGetMinmaxParameteriv", false, &OpenGL12Img);
	glHistogram = MHGLGetProcAddress ("glHistogram", false, &OpenGL12Img);
	glMinmax = MHGLGetProcAddress ("glMinmax", false, &OpenGL12Img);
	glResetHistogram = MHGLGetProcAddress ("glResetHistogram", false, &OpenGL12Img);
	glResetMinmax = MHGLGetProcAddress ("glResetMinmax", false, &OpenGL12Img);
	glTexImage3D = MHGLGetProcAddress ("glTexImage3D", true, &OpenGL12);
	glTexSubImage3D = MHGLGetProcAddress ("glTexSubImage3D", true, &OpenGL12);
	glCopyTexSubImage3D = MHGLGetProcAddress ("glCopyTexSubImage3D", true, &OpenGL12);

	// OpenGL 1.3 Prototypes
	glActiveTexture = MHGLGetProcAddress ("glActiveTexture", true, &OpenGL13);
	glClientActiveTexture = MHGLGetProcAddress ("glClientActiveTexture", true, &OpenGL13);
	glMultiTexCoord1d = MHGLGetProcAddress ("glMultiTexCoord1d", true, &OpenGL13);
	glMultiTexCoord1dv = MHGLGetProcAddress ("glMultiTexCoord1dv", true, &OpenGL13);
	glMultiTexCoord1f = MHGLGetProcAddress ("glMultiTexCoord1f", true, &OpenGL13);
	glMultiTexCoord1fv = MHGLGetProcAddress ("glMultiTexCoord1fv", true, &OpenGL13);
	glMultiTexCoord1i = MHGLGetProcAddress ("glMultiTexCoord1i", true, &OpenGL13);
	glMultiTexCoord1iv = MHGLGetProcAddress ("glMultiTexCoord1iv", true, &OpenGL13);
	glMultiTexCoord1s = MHGLGetProcAddress ("glMultiTexCoord1s", true, &OpenGL13);
	glMultiTexCoord1sv = MHGLGetProcAddress ("glMultiTexCoord1sv", true, &OpenGL13);
	glMultiTexCoord2d = MHGLGetProcAddress ("glMultiTexCoord2d", true, &OpenGL13);
	glMultiTexCoord2dv = MHGLGetProcAddress ("glMultiTexCoord2dv", true, &OpenGL13);
	glMultiTexCoord2f = MHGLGetProcAddress ("glMultiTexCoord2f", true, &OpenGL13);
	glMultiTexCoord2fv = MHGLGetProcAddress ("glMultiTexCoord2fv", true, &OpenGL13);
	glMultiTexCoord2i = MHGLGetProcAddress ("glMultiTexCoord2i", true, &OpenGL13);
	glMultiTexCoord2iv = MHGLGetProcAddress ("glMultiTexCoord2iv", true, &OpenGL13);
	glMultiTexCoord2s = MHGLGetProcAddress ("glMultiTexCoord2s", true, &OpenGL13);
	glMultiTexCoord2sv = MHGLGetProcAddress ("glMultiTexCoord2sv", true, &OpenGL13);
	glMultiTexCoord3d = MHGLGetProcAddress ("glMultiTexCoord3d", true, &OpenGL13);
	glMultiTexCoord3dv = MHGLGetProcAddress ("glMultiTexCoord3dv", true, &OpenGL13);
	glMultiTexCoord3f = MHGLGetProcAddress ("glMultiTexCoord3f", true, &OpenGL13);
	glMultiTexCoord3fv = MHGLGetProcAddress ("glMultiTexCoord3fv", true, &OpenGL13);
	glMultiTexCoord3i = MHGLGetProcAddress ("glMultiTexCoord3i", true, &OpenGL13);
	glMultiTexCoord3iv = MHGLGetProcAddress ("glMultiTexCoord3iv", true, &OpenGL13);
	glMultiTexCoord3s = MHGLGetProcAddress ("glMultiTexCoord3s", true, &OpenGL13);
	glMultiTexCoord3sv = MHGLGetProcAddress ("glMultiTexCoord3sv", true, &OpenGL13);
	glMultiTexCoord4d = MHGLGetProcAddress ("glMultiTexCoord4d", true, &OpenGL13);
	glMultiTexCoord4dv = MHGLGetProcAddress ("glMultiTexCoord4dv", true, &OpenGL13);
	glMultiTexCoord4f = MHGLGetProcAddress ("glMultiTexCoord4f", true, &OpenGL13);
	glMultiTexCoord4fv = MHGLGetProcAddress ("glMultiTexCoord4fv", true, &OpenGL13);
	glMultiTexCoord4i = MHGLGetProcAddress ("glMultiTexCoord4i", true, &OpenGL13);
	glMultiTexCoord4iv = MHGLGetProcAddress ("glMultiTexCoord4iv", true, &OpenGL13);
	glMultiTexCoord4s = MHGLGetProcAddress ("glMultiTexCoord4s", true, &OpenGL13);
	glMultiTexCoord4sv = MHGLGetProcAddress ("glMultiTexCoord4sv", true, &OpenGL13);
	glLoadTransposeMatrixf = MHGLGetProcAddress ("glLoadTransposeMatrixf", true, &OpenGL13);
	glLoadTransposeMatrixd = MHGLGetProcAddress ("glLoadTransposeMatrixd", true, &OpenGL13);
	glMultTransposeMatrixf = MHGLGetProcAddress ("glMultTransposeMatrixf", true, &OpenGL13);
	glMultTransposeMatrixd = MHGLGetProcAddress ("glMultTransposeMatrixd", true, &OpenGL13);
	glSampleCoverage = MHGLGetProcAddress ("glSampleCoverage", true, &OpenGL13);
	glCompressedTexImage3D = MHGLGetProcAddress ("glCompressedTexImage3D", true, &OpenGL13);
	glCompressedTexImage2D = MHGLGetProcAddress ("glCompressedTexImage2D", true, &OpenGL13);
	glCompressedTexImage1D = MHGLGetProcAddress ("glCompressedTexImage1D", true, &OpenGL13);
	glCompressedTexSubImage3D = MHGLGetProcAddress ("glCompressedTexSubImage3D", true, &OpenGL13);
	glCompressedTexSubImage2D = MHGLGetProcAddress ("glCompressedTexSubImage2D", true, &OpenGL13);
	glCompressedTexSubImage1D = MHGLGetProcAddress ("glCompressedTexSubImage1D", true, &OpenGL13);
	glGetCompressedTexImage = MHGLGetProcAddress ("glGetCompressedTexImage", true, &OpenGL13);

	// OpenGL 1.4 Prototypes
	glBlendFuncSeparate = MHGLGetProcAddress ("glBlendFuncSeparate", true, &OpenGL14);
	glFogCoordf = MHGLGetProcAddress ("glFogCoordf", true, &OpenGL14);
	glFogCoordfv = MHGLGetProcAddress ("glFogCoordfv", true, &OpenGL14);
	glFogCoordd = MHGLGetProcAddress ("glFogCoordd", true, &OpenGL14);
	glFogCoorddv = MHGLGetProcAddress ("glFogCoorddv", true, &OpenGL14);
	glFogCoordPointer = MHGLGetProcAddress ("glFogCoordPointer", true, &OpenGL14);
	glMultiDrawArrays = MHGLGetProcAddress ("glMultiDrawArrays", true, &OpenGL14);
	glMultiDrawElements = MHGLGetProcAddress ("glMultiDrawElements", true, &OpenGL14);
	// glPointParameteri and glPointParameteriv are also valid for 1.4, but were not implemented by ATI last time I looked.
	glPointParameterf = MHGLGetProcAddress ("glPointParameterf", true, &OpenGL14);
	glPointParameterfv = MHGLGetProcAddress ("glPointParameterfv", true, &OpenGL14);
	glSecondaryColor3b = MHGLGetProcAddress ("glSecondaryColor3b", true, &OpenGL14);
	glSecondaryColor3bv = MHGLGetProcAddress ("glSecondaryColor3bv", true, &OpenGL14);
	glSecondaryColor3d = MHGLGetProcAddress ("glSecondaryColor3d", true, &OpenGL14);
	glSecondaryColor3dv = MHGLGetProcAddress ("glSecondaryColor3dv", true, &OpenGL14);
	glSecondaryColor3f = MHGLGetProcAddress ("glSecondaryColor3f", true, &OpenGL14);
	glSecondaryColor3fv = MHGLGetProcAddress ("glSecondaryColor3fv", true, &OpenGL14);
	glSecondaryColor3i = MHGLGetProcAddress ("glSecondaryColor3i", true, &OpenGL14);
	glSecondaryColor3iv = MHGLGetProcAddress ("glSecondaryColor3iv", true, &OpenGL14);
	glSecondaryColor3s = MHGLGetProcAddress ("glSecondaryColor3s", true, &OpenGL14);
	glSecondaryColor3sv = MHGLGetProcAddress ("glSecondaryColor3sv", true, &OpenGL14);
	glSecondaryColor3ub = MHGLGetProcAddress ("glSecondaryColor3ub", true, &OpenGL14);
	glSecondaryColor3ubv = MHGLGetProcAddress ("glSecondaryColor3ubv", true, &OpenGL14);
	glSecondaryColor3ui = MHGLGetProcAddress ("glSecondaryColor3ui", true, &OpenGL14);
	glSecondaryColor3uiv = MHGLGetProcAddress ("glSecondaryColor3uiv", true, &OpenGL14);
	glSecondaryColor3us = MHGLGetProcAddress ("glSecondaryColor3us", true, &OpenGL14);
	glSecondaryColor3usv = MHGLGetProcAddress ("glSecondaryColor3usv", true, &OpenGL14);
	glSecondaryColorPointer = MHGLGetProcAddress ("glSecondaryColorPointer", true, &OpenGL14);
	glWindowPos2d = MHGLGetProcAddress ("glWindowPos2d", true, &OpenGL14);
	glWindowPos2dv = MHGLGetProcAddress ("glWindowPos2dv", true, &OpenGL14);
	glWindowPos2f = MHGLGetProcAddress ("glWindowPos2f", true, &OpenGL14);
	glWindowPos2fv = MHGLGetProcAddress ("glWindowPos2fv", true, &OpenGL14);
	glWindowPos2i = MHGLGetProcAddress ("glWindowPos2i", true, &OpenGL14);
	glWindowPos2iv = MHGLGetProcAddress ("glWindowPos2iv", true, &OpenGL14);
	glWindowPos2s = MHGLGetProcAddress ("glWindowPos2s", true, &OpenGL14);
	glWindowPos2sv = MHGLGetProcAddress ("glWindowPos2sv", true, &OpenGL14);
	glWindowPos3d = MHGLGetProcAddress ("glWindowPos3d", true, &OpenGL14);
	glWindowPos3dv = MHGLGetProcAddress ("glWindowPos3dv", true, &OpenGL14);
	glWindowPos3f = MHGLGetProcAddress ("glWindowPos3f", true, &OpenGL14);
	glWindowPos3fv = MHGLGetProcAddress ("glWindowPos3fv", true, &OpenGL14);
	glWindowPos3i = MHGLGetProcAddress ("glWindowPos3i", true, &OpenGL14);
	glWindowPos3iv = MHGLGetProcAddress ("glWindowPos3iv", true, &OpenGL14);
	glWindowPos3s = MHGLGetProcAddress ("glWindowPos3s", true, &OpenGL14);
	glWindowPos3sv = MHGLGetProcAddress ("glWindowPos3sv", true, &OpenGL14);

	if (OpenGL12) Con_Printf ("OpenGL 1.2 OK\n"); else Con_Printf ("OpenGL 1.2 Not fully supported\n");
	/* Don't care... */  // if (OpenGL12Img) Con_Printf ("OpenGL Imaging OK\n"); else Con_Printf ("OpenGL Imaging Not fully supported\n");
	if (OpenGL13) Con_Printf ("OpenGL 1.3 OK\n"); else Con_Printf ("OpenGL 1.3 Not fully supported\n");
	if (OpenGL14) Con_Printf ("OpenGL 1.4 OK\n"); else Con_Printf ("OpenGL 1.4 Not fully supported\n");
}
#pragma warning (default: 4113)


int		texture_mode = GL_LINEAR;


/*
===============
GL_Init
===============
*/
qboolean IntelIntegrated;
qboolean IntelPicmip;

void GL_Init (void)
{
	// check for intel integrated graphics.  because these share system ram, we can't use high res graphics
	// at all, otherwise memory trashing gets pretty severe.
	IntelIntegrated = false;
	IntelPicmip = false;

	gl_vendor = glGetString (GL_VENDOR);
	Con_Printf ("GL_VENDOR: %s\n", gl_vendor);

	if (!strnicmp (gl_vendor, "intel", 5))
	{
		if (COM_CheckParm ("-force32"))
		{
			// force support for 32 bit textures...
			IntelPicmip = true;
			Con_Printf (" (Forcing 32 bit texture support for integrated graphics)\n");
		}
		else
		{
			// it's a safe bet that if it's intel then it's integrated...
			IntelIntegrated = true;
			Con_Printf (" (Downgrading texture support for integrated graphics)\n");
		}
	}

	gl_renderer = glGetString (GL_RENDERER);
	Con_Printf ("GL_RENDERER: %s\n", gl_renderer);

	gl_version = glGetString (GL_VERSION);
	Con_Printf ("GL_VERSION: %s\n", gl_version);

	// we get these, but we're not interested in reporting them back.  extensions are over-rated anyway...
	gl_extensions = glGetString (GL_EXTENSIONS);

	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_max_size);

	// if the texture size is above 256, it's most likely something totally impractical like 4096.  textures that are larger than the
	// lowest of the screen height and width can have visual artefacts when rendering, so here we scale it back to a more reasonable
	// size - in this case, the power of two *below* the lowest of height and width.  we don't go below 256 however.
	if (gl_max_size > 256)
	{
		int maxout = 99999999;
		int width;
		int height;
		int i;
		HDC hdc;

		// get the desktop size
		hdc = GetDC (NULL);
		width = GetDeviceCaps (hdc, HORZRES);
		height = GetDeviceCaps (hdc, VERTRES);
		ReleaseDC (NULL, hdc);

		// take the lowest of the two
		if (width < maxout) maxout = width;
		if (height < maxout) maxout = height;

		// take the power of two above the lowest
		for (i = 1; ; i *= 2)
			if (i > maxout) break;

		// set to half
		maxout = i / 2;

		// take the lowest of the two again
		if (gl_max_size > maxout) gl_max_size = maxout;

		// never less than 256
		if (gl_max_size < 256) gl_max_size = 256;
	}

	// never go above this limit
	if (gl_max_size > MAX_QUAKE_TEXTURE_SIZE) gl_max_size = MAX_QUAKE_TEXTURE_SIZE;

	Con_Printf ("Maximum Usable Texture Size: %i x %i\n", gl_max_size, gl_max_size);

	// Quake doesn't have a 1 TMU render path any more, so...
	glGetIntegerv (GL_MAX_TEXTURE_UNITS, &gl_num_tmus);
	Con_Printf ("Found %i Texture Memory Units\n\n", gl_num_tmus);

	// mhquake requires 3 tmus.  if an intel integrated can have 8, you really have no excuse
	if (gl_num_tmus < 3) Sys_Error ("Not enough TMUs");

	// now we get our entry points up to and including 1.4 - we crash if we fail to get one, cos life's like that sometimes.
	// BIG IMPORTANT NOTE.  these are NOT "extensions".  They are just core OpenGL 1.4 entry points missing from MS's platform SDK.
	// People should stop calling them "extensions", and stop treating them as such.  If your GL_VERSION is 1.4 or higher, you
	// have these natively in your implementation without needing to check for extensions.
	GL_GetFunctionPointers ();

	glClearColor (0, 0, 0, 0);
	glCullFace(GL_FRONT);
	glEnable(GL_TEXTURE_2D);

	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.666f);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
	*x = *y = 0;
	*width = WindowRect.right - WindowRect.left;
	*height = WindowRect.bottom - WindowRect.top;
}


void GL_EndRendering (void)
{
	glFinish ();

	if (!scr_skipupdate || block_drawing)
		SwapBuffers (maindc);

	// handle the mouse state when windowed if that's changed
	if (modestate == MS_WINDOWED)
	{
		if (!_windowed_mouse.value)
		{
			if (windowed_mouse)
			{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
				windowed_mouse = false;
			}
		}
		else
		{
			windowed_mouse = true;
			if (key_dest == key_game && !mouseactive && ActiveApp)
			{
				IN_ActivateMouse ();
				IN_HideMouse ();
			}
			else if (mouseactive && key_dest != key_game)
			{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
			}
		}
	}
}


void VID_SetPalette (unsigned char *palette)
{
	byte	*pal;
	unsigned r,g,b;
	unsigned v;
	unsigned short i;
	unsigned	*table;

	pal = palette;
	table = d_8to24table;

	for (i = 0; i < 256; i++)
	{
		r = pal[0];
		g = pal[1];
		b = pal[2];
		pal += 3;

		v = (255 << 24) + (r << 0) + (g << 8) + (b << 16);
		*table++ = v;
	}

	// set proper alpha colour
	d_8to24table[255] = 0;
}


void VID_ShiftPalette (unsigned char *palette)
{
}


void VID_SetDefaultMode (void)
{
	IN_DeactivateMouse ();
}


void VID_Shutdown (void)
{
   	HGLRC hRC;
   	HDC	  hDC;

	if (vid_initialized)
	{
		vid_canalttab = false;
		hRC = wglGetCurrentContext ();
    	hDC = wglGetCurrentDC ();

    	wglMakeCurrent (NULL, NULL);

    	if (hRC) wglDeleteContext (hRC);
		if (hDC && dibwindow) ReleaseDC (dibwindow, hDC);
		if (modestate == MS_FULLDIB) ChangeDisplaySettings (NULL, 0);
		if (maindc && dibwindow) ReleaseDC (dibwindow, maindc);

		AppActivate (false, false);
	}
}


//==========================================================================


BOOL bSetupPixelFormat (HDC hDC)
{
	static PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof (PIXELFORMATDESCRIPTOR),		// size of this pfd
		1,									// version number
		PFD_DRAW_TO_WINDOW 					// support window
		| PFD_SUPPORT_OPENGL 				// support OpenGL
		| PFD_DOUBLEBUFFER ,				// double buffered
		PFD_TYPE_RGBA,						// RGBA type
		24,									// 24-bit color depth
		0, 0, 0, 0, 0, 0,					// color bits ignored
		0,									// no alpha buffer
		0,									// shift bit ignored
		0,									// no accumulation buffer
		0, 0, 0, 0, 						// accum bits ignored
		32,									// 32-bit z-buffer	
		0,									// no stencil buffer
		0,									// no auxiliary buffer
		PFD_MAIN_PLANE,						// main layer
		0,									// reserved
		0, 0, 0								// layer masks ignored
	};

	int pixelformat;

	if ((pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0)
	{
		MessageBox (NULL, "ChoosePixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	if (SetPixelFormat (hDC, pixelformat, &pfd) == FALSE)
	{
		MessageBox (NULL, "SetPixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	return TRUE;
}



byte        scantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '1',    '2',    '3',    '4',    '5',    '6', 
	'7',    '8',    '9',    '0',    '-',    '=',    K_BACKSPACE, 9, // 0 
	'q',    'w',    'e',    'r',    't',    'y',    'u',    'i', 
	'o',    'p',    '[',    ']',    13 ,    K_CTRL,'a',  's',      // 1 
	'd',    'f',    'g',    'h',    'j',    'k',    'l',    ';', 
	'\'' ,    '`',    K_SHIFT,'\\',  'z',    'x',    'c',    'v',      // 2 
	'b',    'n',    'm',    ',',    '.',    '/',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'-',K_LEFTARROW,'5',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 

byte        shiftscantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '!',    '@',    '#',    '$',    '%',    '^', 
	'&',    '*',    '(',    ')',    '_',    '+',    K_BACKSPACE, 9, // 0 
	'Q',    'W',    'E',    'R',    'T',    'Y',    'U',    'I', 
	'O',    'P',    '{',    '}',    13 ,    K_CTRL,'A',  'S',      // 1 
	'D',    'F',    'G',    'H',    'J',    'K',    'L',    ':', 
	'"' ,    '~',    K_SHIFT,'|',  'Z',    'X',    'C',    'V',      // 2 
	'B',    'N',    'M',    '<',    '>',    '?',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'_',K_LEFTARROW,'%',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 


/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
	key = (key>>16)&255;
	if (key > 127)
		return 0;
	if (scantokey[key] == 0)
		Con_DPrintf("key 0x%02x has no translation\n", key);
	return scantokey[key];
}

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
	int		i;
	
// send an up event for each key, to make sure the server clears them all
	for (i=0 ; i<256 ; i++)
	{
		Key_Event (i, false);
	}

	Key_ClearStates ();
	IN_ClearStates ();
}

void AppActivate(BOOL fActive, BOOL minimize)
/****************************************************************************
*
* Function:     AppActivate
* Parameters:   fActive - True if app is activating
*
* Description:  If the application is activating, then swap the system
*               into SYSPAL_NOSTATIC mode so that our palettes will display
*               correctly.
*
****************************************************************************/
{
	static BOOL	sound_active;

	ActiveApp = fActive;
	Minimized = minimize;

// enable/disable sound on focus gain/loss
	if (!ActiveApp && sound_active)
	{
		S_BlockSound ();
		sound_active = false;
	}
	else if (ActiveApp && !sound_active)
	{
		S_UnblockSound ();
		sound_active = true;
	}

	if (fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
			if (vid_canalttab && vid_wassuspended) {
				vid_wassuspended = false;
				ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN);
				ShowWindow(mainwindow, SW_SHOWNORMAL);
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value && key_dest == key_game)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
	}

	if (!fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			if (vid_canalttab)
			{
				ChangeDisplaySettings (NULL, 0);
				vid_wassuspended = true;
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
		}
	}
}


/* main window procedure */
LONG CDAudio_MessageHandler (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

#ifndef WM_GRAPHNOTIFY
#define WM_GRAPHNOTIFY  WM_USER + 13
#endif

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    LONG    lRet = 1;
	int		fActive, fMinimized, temp;
	extern unsigned int uiWheelMessage;

	if (uMsg == uiWheelMessage)
		uMsg = WM_MOUSEWHEEL;

    switch (uMsg)
    {
		case WM_KILLFOCUS:
			if (modestate == MS_FULLDIB)
				ShowWindow(mainwindow, SW_SHOWMINNOACTIVE);
			break;

		case WM_CREATE:
			break;

		case WM_MOVE:
			window_x = (int) LOWORD(lParam);
			window_y = (int) HIWORD(lParam);
			VID_UpdateWindowStatus ();
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			Key_Event (MapKey(lParam), true);
			break;

		case WM_KEYUP:
		case WM_SYSKEYUP:
			Key_Event (MapKey(lParam), false);
			break;

		case WM_SYSCHAR:
			// keep Alt-Space from happening
			break;

		// this is complicated because Win32 seems to pack multiple mouse events into
		// one update sometimes, so we always check all states and look for events
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_MOUSEMOVE:
			temp = 0;

			if (wParam & MK_LBUTTON)
				temp |= 1;

			if (wParam & MK_RBUTTON)
				temp |= 2;

			if (wParam & MK_MBUTTON)
				temp |= 4;

			IN_MouseEvent (temp);

			break;

		// JACK: This is the mouse wheel with the Intellimouse
		// Its delta is either positive or neg, and we generate the proper
		// Event.
		case WM_MOUSEWHEEL: 
			if ((short) HIWORD(wParam) > 0)
			{
				Key_Event(K_MWHEELUP, true);
				Key_Event(K_MWHEELUP, false);
			}
			else
			{
				Key_Event(K_MWHEELDOWN, true);
				Key_Event(K_MWHEELDOWN, false);
			}
			break;

    	case WM_SIZE:
            break;

   	    case WM_CLOSE:
			if (MessageBox (mainwindow, "Are you sure you want to quit?", "Confirm Exit",
						MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
			{
				Sys_Quit ();
			}

	        break;

		case WM_ACTIVATE:
			fActive = LOWORD(wParam);
			fMinimized = (BOOL) HIWORD(wParam);
			AppActivate(!(fActive == WA_INACTIVE), fMinimized);

			// fix the leftover Alt from any Alt-Tab or the like that switched us away
			ClearAllStates ();

			break;

   	    case WM_DESTROY:
        {
			if (dibwindow)
				DestroyWindow (dibwindow);

            PostQuitMessage (0);
        }
        break;

		case WM_GRAPHNOTIFY:
            lRet = CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
			break;

    	default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}


/*
=================
VID_NumModes
=================
*/
int VID_NumModes (void)
{
	int i;
	gl_vmode_t *mode;

	for (mode = gl_modelist, i = 0; mode; mode = mode->next)
	{
		if (mode->type == MS_WINDOWED) continue;

		i++;
	}

	return i;
}


/*
=================
VID_DescribeCurrentMode_f
=================
*/
void VID_DescribeCurrentMode_f (void)
{
	Con_Printf ("%s\n", gl_current_mode->modedesc);
}


/*
=================
VID_NumModes_f
=================
*/
void VID_NumModes_f (void)
{
	if (VID_NumModes () == 1)
		Con_Printf ("%d video mode is available\n", VID_NumModes ());
	else
		Con_Printf ("%d video modes are available\n", VID_NumModes ());
}


/*
=================
VID_DescribeModes_f
=================
*/
void VID_DescribeModes_f (void)
{
	int			i;
	gl_vmode_t	*mode;

	for (i = 1, mode = gl_modelist; mode; mode = mode->next, i++)
	{
		if (mode->type == MS_WINDOWED) continue;

		Con_Printf ("%2d: %s\n", i, mode->modedesc);
	}
}


void VID_InitDIB (HINSTANCE hInstance)
{
	WNDCLASS		wc;

	/* Register the frame class */
    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC) MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
	wc.hbrBackground = NULL;
    wc.lpszMenuName  = 0;
    wc.lpszClassName = "WinQuake";

    if (!RegisterClass (&wc)) Sys_Error ("Couldn't register window class");

	// set up mode 0 as the windowed mode
	gl_modelist = (gl_vmode_t *) malloc (sizeof (gl_vmode_t));

	gl_modelist->type = MS_WINDOWED;

	// set windowed mode dimensions
	if (COM_CheckParm ("-width"))
		gl_modelist->width = Q_atoi (com_argv[COM_CheckParm ("-width") + 1]);
	else gl_modelist->width = 640;

	if (gl_modelist->width < 640) gl_modelist->width = 640;

	if (COM_CheckParm ("-height"))
		gl_modelist->height= Q_atoi (com_argv[COM_CheckParm ("-height") + 1]);
	else gl_modelist->height = gl_modelist->width * 240 / 320;

	if (gl_modelist->height < 240) gl_modelist->height = 240;

	// fill in the rest
	sprintf (gl_modelist->modedesc, "%d x %d", gl_modelist->width, gl_modelist->height);

	gl_modelist->modenum = MODE_WINDOWED;
	gl_modelist->dib = 1;
	gl_modelist->fullscreen = 0;
	gl_modelist->halfscreen = 0;
	gl_modelist->bpp = 0;

	// points to nothing
	gl_modelist->next = NULL;

	// set the windowed mode
	gl_windowed_mode = gl_modelist;

	nummodes = 1;
}


/*
=================
VID_InitFullDIB
=================
*/
void VID_InitFullDIB (HINSTANCE hInstance)
{
	DEVMODE	devmode;
	int		modenum, originalnummodes;
	BOOL	stat;
	gl_vmode_t *mode = NULL;
	gl_vmode_t *findmode = NULL;

	// enumerate > 8 bpp modes
	originalnummodes = nummodes;
	modenum = 0;

	do
	{
		stat = EnumDisplaySettings (NULL, modenum, &devmode);

		// don't bother if the mode is too small or has insufficient bpp
		if ((devmode.dmBitsPerPel >= 16) && (devmode.dmPelsWidth >= 640) && (devmode.dmPelsHeight >= 480))
		{
			devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

			if (ChangeDisplaySettings (&devmode, CDS_TEST | CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL)
			{
				// search for it in the mode list
				for (mode = gl_modelist; mode; mode = mode->next)
				{
					// skip over windowed modes
					if (mode->type == MS_WINDOWED) continue;

					// check for a match - we're not worried about refresh rates here
					if (mode->bpp == devmode.dmBitsPerPel && mode->width == devmode.dmPelsWidth && mode->height == devmode.dmPelsHeight)
					{
						// found a match
						break;
					}
				}

				// did we find it?
				if (!mode)
				{
					// no - allocate a new mode
					mode = (gl_vmode_t *) malloc (sizeof (gl_vmode_t));

					// fill it in
					mode->type = MS_FULLDIB;
					mode->width = devmode.dmPelsWidth;
					mode->height = devmode.dmPelsHeight;
					mode->modenum = 0;
					mode->halfscreen = 0;
					mode->dib = 1;
					mode->fullscreen = 1;
					mode->bpp = devmode.dmBitsPerPel;
					sprintf (mode->modedesc, "%4d x %4d x %2d", devmode.dmPelsWidth, devmode.dmPelsHeight, devmode.dmBitsPerPel);

					// link it into the list
					if (!gl_modelist)
					{
						// no items so make this the first
						gl_modelist = mode;
						mode->next = NULL;
					}
					else if (!gl_modelist->next)
					{
						// only one item so insert after that
						gl_modelist->next = mode;
						mode->next = NULL;
					}
					else
					{
						// find the end of the list and put it there
						for (findmode = gl_modelist; findmode; findmode = findmode->next)
						{
							// ignore windowed
							if (findmode->type == MS_WINDOWED) continue;

							// nothing more (end of list)
							if (!findmode->next) break;
						}

						if (findmode)
						{
							// insert after findmode
							mode->next = findmode->next;
							findmode->next = mode;
						}
						else
						{
							// just put it at the start of the list
							mode->next = gl_modelist;
							gl_modelist = mode;
						}
					}
				}
			}
		}

		modenum++;
	} while (stat);

	if (!gl_modelist) Con_SafePrintf ("No fullscreen DIB modes found\n");
}


static void Check_Gamma (unsigned char *pal)
{
	float	f, inf;
	unsigned char	palette[768];
	int		i;

	if ((i = COM_CheckParm ("-gamma")) == 0)
		vid_gamma = 1;
	else
		vid_gamma = Q_atof (com_argv[i + 1]);

	for (i = 0; i < 768; i++)
	{
		f = pow ((pal[i] + 1) / 256.0, vid_gamma );

		inf = f * 255 + 0.5;

		if (inf < 0) inf = 0;
		if (inf > 255) inf = 255;

		palette[i] = inf;
	}

	memcpy (pal, palette, sizeof (palette));
}

/*
===================
VID_Init
===================
*/
void VID_Init (unsigned char *palette)
{
	int		i;
	int		basenummodes, width, height, bpp;
	HDC		hdc;
	DEVMODE	devmode;
	gl_vmode_t *mode = gl_windowed_mode;

	memset (&devmode, 0, sizeof (devmode));

	Cvar_RegisterVariable (&vid_mode);
	Cvar_RegisterVariable (&vid_wait);
	Cvar_RegisterVariable (&vid_nopageflip);
	Cvar_RegisterVariable (&_vid_wait_override);
	Cvar_RegisterVariable (&_vid_default_mode);
	Cvar_RegisterVariable (&_vid_default_mode_win);
	Cvar_RegisterVariable (&vid_config_x);
	Cvar_RegisterVariable (&vid_config_y);
	Cvar_RegisterVariable (&vid_stretch_by_2);
	Cvar_RegisterVariable (&_windowed_mouse);

	Cmd_AddCommand ("vid_nummodes", VID_NumModes_f);
	Cmd_AddCommand ("vid_describecurrentmode", VID_DescribeCurrentMode_f);
	Cmd_AddCommand ("vid_describemodes", VID_DescribeModes_f);

	hIcon = LoadIcon (global_hInstance, MAKEINTRESOURCE (IDI_ICON2));

	InitCommonControls ();

	VID_InitDIB (global_hInstance);
	basenummodes = nummodes = 1;

	VID_InitFullDIB (global_hInstance);

	if (COM_CheckParm ("-window"))
	{
		// set windowed mode
		hdc = GetDC (NULL);

		if (GetDeviceCaps (hdc, RASTERCAPS) & RC_PALETTE) Sys_Error ("Can't run in non-RGB mode");

		ReleaseDC (NULL, hdc);

		windowed = true;
		leavecurrentmode = true;

		vid_default = MODE_WINDOWED;

		mode = gl_windowed_mode;
	}
	else
	{
		windowed = false;

		// get the desktop device context
		hdc = GetDC (NULL);

		// check for command-line parameters, set to desktop resolution and colour depth if not specified
		if (COM_CheckParm ("-width"))
			width = Q_atoi (com_argv[COM_CheckParm ("-width") + 1]);
		else width = GetDeviceCaps (hdc, HORZRES);

		if (COM_CheckParm ("-height"))
			height = Q_atoi (com_argv[COM_CheckParm ("-height") + 1]);
		else height = GetDeviceCaps (hdc, VERTRES);

		if (COM_CheckParm ("-bpp"))
			bpp = Q_atoi (com_argv[COM_CheckParm ("-bpp") + 1]);
		else bpp = GetDeviceCaps (hdc, BITSPIXEL);

		// check do we need to change
		if (width == GetDeviceCaps (hdc, HORZRES) && height == GetDeviceCaps (hdc, VERTRES) && bpp == GetDeviceCaps (hdc, BITSPIXEL))
			leavecurrentmode = true;
		else leavecurrentmode = false;
			
		// release the device context
		ReleaseDC (NULL, hdc);

		// find a mode that matches the selection
		for (mode = gl_modelist; mode; mode = mode->next)
		{
			// ignore windowed modes
			if (mode->type == MS_WINDOWED) continue;

			// check the match
			if (mode->width == width && mode->height == height && mode->bpp == bpp) break;
		}

		// no matching mode found
		if (!mode)
		{
			// take the first from the modelist that matches anything at all...
			for (mode = gl_modelist; mode; mode = mode->next)
			{
				// ignore windowed modes
				if (mode->type == MS_WINDOWED) continue;

				// check the match
				if (mode->width == width || mode->height == height || mode->bpp == bpp) break;
			}
		}

		// no matching mode found even with that, so we have a crazy command line.
		if (!mode)
		{
			// just take the first fullscreen mode available
			for (mode = gl_modelist; mode; mode = mode->next)
			{
				// ignore windowed modes
				if (mode->type == MS_WINDOWED) continue;

				break;
			}
		}

		// no fullscreen modes!
		if (!mode) Sys_Error ("No RGB Fullscreen modes available!");
	}

	vid_initialized = true;

	// set console width to the mode width unless specified otherwise
	if ((i = COM_CheckParm ("-conwidth")) != 0)
		vid.conwidth = Q_atoi (com_argv[i + 1]);
	else
		vid.conwidth = mode->width;

	vid.conwidth &= 0xfff8; // make it a multiple of eight

	if (vid.conwidth < 320) vid.conwidth = 320;

	// pick a conheight that matches with correct aspect
	vid.conheight = vid.conwidth * 3 / 4;

	if ((i = COM_CheckParm ("-conheight")) != 0) vid.conheight = Q_atoi (com_argv[i + 1]);
	if (vid.conheight < 200) vid.conheight = 200;

	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *) vid.colormap + 2048));

	Check_Gamma (palette);
	VID_SetPalette (palette);

	DestroyWindow (hwnd_dialog);

	// set the selected mode
	VID_SetMode (mode, palette);

	// save out current mode
	gl_current_mode = mode;

    maindc = GetDC (mainwindow);
	bSetupPixelFormat (maindc);

    baseRC = wglCreateContext (maindc);
	if (!baseRC) Sys_Error ("Could not initialize GL (wglCreateContext failed).\n\nMake sure you in are 65535 color mode, and try running -window.");
    if (!wglMakeCurrent (maindc, baseRC)) Sys_Error ("wglMakeCurrent failed");

	GL_Init ();

	vid_realmode = vid_modenum;

	vid_menudrawfn = VID_MenuDraw;
	vid_menukeyfn = VID_MenuKey;

	vid_canalttab = true;

	if (COM_CheckParm ("-fullsbar")) fullsbardraw = true;
}


//========================================================
// Video menu stuff
//========================================================

extern void M_Menu_Options_f (void);
extern void M_Print (int cx, int cy, char *str);
extern void M_PrintWhite (int cx, int cy, char *str);
extern void M_DrawCharacter (int cx, int line, int num);
extern void M_DrawTransPic (int x, int y, qpic_t *pic);
extern void M_DrawPic (int x, int y, qpic_t *pic);

static int	vid_line, vid_wmodes;

typedef struct
{
	int		modenum;
	char	*desc;
	int		iscur;
} modedesc_t;

#define MAX_COLUMN_SIZE		15
#define MODE_AREA_HEIGHT	(MAX_COLUMN_SIZE + 2)
#define MAX_MODEDESCS		(MAX_COLUMN_SIZE*3)

static modedesc_t	modedescs[MAX_MODEDESCS];

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	qpic_t		*p;
	int			column, row;
	gl_vmode_t	*mode;
	int colnum;

	p = Draw_CachePic ("gfx/vidmodes.lmp");

	M_DrawPic ((320 - p->width) / 2, 4, p);

	M_Print (4, 36 + 0 * 8, "Fullscreen Modes (WIDTH x HEIGHT x BPP)");

	column = 8;
	row = 36 + 2 * 8;
	colnum = 0;

	for (mode = gl_modelist; mode; mode = mode->next)
	{
		// don't list windowed modes
		if (mode->type == MS_WINDOWED) continue;

		if (mode == gl_current_mode)
			M_PrintWhite (column, row, mode->modedesc);
		else M_Print (column, row, mode->modedesc);

		// advance to the next column
		column += 20 * 8;
		colnum++;

		// check row advance
		if (colnum == 2)
		{
			// advance to the next row
			column = 8;
			row += 8;
			colnum = 0;
		}
	}

	M_Print (3 * 8, 36 + MODE_AREA_HEIGHT * 8 + 8 * 2, "Video modes must be set from the");
	M_Print (3 * 8, 36 + MODE_AREA_HEIGHT * 8 + 8 * 3, "command line with -width <width>");
	M_Print (3 * 8, 36 + MODE_AREA_HEIGHT * 8 + 8 * 4, "and -bpp <bits-per-pixel>");
	M_Print (3 * 8, 36 + MODE_AREA_HEIGHT * 8 + 8 * 6, "Select windowed mode with -window");
}


/*
================
VID_MenuKey
================
*/
void VID_MenuKey (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu1.wav");
		M_Menu_Options_f ();
		break;

	default:
		break;
	}
}
