/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 

#include "gl_local.h"

image_t		gltextures[MAX_GLTEXTURES];
int			numgltextures;
int			base_textureid;		// gltextures[i] = base_textureid + i

static byte			 intensitytable[256];
static unsigned char gammatable[256];
unsigned char lmgammatable[256];

cvar_t		*intensity;

unsigned	d_8to24table[256];
byte		lumatable[256];
byte		piclumatable[256] = {255};
byte		mdllumatable[256];
byte		heighttable[256];

#define STAGE_DIFF		0
#define STAGE_NORM		1
#define STAGE_SPEC		2
#define STAGE_LUMA		3
#define STAGE_NOSCALE	4
#define STAGE_MDLLUMA	5


#define MAX_SUPPORTED_TEXTURE_WIDTH		512
#define MAX_SUPPORTED_TEXTURE_HEIGHT	256

void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean is_sky, int stage, qboolean is_wall);
void GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap);
void GL_UploadTexture (unsigned *data, int width, int height, qboolean mipmap, int bpp, int stage) ;


int		gl_solid_format = 3;
int		gl_alpha_format = 4;

int		gl_tex_solid_format = 3;
int		gl_tex_alpha_format = 4;

int		gl_filter_min = GL_LINEAR_MIPMAP_NEAREST;
int		gl_filter_max = GL_LINEAR;

void GL_SetTexturePalette (unsigned palette[256]) 
{
}

void GL_Enable666Multitexture (qboolean enable) 
{
	if (!glActiveTexture) return;

	if (enable) 
	{
		GL_SelectTexture (GL_TEXTURE1);
		glEnable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}
	else
	{
		GL_SelectTexture (GL_TEXTURE1);
		glDisable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	GL_SelectTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

void GL_SelectTexture (GLenum texture) 
{
	int tmu;

	if (!glActiveTexture) return;

	if (texture == GL_TEXTURE0) 
	{
		tmu = 0;
	}
	else
	{
		tmu = 1;
	}

	if (tmu == gl_state.currenttmu) 
	{
		return;
	}

	gl_state.currenttmu = tmu;

	if (glActiveTexture) 
	{
		glActiveTexture (texture);
		glClientActiveTexture (texture);
	}
}

void GL_Tex666Env (GLenum mode) 
{
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode);
}


typedef struct
{
	char *name;
	int	minimize, maximize;
} glmode_t;

glmode_t modes[] = {
	{"GL_NEAREST", GL_NEAREST, GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR}
};

#define NUM_GL_MODES (sizeof (modes) / sizeof (glmode_t)) 

typedef struct
{
	char *name;
	int mode;
} gltmode_t;

gltmode_t gl_alpha_modes[] = {
	{"default", 4},
	{"GL_RGBA", GL_RGBA},
	{"GL_RGBA8", GL_RGBA8},
	{"GL_RGB5_A1", GL_RGB5_A1},
	{"GL_RGBA4", GL_RGBA4},
	{"GL_RGBA2", GL_RGBA2},
};

#define NUM_GL_ALPHA_MODES (sizeof (gl_alpha_modes) / sizeof (gltmode_t)) 

gltmode_t gl_solid_modes[] = {
	{"default", 3},
	{"GL_RGB", GL_RGB},
	{"GL_RGB8", GL_RGB8},
	{"GL_RGB5", GL_RGB5},
	{"GL_RGB4", GL_RGB4},
	{"GL_R3_G3_B2", GL_R3_G3_B2},
#ifdef GL_RGB2_EXT
	{"GL_RGB2", GL_RGB2_EXT},
#endif
};

#define NUM_GL_SOLID_MODES (sizeof (gl_solid_modes) / sizeof (gltmode_t)) 

/*
=============== 
GL_TextureMode
=============== 
*/ 
void GL_TextureMode (char *string) 
{
	int		i;
	image_t	*glt;

	for (i = 0; i< NUM_GL_MODES; i++) 
	{
		if (!Q_stricmp (modes[i].name, string)) 
			break;
	}

	if (i == NUM_GL_MODES) 
	{
		VID_Printf (PRINT_ALL, "bad filter name\n");
		return;
	}

	gl_filter_min = modes[i].minimize;
	gl_filter_max = modes[i].maximize;

	// change all the existing mipmap texture objects
	for (i = 0, glt = gltextures; i<numgltextures; i++ , glt++) 
	{
		if (glt->type != it_pic && glt->type != it_sky) 
		{
			glBindTexture (GL_TEXTURE_2D, glt->texnum);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
		}
	}
}


void GL_ApplyAniso (int texnum)
{
	if (texnum)
	{
		glBindTexture (GL_TEXTURE_2D, texnum);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, gl_anisotropic_filter->value);
	}
}


void GL_AnisotropicFilter (void) 
{
	int i;
	image_t	*glt;

	// no anisotropy available
	if (gl_config.maxanisotropy < 2) return;

	if (gl_anisotropic_filter->value < 1) gl_anisotropic_filter->value = 1;

	// get the next power of 2
	for (i = 1; i < gl_anisotropic_filter->value; i *= 2);

	if (i > gl_config.maxanisotropy) 
		i = gl_config.maxanisotropy;

	gl_anisotropic_filter->value = i;

	// change all the existing mipmap texture objects
	for (i = 0, glt = gltextures; i < numgltextures; i++ , glt++) 
	{
		if (glt->type != it_pic && glt->type != it_sky) 
		{
			GL_ApplyAniso (glt->texnum);
			GL_ApplyAniso (glt->lumatex);
			GL_ApplyAniso (glt->normtex);
		}
	}

	gl_anisotropic_filter->modified = false;
}


/*
=============== 
GL_TextureAlphaMode
=============== 
*/ 
void GL_TextureAlphaMode (char *string) 
{
	int		i;

	for (i = 0; i< NUM_GL_ALPHA_MODES; i++) 
	{
		if (!Q_stricmp (gl_alpha_modes[i].name, string)) 
			break;
	}

	if (i == NUM_GL_ALPHA_MODES) 
	{
		VID_Printf (PRINT_ALL, "bad alpha texture mode name\n");
		return;
	}

	gl_tex_alpha_format = gl_alpha_modes[i].mode;
}

/*
=============== 
GL_TextureSolidMode
=============== 
*/ 
void GL_TextureSolidMode (char *string) 
{
	int		i;

	for (i = 0; i< NUM_GL_SOLID_MODES; i++) 
	{
		if (!Q_stricmp (gl_solid_modes[i].name, string)) 
			break;
	}

	if (i == NUM_GL_SOLID_MODES) 
	{
		VID_Printf (PRINT_ALL, "bad solid texture mode name\n");
		return;
	}

	gl_tex_solid_format = gl_solid_modes[i].mode;
}

/*
=============== 
GL_ImageList_f
=============== 
*/ 
void	GL_ImageList_f (void) 
{
}


/*
============================================================================= 

scrap allocation

Allocate all the little status bar obejcts into a single texture
to crutch up inefficient hardware / drivers

============================================================================= 
*/ 

#define	MAX_SCRAPS		1
#define	BLOCK_WIDTH		256
#define	BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][BLOCK_WIDTH*BLOCK_HEIGHT];
qboolean	scrap_dirty;

// returns a texture number and the position inside it
int Scrap_AllocBlock (int w, int h, int *x, int *y) 
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum = 0; texnum<MAX_SCRAPS; texnum++) 
	{
		best = BLOCK_HEIGHT;

		for (i = 0; i<BLOCK_WIDTH - w; i++) 
		{
			best2 = 0;

			for (j = 0; j<w; j++) 
			{
				if (scrap_allocated[texnum][i + j]>= best) 
					break;
				if (scrap_allocated[texnum][i + j]> best2) 
					best2 = scrap_allocated[texnum][i + j];
			}
			if (j == w) 
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT) 
			continue;

		for (i = 0; i<w; i++) 
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	return - 1;
	// 	VID_Error ("Scrap_AllocBlock: full");
}

int	scrap_uploads;

void Scrap_Upload (void) 
{
	int i;

	scrap_uploads++;

	for (i = 0; i < MAX_SCRAPS; i++)
	{
		glBindTexture (GL_TEXTURE_2D, TEXNUM_SCRAPS + i);
		GL_UploadTexture ((unsigned *) scrap_texels[i], BLOCK_WIDTH, BLOCK_HEIGHT, false, 8, STAGE_DIFF);
	}

	scrap_dirty = false;
}


void Scrap_Init (void)
{
	int i;

	for (i = 0; i < MAX_SCRAPS; i++)
	{
		memset (scrap_texels[i], 0, BLOCK_WIDTH * BLOCK_HEIGHT);
	}

	scrap_dirty = true;
}


/*
================================================================= 

PCX LOADING

================================================================= 
*/ 


/*
============== 
LoadPCX
============== 
*/ 
void LoadPCX (char *filename, byte **pic, byte **palette, int *width, int *height) 
{
	byte	*raw;
	pcx_t	*pcx;
	int		x, y;
	int		len;
	int		dataByte, runLength;
	byte	*out, *pix;

	*pic = NULL;
	*palette = NULL;

	// 
	// load the file
	// 
	len = FS_LoadFile (filename, (void **) &raw);
	if (!raw) 
	{
		VID_Printf (PRINT_DEVELOPER, "Bad pcx file %s\n", filename);
		return;
	}

	// 
	// parse the PCX file
	// 
	pcx = (pcx_t *) raw;

	pcx->xmin = LittleShort (pcx->xmin);
	pcx->ymin = LittleShort (pcx->ymin);
	pcx->xmax = LittleShort (pcx->xmax);
	pcx->ymax = LittleShort (pcx->ymax);
	pcx->hres = LittleShort (pcx->hres);
	pcx->vres = LittleShort (pcx->vres);
	pcx->bytes_per_line = LittleShort (pcx->bytes_per_line);
	pcx->palette_type = LittleShort (pcx->palette_type);

	raw = &pcx->data;

	if (pcx->manufacturer != 0x0a
		|| pcx->version != 5
		|| pcx->encoding != 1
		|| pcx->bits_per_pixel != 8
		|| pcx->xmax >= 640
		|| pcx->ymax >= 480) 
	{
		VID_Printf (PRINT_ALL, "Bad pcx file %s\n", filename);
		return;
	}

	out = malloc ((pcx->ymax + 1) * (pcx->xmax + 1));

	*pic = out;

	pix = out;

	if (palette) 
	{
		*palette = malloc (768);
		memcpy (*palette, (byte *) pcx + len - 768, 768);
	}

	if (width) 
		*width = pcx->xmax + 1;
	if (height) 
		*height = pcx->ymax + 1;

	for (y = 0; y<= pcx->ymax; y++ , pix += pcx->xmax + 1) 
	{
		for (x = 0; x<= pcx->xmax;) 
		{
			dataByte = *raw++;

			if ((dataByte & 0xC0) == 0xC0) 
			{
				runLength = dataByte & 0x3F;
				dataByte = *raw++;
			}
			else
				runLength = 1;

			while (runLength-- > 0) 
				pix[x++] = dataByte;
		}

	}

	if (raw - (byte *) pcx > len) 
	{
		VID_Printf (PRINT_DEVELOPER, "PCX file %s was malformed", filename);
		free (*pic);
		*pic = NULL;
	}

	FS_FreeFile (pcx);
}


/*
========================================================= 

TARGA LOADING

========================================================= 
*/ 

typedef struct _TargaHeader {
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} TargaHeader;


/*
============= 
LoadTGA
============= 
*/ 
void LoadTGA (char *name, byte **pic, int *width, int *height) 
{
	int		columns, rows, numPixels;
	byte	*pixbuf;
	int		row, column;
	byte	*buf_p;
	byte	*buffer;
	int		length;
	TargaHeader		targa_header;
	byte			*targa_rgba;
	byte tmp[2];

	*pic = NULL;

	// load the file
	length = FS_LoadFile (name, (void **) &buffer);
	if (!buffer) 
	{
		VID_Printf (PRINT_DEVELOPER, "Bad tga file %s\n", name);
		return;
	}

	buf_p = buffer;

	targa_header.id_length = *buf_p++;
	targa_header.colormap_type = *buf_p++;
	targa_header.image_type = *buf_p++;

	tmp[0] = buf_p[0];
	tmp[1] = buf_p[1];
	targa_header.colormap_index = LittleShort (* ((short *) tmp));
	buf_p += 2;
	tmp[0] = buf_p[0];
	tmp[1] = buf_p[1];
	targa_header.colormap_length = LittleShort (* ((short *) tmp));
	buf_p += 2;
	targa_header.colormap_size = *buf_p++;
	targa_header.x_origin = LittleShort (* ((short *) buf_p));
	buf_p += 2;
	targa_header.y_origin = LittleShort (* ((short *) buf_p));
	buf_p += 2;
	targa_header.width = LittleShort (* ((short *) buf_p));
	buf_p += 2;
	targa_header.height = LittleShort (* ((short *) buf_p));
	buf_p += 2;
	targa_header.pixel_size = *buf_p++;
	targa_header.attributes = *buf_p++;

	if (targa_header.image_type!= 2 
		&& targa_header.image_type!= 10) 
		VID_Error (ERR_DROP, "LoadTGA: Only type 2 and 10 targa RGB images supported\n");

	if (targa_header.colormap_type != 0 
		|| (targa_header.pixel_size!= 32 && targa_header.pixel_size!= 24)) 
		VID_Error (ERR_DROP, "LoadTGA: Only 32 or 24 bit images supported (no colormaps) \n");

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	if (width) 
		*width = columns;
	if (height) 
		*height = rows;

	targa_rgba = malloc (numPixels*4);
	*pic = targa_rgba;

	if (targa_header.id_length != 0) 
		buf_p += targa_header.id_length;// skip TARGA image comment

	if (targa_header.image_type == 2) {// Uncompressed, RGB images
		for (row = rows - 1; row>= 0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for (column = 0; column<columns; column++) {
				unsigned char red,green,blue,alphabyte;
				switch (targa_header.pixel_size) {
					case 24:

						blue = *buf_p++;
						green = *buf_p++;
						red = *buf_p++;
						*pixbuf++= red;
						*pixbuf++= green;
						*pixbuf++= blue;
						*pixbuf++= 255;
						break;
					case 32:
						blue = *buf_p++;
						green = *buf_p++;
						red = *buf_p++;
						alphabyte = *buf_p++;
						*pixbuf++= red;
						*pixbuf++= green;
						*pixbuf++= blue;
						*pixbuf++= alphabyte;
						break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10) {// Runlength encoded RGB images
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for (row = rows - 1; row>= 0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for (column = 0; column<columns;) {
				packetHeader = *buf_p++;
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) {// run - length packet
					switch (targa_header.pixel_size) {
						case 24:
							blue = *buf_p++;
							green = *buf_p++;
							red = *buf_p++;
							alphabyte = 255;
							break;
						case 32:
							blue = *buf_p++;
							green = *buf_p++;
							red = *buf_p++;
							alphabyte = *buf_p++;
							break;
					}

					for (j = 0;j<packetSize;j++) {
						*pixbuf++= red;
						*pixbuf++= green;
						*pixbuf++= blue;
						*pixbuf++= alphabyte;
						column++;
						if (column == columns) {// run spans across rows
							column = 0;
							if (row>0) 
								row--;
							else
								goto breakOut;
							pixbuf = targa_rgba + row*columns*4;
						}
					}
				}
				else {
					// non run - length packet
					for (j = 0;j<packetSize;j++) {
						switch (targa_header.pixel_size) {
							case 24:
								blue = *buf_p++;
								green = *buf_p++;
								red = *buf_p++;
								*pixbuf++= red;
								*pixbuf++= green;
								*pixbuf++= blue;
								*pixbuf++= 255;
								break;
							case 32:
								blue = *buf_p++;
								green = *buf_p++;
								red = *buf_p++;
								alphabyte = *buf_p++;
								*pixbuf++= red;
								*pixbuf++= green;
								*pixbuf++= blue;
								*pixbuf++= alphabyte;
								break;
						}
						column++;
						if (column == columns) {// pixel packet run spans across rows
							column = 0;
							if (row>0) 
								row--;
							else
								goto breakOut;
							pixbuf = targa_rgba + row*columns*4;
						}						
					}
				}
			}
breakOut:;
		}
	}

	FS_FreeFile (buffer);
}


/*
==================================================================== 

IMAGE FLOOD FILLING

==================================================================== 
*/ 


/*
================= 
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes
================= 
*/ 

typedef struct
{
	short		x, y;
} floodfill_t;

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1) 

#define FLOODFILL_STEP(off, dx, dy) \
{\
	if (pos[off] == fillcolor) \
{\
	pos[off] = 255; \
	fifo[inpt].x = x + (dx) , fifo[inpt].y = y + (dy); \
	inpt = (inpt + 1) & FLOODFILL_FIFO_MASK; \
} \
	else if (pos[off]!= 255) fdc = pos[off]; \
}

void R_FloodFillSkin (byte *skin, int skinwidth, int skinheight) 
{
	byte				fillcolor = *skin;// assume this is the pixel to fill
	floodfill_t			fifo[FLOODFILL_FIFO_SIZE];
	int					inpt = 0, outpt = 0;
	int					filledcolor = - 1;
	int					i;

	if (filledcolor == - 1) 
	{
		filledcolor = 0;
		// attempt to find opaque black
		for (i = 0; i < 256; ++ i) 
			if (d_8to24table[i] == (255 << 0))// alpha 1.0
			{
				filledcolor = i;
				break;
			}
	}

	// can't fill to filled color or to transparent color (used as visited marker) 
	if ((fillcolor == filledcolor) || (fillcolor == 255)) 
	{
		// printf ("not filling skin from %d to %d\n", fillcolor, filledcolor);
		return;
	}

	fifo[inpt].x = 0, fifo[inpt].y = 0;
	inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

	while (outpt != inpt) 
	{
		int			x = fifo[outpt].x, y = fifo[outpt].y;
		int			fdc = filledcolor;
		byte		*pos = &skin[x + skinwidth * y];

		outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

		if (x > 0) 				FLOODFILL_STEP( - 1, - 1, 0);
		if (x < skinwidth - 1) 	FLOODFILL_STEP(1, 1, 0);
		if (y > 0) 				FLOODFILL_STEP( - skinwidth, 0, - 1);
		if (y < skinheight - 1) 	FLOODFILL_STEP(skinwidth, 0, 1);
		skin[x + skinwidth * y] = fdc;
	}
}

// ======================================================= 


/*
==============================================

	NORMAL MAP GENERATION

==============================================
*/
// quickie table for deriving square roots
float sqrttable[256] = {-1};

float GSPixels[MAX_SUPPORTED_TEXTURE_WIDTH * MAX_SUPPORTED_TEXTURE_HEIGHT];

/*
=============================
MakeDOT3

Convert a height field image into a normal map. This involves
differencing each texel with its right and upper neighbor, then
normalizing the cross product of the two difference vectors.

Original code from "Practical and robust bumpmapping" hacked to death by me...
=============================
*/
void MakeDOT3 (unsigned int *dst, int w, int h)
{
	int i, j;
	int di, dj;
	int iw, diw;
	int i2hw, di2hw;

	float sqlen, reciplen, nx, ny, nz, c, cx, cy, dcx, dcy;
	float scale = 8;
	byte *dstrgba = (byte *) dst;

	// downsample with a box filter at the same time as we're building the normal map
	for (i = 0; i < h; i += 2)
	{
		di = i + 1;
		iw = i * w;
		diw = di * w;
		i2hw = ((i + 2) % h) * w;
		di2hw = ((di + 2) % h) * w;

		for (j = 0; j < w; j += 2, dstrgba += 4)
		{
			// don't bother making j2w as it's only saving half the calculations per iteration
			// the outside loop stuff was critical however...
			dj = j + 1;

			// Expand [0,255] texel values to the [0,1] range.
			c = GSPixels[iw + j];
			c += GSPixels[diw + j];
			c += GSPixels[iw + dj];
			c += GSPixels[diw + dj];

			// Expand the texel to its right.
			cx = GSPixels[iw + (j + 2) % w];
			cx += GSPixels[diw + (j + 2) % w];
			cx += GSPixels[iw + (dj + 2) % w];
			cx += GSPixels[diw + (dj + 2) % w];

			// Expand the texel one up.
			cy = GSPixels[i2hw + j];
			cy += GSPixels[di2hw + j];
			cy += GSPixels[i2hw + dj];
			cy += GSPixels[di2hw + dj];

			dcx = scale * (c - cx);
			dcy = scale * (c - cy);

			// Normalize the vector.  (nz is 1 so add 1 * 1 here...)
			sqlen = dcx * dcx + dcy * dcy + 1;
			reciplen = 1.0f / sqrt (sqlen);

			nx = dcy * reciplen;
			ny = -dcx * reciplen;
			nz = reciplen;

			// Repack the normalized vector into an RGB unsigned byte
			// vector in the normal map image.
			dstrgba[0] = 128 + 127 * nx;
			dstrgba[1] = 128 + 127 * ny;
			dstrgba[2] = 128 + 127 * nz;
		}
	}
}


void R_InitSqrtTable (void)
{
	int i;
	float f;

	for (i = 0; i < 256; i++)
	{
		f = i;
		f /= 255.0f;
		f = sqrt (f);
		f *= 255.0f;
		sqrttable[i] = f;
	}
}

/*
================ 
GL_ScaleTextureForStage

Scale up the pixel values in a texture to increase the lighting range.
================ 
*/ 
void GL_ScaleTextureForStage (unsigned *in, int inwidth, int inheight, qboolean mipmap, int stage) 
{
	int i;
	int c;
	byte *p = (byte *) in;

	c = inwidth * inheight;

	if (piclumatable[0] == 255)
	{
		for (i = 0; i < 256; i++)
		{
			c = (((int) gammatable[i] * 3) / 4) + lumatable[i];

			if (c > 255)
				piclumatable[i] = 255;
			else piclumatable[i] = c;
		}
	}

	if (!mipmap) 
	{
		// flat graphic so just apply gamma
		for (i = 0; i < c; i++, p += 4) 
		{
			p[0] = piclumatable[p[0]];
			p[1] = piclumatable[p[1]];
			p[2] = piclumatable[p[2]];
		}
	}
	else if (stage == STAGE_NOSCALE)
	{
		// explicitly not lightscaled
		for (i = 0; i < c; i++, p += 4) 
		{
			p[0] = piclumatable[p[0]];
			p[1] = piclumatable[p[1]];
			p[2] = piclumatable[p[2]];
		}
	}
	else if (stage == STAGE_NORM)
	{
		if (sqrttable[0] < 0) R_InitSqrtTable ();

		// apply the heightmap version
		for (i = 0; i < c; i++, p += 4) 
		{
			// save out to the greyscale buffer
			GSPixels[i] = sqrttable[((int) heighttable[p[0]] * 30 + (int) heighttable[p[1]] * 59 + (int) heighttable[p[2]] * 11) / 100];

			// convert to 0 to 1 range
			// because we'll be averaging 4 pixels here we'll divide this by 4 as well as by 255
			GSPixels[i] /= 1020.0f;
		}

		// now convert it to a half-size normal map
		MakeDOT3 (in, inwidth, inheight);
	}
	else if (stage == STAGE_LUMA)
	{
		// aply the double-squared luminescent version
		for (i = 0; i < c; i++, p += 4) 
		{
			p[0] = lumatable[p[0]];
			p[1] = lumatable[p[1]];
			p[2] = lumatable[p[2]];
		}
	}
	else if (stage == STAGE_MDLLUMA)
	{
		// aply the double-squared luminescent version
		for (i = 0; i < c; i++, p += 4) 
		{
			p[0] = mdllumatable[p[0]];
			p[1] = mdllumatable[p[1]];
			p[2] = mdllumatable[p[2]];
		}
	}
	else
	{
		// diffuse stage
		for (i = 0; i < c; i++, p += 4) 
		{
			p[0] = gammatable[intensitytable[p[0]]];
			p[1] = gammatable[intensitytable[p[1]]];
			p[2] = gammatable[intensitytable[p[2]]];
		}
	}
}


typedef struct resamplepoint_s
{
	int p1;
	int p2;
	int p1percent;
	int p2percent;
} resamplepoint_t;


resamplepoint_t resamplew[MAX_SUPPORTED_TEXTURE_WIDTH];
resamplepoint_t resampleh[MAX_SUPPORTED_TEXTURE_HEIGHT];

void MakeResamplePoint (resamplepoint_t *resamp, int inscale, int outscale, int outmult)
{
	int i;

	// set up the resample points across
	for (i = 0; i < outscale; i++)
	{
		// calculate the equalivalent point in the "in" buffer
		float fp = (float) (inscale - 1) * ((float) i / (float) (outscale - 1));
		float decpart;

		resamp[i].p1 = (int) floor (fp);
		resamp[i].p2 = (int) ceil (fp);

		// the decimal part is the percentage of p2 used
		resamp[i].p2percent = (int) ((fp - (float) resamp[i].p1) * 100.0f);

		// the rest is the percentage of p2 used
		resamp[i].p1percent = 100 - resamp[i].p2percent;

		resamp[i].p1 *= outmult;
		resamp[i].p2 *= outmult;
	}
}


void GL_ResampleTexture (byte *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight, int bpp)
{
	int h;
	int w;
	int outpoint;
	byte *indata[4];
	byte *outdata;
	int i;
	int inrgba[4];
	int inpercents[4];
	int resampmult;

	if (bpp == 8)
		resampmult = 1;
	else resampmult = 4;

	// set up our resample points in 2 lookup tables for added speed
	MakeResamplePoint (resamplew, inwidth, outwidth, 1 * resampmult);
	MakeResamplePoint (resampleh, inheight, outheight, inwidth * resampmult);

	for (h = 0, outpoint = 0; h < outheight; h++)
	{
		for (w = 0; w < outwidth; w++, outpoint++)
		{
			// get a valid pointer to the outdata
			outdata = (byte *) &out[outpoint];

			// set up the percents
			inpercents[0] = (resamplew[w].p1percent + resampleh[h].p1percent);
			inpercents[1] = (resamplew[w].p2percent + resampleh[h].p1percent);
			inpercents[2] = (resamplew[w].p1percent + resampleh[h].p2percent);
			inpercents[3] = (resamplew[w].p2percent + resampleh[h].p2percent);

			// initially 0
			inrgba[0] = inrgba[1] = inrgba[2] = inrgba[3] = 0;

			// set up the data points
			if (bpp == 8)
			{
				indata[0] = (byte *) &d_8to24table[in[resamplew[w].p1 + resampleh[h].p1]];
				indata[1] = (byte *) &d_8to24table[in[resamplew[w].p2 + resampleh[h].p1]];
				indata[2] = (byte *) &d_8to24table[in[resamplew[w].p1 + resampleh[h].p2]];
				indata[3] = (byte *) &d_8to24table[in[resamplew[w].p2 + resampleh[h].p2]];
			}
			else
			{
				indata[0] = &in[resamplew[w].p1 + resampleh[h].p1];
				indata[1] = &in[resamplew[w].p2 + resampleh[h].p1];
				indata[2] = &in[resamplew[w].p1 + resampleh[h].p2];
				indata[3] = &in[resamplew[w].p2 + resampleh[h].p2];
			}

			// accumulate the data for each point
			inrgba[0] += (int) indata[0][0] * inpercents[0];
			inrgba[1] += (int) indata[0][1] * inpercents[0];
			inrgba[2] += (int) indata[0][2] * inpercents[0];
			inrgba[3] += (int) indata[0][3] * inpercents[0];

			inrgba[0] += (int) indata[1][0] * inpercents[1];
			inrgba[1] += (int) indata[1][1] * inpercents[1];
			inrgba[2] += (int) indata[1][2] * inpercents[1];
			inrgba[3] += (int) indata[1][3] * inpercents[1];

			inrgba[0] += (int) indata[2][0] * inpercents[2];
			inrgba[1] += (int) indata[2][1] * inpercents[2];
			inrgba[2] += (int) indata[2][2] * inpercents[2];
			inrgba[3] += (int) indata[2][3] * inpercents[2];

			inrgba[0] += (int) indata[3][0] * inpercents[3];
			inrgba[1] += (int) indata[3][1] * inpercents[3];
			inrgba[2] += (int) indata[3][2] * inpercents[3];
			inrgba[3] += (int) indata[3][3] * inpercents[3];

			// save out
			outdata[0] = (byte) (inrgba[0] / 400);
			outdata[1] = (byte) (inrgba[1] / 400);
			outdata[2] = (byte) (inrgba[2] / 400);
			outdata[3] = (byte) (inrgba[3] / 400);
		}
	}
}


void GL_DeleteTexture (int texnum)
{
	if (texnum)
	{
		glDeleteTextures (1, &texnum);
		texnum = 0;
	}
}


/*
=============== 
GL_UploadTexture

=============== 
*/
unsigned int scaled[MAX_SUPPORTED_TEXTURE_WIDTH * MAX_SUPPORTED_TEXTURE_HEIGHT];

int upload_width;
int upload_height;

void GL_UploadTexture (unsigned *data, int width, int height, qboolean mipmap, int bpp, int stage) 
{
	int i;
	int scaled_width, scaled_height;

	// get power of 2 dimensions
	for (scaled_width = 1; scaled_width < width; scaled_width *= 2);
	for (scaled_height = 1; scaled_height < height; scaled_height *= 2);

	// bound the width and height
	if (scaled_width > gl_config.maxtexsize) scaled_width = gl_config.maxtexsize;
	if (scaled_height > gl_config.maxtexsize) scaled_height = gl_config.maxtexsize;

	if (scaled_width > MAX_SUPPORTED_TEXTURE_WIDTH) scaled_width = MAX_SUPPORTED_TEXTURE_WIDTH;
	if (scaled_height > MAX_SUPPORTED_TEXTURE_HEIGHT) scaled_height = MAX_SUPPORTED_TEXTURE_HEIGHT;

	// should never happen...
	if (scaled_width < 1) scaled_width = 1;
	if (scaled_height < 1) scaled_height = 1;

	if (bpp == 32 && scaled_width == width && scaled_height == height)
	{
		// copy it straight in
		for (i = 0; i < width * height; i++) 
		{
			scaled[i] = data[i];
		}
	}
	else if (bpp == 8 && scaled_width == width && scaled_height == height)
	{
		// expand 8bpp to 32 bpp
		byte *data8bpp = (byte *) data;

		for (i = 0; i < width * height; i++) 
		{
			scaled[i] = d_8to24table[data8bpp[i]];
		}
	}
	else GL_ResampleTexture ((byte *) data, width, height, scaled, scaled_width, scaled_height, bpp);

	// apply lightscaling here before we upload.  this will also handle the scaling for the stage.
	// we copied everything into "scaled" above so that subsequent passes through here for other stages
	// won't have to work on pre-lightscaled data (which can give unwanted results)
	GL_ScaleTextureForStage (scaled, scaled_width, scaled_height, mipmap, stage);

	// if it's the normal map stage we must halve the sizes again
	if (stage == STAGE_NORM)
	{
		scaled_width /= 2;
		scaled_height /= 2;
	}
	else if (stage == STAGE_DIFF)
	{
		upload_width = scaled_width;
		upload_height = scaled_height;
	}

	// hardware mipmapping (OpenGL 1.4 or higher an absolute must) 
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, mipmap ? GL_TRUE : GL_FALSE);

	// don't bother compressing cos Quake 2 textures are fairly light on the old video ram
	// always take it up in RGBA format cos that's what OpenGL uses (or at least is supposed
	// to use, according to the specs) internally anyway.
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mipmap ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


/*
================ 
GL_LoadPic

This is also used as an entry point for the generated r_notexture
================ 
*/ 
image_t *GL_LoadPic (char *name, byte *pic, int width, int height, imagetype_t type, int bits) 
{
	image_t		*image;
	int			i;
	qboolean mipmap;

	// find a free image_t
	for (i = 0, image = gltextures; i<numgltextures; i++ ,image++) 
	{
		if (!image->texnum) 
			break;
	}

	if (i == numgltextures) 
	{
		if (numgltextures == MAX_GLTEXTURES) 
			VID_Error (ERR_DROP, "MAX_GLTEXTURES");
		numgltextures++;
	}

	image = &gltextures[i];

	if (strlen (name) >= sizeof (image->name)) 
		VID_Error (ERR_DROP, "Draw_LoadPic: \"%s\" is too long", name);

	strcpy (image->name, name);
	image->registration_sequence = registration_sequence;

	image->width = width;
	image->height = height;
	image->type = type;

	if (type == it_skin && bits == 8) R_FloodFillSkin (pic, width, height);

	// load little pics into the scrap
	if (image->type == it_pic && bits == 8 && image->width < 64 && image->height < 64) 
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

#define SCRAP_BORDER 8
		// look for a block SCRAP_BORDER texels greater than the image dimensions to ensure that linear
		// filtering doesn't fuck up the edges of the image in the scrap...
		texnum = Scrap_AllocBlock (image->width + SCRAP_BORDER, image->height + SCRAP_BORDER, &x, &y);

		if (texnum == -1)
		{
			VID_Printf (PRINT_DEVELOPER, "Failed to Allocate Scrap for %s\n", name);
			goto nonscrap;
		}

		scrap_dirty = true;

		// set the scrap block to all alpha first to keep edges safe
		// include the SCRAP_BORDER texel boundary here...
		for (i = 0; i < image->height + SCRAP_BORDER; i++) 
			for (j = 0; j < image->width + SCRAP_BORDER; j++) 
				scrap_texels[texnum][(y + i) *BLOCK_WIDTH + x + j] = 255;

		// now increase x and y by (SCRAP_BORDER / 2) each to center it in the newly allocated area
		x += (SCRAP_BORDER / 2);
		y += (SCRAP_BORDER / 2);
#undef SCRAP_BORDER

		// now copy the texels into the scrap block at the new x and y points
		for (i = 0, k = 0; i < image->height; i++) 
			for (j = 0; j < image->width; j++, k++) 
				scrap_texels[texnum][(y + i) *BLOCK_WIDTH + x + j] = pic[k];

		image->texnum = TEXNUM_SCRAPS + texnum;
		image->scrap = true;
		image->sl = ((float) x + 0.01) / (float) BLOCK_WIDTH;
		image->sh = ((float) x + (float) image->width - 0.01) / (float) BLOCK_WIDTH;
		image->tl = ((float) y + 0.01) / (float) BLOCK_WIDTH;
		image->th = ((float) y + (float) image->height - 0.01) / (float) BLOCK_WIDTH;
	}
	else
	{
nonscrap:
		// free any texes from the last use of this struct
		GL_DeleteTexture (image->texnum);
		GL_DeleteTexture (image->lumatex);
		GL_DeleteTexture (image->normtex);

		image->scrap = false;
		image->texnum = TEXNUM_IMAGES + (image - gltextures);

		mipmap = (image->type != it_pic && image->type != it_sky);

		glBindTexture (GL_TEXTURE_2D, image->texnum);

		// diffuse stage
		GL_UploadTexture ((unsigned *) pic, width, height, mipmap, bits, STAGE_DIFF);

		if (image->type == it_wall || image->type == it_skin)
		{
			// save out
			image->upload_width = upload_width;
			image->upload_height = upload_height;

			// luminescent stage
			image->lumatex = TEXNUM_LUMAS + (image - gltextures);
			glBindTexture (GL_TEXTURE_2D, image->lumatex);
			GL_UploadTexture ((unsigned *) pic, width, height, mipmap, bits, STAGE_LUMA);

			// STAGE_NORM
			image->normtex = TEXNUM_NORMS + (image - gltextures);
			glBindTexture (GL_TEXTURE_2D, image->normtex);
			GL_UploadTexture ((unsigned *) pic, width, height, mipmap, bits, STAGE_NORM);
		}
		else
		{
			// probably not necessary...
			image->lumatex = 0;
			image->normtex = 0;
		}

		image->sl = 0;
		image->sh = 1;
		image->tl = 0;
		image->th = 1;
	}

	return image;
}


/*
================ 
GL_LoadWal
================ 
*/ 
image_t *GL_LoadWal (char *name, imagetype_t type) 
{
	miptex_t	*mt;
	int			width, height, ofs;
	image_t		*image;

	FS_LoadFile (name, (void **) &mt);
	if (!mt) 
	{
		VID_Printf (PRINT_ALL, "GL_FindImage: can't load %s\n", name);
		return r_notexture;
	}

	width = LittleLong (mt->width);
	height = LittleLong (mt->height);
	ofs = LittleLong (mt->offsets[0]);

	// check for mt->flags & SURF_LIGHT
	image = GL_LoadPic (name, (byte *) mt + ofs, width, height, type, 8);

	FS_FreeFile ((void *) mt);

	return image;
}


/*
=============== 
GL_FindImage

Finds or loads the given image
=============== 
*/ 
image_t	*GL_FindImage (char *name, imagetype_t type) 
{
	image_t	*image;
	int		i, len;
	byte	*pic, *palette;
	int		width, height;

	if (!name) return NULL;
	len = strlen (name);
	if (len<5) return NULL;

	// look for it
	for (i = 0, image = gltextures; i<numgltextures; i++ ,image++) 
	{
		if (!strcmp (name, image->name)) 
		{
			image->registration_sequence = registration_sequence;
			return image;
		}
	}

	// 
	// load the pic from disk
	// 
	pic = NULL;
	palette = NULL;
	if (!strcmp (name + len - 4, ".pcx")) 
	{
		LoadPCX (name, &pic, &palette, &width, &height);
		if (!pic) 
			return NULL;// VID_Error (ERR_DROP, "GL_FindImage: can't load %s", name);

		image = GL_LoadPic (name, pic, width, height, type, 8);
	}
	else if (!strcmp (name + len - 4, ".wal")) 
	{
		image = GL_LoadWal (name, type);
	}
	else if (!strcmp (name + len - 4, ".tga")) 
	{
		LoadTGA (name, &pic, &width, &height);
		if (!pic) 
			return NULL;// VID_Error (ERR_DROP, "GL_FindImage: can't load %s", name);
		image = GL_LoadPic (name, pic, width, height, type, 32);
	}
	else
		return NULL;	// 	VID_Error (ERR_DROP, "GL_FindImage: bad extension on: %s", name);


	if (pic) 
		free (pic);
	if (palette) 
		free (palette);

	return image;
}



/*
=============== 
R_RegisterSkin
=============== 
*/ 
struct image_s *R_RegisterSkin (char *name) 
{
	return GL_FindImage (name, it_skin);
}


/*
================ 
GL_FreeUnusedImages

Any image that was not touched on this registration sequence
will be freed.
================ 
*/ 
void GL_FreeUnusedImages (void) 
{
	int		i;
	image_t	*image;

	// never free r_notexture or particle texture
	r_notexture->registration_sequence = registration_sequence;
	r_particletexture->registration_sequence = registration_sequence;

	// and our other misc textures
	r_bloomtexture->registration_sequence = registration_sequence;
	r_hazetexture->registration_sequence = registration_sequence;
	r_cubemap->registration_sequence = registration_sequence;
	r_detailtexture->registration_sequence = registration_sequence;
	r_blacktexture->registration_sequence = registration_sequence;
	r_nonormaltexture->registration_sequence = registration_sequence;

	for (i = 0, image = gltextures; i<numgltextures; i++ , image++) 
	{
		if (image->registration_sequence == registration_sequence) continue;	// used this sequence
		if (!image->registration_sequence) continue;							// free image_t slot
		if (image->type == it_pic) continue;									// don't free pics

		// free it
		GL_DeleteTexture (image->texnum);
		GL_DeleteTexture (image->normtex);
		GL_DeleteTexture (image->lumatex);

		memset (image, 0, sizeof (*image));
	}

	// make sure that anisotropic filtering is properly applied!!!
	GL_AnisotropicFilter ();
}


/*
=============== 
Draw_GetPalette
=============== 
*/ 
int Draw_GetPalette (void) 
{
	int		i;
	int		j;
	int		rgb[3];
	unsigned	v;
	byte	*pic, *pal;
	int		width, height;
	float	f;

	// get the palette
	LoadPCX ("pics/colormap.pcx", &pic, &pal, &width, &height);

	if (!pal) VID_Error (ERR_FATAL, "Couldn't load pics/colormap.pcx");

	for (i = 0; i < 256; i++) 
	{
		rgb[0] = pal[i * 3 + 0];
		rgb[1] = pal[i * 3 + 1];
		rgb[2] = pal[i * 3 + 2];

		v = (255 << 24) + (rgb[0] << 0) + (rgb[1] << 8) + (rgb[2] << 16);
		d_8to24table[i] = LittleLong (v);
	}

	d_8to24table[255] = 0;	// 255 is transparent

	free (pic);
	free (pal);

	// make a luma table by squaring the intensity twice
	for (i = 0; i < 256; i++)
	{
		f = (float) i / 255.0f;

		f *= f;
		f *= 2;
		f *= f;
		f *= 2;

		if (f < 0) f = 0;
		if (f > 1) f = 1;

		lumatable[i] = (byte) (f * 255.0f);
	}

	// identify first use
	piclumatable[0] = 255;

	// make a luma table by squaring the intensity twice
	for (i = 0; i < 256; i++)
	{
		f = (float) i / 255.0f;

		f *= f;
		f *= 2;
		f *= f;
		f *= 2;
		f *= f;
		f *= 2;

		if (f < 0) f = 0;
		if (f > 1) f = 1;

		mdllumatable[i] = (byte) (f * 255.0f);
	}

	// make a height table by squaring the intensity once
	for (i = 0; i < 256; i++)
	{
		f = (float) i / 255.0f;

		f *= f;

		if (f < 0) f = 0;
		if (f > 1) f = 1;

		heighttable[i] = (byte) (f * 255.0f);
	}

	return 0;
}


/*
=============== 
GL_InitImages
=============== 
*/ 
void GL_InitImages (void) 
{
	image_t	*image;
	int		i, j;
	float	g = vid_gamma->value;

	registration_sequence = 1;

	// init intensity conversions
	// made default intensity 1 and did an RGB scale of 2 on the texture blend, which replicates the
	// effect but does not reduce the texture to 7BPP
	intensity = Cvar_Get ("intensity", "1", 0);

	if (intensity->value <= 1) Cvar_Set ("intensity", "1");

	gl_state.inverse_intensity = 1 / intensity->value;

	Draw_GetPalette ();

	for (i = 0; i < 256; i++) 
	{
		if (g == 1) 
		{
			gammatable[i] = i;
		}
		else
		{
			float inf;

			inf = 255 * pow ((i + 0.5) / 255.5 , g) + 0.5;

			if (inf < 0) inf = 0;
			if (inf > 255) inf = 255;

			gammatable[i] = inf;
		}
	}

	g = lm_gamma->value;

	for (i = 0; i < 256; i++) 
	{
		if (g == 1) 
		{
			lmgammatable[i] = i;
		}
		else
		{
			float inf;

			inf = 255 * pow ((i + 0.5) / 255.5 , g) + 0.5;

			if (inf < 0) inf = 0;
			if (inf > 255) inf = 255;

			lmgammatable[i] = inf;
		}
	}

	for (i = 0; i < 256; i++) 
	{
		// allow floating point intensity values.
		// to be sure to be sure
		j = (int) ((float) i * intensity->value);

		if (j > 255) j = 255;

		intensitytable[i] = j;
	}

	// clear all of our image_t structures to 0
	for (i = 0, image = gltextures; i < MAX_GLTEXTURES; i++ , image++) 
	{
		memset (image, 0, sizeof (*image));
	}

	// set the scrap to all black
	Scrap_Init ();

	numgltextures = 0;
}


/*
=============== 
GL_ShutdownImages
=============== 
*/ 
void	GL_ShutdownImages (void) 
{
	int		i;
	image_t	*image;

	for (i = 0, image = gltextures; i < numgltextures; i++ , image++) 
	{
		// free image_t slot
		if (!image->registration_sequence) continue;

		// free it
		GL_DeleteTexture (image->texnum);
		GL_DeleteTexture (image->lumatex);
		GL_DeleteTexture (image->normtex);

		memset (image, 0, sizeof (*image));
	}
}


/*
================
GL_LoadBitmapResourceTexture

Loads a BMP resource texture

only bitmaps can be loaded from resources, cos versions of Windows prior to XP don't support other
image formats in resources.
================
*/
image_t *GL_LoadBitmapResourceTexture (int ResourceID, qboolean alphamask)
{
	unsigned char *ImageRGBAData;
	int i, j;
	int run;
	int runlength;
	unsigned char *bmIn;
	BITMAPINFO *bmInfo;
	BITMAPFILEHEADER bmHeader;
	RGBQUAD bmPalette[256];
	RGBQUAD *PaletteIn;
	unsigned char *bmData;
	unsigned char *bmRaw;
	int pad;
	HINSTANCE hInstance;
	HANDLE hBitmap;
	char imgname[16];
	image_t *image;

	// derive a name from the resource id
	sprintf (imgname, "Bitmap_%3i", ResourceID);

	// get a handle to the MHQuake II executable
	hInstance = GetModuleHandle (NULL);

	// load the resource
	hBitmap = LoadResource (hInstance, FindResource (hInstance, MAKEINTRESOURCE (ResourceID), RT_BITMAP));

	// lock it
	bmRaw = (byte *) LockResource (hBitmap);

	// get the bm info header
	// read the header
	bmInfo = (BITMAPINFO *) bmRaw;

	// skip past the header (for 24 BPP RGB images the pixel data immediately follows, otherwise a palette is next)
	bmRaw += sizeof (BITMAPINFO);

	if (bmInfo->bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo->bmiHeader.biCompression == BI_RLE8 && bmInfo->bmiHeader.biBitCount == 8))
			goto BadBMP;
	}

	// support 4, 8 and 24 BPP modes
	if (bmInfo->bmiHeader.biBitCount != 4 && 
		bmInfo->bmiHeader.biBitCount != 8 && 
		bmInfo->bmiHeader.biBitCount != 24) goto BadBMP;

	// allocate a buffer and set the dynamic data pointer
	ImageRGBAData = malloc (bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight) * 4);
	bmData = ImageRGBAData;

	if (bmInfo->bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo->bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo->bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo->bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo->bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo->bmiHeader.biClrUsed; i++)
		{
			PaletteIn = (RGBQUAD *) bmRaw;

			bmPalette[i].rgbBlue = PaletteIn->rgbBlue;
			bmPalette[i].rgbGreen = PaletteIn->rgbGreen;
			bmPalette[i].rgbRed = PaletteIn->rgbRed;
			bmPalette[i].rgbReserved = PaletteIn->rgbReserved;

			bmRaw += sizeof (RGBQUAD);
		}
	}

	// each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
	// out how much padding we need (this may not work with RLE images)
	for (i = 0; ; i++)
	{
		pad = i * 4;
		if (pad >= (bmInfo->bmiHeader.biWidth * 3)) break;
	}

	pad -= bmInfo->bmiHeader.biWidth * 3;

	// load the data
	for (i = 0; i < bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight); i++)
	{
		if (bmInfo->bmiHeader.biBitCount == 4)
		{
			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[*bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[*bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[*bmIn % 16].rgbRed;
			bmData[5] = bmPalette[*bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[*bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo->bmiHeader.biBitCount == 8)
		{
			if (bmInfo->bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = bmRaw++;

				if (*bmIn == 0)
				{
					bmIn = bmRaw++;
					i++;

					switch (*bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * 4;

						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * bmInfo->bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = *bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = bmRaw++;
							i++;

							bmData[0] = bmPalette[*bmIn].rgbRed;
							bmData[1] = bmPalette[*bmIn].rgbGreen;
							bmData[2] = bmPalette[*bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = bmRaw++;
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = *bmIn;

					bmIn = bmRaw++;
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[*bmIn].rgbRed;
						bmData[1] = bmPalette[*bmIn].rgbGreen;
						bmData[2] = bmPalette[*bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn].rgbRed;
			bmData[1] = bmPalette[*bmIn].rgbGreen;
			bmData[2] = bmPalette[*bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else
		{
			// 24 BPP
			bmData[2] = *bmRaw++;
			bmData[1] = *bmRaw++;
			bmData[0] = *bmRaw++;
			bmData[3] = 255;

			bmData += 4;
		}

		// read the padding
		for (j = 0; j < pad; j++) bmRaw++;
	}

bmDone:;

	// check orientation
	if (bmInfo->bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < abs (bmInfo->bmiHeader.biHeight) / 2; i++)
		{
			for (j = 0; j < bmInfo->bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo->bmiHeader.biWidth) + j;
				unsigned int dstpos = (((abs (bmInfo->bmiHeader.biHeight) - 1) - i) * bmInfo->bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) ImageRGBAData)[dstpos];
				((unsigned int *) ImageRGBAData)[dstpos] = ((unsigned int *) ImageRGBAData)[srcpos];
				((unsigned int *) ImageRGBAData)[srcpos] = tmp;
			}
		}
	}

	// because this doesn't happen at runtime, and only happens a couple of times per refresh load, we can very
	// safely do it slowly and inelegantly
	if (alphamask)
	{
		bmData = ImageRGBAData;

		for (i = 0; i < abs (bmInfo->bmiHeader.biHeight); i++)
		{
			for (j = 0; j < bmInfo->bmiHeader.biWidth; j++, bmData += 4)
			{
				// set alpha channel to the value of the red channel
				// (it's assumed these are greyscale)
				bmData[3] = bmData[0];

				// set rgb to white
				bmData[0] = 255;
				bmData[1] = 255;
				bmData[2] = 255;
			}
		}
	}

	// load it as a pic so it won't get unregistered, and always 32 bpp
	image = GL_LoadPic (imgname, ImageRGBAData, bmInfo->bmiHeader.biWidth, abs (bmInfo->bmiHeader.biHeight), it_pic, 32);

	// done with the data
	free (ImageRGBAData);

	// let the calling function handle success or failure (in most cases it's safe to assume success
	// otherwise (a) the programmer has fucked up and specified an invalid resource id, or (b) the
	// executable is corrupt, in which case not being able to load a resource is the least of our worries)
	return image;

BadBMP:;
	return NULL;
}

