/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 

// particle sprite rendering

#include "gl_local.h"


extern image_t *r_particlefont;


float particlest[][4] =
{
//	 sbase  tbase  smax   tmax
	{0.000, 0.000, 0.250, 0.500},		// 0: smoke
	{0.250, 0.000, 0.375, 0.250},		// 1: dot
	{0.375, 0.000, 0.500, 0.250},		// 2: blood 1
	{0.500, 0.000, 0.625, 0.250},		// 3: blood 2
	{0.625, 0.000, 0.750, 0.250},		// 4: blood 3
	{0.750, 0.000, 0.875, 0.250},		// 5: blood 4
	{0.875, 0.000, 1.000, 0.250},		// 6: spark
	{0.250, 0.250, 0.375, 0.500},		// 7: mini-blast
	{0.375, 0.250, 0.500, 0.500},		// 8: bubble
	{0.500, 0.250, 0.625, 0.500},		// 9: raindrop
	{0.000, 0.500, 0.250, 1.000},		// 10: blast
	{0.250, 0.500, 0.500, 1.000},		// 11: flare 1
	{0.500, 0.500, 0.750, 1.000},		// 12: flare 2
	{0.750, 0.500, 1.000, 1.000},		// 13: fire cinder
};

#define P_SMOKE			0
#define P_DOT			1
#define P_BLOOD1		2
#define P_BLOOD2		3
#define P_BLOOD3		4
#define P_BLOOD4		5
#define P_SPARK			6
#define P_MINIBLAST		7
#define P_BUBBLR		8
#define P_RAINDROP		9
#define P_BLAST			10
#define P_FLARE1		11
#define P_FLARE2		12
#define P_CINDER		13

typedef struct rparticle_s
{
	// render fields
	float renderscale;
	vec3_t renderorg;
	float renderalpha;

	// higher is less attenuation, 0.004 is the same as regular Quake particles
	float atten;

	float scale;
	int type;

	// linked list management
	struct rparticle_s *next;

	// chained for drawing
	struct rparticle_s *chain;

	// fields duplicated from cparticle_s
	float		time;
	vec3_t		org;
	vec3_t		vel;
	vec3_t		accel;
	float		color;
	float		colorvel;
	float		alpha;
	float		alphavel;
} rparticle_t;


// copy in some defines from client.h
#define	R_PARTICLE_GRAVITY			40
#define R_BLASTER_PARTICLE_COLOR	0xe0
#define R_INSTANT_PARTICLE			-10000.0

// this is our new max_particles
#define MAX_R_PARTICLES		MAX_PARTICLES * 4

rparticle_t R_Particles[MAX_R_PARTICLES];

rparticle_t *FreeParticles = &R_Particles[0];
rparticle_t *ActiveParticles = NULL;

// 2 chains for the 2 different blend funcs
rparticle_t *PDrawChain[2];

float ParticleCurrTime;
float ParticleFrameTime;

#define NUM_SMOKE_PARTICLES 1024

rparticle_t R_SmokeParticles[NUM_SMOKE_PARTICLES];


void R_UpdateSmokeParticle (rparticle_t *p, float updatetime)
{
	p->org[2] += p->vel[2] * updatetime;
	p->scale += (updatetime * 30);
	p->vel[2] += (updatetime * 45);
	p->alpha -= (updatetime / 2);
}


void R_InitSmokeParticle (rparticle_t *p)
{
	// set the base particle state
	if (rand () & 1)
		p->color = 0 + (rand () % 7);
	else p->color = 57 + (rand () % 6);

	p->org[0] = -2 + (rand () % 5);
	p->org[1] = -2 + (rand () % 5);
	p->org[2] = -2 + (rand () % 5);

	p->vel[0] = p->vel[1] = p->vel[2] = 0;

	p->type = P_SMOKE;
	p->scale = 3;
	p->alpha = 0.666;
}


void R_UpdateSmokeParticles (float time)
{
	int i;
	float lowestalpha = 666;
	int lowestalphanum = 0;

	for (i = 0; i < NUM_SMOKE_PARTICLES; i++)
	{
		R_UpdateSmokeParticle (&R_SmokeParticles[i], time);

		// find the particle with the lowest alpha
		if (R_SmokeParticles[i].alpha < lowestalpha)
		{
			lowestalpha = R_SmokeParticles[i].alpha;
			lowestalphanum = i;
		}
	}

	// add a new particle to the list to replace the particle with the lowest alpha
	// done here to ensure that only one new particle is added each frame, otherwise the
	// smoke animation experiences some strange blips every now and then
	R_InitSmokeParticle (&R_SmokeParticles[lowestalphanum]);
}


void R_InitSmokeParticles (void)
{
	int i;
	int j;
	particle_t *p;
	float updatetime;

	for (i = 0; i < NUM_SMOKE_PARTICLES; i++)
	{
		R_InitSmokeParticle (&R_SmokeParticles[i]);

		// run a fake update to simulate adding an extra particle each frame
		for (j = 0; j < i; j++)
		{
			updatetime = 6 + (rand () % 9);
			updatetime /= 1000.0;

			R_UpdateSmokeParticle (&R_SmokeParticles[j], updatetime);
		}
	}
}



void R_ClearParticleSprites (void)
{
	int i;

	FreeParticles = &R_Particles[0];
	ActiveParticles = NULL;

	for (i = 0; i < MAX_R_PARTICLES - 1; i++)
		R_Particles[i].next = &R_Particles[i + 1];

	R_Particles[MAX_R_PARTICLES - 1].next = NULL;

	// and the smokes too.
	R_InitSmokeParticles ();

	// now run through 512 fake frames to get a reasonably random and advanced baseline
	for (i = 0; i < 512; i++)
	{
		float faketime = 6 + (rand () % 9);
		faketime /= 1000.0f;
		R_UpdateSmokeParticles (faketime);
	}
}


__inline rparticle_t *R_LinkParticle (void)
{
	rparticle_t *p;

	if (!FreeParticles) return NULL;

	p = FreeParticles;
	FreeParticles = p->next;
	p->next = ActiveParticles;
	ActiveParticles = p;

	p->time = ParticleCurrTime;

	return p;
}


void R_AddBloodSprites (vec3_t org, vec3_t dir, int color, int count)
{
	int i;
	int j;
	float d;
	rparticle_t *p;

	for (i = 0; i < count; i++)
	{
		if (!(p = R_LinkParticle ())) return;

		p->color = color + (rand () &7);

		d = rand () & 31;

		for (j = 0; j < 3; j++) 
		{
			p->org[j] = org[j] + ((rand () & 7) - 4) + d * dir[j];
			p->vel[j] = crand () * 20;
		}

		p->accel[0] = p->accel[1] = 0;
		p->accel[2] = -R_PARTICLE_GRAVITY;

		p->alpha = 1.0;
		p->alphavel = - 1.0 / (0.5 + frand () * 0.3);

		switch (rand () % 8)
		{
		default:
		case 0:
		case 1:
		case 2:
			p->type = P_BLOOD1;
			break;

		case 3:
		case 4:
		case 5:
			p->type = P_BLOOD2;
			break;

		case 6:
			p->type = P_BLOOD3;
			break;

		case 7:
			p->type = P_BLOOD4;
			break;
		}

		p->scale = 4 + (rand () % 4);
		p->atten = 0.001;
	}
}


void R_SingleBloodSprite (vec3_t move, float orgscale, float velscale, int color)
{
	int j;
	rparticle_t *p;

	if (!(p = R_LinkParticle ())) return;

	p->color = color + (rand () &7);

	for (j = 0; j < 3; j++) 
	{
		p->org[j] = move[j] + crand () * orgscale;
		p->vel[j] = crand () * velscale;
		p->accel[j] = 0;
	}

	p->vel[2] -= R_PARTICLE_GRAVITY;

	p->alpha = 1.0;
	p->alphavel = -1.0 / (1 + frand () * 0.4);

	switch (rand () % 6)
	{
	default:
	case 0:
	case 1:
	case 2:
		p->type = P_BLOOD1;
		break;

	case 3:
	case 4:
	case 5:
		p->type = P_BLOOD2;
		break;
	}

	p->scale = 4 + (rand () % 4);

	// attenuate these less as there's more of them
	p->atten = 0.002;
}


void R_SingleSmokeSprite (vec3_t move, float orgscale, float velscale, int color)
{
	int j;
	rparticle_t *p;

	// hack - because smoke is bigger, we don't need so many particles with sprites as we do with points
	static qboolean SkipMe = true;

	SkipMe = !SkipMe;

	if (SkipMe) return;

	if (!(p = R_LinkParticle ())) return;

	p->color = color + (rand () & 7);

	for (j = 0; j < 3; j++) 
	{
		p->org[j] = move[j] + crand () * orgscale;
		p->vel[j] = crand () * velscale;
	}

	p->accel[0] = p->accel[1] = 0;
	p->accel[2] = 20;

	p->alpha = 0.5;
	p->alphavel = (-1.0 / (1 + frand () * 0.2)) * 0.5;

	p->type = P_SMOKE;

	p->scale = 14 + (rand () % 6);

	// attenuate less
	p->atten = 0.003;
}


// generic shell for particle adding
void R_AddParticleGeneric (vec3_t org, vec3_t dir, int color, int count)
{
	int i;
	rparticle_t *p;

	for (i = 0; i < count; i++)
	{
		if (!(p = R_LinkParticle ())) return;
	}
}


// this should have the same value as MAX_VARRAY_VERTS in gl_varrays.c
#define MAX_PART_VERTS MAX_VERTS + 2
#define PARTICLE_VERTEX_SIZE 9

// share verts with the main render but draw in different sized batches
extern float VArrayVerts[];

float ParticleRGBTable[256][3] =
{
	// invalid value at startup so we can identify that it hasn't been set yet
	{-1, -1, -1},
};


// this is the only vertex array code not in gl_varrays.c
void R_RenderParticleSpriteChain (rparticle_t *chain, GLenum MH_GL_BLENDFUNC_SRC_FUNC, GLenum MH_GL_BLENDFUNC_DST_FUNC)
{
	rparticle_t		*p;
	vec3_t			up, right;
	float			scale;
	float			*pv = VArrayVerts;
	int				PrimitiveCount = 0;

	if (ParticleRGBTable[0][0] < 0)
	{
		// startup so hit the RGB table
		int i;

		for (i = 0; i < 256; i++)
		{
			ParticleRGBTable[i][0] = (float) ((byte *) &d_8to24table[i])[0] / 255.0f;
			ParticleRGBTable[i][1] = (float) ((byte *) &d_8to24table[i])[1] / 255.0f;
			ParticleRGBTable[i][2] = (float) ((byte *) &d_8to24table[i])[2] / 255.0f;
		}
	}

	if (!chain) return;

	glBlendFunc (MH_GL_BLENDFUNC_SRC_FUNC, MH_GL_BLENDFUNC_DST_FUNC);

	for (p = chain; p; p = p->chain)
	{
		// hack a scale up to keep particles from disapearing
		scale = (p->renderorg[0] - r_origin[0]) * vpn[0] + (p->renderorg[1] - r_origin[1]) * vpn[1] + (p->renderorg[2] - r_origin[2]) * vpn[2];

		if (scale < 20) 
			scale = 1;
		else
			scale = 1.0f + scale * p->atten;

		// do this per particle as they can have different scales.
		// do it with inline code rather than sending through VectorScale
		up[0] = vup[0] * p->renderscale * scale;
		up[1] = vup[1] * p->renderscale * scale;
		up[2] = vup[2] * p->renderscale * scale;

		right[0] = vright[0] * p->renderscale * scale;
		right[1] = vright[1] * p->renderscale * scale;
		right[2] = vright[2] * p->renderscale * scale;

		// verts
		// the particle will be off-center at the specified origin, so correct that
		pv[0] = p->renderorg[0] - (right[0] + up[0]) * 0.5f;
		pv[1] = p->renderorg[1] - (right[1] + up[1]) * 0.5f;
		pv[2] = p->renderorg[2] - (right[2] + up[2]) * 0.5f;

		pv[9]  = pv[0] + up[0];
		pv[10] = pv[1] + up[1];
		pv[11] = pv[2] + up[2];

		pv[18] = pv[9]  + right[0];
		pv[19] = pv[10] + right[1];
		pv[20] = pv[11] + right[2];

		pv[27] = pv[0] + right[0];
		pv[28] = pv[1] + right[1];
		pv[29] = pv[2] + right[2];

		// texcoords
		pv[3] = particlest[p->type][0];
		pv[4] = particlest[p->type][1];

		pv[12] = particlest[p->type][0];
		pv[13] = particlest[p->type][3];

		pv[21] = particlest[p->type][2];
		pv[22] = particlest[p->type][3];

		pv[30] = particlest[p->type][2];
		pv[31] = particlest[p->type][1];

		// rgba
		pv[5] = pv[14] = pv[23] = pv[32] = ParticleRGBTable[(int) p->color][0];
		pv[6] = pv[15] = pv[24] = pv[33] = ParticleRGBTable[(int) p->color][1];
		pv[7] = pv[16] = pv[25] = pv[34] = ParticleRGBTable[(int) p->color][2];
		pv[8] = pv[17] = pv[26] = pv[35] = p->renderalpha;

		pv += (4 * PARTICLE_VERTEX_SIZE);
		PrimitiveCount += 4;

		if (PrimitiveCount >= MAX_PART_VERTS)
		{
			// draw this number of primitives
			glDrawArrays (GL_QUADS, 0, MAX_PART_VERTS);

			// reset the pointer and counter
			PrimitiveCount = 0;
			pv = VArrayVerts;
		}
	}

	// draw anything left over
	if (PrimitiveCount) glDrawArrays (GL_QUADS, 0, PrimitiveCount);
}


float	rp_turbsin[] = 
{
#include "warpsin.h"
};
#define P_TURBSCALE (256.0 / (2 * M_PI)) 

#define LAVA_SMOKE_SCALE 12.0f
#define LAVA_RENDER_SCALE 72.0f

void R_RenderLavaSmokeParticle (vec3_t org, int offset)
{
	int i;
	rparticle_t *sp;
	vec3_t up, right;
	float *pv = VArrayVerts;
	float scale;

	// we could make this a global across all functions, but in practice this one draws in batches of 256 anyway,
	// which is the max we'll use for it, so we won't bother
	int PrimitiveCount = 0;

	// shouldn't happen, but y' know...
	offset = abs (offset);

	// calculate the scaled vertex offset factors
	// only once for lava
	// anything higher than this and we start getting performance issues up close
	up[0] = vup[0] * LAVA_RENDER_SCALE;
	up[1] = vup[1] * LAVA_RENDER_SCALE;
	up[2] = vup[2] * LAVA_RENDER_SCALE;

	right[0] = vright[0] * LAVA_RENDER_SCALE;
	right[1] = vright[1] * LAVA_RENDER_SCALE;
	right[2] = vright[2] * LAVA_RENDER_SCALE;

	// use this blendfunc for smokes
	glBlendFunc (GL_SRC_ALPHA, GL_ONE); //GL_ONE_MINUS_SRC_ALPHA);

	for (i = offset; i < NUM_SMOKE_PARTICLES; i += 16)
	{
		// get a pointer to the particle so that things don't get messy
		sp = &R_SmokeParticles[i];

		// don't draw if alpha is too low.  we accept negative alpha then scale back up later
		// for a better height range
		if (sp->alpha < -0.25) continue;

		// set particle origin
		sp->renderorg[0] = org[0] + sp->org[0] * LAVA_SMOKE_SCALE;
		sp->renderorg[1] = org[1] + sp->org[1] * LAVA_SMOKE_SCALE;

		// offset z negatively so it pushes below the surface
		sp->renderorg[2] = org[2] + sp->org[2] * (LAVA_SMOKE_SCALE * 0.25f) - offset;

		// too low
		if ((sp->renderorg[2] + LAVA_RENDER_SCALE) < org[2]) continue;

		scale = (sp->renderorg[0] - r_origin[0]) * vpn[0] + (sp->renderorg[1] - r_origin[1]) * vpn[1] + (sp->renderorg[2] - r_origin[2]) * vpn[2];

		// too near the screen or behind the current viewpoint
		if (scale < 20) continue;

		// now add 0.25 to alpha to compensate for the negative alpha above
		sp->renderalpha = sp->alpha + 0.25;

		// clamp the range
		if (sp->renderalpha > 1) sp->renderalpha = 1;
		if (sp->renderalpha < 0) sp->renderalpha = 0;

		// verts
		// the particle will be off-center at the specified origin, so correct that
		pv[0] = sp->renderorg[0] - (right[0] + up[0]) * 0.5f;
		pv[1] = sp->renderorg[1] - (right[1] + up[1]) * 0.5f;
		pv[2] = sp->renderorg[2] - (right[2] + up[2]) * 0.5f;

		pv[9]  = pv[0] + up[0];
		pv[10] = pv[1] + up[1];
		pv[11] = pv[2] + up[2];

		pv[18] = pv[9]  + right[0];
		pv[19] = pv[10] + right[1];
		pv[20] = pv[11] + right[2];

		pv[27] = pv[0] + right[0];
		pv[28] = pv[1] + right[1];
		pv[29] = pv[2] + right[2];

		// texcoords - always smoke
		pv[3] = particlest[P_SMOKE][0];
		pv[4] = particlest[P_SMOKE][1];

		pv[12] = particlest[P_SMOKE][0];
		pv[13] = particlest[P_SMOKE][3];

		pv[21] = particlest[P_SMOKE][2];
		pv[22] = particlest[P_SMOKE][3];

		pv[30] = particlest[P_SMOKE][2];
		pv[31] = particlest[P_SMOKE][1];

		// rgba
		pv[5] = pv[14] = pv[23] = pv[32] = ParticleRGBTable[(int) sp->color][0];
		pv[6] = pv[15] = pv[24] = pv[33] = ParticleRGBTable[(int) sp->color][1];
		pv[7] = pv[16] = pv[25] = pv[34] = ParticleRGBTable[(int) sp->color][2];
		pv[8] = pv[17] = pv[26] = pv[35] = sp->renderalpha;

		pv += (4 * PARTICLE_VERTEX_SIZE);
		PrimitiveCount += 4;

		if (PrimitiveCount >= MAX_PART_VERTS)
		{
			// draw this number of primitives
			glDrawArrays (GL_QUADS, 0, MAX_PART_VERTS);

			// reset the pointer and counter
			PrimitiveCount = 0;
			pv = VArrayVerts;
		}
	}

	// draw anything left over
	if (PrimitiveCount) glDrawArrays (GL_QUADS, 0, PrimitiveCount);
}


void R_RenderLavaSmoke (qboolean StateChange);

void R_ParticleStateChange (qboolean Init)
{
	if (Init)
	{
		glDepthMask (GL_FALSE);
		glEnable (GL_BLEND);

		glBindTexture (GL_TEXTURE_2D, r_particlefont->texnum);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glEnableClientState (GL_VERTEX_ARRAY);
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glEnableClientState (GL_COLOR_ARRAY);

		glVertexPointer (3, GL_FLOAT, sizeof (float) * PARTICLE_VERTEX_SIZE, &VArrayVerts[0]);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * PARTICLE_VERTEX_SIZE, &VArrayVerts[3]);
		glColorPointer (4, GL_FLOAT, sizeof (float) * PARTICLE_VERTEX_SIZE, &VArrayVerts[5]);
	}
	else
	{
		glDisableClientState (GL_VERTEX_ARRAY);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		glDisableClientState (GL_COLOR_ARRAY);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glColor4f (1, 1, 1, 1);

		glDisable (GL_BLEND);
		glDepthMask (GL_TRUE);
	}
}


// seperated out cos we need the updated positions for the stencil haze state
void R_SetSmokeFrameState (void)
{
	// always update even if there's nothing to draw
	R_UpdateSmokeParticles (ParticleFrameTime);
}


void R_RenderParticleSprites (void)
{
	rparticle_t		*p, *next;
	float			alpha;
	float			time, time2;
	rparticle_t		*active, *tail;

	active = NULL;
	tail = NULL;

	// send through this to get a valid colour table
	R_RenderParticleSpriteChain (NULL, 0, 0);

	// nothing to draw
	if (!ActiveParticles)
	{
		R_RenderLavaSmoke (true);
		return;
	}

	PDrawChain[0] = PDrawChain[1] = NULL;

	for (p = ActiveParticles; p; p = next)
	{
		next = p->next;

		// INSTANT_PARTICLE handling for heat beam
		// (MH - and other 1 frame particles...)
		if (p->alphavel != R_INSTANT_PARTICLE) 
		{
			time = (ParticleCurrTime - p->time) * 0.001;
			alpha = p->alpha + time * p->alphavel;

			if (alpha <= 0)
			{
				// faded out
				p->next = FreeParticles;
				FreeParticles = p;
				continue;
			}
		}
		else
		{
			time = 0;
			alpha = p->alpha;
		}

		p->next = NULL;

		if (!tail)
			active = tail = p;
		else
		{
			tail->next = p;
			tail = p;
		}

		if (alpha > 1.0) alpha = 1;

		time2 = time * time;

		p->renderorg[0] = p->org[0] + p->vel[0] * time + p->accel[0] * time2;
		p->renderorg[1] = p->org[1] + p->vel[1] * time + p->accel[1] * time2;
		p->renderorg[2] = p->org[2] + p->vel[2] * time + p->accel[2] * time2;

		p->renderalpha = alpha;

		switch (p->type)
		{
		case P_SMOKE:
			// modulate scale and origin by time
			// simulate brownian motion
			p->renderorg[0] += (float) ((rand () & 2) - 1) / 2;
			p->renderorg[1] += (float) ((rand () & 2) - 1) / 2;
			p->renderorg[2] += (float) ((rand () & 2) - 1) / 2;

			p->renderscale = p->scale + rp_turbsin[Q_ftol (0.125 * ((p->renderorg[0] + r_newrefdef.time) * P_TURBSCALE)) & 255];

			p->chain = PDrawChain[1];
			PDrawChain[1] = p;
			break;

		case P_BLOOD1:
		case P_BLOOD2:
		case P_BLOOD3:
		case P_BLOOD4:
			// increase scale with time for blood particles
			p->renderscale = p->scale + time * 20;

			p->chain = PDrawChain[1];
			PDrawChain[1] = p;
			break;

		default:
			// default renderscale and chaining
			p->renderscale = p->scale;

			p->chain = PDrawChain[1];
			PDrawChain[1] = p;

			// handle unspecified particle types
			if (p->type < P_SMOKE || p->type > P_CINDER) p->type = P_DOT;
			break;
		}

		if (p->alphavel == R_INSTANT_PARTICLE) 
		{
			p->alphavel = 0.0;
			p->alpha = 0.0;
		}
	}

	ActiveParticles = active;

	R_ParticleStateChange (true);

	R_RenderParticleSpriteChain (PDrawChain[0], GL_SRC_ALPHA, GL_ONE);
	R_RenderParticleSpriteChain (PDrawChain[1], GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	R_RenderLavaSmoke (false);

	R_ParticleStateChange (false);
}

