



#include "quakedef.h"

gltexture_t *BloodTextures[4];
gltexture_t *MuzzleFlash;


typedef struct glnewpart_s
{
	vec3_t	org;
	vec3_t	activeorg;

	vec3_t	vel;
	vec3_t	activevel;

	float	alpha;
	float	activealpha;

	struct glnewpart_s *next;
} glnewpart_t;


typedef struct glparticleeffect_s
{
	// particles contained by this effect
	glnewpart_t *particles;

	// how long the effect has been running for
	float age;

	// how long the effect will last for.  if age > die then the effect is not rendered.
	float die;

	// texture used by the effect
	int texnum;

	// alpha fade
	float fade;

	// next instance of this effect type
	struct glparticleeffect_s *next;
} glparticleeffect_t;


// effects
glparticleeffect_t *MFlash = NULL;


glparticleeffect_t *AllocParticleEffect (int NumEffects, int ParticlesPerEffect, int texnum, float fade)
{
	glnewpart_t *p = NULL;
	glnewpart_t *Part = NULL;

	glparticleeffect_t *ef = NULL;
	glparticleeffect_t *Effect = NULL;

	int i;

	// create the particles.  all instances of each effect share the same particle list,
	// which contains master copies of the various particle parameters (which will never
	// change) as well as active copies which are updated as required at render time.
	for (i = 0; i < ParticlesPerEffect; i++)
	{
		Part = (glnewpart_t *) malloc (sizeof (glnewpart_t));
		Part->next = p;
		p = Part;
	}

	// create the effects
	for (i = 0; i < NumEffects; i++)
	{
		// create an instance of the effect
		Effect = (glparticleeffect_t *) malloc (sizeof (glparticleeffect_t *));

		// set particles
		Effect->particles = Part;

		// initial params
		Effect->age = 0;
		Effect->die = -1;
		Effect->fade = fade;

		// link in
		Effect->next = ef;
		ef = Effect;
	}

	return Effect;
}


glparticleeffect_t *FindEffect (glparticleeffect_t *base)
{
	glparticleeffect_t *ef;

	for (ef = base; ef; ef = ef->next)
		if (ef->die < 0)
			return ef;

	return NULL;
}


/*
=========================================================================================================

	MUZZLEFLASH

=========================================================================================================
*/

void R_MuzzleFlashEffect (vec3_t origin, int count)
{
	int i;
	glparticleeffect_t *ef = FindEffect (MFlash);
	glnewpart_t *p;

	if (!ef) return;

	// initial parameters
	ef->die = 0.05;

	for (i = 0, p = MFlash->particles; i < count && p; i++, p = p->next)
	{
	}
}

/*
=========================================================================================================

	SETUP

=========================================================================================================
*/


qboolean R_InitNewParticles (void)
{
	// get external textures
	if (!(BloodTextures[0] = GL_LoadTexture ("particles/blood_screen_1", 8, 8, NULL, (TEX_32BPP | TEX_MIPMAP)))) return false;
	if (!(BloodTextures[1] = GL_LoadTexture ("particles/blood_screen_2", 8, 8, NULL, (TEX_32BPP | TEX_MIPMAP)))) return false;
	if (!(BloodTextures[2] = GL_LoadTexture ("particles/blood_screen_3", 8, 8, NULL, (TEX_32BPP | TEX_MIPMAP)))) return false;
	if (!(BloodTextures[3] = GL_LoadTexture ("particles/blood_screen_4", 8, 8, NULL, (TEX_32BPP | TEX_MIPMAP)))) return false;
	if (!(MuzzleFlash = GL_LoadTexture ("particles/mflash", 8, 8, NULL, (TEX_32BPP)))) return false;

	// initialize effects
	//MFlash = AllocParticleEffect (2, 12, MuzzleFlash->texnum, 0.01);

	return true;
}


