/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sbar.c -- status bar code

#include "quakedef.h"


int			sb_updates;		// if >= vid.numpages, no update needed

#define STAT_MINUS		10	// num frame for '-' stats digit
qpic_t		*sb_nums[2][11];
qpic_t		*sb_colon, *sb_slash;

qpic_t      *sb_weapons[7][8];   // 0 is active, 1 is owned, 2-5 are flashes
qpic_t      *sb_ammo[4];
qpic_t		*sb_sigil[4];
qpic_t		*sb_armor[3];
qpic_t		*sb_items[32];

qpic_t	*sb_faces[7][2];		// 0 is gibbed, 1 is dead, 2-6 are alive
							// 0 is static, 1 is temporary animation
qpic_t	*sb_face_invis;
qpic_t	*sb_face_quad;
qpic_t	*sb_face_invuln;
qpic_t	*sb_face_invis_invuln;

qboolean	sb_showscores;

int			sb_lines;			// scan lines to draw

qpic_t      *rsb_weapons[5];
qpic_t      *rsb_items[2];
qpic_t      *rsb_ammo[3];

//MED 01/04/97 added two more weapons + 3 alternates for grenade launcher
qpic_t      *hsb_weapons[7][5];   // 0 is active, 1 is owned, 2-5 are flashes
//MED 01/04/97 added array to simplify weapon parsing
int         hipweapons[4] = {HIT_LASER_CANNON_BIT,HIT_MJOLNIR_BIT,4,HIT_PROXIMITY_GUN_BIT};
//MED 01/04/97 added hipnotic items array
qpic_t      *hsb_items[2];

void Sbar_MiniDeathmatchOverlay (void);
void Sbar_DeathmatchOverlay (void);
void M_DrawPic (int x, int y, qpic_t *pic);

// support up to 20 crosshairs
gltexture_t *sb_crosshairs[20];
extern	cvar_t	crosshair;

void Draw_Crosshair (int texnum);


/*
===============
Sbar_ShowScores

Tab key down
===============
*/
void Sbar_ShowScores (void)
{
	if (sb_showscores)
	{
		sb_showscores = false;
		sb_updates = 0;
	}
	else
	{
		sb_showscores = true;
		sb_updates = 0;
	}
}

/*
===============
Sbar_DontShowScores

Tab key up
===============
*/
void Sbar_DontShowScores (void)
{
}

/*
===============
Sbar_Changed
===============
*/
void Sbar_Changed (void)
{
	sb_updates = 0;	// update next frame
}

/*
===============
Sbar_Init
===============
*/
void Sbar_Init (void)
{
	int		i;

	for (i=0 ; i<10 ; i++)
	{
		sb_nums[0][i] = Draw_PicFromWad ("num_%i",i);
		sb_nums[1][i] = Draw_PicFromWad ("anum_%i",i);
	}

	sb_nums[0][10] = Draw_PicFromWad ("num_minus");
	sb_nums[1][10] = Draw_PicFromWad ("anum_minus");

	sb_colon = Draw_PicFromWad ("num_colon");
	sb_slash = Draw_PicFromWad ("num_slash");

	// hack the lightning gun width so its the same as the rest
	sb_weapons[0][0] = Draw_PicFromWad ("inv_shotgun");
	sb_weapons[0][1] = Draw_PicFromWad ("inv_sshotgun");
	sb_weapons[0][2] = Draw_PicFromWad ("inv_nailgun");
	sb_weapons[0][3] = Draw_PicFromWad ("inv_snailgun");
	sb_weapons[0][4] = Draw_PicFromWad ("inv_rlaunch");
	sb_weapons[0][5] = Draw_PicFromWad ("inv_srlaunch");
	sb_weapons[0][6] = Draw_PicFromWad ("inv_lightng");
	sb_weapons[0][6]->width = 24;

	sb_weapons[1][0] = Draw_PicFromWad ("inv2_shotgun");
	sb_weapons[1][1] = Draw_PicFromWad ("inv2_sshotgun");
	sb_weapons[1][2] = Draw_PicFromWad ("inv2_nailgun");
	sb_weapons[1][3] = Draw_PicFromWad ("inv2_snailgun");
	sb_weapons[1][4] = Draw_PicFromWad ("inv2_rlaunch");
	sb_weapons[1][5] = Draw_PicFromWad ("inv2_srlaunch");
	sb_weapons[1][6] = Draw_PicFromWad ("inv2_lightng");
	sb_weapons[1][6]->width = 24;

	for (i=0 ; i<5 ; i++)
	{
		sb_weapons[2+i][0] = Draw_PicFromWad ("inva%i_shotgun", i+1);
		sb_weapons[2+i][1] = Draw_PicFromWad ("inva%i_sshotgun",i+1);
		sb_weapons[2+i][2] = Draw_PicFromWad ("inva%i_nailgun",i+1);
		sb_weapons[2+i][3] = Draw_PicFromWad ("inva%i_snailgun",i+1);
		sb_weapons[2+i][4] = Draw_PicFromWad ("inva%i_rlaunch",i+1);
		sb_weapons[2+i][5] = Draw_PicFromWad ("inva%i_srlaunch",i+1);
		sb_weapons[2+i][6] = Draw_PicFromWad ("inva%i_lightng",i+1);
		sb_weapons[2+i][6]->width = 24;
	}

	sb_ammo[0] = Draw_PicFromWad ("sb_shells");
	sb_ammo[1] = Draw_PicFromWad ("sb_nails");
	sb_ammo[2] = Draw_PicFromWad ("sb_rocket");
	sb_ammo[3] = Draw_PicFromWad ("sb_cells");

	sb_armor[0] = Draw_PicFromWad ("sb_armor1");
	sb_armor[1] = Draw_PicFromWad ("sb_armor2");
	sb_armor[2] = Draw_PicFromWad ("sb_armor3");

	sb_items[0] = Draw_PicFromWad ("sb_key1");
	sb_items[1] = Draw_PicFromWad ("sb_key2");
	sb_items[2] = Draw_PicFromWad ("sb_invis");
	sb_items[3] = Draw_PicFromWad ("sb_invuln");
	sb_items[4] = Draw_PicFromWad ("sb_suit");
	sb_items[5] = Draw_PicFromWad ("sb_quad");

	sb_sigil[0] = Draw_PicFromWad ("sb_sigil1");
	sb_sigil[1] = Draw_PicFromWad ("sb_sigil2");
	sb_sigil[2] = Draw_PicFromWad ("sb_sigil3");
	sb_sigil[3] = Draw_PicFromWad ("sb_sigil4");

	sb_faces[4][0] = Draw_PicFromWad ("face1");
	sb_faces[4][1] = Draw_PicFromWad ("face_p1");
	sb_faces[3][0] = Draw_PicFromWad ("face2");
	sb_faces[3][1] = Draw_PicFromWad ("face_p2");
	sb_faces[2][0] = Draw_PicFromWad ("face3");
	sb_faces[2][1] = Draw_PicFromWad ("face_p3");
	sb_faces[1][0] = Draw_PicFromWad ("face4");
	sb_faces[1][1] = Draw_PicFromWad ("face_p4");
	sb_faces[0][0] = Draw_PicFromWad ("face5");
	sb_faces[0][1] = Draw_PicFromWad ("face_p5");

	sb_face_invis = Draw_PicFromWad ("face_invis");
	sb_face_invuln = Draw_PicFromWad ("face_invul2");
	sb_face_invis_invuln = Draw_PicFromWad ("face_inv2");
	sb_face_quad = Draw_PicFromWad ("face_quad");

	for (i = 0; i < 20; i++)
	{
		char chname[64];

		sprintf (chname, "wad/crosshair%i", i);

		sb_crosshairs[i] = GL_LoadTexture (chname, 32, 32, NULL, 0);
	}

	Cmd_AddCommand ("+showscores", Sbar_ShowScores);
	Cmd_AddCommand ("-showscores", Sbar_DontShowScores);

//MED 01/04/97 added new hipnotic weapons
	if (hipnotic)
	{
	  hsb_weapons[0][0] = Draw_PicFromWad ("inv_laser");
	  hsb_weapons[0][1] = Draw_PicFromWad ("inv_mjolnir");
	  hsb_weapons[0][2] = Draw_PicFromWad ("inv_gren_prox");
	  hsb_weapons[0][3] = Draw_PicFromWad ("inv_prox_gren");
	  hsb_weapons[0][4] = Draw_PicFromWad ("inv_prox");

	  hsb_weapons[1][0] = Draw_PicFromWad ("inv2_laser");
	  hsb_weapons[1][1] = Draw_PicFromWad ("inv2_mjolnir");
	  hsb_weapons[1][2] = Draw_PicFromWad ("inv2_gren_prox");
	  hsb_weapons[1][3] = Draw_PicFromWad ("inv2_prox_gren");
	  hsb_weapons[1][4] = Draw_PicFromWad ("inv2_prox");

	  for (i=0 ; i<5 ; i++)
	  {
		 hsb_weapons[2+i][0] = Draw_PicFromWad ("inva%i_laser",i+1);
		 hsb_weapons[2+i][1] = Draw_PicFromWad ("inva%i_mjolnir",i+1);
		 hsb_weapons[2+i][2] = Draw_PicFromWad ("inva%i_gren_prox",i+1);
		 hsb_weapons[2+i][3] = Draw_PicFromWad ("inva%i_prox_gren",i+1);
		 hsb_weapons[2+i][4] = Draw_PicFromWad ("inva%i_prox",i+1);
	  }

	  hsb_items[0] = Draw_PicFromWad ("sb_wsuit");
	  hsb_items[1] = Draw_PicFromWad ("sb_eshld");
	}

	if (rogue)
	{
		rsb_weapons[0] = Draw_PicFromWad ("r_lava");
		rsb_weapons[1] = Draw_PicFromWad ("r_superlava");
		rsb_weapons[2] = Draw_PicFromWad ("r_gren");
		rsb_weapons[3] = Draw_PicFromWad ("r_multirock");
		rsb_weapons[4] = Draw_PicFromWad ("r_plasma");

		rsb_items[0] = Draw_PicFromWad ("r_shield1");
        rsb_items[1] = Draw_PicFromWad ("r_agrav1");

		rsb_ammo[0] = Draw_PicFromWad ("r_ammolava");
		rsb_ammo[1] = Draw_PicFromWad ("r_ammomulti");
		rsb_ammo[2] = Draw_PicFromWad ("r_ammoplasma");
	}
}


//=============================================================================

// drawing routines are relative to the status bar location

/*
=============
Sbar_DrawPic
=============
*/
void Sbar_DrawPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x, (vid.height - y - SBAR_HEIGHT), pic);
}

/*
=============
Sbar_DrawTransPic
=============
*/
void Sbar_DrawTransPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x, (vid.height - y - SBAR_HEIGHT), pic);
}

/*
================
Sbar_DrawCharacter

Draws one solid graphics character
================
*/
void Sbar_DrawCharacter (int x, int y, int num)
{
	Draw_Character (x, (vid.height - y - SBAR_HEIGHT), num);
}

/*
================
Sbar_DrawString
================
*/
void Sbar_DrawString (int x, int y, char *str)
{
	Draw_String (x, (vid.height - y - SBAR_HEIGHT), str);
}

/*
=============
Sbar_itoa
=============
*/
int Sbar_itoa (int num, char *buf)
{
	char	*str;
	int		pow10;
	int		dig;

	str = buf;

	if (num < 0)
	{
		*str++ = '-';
		num = -num;
	}

	for (pow10 = 10 ; num >= pow10 ; pow10 *= 10)
	;

	do
	{
		pow10 /= 10;
		dig = num/pow10;
		*str++ = '0'+dig;
		num -= dig*pow10;
	} while (pow10 != 1);

	*str = 0;

	return str-buf;
}


/*
=============
Sbar_DrawNum
=============
*/
void Sbar_DrawNum (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				frame;

	// bound number so it doesn't overflow
	if (num > 666) num = 666;
	if (num < -99) num = -99;

	sprintf (str, "%i", num);
	ptr = str;

	if (num > 99)
		;	// 3 digits
	else if (num < -9)
		;	// 3 digits
	else if (num > 9)
		x += 24;	// 2 digits
	else if (num >= 0)
		x += 48;	// 1 digit
	else if (num < 0)
		x += 24;	// 1 digit

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Sbar_DrawPic (x, y, sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}

//=============================================================================

int		fragsort[MAX_SCOREBOARD];

char	scoreboardtext[MAX_SCOREBOARD][20];
int		scoreboardtop[MAX_SCOREBOARD];
int		scoreboardbottom[MAX_SCOREBOARD];
int		scoreboardcount[MAX_SCOREBOARD];
int		scoreboardlines;

/*
===============
Sbar_SortFrags
===============
*/
void Sbar_SortFrags (void)
{
	int		i, j, k;

// sort by frags
	scoreboardlines = 0;
	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.scores[i].name[0])
		{
			fragsort[scoreboardlines] = i;
			scoreboardlines++;
		}
	}

	for (i=0 ; i<scoreboardlines ; i++)
		for (j=0 ; j<scoreboardlines-1-i ; j++)
			if (cl.scores[fragsort[j]].frags < cl.scores[fragsort[j+1]].frags)
			{
				k = fragsort[j];
				fragsort[j] = fragsort[j+1];
				fragsort[j+1] = k;
			}
}

int	Sbar_ColorForMap (int m)
{
	return m < 128 ? m + 8 : m + 8;
}

/*
===============
Sbar_UpdateScoreboard
===============
*/
void Sbar_UpdateScoreboard (void)
{
	int		i, k;
	int		top, bottom;
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	memset (scoreboardtext, 0, sizeof(scoreboardtext));

	for (i=0 ; i<scoreboardlines; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		sprintf (&scoreboardtext[i][1], "%3i %s", s->frags, s->name);

		top = s->colors & 0xf0;
		bottom = (s->colors & 15) <<4;
		scoreboardtop[i] = Sbar_ColorForMap (top);
		scoreboardbottom[i] = Sbar_ColorForMap (bottom);
	}
}



/*
===============
Sbar_SoloScoreboard
===============
*/
void Sbar_SoloScoreboard (void)
{
	char str[80];
	int minutes, seconds, tens, units;
	int l;
	int i;
	int SBX;
	int SBY;

	// calculate time
	minutes = cl.time / 60;
	seconds = cl.time - 60 * minutes;
	tens = seconds / 10;
	units = seconds - 10 * tens;

	// base X
	SBX = vid.width / 2;

	// starting Y
	SBY = vid.height * 0.25;

	// level name
	l = strlen (cl.levelname);
	Draw_String (SBX - l * 4, SBY, cl.levelname);

	for (i = 0; i < l; i++)
	{
		if (!i)
			str[i] = 32;
		else if (i == 1)
			str[i] = 29;
		else if (i == l - 2)
			str[i] = 31;
		else if (i == l - 1)
			str[i] = 32;
		else str[i] = 30;
	}

	str[l] = 0;
	SBY += 12;
	Draw_String (SBX - l * 4, SBY, str);

	// kill count
	sprintf (str, "Monsters: %i/%i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
	l = strlen (str);
	SBY += 12;
	Draw_String (SBX - l * 4, SBY, str);

	// secrets count
	sprintf (str, "Secrets: %i/%i", cl.stats[STAT_SECRETS], cl.stats[STAT_TOTALSECRETS]);
	l = strlen (str);
	SBY += 12;
	Draw_String (SBX - l * 4, SBY, str);

	// time elapsed
	sprintf (str, "Time: %i:%i%i", minutes, tens, units);
	l = strlen (str);
	SBY += 12;
	Draw_String (SBX - l * 4, SBY, str);

	// skill
	switch ((int) skill.value)
	{
	case 0:
		strcpy (str, "Skill: Easy");
		break;

	case 1:
		strcpy (str, "Skill: Normal");
		break;

	case 2:
		strcpy (str, "Skill: Hard");
		break;

	case 3:
		strcpy (str, "Skill: NIGHTMARE");
		break;
	}

	l = strlen (str);
	SBY += 12;
	Draw_String (SBX - l * 4, SBY, str);
}

/*
===============
Sbar_DrawScoreboard
===============
*/
void Sbar_DrawScoreboard (void)
{
	Sbar_SoloScoreboard ();
	if (cl.gametype == GAME_DEATHMATCH)
		Sbar_DeathmatchOverlay ();
}

//=============================================================================


void Sbar_DrawAmmoCount (int x, int y, int count)
{
	char	num[6];

	sprintf (num, "%3i", cl.stats[count]);

	if (num[0] != ' ') Draw_Character (vid.width - x + 0, y, 18 + num[0] - '0');
	if (num[1] != ' ') Draw_Character (vid.width - x + 8, y, 18 + num[1] - '0');
	if (num[2] != ' ') Draw_Character (vid.width - x + 16, y, 18 + num[2] - '0');
}


/*
===============
Sbar_DrawInventory
===============
*/
void Sbar_DrawInventory (void)
{
	int		i;
	float	time;
	int		flashon;
	int		itemloc;
	int		firstitem;

	// weaponses
	for (i = 0; i < 7; i++)
	{
		if (cl.items & (IT_SHOTGUN << i))
		{
			time = cl.item_gettime[i];

			flashon = (int) ((cl.time - time) * 10);

			if (flashon >= 10)
			{
				if (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN << i))
					flashon = 1;
				else
					flashon = 0;
			}
			else flashon = (flashon % 5) + 2;

			Draw_Pic (vid.width - 40 - sb_weapons[flashon][i]->width, 8 + (i * 17), sb_weapons[flashon][i]);

			// force update to remove flash
			if (flashon > 1) sb_updates = 0;
		}
	}

	// hipnotic weaponses
	if (hipnotic)
	{
		int grenadeflashing = 0;

		for (i = 0; i < 4; i++)
		{
			if (cl.items & (1 << hipweapons[i]))
			{
				time = cl.item_gettime[hipweapons[i]];
				flashon = (int) ((cl.time - time) * 10);

				if (flashon >= 10)
				{
					if (cl.stats[STAT_ACTIVEWEAPON] == (1 << hipweapons[i]))
						flashon = 1;
					else flashon = 0;
				}
				else flashon = (flashon % 5) + 2;

				// check grenade launcher
				if (i == 2)
				{
					if (cl.items & HIT_PROXIMITY_GUN)
					{
						if (flashon)
						{
							grenadeflashing = 1;
							Draw_Pic (vid.width - 40 - hsb_weapons[flashon][2]->width, 76, hsb_weapons[flashon][2]);
						}
					}
				}
				else if (i == 3)
				{
					if (cl.items & (IT_SHOTGUN << 4))
					{
						if (flashon && !grenadeflashing)
						{
							Draw_Pic (vid.width - 40 - hsb_weapons[flashon][3]->width, 76, hsb_weapons[flashon][3]);
						}
						else if (!grenadeflashing)
						{
							Draw_Pic (vid.width - 40 - hsb_weapons[0][3]->width, 76, hsb_weapons[0][3]);
						}
					}
					else Draw_Pic (vid.width - 40 - hsb_weapons[flashon][4]->width, 76, hsb_weapons[flashon][4]);
				}
				else Draw_Pic (vid.width - 40 - hsb_weapons[flashon][i]->width, 127 + (i * 17), hsb_weapons[flashon][i]);

				// force update to remove flash
				if (flashon > 1) sb_updates = 0;
			}
		}
	}

	if (rogue)
	{
		// check for powered up weapon.
		if (cl.stats[STAT_ACTIVEWEAPON] >= RIT_LAVA_NAILGUN)
		{
			for (i = 0; i < 5; i++)
			{
				if (cl.stats[STAT_ACTIVEWEAPON] == (RIT_LAVA_NAILGUN << i))
				{
					Draw_Pic (vid.width - 40 - rsb_weapons[i]->width, 42 + (i * 17), rsb_weapons[i]);
				}
			}
		}
	}

	// ammo counts
	// shells
	Sbar_DrawAmmoCount (32, 10, STAT_SHELLS);
	Sbar_DrawAmmoCount (32, 27, STAT_SHELLS);

	// nails
	Sbar_DrawAmmoCount (32, 44, STAT_NAILS);
	Sbar_DrawAmmoCount (32, 61, STAT_NAILS);

	// rockets
	Sbar_DrawAmmoCount (32, 78, STAT_ROCKETS);
	Sbar_DrawAmmoCount (32, 95, STAT_ROCKETS);

	// cells
	Sbar_DrawAmmoCount (32, 112, STAT_CELLS);

	if (hipnotic)
	{
		Sbar_DrawAmmoCount (32, 129, STAT_CELLS);
		Sbar_DrawAmmoCount (32, 146, STAT_CELLS);
	}

	flashon = 0;
	itemloc = 32;

	// sigilses
	if (!hipnotic && !rogue)
	{
		qboolean hassigilses = false;

		for (i = 0; i < 4; i++)
		{
			if (cl.items & (1 << (28 + i)))
			{
				Sbar_DrawPic (vid.width - itemloc, 36, sb_sigil[i]);
				itemloc += 12;
				hassigilses = true;
			}
		}

		if (hassigilses) itemloc += 8;
	}

	// keyses
	// have to be done differently because hipnotic messes things up (yet again - sigh)
	if (hipnotic)
	{
		if (cl.items & IT_KEY1) Sbar_DrawPic (vid.width - itemloc, 28, sb_items[0]);
		if (cl.items & IT_KEY2) Sbar_DrawPic (vid.width - itemloc, 38, sb_items[1]);

		if ((cl.items & IT_KEY1) || (cl.items & IT_KEY2)) itemloc += 20;

		// starts at 2 because 0 and 1 are the keys
		firstitem = 2;
	}
	else firstitem = 0;

	// itemses
	for (i = firstitem; i < 6; i++)
	{
		if (cl.items & (1 << (17 + i)))
		{
			Sbar_DrawPic (vid.width - itemloc, 36, sb_items[i]);
			itemloc += 20;
		}
	}

	// hipnotic itemses
	if (hipnotic)
	{
		for (i = 0; i < 2; i++)
		{
			if (cl.items & (1 << (24 + i)))
			{
				Sbar_DrawPic (vid.width - itemloc, 36, hsb_items[i]);
				itemloc += 20;
			}
		}
	}

	// rogue itemses
	if (rogue)
	{
		for (i = 0; i < 2; i++)
		{
			if (cl.items & (1 << (29 + i)))
			{
				Sbar_DrawPic (vid.width - itemloc, 36, rsb_items[i]);
				itemloc += 20;
			}
		}
	}
}

//=============================================================================

/*
===============
Sbar_DrawFrags
===============
*/
void Sbar_DrawFrags (void)
{
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	int				xofs;
	char			num[12];
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines <= 4 ? scoreboardlines : 4;

	x = 23;
	if (cl.gametype == GAME_DEATHMATCH)
		xofs = 0;
	else
		xofs = (vid.width - 320)>>1;
	y = vid.height - SBAR_HEIGHT - 23;

	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill (xofs + x*8 + 10, y, 28, 4, top);
		Draw_Fill (xofs + x*8 + 10, y+4, 28, 3, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Sbar_DrawCharacter ( (x+1)*8 , -24, num[0]);
		Sbar_DrawCharacter ( (x+2)*8 , -24, num[1]);
		Sbar_DrawCharacter ( (x+3)*8 , -24, num[2]);

		if (k == cl.viewentity - 1)
		{
			Sbar_DrawCharacter (x*8+2, -24, 16);
			Sbar_DrawCharacter ( (x+4)*8-4, -24, 17);
		}
		x+=4;
	}
}

//=============================================================================


/*
===============
Sbar_DrawFace
===============
*/
void Sbar_DrawFace (void)
{
	int		f, anim;

	if ( (cl.items & (IT_INVISIBILITY | IT_INVULNERABILITY) ) == (IT_INVISIBILITY | IT_INVULNERABILITY) )
	{
		Sbar_DrawPic (8, 8, sb_face_invis_invuln);
		return;
	}
	if (cl.items & IT_QUAD)
	{
		Sbar_DrawPic (8, 8, sb_face_quad);
		return;
	}
	if (cl.items & IT_INVISIBILITY)
	{
		Sbar_DrawPic (8, 8, sb_face_invis);
		return;
	}
	if (cl.items & IT_INVULNERABILITY)
	{
		Sbar_DrawPic (8, 8, sb_face_invuln);
		return;
	}

	if (cl.stats[STAT_HEALTH] >= 100)
		f = 4;
	else if (cl.stats[STAT_HEALTH] <= 0)
		f = 0;
	else f = cl.stats[STAT_HEALTH] / 20;

	if (cl.time <= cl.faceanimtime)
	{
		anim = 1;
		sb_updates = 0;		// make sure the anim gets drawn over
	}
	else
		anim = 0;

	Sbar_DrawPic (8, 8, sb_faces[f][anim]);
}

/*
===============
Sbar_Draw
===============
*/
void Sbar_Draw (void)
{
	qboolean hasammo = true;

	// console is full screen
	if (scr_con_current == vid.height) return;

	// no sbar in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	scr_copyeverything = 1;

	sb_updates++;

	// viewsize 120 = no sbar
	if (scr_viewsize.value > 110) return;

	if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
	{
		Sbar_DrawScoreboard ();
		sb_updates = 0;
	}

	// viewsize 100 = include inventory
	if (scr_viewsize.value < 110)
	{
		Sbar_DrawInventory ();

		if (cl.maxclients != 1)
		Sbar_DrawFrags ();
	}

	// armor
	if (cl.items & IT_INVULNERABILITY)
	{
		Sbar_DrawNum (24, 0, 666, 3, 1);
		Sbar_DrawPic (0, 0, draw_disc);
	}
	else
	{
		int armx = 8;
		int army = 40;

		if (rogue && cl.stats[STAT_ARMOR])
		{
			Sbar_DrawNum (armx + 32, army, cl.stats[STAT_ARMOR], 3, cl.stats[STAT_ARMOR] <= 25);

			if (cl.items & RIT_ARMOR3) Sbar_DrawPic (armx, army, sb_armor[2]);
			else if (cl.items & RIT_ARMOR2) Sbar_DrawPic (armx, army, sb_armor[1]);
			else if (cl.items & RIT_ARMOR1) Sbar_DrawPic (armx, army, sb_armor[0]);
		}
		else if (cl.stats[STAT_ARMOR])
		{
			Sbar_DrawNum (armx + 32, army, cl.stats[STAT_ARMOR], 3, cl.stats[STAT_ARMOR] <= 25);

			if (cl.items & IT_ARMOR3) Sbar_DrawPic (armx, army, sb_armor[2]);
			else if (cl.items & IT_ARMOR2) Sbar_DrawPic (armx, army, sb_armor[1]);
			else if (cl.items & IT_ARMOR1) Sbar_DrawPic (armx, army, sb_armor[0]);
		}
	}

	// face
	Sbar_DrawFace ();

	// health
	Sbar_DrawNum (40, 8, cl.stats[STAT_HEALTH], 3, cl.stats[STAT_HEALTH] <= 25);

	// ammo icon
	if (rogue)
	{
		if (cl.items & RIT_SHELLS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[0]);
		else if (cl.items & RIT_NAILS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[1]);
		else if (cl.items & RIT_ROCKETS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[2]);
		else if (cl.items & RIT_CELLS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[3]);
		else if (cl.items & RIT_LAVA_NAILS)
			Sbar_DrawPic (vid.width - 32, 8, rsb_ammo[0]);
		else if (cl.items & RIT_PLASMA_AMMO)
			Sbar_DrawPic (vid.width - 32, 8, rsb_ammo[1]);
		else if (cl.items & RIT_MULTI_ROCKETS)
			Sbar_DrawPic (vid.width - 32, 8, rsb_ammo[2]);
		else hasammo = false;
	}
	else
	{
		if (cl.items & IT_SHELLS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[0]);
		else if (cl.items & IT_NAILS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[1]);
		else if (cl.items & IT_ROCKETS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[2]);
		else if (cl.items & IT_CELLS)
			Sbar_DrawPic (vid.width - 32, 8, sb_ammo[3]);
		else hasammo = false;
	}

	if (hasammo)
	{
		// ammo number
		Sbar_DrawNum (vid.width - 112, 8, cl.stats[STAT_AMMO], 3, cl.stats[STAT_AMMO] <= 10);

		// crosshair should go here too...
		if (crosshair.value >= 1)
		{
			if (!sb_crosshairs[0] && crosshair.value)
				Draw_Character (vid.width / 2 - 4, vid.height / 2 + 4, '+');
			else
			{
				int cnum = crosshair.value;

				while (cnum >= 20) cnum--;

				while (cnum)
				{
					if (sb_crosshairs[cnum]) break;

					cnum--;
				}

				if (cnum) Draw_Crosshair (sb_crosshairs[cnum]->texnum);
			}
		}
	}

	if (cl.gametype == GAME_DEATHMATCH) Sbar_MiniDeathmatchOverlay ();
}

//=============================================================================

/*
==================
Sbar_IntermissionNumber

==================
*/
void Sbar_IntermissionNumber (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Draw_TransPic (x,y,sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_DeathmatchOverlay (void)
{
	qpic_t			*pic;
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	pic = Draw_CachePic ("gfx/ranking.lmp");
	M_DrawPic ((320-pic->width)/2, 8, pic);

// scores
	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines;

	x = 80 + ((vid.width - 320)>>1);
	y = 40;
	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill ( x, y, 40, 4, top);
		Draw_Fill ( x, y+4, 40, 4, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Draw_Character ( x+8 , y, num[0]);
		Draw_Character ( x+16 , y, num[1]);
		Draw_Character ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1)
			Draw_Character ( x - 8, y, 12);

#if 0
{
	int				total;
	int				n, minutes, tens, units;

	// draw time
		total = cl.completed_time - s->entertime;
		minutes = (int)total/60;
		n = total - minutes*60;
		tens = n/10;
		units = n%10;

		sprintf (num, "%3i:%i%i", minutes, tens, units);

		Draw_String ( x+48 , y, num);
}
#endif

	// draw name
		Draw_String (x+64, y, s->name);

		y += 10;
	}
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_MiniDeathmatchOverlay (void)
{
	int				i, k;
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;

	// scores
	Sbar_SortFrags ();

	// draw the text
	y = vid.height - 80;
	x = 8;

	for (i = 0; i < scoreboardlines && y > 12; i++)
	{
		// get current entry
		k = fragsort[i];
		s = &cl.scores[k];

		// no name!
		if (!s->name[0]) continue;

		// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15) << 4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill (x, y + 1, 40, 3, top);
		Draw_Fill (x, y + 4, 40, 4, bottom);

		// draw number
		f = s->frags;
		sprintf (num, "%3i", f);

		Draw_Character (x + 8, y, num[0]);
		Draw_Character (x + 16, y, num[1]);
		Draw_Character (x + 24, y, num[2]);

		// mark us
		if (k == cl.viewentity - 1)
		{
			Draw_Character (x, y, 16);
			Draw_Character (x + 32, y, 17);
		}

		// draw name
		Draw_String (x + 48, y, s->name);

		y -= 12;
	}}

/*
==================
Sbar_IntermissionOverlay

==================
*/
void Sbar_IntermissionOverlay (void)
{
	qpic_t	*pic;
	int		dig;
	int		num;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	if (cl.gametype == GAME_DEATHMATCH)
	{
		Sbar_DeathmatchOverlay ();
		return;
	}

	pic = Draw_CachePic ("gfx/complete.lmp");
	Draw_Pic (64, 24, pic);

	pic = Draw_CachePic ("gfx/inter.lmp");
	Draw_TransPic (0, 56, pic);

// time
	dig = cl.completed_time/60;
	Sbar_IntermissionNumber (160, 64, dig, 3, 0);
	num = cl.completed_time - dig*60;
	Draw_TransPic (234,64,sb_colon);
	Draw_TransPic (246,64,sb_nums[0][num/10]);
	Draw_TransPic (266,64,sb_nums[0][num%10]);

	Sbar_IntermissionNumber (160, 104, cl.stats[STAT_SECRETS], 3, 0);
	Draw_TransPic (232,104,sb_slash);
	Sbar_IntermissionNumber (240, 104, cl.stats[STAT_TOTALSECRETS], 3, 0);

	Sbar_IntermissionNumber (160, 144, cl.stats[STAT_MONSTERS], 3, 0);
	Draw_TransPic (232,144,sb_slash);
	Sbar_IntermissionNumber (240, 144, cl.stats[STAT_TOTALMONSTERS], 3, 0);

}


/*
==================
Sbar_FinaleOverlay

==================
*/
void Sbar_FinaleOverlay (void)
{
	qpic_t	*pic;

	scr_copyeverything = 1;

	pic = Draw_CachePic ("gfx/finale.lmp");
	Draw_TransPic ( (vid.width-pic->width)/2, 16, pic);
}
