
#include "quakedef.h"


typedef struct lightent_s
{
	int intensity;
	float colour[3];
	int style;
	float origin[3];
	struct mleaf_s *lightleaf;

	// contains all surfaces which have been hit by this light
	msurface_t **lightsurfaces;
	int numlightsurfaces;

	struct lightent_s *next;
} lightent_t;

lightent_t *lightents = NULL;


int numlights;


/*
========================================================================================================================

				LIGHT LOADING FROM ENTITIES

========================================================================================================================
*/

void StringToVector (char *TheString, float *TheVector)
{
	char *v, *w;
	int i;

	v = TheString;
	w = TheString;

	for (i = 0; i < 3; i++)
	{
		while (*v && *v != ' ')
			v++;

		*v = 0;

		TheVector[i] = atof (w);
		w = v = v + 1;
	}
}


char *R_ParseEntity (char *entities)
{
	char classname[64];
	char origin[64];
	char light[64];
	char style[64];
	char key[64];
	lightent_t *le;

	classname[0] = 0;
	origin[0] = 0;
	style[0] = 0;

	for (;;)
	{
		entities = COM_Parse (entities);

		if (com_token[0] == '}') break;

		strcpy (key, com_token);

		entities = COM_Parse (entities);

		// save out the fields we want to use
		if (!strnicmp (key, "classname", 9)) strcpy (classname, com_token);
		if (!strnicmp (key, "origin", 6)) strcpy (origin, com_token);
		if (!strnicmp (key, "style", 5)) strcpy (style, com_token);
		if (!strnicmp (key, "light", 5)) strcpy (light, com_token);
	}

	// check for a light
	if (!strnicmp (classname, "light", 5))
	{
		// spawn a new light ent
		le = (lightent_t *) malloc (sizeof (lightent_t));
		le->next = lightents;
		lightents = le;

		// fill in the light entity
		StringToVector (origin, le->origin);
		le->intensity = atoi (light);
		le->style = atoi (style);
		le->lightleaf = Mod_PointInLeaf (le->origin, cl.worldmodel);

		// set default intensity (matches light.exe default)
		if (!le->intensity) le->intensity = 300;

		if (!le->style)
		{
			// hack in a style based on the light name
			if (!strnicmp (classname, "light_flame", 11)) le->style = 1;
			if (!strnicmp (classname, "light_torch", 11)) le->style = 1;
		}

		// these will get filled in later
		le->lightsurfaces = NULL;

		// hack in colour for certain light types
		if (!strnicmp (classname, "light_flame", 11))
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else if (!strnicmp (classname, "light_torch", 11))
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else if (le->style == 1 || le->style == 6 || le->style == 3 || le->style == 7 || le->style == 8)
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else
		{
			le->colour[0] = -1.5f;
			le->colour[1] = -1.0f;
			le->colour[2] = -0.5f;
		}

		// increment the light counter
		numlights++;
	}

	return entities;
}


// the entities have already been validated in PR_Edict.c, so we'll be loose on the error checking
void R_ReadLights (void)
{
	char *entities = cl.worldmodel->entities;
	int i;
	lightent_t *le;

	if (lightents)
	{
		for (i = 0; ;)
		{
			le = lightents->next;
			if (lightents->lightsurfaces) free (lightents->lightsurfaces);
			free (lightents);
			lightents = le;

			i++;

			if (!lightents) break;
		}

		Con_DPrintf ("released %i lights\n", i);
	}

	numlights = 0;

	for (;;)
	{
		entities = COM_Parse (entities);
		if (!entities) break;

		entities = R_ParseEntity (entities);
	}

	Con_DPrintf ("read %i lights\n", numlights);
}


/*
========================================================================================================================

				LIGHT MARKING (Most code adapted from light.exe)

========================================================================================================================
*/


qboolean SV_RecursiveHullCheck (hull_t *hull, int num, float p1f, float p2f, vec3_t p1, vec3_t p2, trace_t *trace);

qboolean TestLine (vec3_t start, vec3_t end, float fudgefactor)
{
	trace_t	trace;
	vec3_t impact;

	memset (&trace, 0, sizeof(trace));
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, start, end, &trace);

	VectorCopy (trace.endpos, impact);

	// allow a bit of fudgery for midpoints that may be reachable
	if (impact[0] > end[0] + fudgefactor) return false;
	if (impact[1] > end[1] + fudgefactor) return false;
	if (impact[2] > end[2] + fudgefactor) return false;
	if (impact[0] < end[0] - fudgefactor) return false;
	if (impact[1] < end[1] - fudgefactor) return false;
	if (impact[2] < end[2] - fudgefactor) return false;

	// hit
	return true;
}


/*
============
CastRay

Returns the distance between the points, or -1 if blocked
=============
*/
vec_t CastRay (vec3_t p1, vec3_t p2, float extents)
{
	int		i;
	vec_t	t;
	qboolean	trace;

	trace = TestLine (p1, p2, extents);

	// blocked
	if (!trace) return -1;

	t = 0;

	for (i = 0; i < 3; i++)
		t += (p2[i] - p1[i]) * (p2[i] - p1[i]);

	if (t < 1) t = 1;

	return sqrt (t);
}


qboolean R_TestLightFace (msurface_t *surf, lightent_t *le)
{
	int i;
	int v;
	vec3_t facenormal;
	vec_t facedist;
	float dist;
	vec3_t surfmins = {9999999, 9999999, 9999999};
	vec3_t surfmaxs = {-9999999, -9999999, -9999999};
	float extents;
	qboolean inrange;

	// these surface types are not lit
	if (surf->flags & SURF_DRAWTURB) return false;
	if (surf->flags & SURF_DRAWSKY) return false;

	// paranoia!
	if (!surf->polys) return false;

	// set up the normal
	facenormal[0] = surf->plane->normal[0];
	facenormal[1] = surf->plane->normal[1];
	facenormal[2] = surf->plane->normal[2];
	facedist = surf->plane->dist;

	// check which side it's on
	if (surf->flags & SURF_PLANEBACK)
	{
		facenormal[0] *= -1;
		facenormal[1] *= -1;
		facenormal[2] *= -1;
		facedist *= -1;
	}

	// check distance to light
	dist = DotProduct (le->origin, facenormal) - facedist;

	// wrong side
	if (dist <= 0) return false;

	// too far away
	if (dist > le->intensity) return false;

	inrange = false;

	// calculate surf extents in world space
	for (v = 0; v < surf->polys->numverts; v++)
	{
		// fixme - save these to the msurface_t struct at poly allocation time
		if (surf->polys->verts[v].v[0] < surfmins[0]) surfmins[0] = surf->polys->verts[v].v[0];
		if (surf->polys->verts[v].v[1] < surfmins[1]) surfmins[1] = surf->polys->verts[v].v[1];
		if (surf->polys->verts[v].v[2] < surfmins[2]) surfmins[2] = surf->polys->verts[v].v[2];

		if (surf->polys->verts[v].v[0] > surfmaxs[0]) surfmaxs[0] = surf->polys->verts[v].v[0];
		if (surf->polys->verts[v].v[1] > surfmaxs[1]) surfmaxs[1] = surf->polys->verts[v].v[1];
		if (surf->polys->verts[v].v[2] > surfmaxs[2]) surfmaxs[2] = surf->polys->verts[v].v[2];

		// test texcoord range
		for (i = 0; i < 3; i++)
		{
			extents = ((surf->polys->verts[v].v[i] - le->origin[i]) / le->intensity) / 2.0f + 0.5f;

			if (extents >= 0.0f && extents <= 1.0f) inrange = true;
		}
	}

	// out of texcoord range
	if (!inrange) return false;

	// accumulate into the colour
	if (surf->texinfo->texture->gl_texture->luma && le->colour[0] < 0.015)
	{
		le->colour[0] += (surf->texinfo->texture->gl_texture->colour[0] * le->intensity / dist);
		le->colour[1] += (surf->texinfo->texture->gl_texture->colour[1] * le->intensity / dist);
		le->colour[2] += (surf->texinfo->texture->gl_texture->colour[2] * le->intensity / dist);
	}

	// between the combination of visibility and side checking we now have a reasonably valid hit
	return true;

	/*
	extents = 0;

	if (surfmaxs[0] - surfmins[0] > extents) extents = surfmaxs[0] - surfmins[0];
	if (surfmaxs[1] - surfmins[1] > extents) extents = surfmaxs[1] - surfmins[1];
	if (surfmaxs[2] - surfmins[2] > extents) extents = surfmaxs[2] - surfmins[2];

	// attempt to trace direct from the light to the surf
	dist = CastRay (le->origin, surf->polys->midpoint, extents);

	// light doesn't reach the midpoint.  we must try all the verts now to ensure that we're really OK to throw it away
	if (dist < 0)
	{
		vec3_t avg;

		for (v = 0; v < surf->polys->numverts; v++)
		{
			dist = CastRay (le->origin, surf->polys->verts[v].v, extents);

			if (dist >= 0) break;

			avg[0] = (surf->polys->midpoint[0] + surf->polys->verts[v].v[0]) / 2.0f;
			avg[1] = (surf->polys->midpoint[1] + surf->polys->verts[v].v[1]) / 2.0f;
			avg[2] = (surf->polys->midpoint[2] + surf->polys->verts[v].v[2]) / 2.0f;

			dist = CastRay (le->origin, avg, extents);

			if (dist >= 0) break;

			// take another average with the new average and the midpoint to be more precise
			avg[0] = (avg[0] + surf->polys->midpoint[0]) / 2.0f;
			avg[1] = (avg[1] + surf->polys->midpoint[1]) / 2.0f;
			avg[2] = (avg[2] + surf->polys->midpoint[2]) / 2.0f;

			dist = CastRay (le->origin, avg, extents);

			if (dist >= 0) break;

			// take another average with the new average and the midpoint to be more precise
			avg[0] = (avg[0] + surf->polys->verts[v].v[0]) / 2.0f;
			avg[1] = (avg[1] + surf->polys->verts[v].v[1]) / 2.0f;
			avg[2] = (avg[2] + surf->polys->verts[v].v[2]) / 2.0f;

			dist = CastRay (le->origin, avg, extents);

			if (dist >= 0) break;
		}
	}

	// still didn't reach
	if (dist < 0) return false;

	// nothing to add
	//if ((float) le->intensity - dist < 0) return false;

	return true;
	*/
}


int maxpass = 0;

void R_MarkLightSetup (lightent_t *le, int lnum, mnode_t *node)
{
	int i;
	float dist;
	msurface_t *surf;

	if (node->contents < 0) return;
	if (node->lightvisframe != -lnum) return;

	dist = DotProduct (le->origin, node->plane->normal) - node->plane->dist;

	if (dist > le->intensity)
	{
		R_MarkLightSetup (le, lnum, node->children[0]);
		return;
	}

	if (dist < -le->intensity)
	{
		R_MarkLightSetup (le, lnum, node->children[1]);
		return;
	}

	// mark the polygons
	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		if (!R_TestLightFace (surf, le)) continue;

		// OK, we got us a hit, people
		// mark the surf and increment the number of surfs used
		le->numlightsurfaces++;
		surf->visframe = lnum;
		surf->passframe++;

		if (surf->passframe > maxpass) maxpass = surf->passframe;
	}

	R_MarkLightSetup (le, lnum, node->children[0]);
	R_MarkLightSetup (le, lnum, node->children[1]);
}


void R_LightVisibility (int lnum, lightent_t *light)
{
	int i;
	byte *vis;
	mnode_t *node;

	vis = Mod_LeafPVS (light->lightleaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->lightvisframe == -lnum) break;

				node->lightvisframe = -lnum;
				node = node->parent;
			} while (node);
		}
	}

}


void R_SetupLighting (void)
{
	msurface_t *surf;
	lightent_t *le;
	int i;
	int j;
	int numsurfs;
	float dist;

	R_ReadLights ();

	maxpass = 0;

	// set initial state on the surfaces
	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		// not visible to the light
		surf->visframe = -1;
	}

	for (le = lightents, j = 0; le; le = le->next, j++)
	{
		// no surfs use this light yet
		le->numlightsurfaces = 0;

		if (le->colour[0] < 0)
		{
			le->colour[0] = 0.01;
			le->colour[1] = 0.01;
			le->colour[2] = 0.01;
		}

		R_LightVisibility (j, le);
		R_MarkLightSetup (le, j, cl.worldmodel->nodes);

		{
			le->colour[0] = 0.01;
			le->colour[1] = 0.01;
			le->colour[2] = 0.01;
		}

		VectorNormalize (le->colour);

		le->colour[0] *= 1.5;
		le->colour[1] *= 1.5;
		le->colour[2] *= 1.5;

		// no surfs use this light at all!  (this should rarely happen)
		if (!le->numlightsurfaces) continue;

		// set up surface pointers
		le->lightsurfaces = malloc (le->numlightsurfaces * sizeof (msurface_t *));

		// copy them in
		for (i = 0, numsurfs = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
		{
			if (surf->visframe != j) continue;

			// copy in
			le->lightsurfaces[numsurfs++] = surf;
		}
	}

#if 0
	// overdraw reduction - only the 3 lights closest to a surface can light it
	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
#define NUMCLOSEST 1
		lightent_t *closest[NUMCLOSEST];
		int numclosest = 0;
		int k;

		// these surface types are not lit
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// paranoia!
		if (!surf->polys) continue;

		// find the 3 closest lights
		for (k = 0; k < NUMCLOSEST; k++)
		{
			float closestdist = 9999999;

			closest[k] = NULL;

			for (le = lightents; le; le = le->next)
			{
				// test if this is used
				for (j = 0; j < NUMCLOSEST; j++)
					if (le == closest[j])
						break;

				// already used
				if (j != NUMCLOSEST) continue;

				for (j = 0; j < le->numlightsurfaces; j++)
				{
					// surf not used by this light
					if (surf != le->lightsurfaces[j]) continue;

					dist = fabs (DotProduct (le->origin, surf->plane->normal) - surf->plane->dist);

					if (dist < closestdist)
					{
						closest[k] = le;
						closestdist = dist;
						numclosest = k;
					}
				}
			}
		}

		// got them - now remove this surf from every other light
		for (le = lightents; le; le = le->next)
		{
			// test if this is used
			for (j = 0; j < NUMCLOSEST; j++)
				if (le == closest[j])
					break;

			// already used
			if (j != NUMCLOSEST) continue;

			for (j = 0; j < le->numlightsurfaces; j++)
			{
				// surf not used by this light
				if (surf != le->lightsurfaces[j]) continue;

				le->lightsurfaces[j] = NULL;
			}
		}
	}
#endif

	Con_Printf ("Max overdraw: %i\n", maxpass);

	// clear down everything and set up some stats
	dist = 0;

	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		// these surface types are not lit
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// paranoia!
		if (!surf->polys) continue;

		dist += (float) surf->passframe;

		surf->visframe = -1;
		surf->passframe = -1;
	}

	dist /= (float) cl.worldmodel->numsurfaces;

	Con_Printf ("Average overdraw: %0.1f\n", dist);
}


/*
========================================================================================================================

				RENDERING

========================================================================================================================
*/

extern gltexture_t Atten1D;
extern gltexture_t Atten2D;
int lightcycle = 0;
int currentlight = 0;


void R_RenderSurfaceLight (msurface_t *surf, lightent_t *light)
{
	int j;
	glpoly_t *p;
	glvertex_t *verts, *v;

	for (p = surf->polys; p; p = p->next)
	{
		for (j = 0, verts = surfverts, v = p->verts; j < p->numverts; j++, verts++, v++)
		{
			// build transformed verts directly into the vertex array
			verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
			verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
			verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

			verts->rgba[0] = ((p->verts[j].v[0] - light->origin[0]) / light->intensity) / 2.0f + 0.5f;
			verts->rgba[1] = ((p->verts[j].v[1] - light->origin[1]) / light->intensity) / 2.0f + 0.5f;
			verts->rgba[2] = ((p->verts[j].v[2] - light->origin[2]) / light->intensity) / 2.0f + 0.5f;
		}

		glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
	}
}


void R_RenderSurfaceLights (lightent_t *light)
{
	int i;

	for (i = 0; i < light->numlightsurfaces; i++)
	{
		if (!light->lightsurfaces[i]) continue;

		if (light->lightsurfaces[i]->visframe != r_framecount) continue;
		if (light->lightsurfaces[i]->flags & SURF_DRAWTURB) continue;
		if (light->lightsurfaces[i]->flags & SURF_DRAWSKY) continue;
		if (!light->lightsurfaces[i]->polys->matrix) continue;

		// render it
		R_RenderSurfaceLight (light->lightsurfaces[i], light);
	}
}


void R_RenderNewLights (void)
{
	gltexture_t *tex;
	msurface_t *surf;
	glpoly_t *p;
	glvertex_t *verts, *v;
	int i;
	lightent_t *le;
	lightent_t lent;

	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (glvertex_t), surfverts->v);

	// render everything in black
	glDisable (GL_TEXTURE_2D);
	glColor3f (0, 0, 0);

	// go through all the textures in the world
	for (tex = gltextures; tex; tex = tex->next)
	{
		// no surfs chained for this texture
		if (!tex->texturechain) continue;

		// no lighting info
		if (tex->texturechain->flags & SURF_DRAWTURB) continue;
		if (tex->texturechain->flags & SURF_DRAWSKY) continue;

		// go through the surfs that use this texture
		for (surf = tex->texturechain; surf; surf = surf->texturechain)
		{
			// build a transformed vertex array for the poly
			p = surf->polys;

			for (i = 0, verts = surfverts, v = p->verts; i < p->numverts; i++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}

	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);

	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	glBlendFunc (GL_ONE, GL_ONE);

	// texcoords arrays
	glClientActiveTexture (GL_TEXTURE0);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->rgba);

	// hack - use RGBA for the texcoord pointer here
	glClientActiveTexture (GL_TEXTURE1);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (1, GL_FLOAT, sizeof (glvertex_t), &surfverts->rgba[2]);

	GL_ActivateTexture (GL_TEXTURE1);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_1D);
	glBindTexture (GL_TEXTURE_1D, Atten1D.texnum);

	// modulate mode on TMU 1 - this tmu will store the actual texture colour.
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// switch to tmu 0 for the 2D texture
	GL_ActivateTexture (GL_TEXTURE0);
	GL_BindTexture (Atten2D.texnum);

	// modulate mode on TMU0
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// make sure TMU 1 is active coming in here
	GL_ActivateTexture (GL_TEXTURE1);

	currentlight = 0;

	for (le = lightents; le; le = le->next)
	{
		// light is not visible this frame
		// put this in an extended visibility set
		//if (le->lightleaf->visframe != r_visframecount) continue;

#define LIGHT_DIV 512.0f
		glColor3f (le->colour[0] * ((float) d_lightstylevalue[le->style] / LIGHT_DIV),
			le->colour[1] * ((float) d_lightstylevalue[le->style] / LIGHT_DIV), 
			le->colour[2] * ((float) d_lightstylevalue[le->style] / LIGHT_DIV));

		R_RenderSurfaceLights (le);
	}

	// add in dynamic lights
	for (i = 0; i < MAX_DLIGHTS; i++)
	{
		// dead light
		if (cl_dlights[i].die < cl.time) continue;

		// build a dummy light entity
		lent.origin[0] = cl_dlights[i].origin[0];
		lent.origin[1] = cl_dlights[i].origin[1];
		lent.origin[2] = cl_dlights[i].origin[2];

		lent.intensity = cl_dlights[i].radius;

		lent.lightleaf = Mod_PointInLeaf (lent.origin, cl.worldmodel);
		lent.lightsurfaces = NULL;
		lent.numlightsurfaces = 0;

		glColor3f (cl_dlights[i].rgb[0] / 4.0f, cl_dlights[i].rgb[1] / 4.0f, cl_dlights[i].rgb[2] / 4.0f);

		R_RenderSurfaceLights (&lent);
	}

	// hack to keep tracking up to date
	glDisable (GL_TEXTURE_1D);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable (GL_TEXTURE_2D);
	GL_BindTexture (Atten2D.texnum);
	glDisable (GL_TEXTURE_2D);

	glClientActiveTexture (GL_TEXTURE1);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glColor3f (1, 1, 1);

	GL_ActivateTexture (GL_TEXTURE0);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glDisableClientState (GL_VERTEX_ARRAY);
}


void R_CycleLights_f (void)
{
	lightcycle++;

	if (lightcycle >= numlights) lightcycle = 0;

	Con_Printf ("%i of %i\n", lightcycle, numlights);
}


/*
==================
R_LightVertexPoint

Simulates OpenGL lighting equations in software, by getting all light casting objects within range and visibility of
the specified vertex point and applying the classic light equation from them.

The point should be already specified in world space, and the eventual colour is stored in the l parameter.

This can be run at load time or at run-time (which would be slower).
==================
*/

// for casting around to find a valid empty space
float poffsets[6][3] =
{
	{-2, 0, 0},
	{2, 0, 0},
	{0, -2, 0},
	{0, 2, 0},
	{0, 0, -2},
	{0, 0, 2}
};

void R_LightVertexPoint (float *vpoint, float *l)
{
	int i;
	mleaf_t *leaf = NULL;
	float newvpoint[3];

	// initially black
	l[0] = l[1] = l[2] = 0.0f;

	// cast about in each direction
	for (i = 0; i < 6; i++)
	{
		// cast about by 2 units in this direction
		newvpoint[0] = vpoint[0] + poffsets[i][0];
		newvpoint[1] = vpoint[1] + poffsets[i][1];
		newvpoint[2] = vpoint[2] + poffsets[i][2];

		// get the leaf
		leaf = Mod_PointInLeaf (newvpoint, cl.worldmodel);

		// take the first non-solid leaf
		if (leaf->contents != CONTENTS_SOLID) break;
	}

	// this should never happen
	if (!leaf) return;

	// never found a leaf
	if (leaf->contents == CONTENTS_SOLID) return;

	// now obtain visibility for this point
}
