/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// mathlib.h

typedef float vec_t;
typedef vec_t vec3_t[3];
typedef vec_t vec5_t[5];

typedef struct vector_s
{
	float x;
	float y;
	float z;
} vector_st;

typedef union vector_u
{
	vector_st vec;
	float vecs[3];
} vector_t;

typedef	int	fixed4_t;
typedef	int	fixed8_t;
typedef	int	fixed16_t;

#ifndef M_PI
#define M_PI		3.14159265358979323846	// matches value in gcc v2 math.h
#endif

struct mplane_s;

extern vec3_t vec3_origin;

#define DEG2RAD(a) ((a) * ((float) M_PI / 180.0f))
#define RAD2DEG(a) ((a) * (180.0f / (float) M_PI))
#define ANGLEMOD(a) (((int) ((a) * (65536.0f / 360.0f)) & 65535) * (360.0f / 65536.0f))

#define	NANMASK (255<<23)
#define	IS_NAN(x) (((*(int *)&x)&NANMASK) == NANMASK)

#define DotProduct(x,y)				(x[0]*y[0]+x[1]*y[1]+x[2]*y[2])
#define CrossProduct(a,b,c)			((c)[0]=(a)[1]*(b)[2]-(a)[2]*(b)[1],(c)[1]=(a)[2]*(b)[0]-(a)[0]*(b)[2],(c)[2]=(a)[0]*(b)[1]-(a)[1]*(b)[0])
#define VectorCompare(a,b)			(((a)[0]==(b)[0])&&((a)[1]==(b)[1])&&((a)[2]==(b)[2]))
#define VectorMA(a, scale, b, c)	((c)[0] = (a)[0] + (scale) * (b)[0],(c)[1] = (a)[1] + (scale) * (b)[1],(c)[2] = (a)[2] + (scale) * (b)[2])

float		VectorLenght (vec3_t v);
float		VectorNormalize (vec3_t v);		// returns vector length
void		VectorNormalizeFast (vec3_t v);
void		VectorSubtract (vec3_t veca, vec3_t vecb, vec3_t out);
void		VectorAdd (vec3_t veca, vec3_t vecb, vec3_t out);
void		VectorCopy (vec3_t in, vec3_t out);
void		VectorNegate (vec3_t in, vec3_t out);
void		VectorScale (vec3_t in, vec_t scale, vec3_t out);
void		VectorInverse (vec3_t v);

float		Q_rsqrt (float number);
float		Q_sqrt (float num);
void		Q_randinit (void);
int			Q_rand (void);
int			Q_log2(int val);

void		TurnVector (vec3_t out, const vec3_t forward, const vec3_t side, float angle);
void		R_ConcatRotations (float in1[3][3], float in2[3][3], float out[3][3]);
void		R_ConcatTransforms (float in1[3][4], float in2[3][4], float out[3][4]);

void		FloorDivMod (double numer, double denom, int *quotient, int *rem);
fixed16_t	Invert24To16(fixed16_t val);
int			GreatestCommonDivisor (int i1, int i2);

void		AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up);
int			BoxOnPlaneSide (vec3_t emins, vec3_t emaxs, struct mplane_s *plane);

#define BOX_ON_PLANE_SIDE(emins, emaxs, p)	\
	(((p)->type < 3)?						\
	(										\
		((p)->dist <= (emins)[(p)->type])?	\
			1								\
		:									\
		(									\
			((p)->dist >= (emaxs)[(p)->type])?\
				2							\
			:								\
				3							\
		)									\
	)										\
	:										\
		BoxOnPlaneSide( (emins), (emaxs), (p)))
