/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

extern cvar_t	pausable;
extern int worldtype_found;

cvar_t sv_maxclientstring = {"sv_maxclientstring", "64", true};

unsigned char game_id[GAME_ID_LENGTH];

int	current_skill;

void Mod_Print (void);

qboolean r_mapshot = false;

void GL_LoadTextureReplace (char *texname, char *oldname, qboolean report);


/*
==================
Host_Quit_f
==================
*/

extern void M_Menu_Quit_f (void);

void Host_Quit_f (void)
{
	// somebody complained on qer about f10 in the console, so heeere we go...
	if (key_dest != key_console && cls.state != ca_dedicated)
	{
		M_Menu_Quit_f ();
		return;
	}

	CL_Disconnect ();
	Host_ShutdownServer(false);		

	Sys_Quit ();
}


/*
==================
Host_Status_f
==================
*/
void Host_Status_f (void)
{
	client_t	*client;
	int			seconds;
	int			minutes;
	int			hours = 0;
	int			j;
	void		(*print) (char *fmt, ...);
	
	if (cmd_source == src_command)
	{
		if (!sv.active)
		{
			Cmd_ForwardToServer ();
			return;
		}
		print = Con_Printf;
	}
	else
		print = SV_ClientPrintf;

	print ("host:    %s\n", Cvar_VariableString ("hostname"));
	print ("version: %4.2f\n", VERSION);
	if (tcpipAvailable)
		print ("tcp/ip:  %s\n", my_tcpip_address);
	if (ipxAvailable)
		print ("ipx:     %s\n", my_ipx_address);
	print ("map:     %s\n", sv.name);
	print ("players: %i active (%i max)\n\n", net_activeconnections, svs.maxclients);
	for (j=0, client = svs.clients ; j<svs.maxclients ; j++, client++)
	{
		if (!client->active)
			continue;
		seconds = (int)(net_time - client->netconnection->connecttime);
		minutes = seconds / 60;
		if (minutes)
		{
			seconds -= (minutes * 60);
			hours = minutes / 60;
			if (hours)
				minutes -= (hours * 60);
		}
		else
			hours = 0;
		print ("#%-2u %-16.16s  %3i  %2i:%02i:%02i\n", j+1, client->name, (int)client->edict->v.frags, hours, minutes, seconds);
		print ("   %s\n", client->netconnection->address);
	}
}


/*
==================
Host_God_f

Sets client to godmode
==================
*/
void Host_God_f (void)
{
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	sv_player->v.flags = (int)sv_player->v.flags ^ FL_GODMODE;
	if (!((int)sv_player->v.flags & FL_GODMODE) )
		SV_ClientPrintf ("godmode OFF\n");
	else
		SV_ClientPrintf ("godmode ON\n");
}


void Host_Notarget_f (void)
{
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	sv_player->v.flags = (int)sv_player->v.flags ^ FL_NOTARGET;

	if (!((int)sv_player->v.flags & FL_NOTARGET) )
		SV_ClientPrintf ("notarget OFF\n");
	else
		SV_ClientPrintf ("notarget ON\n");
}


qboolean noclip_anglehack;

void Host_Noclip_f (void)
{
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	if (sv_player->v.movetype != MOVETYPE_NOCLIP)
	{
		noclip_anglehack = true;
		sv_player->v.movetype = MOVETYPE_NOCLIP;
		SV_ClientPrintf ("noclip ON\n");
	}
	else
	{
		noclip_anglehack = false;
		sv_player->v.movetype = MOVETYPE_WALK;
		SV_ClientPrintf ("noclip OFF\n");
	}
}

/*
==================
Host_Fly_f

Sets client to flymode
==================
*/
void Host_Fly_f (void)
{
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	if (sv_player->v.movetype != MOVETYPE_FLY)
	{
		sv_player->v.movetype = MOVETYPE_FLY;
		SV_ClientPrintf ("flymode ON\n");
	}
	else
	{
		sv_player->v.movetype = MOVETYPE_WALK;
		SV_ClientPrintf ("flymode OFF\n");
	}
}


/*
==================
Host_Ping_f

==================
*/
void Host_Ping_f (void)
{
	int		i, j;
	float	total;
	client_t	*client;
	
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	SV_ClientPrintf ("Client ping times:\n");
	for (i=0, client = svs.clients ; i<svs.maxclients ; i++, client++)
	{
		if (!client->active)
			continue;
		total = 0;
		for (j=0 ; j<NUM_PING_TIMES ; j++)
			total+=client->ping_times[j];
		total /= NUM_PING_TIMES;
		SV_ClientPrintf ("%4i %s\n", (int)(total*1000), client->name);
	}
}

/*
===============================================================================

SERVER TRANSITIONS

===============================================================================
*/


/*
GenerateUniqueGameID

Generate a unique game id to enable info_player_return to handle multiple games without
having to clear down the saved states each time a new game is started

a game id is not a checksum or anything fancy like that - it's just a long random string
with a very low chance of duplication.

the string is (of course) alphabetic to make things easier on ourselves, cos game saves
are just plain text ascii files... this reduces the number of combinations quite significantly,
but a chance of 1 in 25**255 of duplication is still pretty high, i reckon.
*/
void GenerateUniqueGameID (void)
{
	int i;

	for (i = 0; i < GAME_ID_LENGTH; i++)
		game_id[i] = 97 + (rand () & 25);

	// terminate the string properly for printf and the like
	game_id[GAME_ID_LENGTH - 1] = '\0';
}


/*
CheckGameID

check the current game id against one just loaded - this is basically just a strcmp!!!!
*/
qboolean CheckGameID (char *oldid)
{
	int i;

	for (i = 0; i < (GAME_ID_LENGTH -1); i++)
		if (game_id[i] != oldid[i])
			return false;

	return true;
}


/*
======================
Host_Map_f

handle a 
map <servername>
command from the console.  Active clients are kicked off.
======================
*/
void R_MapShot (char *mapname);

void Host_Map_f (void)
{
	int		i;
	char	name[MAX_QPATH];

	if (cmd_source != src_command)
		return;

	GL_LoadTextureReplace (va ("mapshots/%s", Cmd_Argv (1)), "conback", false);

	r_mapshot = true;

	GenerateUniqueGameID ();

	cls.demonum = -1;		// stop demo loop in case this fails

	R_MapShot (Cmd_Argv (1));

	CL_Disconnect ();
	Host_ShutdownServer(false);		

	key_dest = key_game;			// remove console or menu
	SCR_BeginLoadingPlaque ();

	cls.mapstring[0] = 0;

	for (i=0 ; i<Cmd_Argc() ; i++)
	{
		strcat (cls.mapstring, Cmd_Argv(i));
		strcat (cls.mapstring, " ");
	}

	strcat (cls.mapstring, "\n");

	svs.serverflags = 0;			// haven't completed an episode yet
	strcpy (name, Cmd_Argv(1));
	worldtype_found = 0;

	SV_SpawnServer (name);

	if (!sv.active)
	{
		r_mapshot = false;
		return;
	}
	
	if (cls.state != ca_dedicated)
	{
		// hmmm - this is being done twice - doesn't do any harm, but it really shouldn't be...
		GenerateUniqueGameID ();

		strcpy (cls.spawnparms, "");

		for (i=2 ; i<Cmd_Argc() ; i++)
		{
			strcat (cls.spawnparms, Cmd_Argv(i));
			strcat (cls.spawnparms, " ");
		}
		
		Cmd_ExecuteString ("connect local", src_command);
	}
}


/*
==================
Host_Changelevel_f

Goes to a new map, taking all clients along

MH - includes 'info_player_return' code which allows the player to return to the same map with
all monsters that have been killed staying dead, and all items that have been picked up staying
gone.  does not bring across keys and powerups.

the hub system will work perfectly with existing game saves and games that do not use it - 
however, it will create a load of extra stuff in the 'save' directory.  this is harmless in these
instances
==================
*/
int CheckSameGame (char *mapname, char *return_point, qboolean save_state)
{
	// if the game id is the same for a given saved state, we will use that state, if it's not
	// the same we won't use it.
	FILE *f;
	int i;
	char test_id[GAME_ID_LENGTH];
	qboolean good_id;

	if (save_state)
		Con_DPrintf ("\nSaving\n");
	else Con_DPrintf ("\nLoading\n");

	for (i = 0; ; i++)
	{
		Con_DPrintf ("  checking s%i", i);

		sprintf (return_point, "%s/save/%s/s%i.hub", com_gamedir, mapname, i);

		f = fopen (return_point, "r");

		if (!f) break;

		// read the game id and compare it with the current one - if it's the same return i,
		// otherwise go to the next value of i.
		fscanf (f, "%s\n", test_id);

		good_id = CheckGameID (test_id);

		fclose (f);

		if (good_id)
		{
			Con_DPrintf (" - valid state\n");
			return i;
		}
		else Con_DPrintf (" - invalid state\n");

		// check for runaway loops - should never happen, but just in case...
		if (i > 32767) break;
	}

	Con_DPrintf (" - invalid state\n  no valid states found");

	if (save_state)
		Con_DPrintf (" - creating a new state\n");
	else Con_DPrintf (" - loading as normal\n");

	// save returns i if nothing found cos it will want to create a new state numbered i,
	// load returns -1 cos 0 is a valid load state number - a -1 return will spawn the
	// server normally rather than load from a saved state
	if (save_state)
		return i;
	else return -1;
}


void MH_SaveState (char *, char *);
void MH_LoadState (char *, char *);
void Host_Reconnect_f (void);
void Host_SavegameComment (char *text);
int SV_ModelIndex (char *);


// for restoring the state
float MH_ammo_shells;
float MH_ammo_rockets;
float MH_ammo_cells;
float MH_ammo_nails;
float MH_armortype;
float MH_armorvalue;
string_t MH_weaponmodel;
float MH_weaponframe;
float MH_frame;
float MH_currentammo;
float MH_weapon;


void MH_SavePlayerState (void)
{
	edict_t *player_state;

	player_state = EDICT_NUM (1);  // player is *always* edict 1...

	MH_ammo_shells = player_state->v.ammo_shells;
	MH_ammo_rockets = player_state->v.ammo_rockets;
	MH_ammo_nails = player_state->v.ammo_nails;
	MH_ammo_cells = player_state->v.ammo_cells;
	MH_armortype = player_state->v.armortype;
	MH_armorvalue = player_state->v.armorvalue;
	MH_weaponmodel = player_state->v.weaponmodel;
	MH_currentammo = player_state->v.currentammo;
	MH_weaponframe = player_state->v.weaponframe;
	MH_frame = player_state->v.frame;
	MH_weapon = player_state->v.weapon;
}


void MH_RestorePlayerState (void)
{
	edict_t *player_state;

	player_state = EDICT_NUM (1);  // player is *always* edict 1...

	sv_player->v.ammo_shells = MH_ammo_shells;
	sv_player->v.ammo_rockets = MH_ammo_rockets;
	sv_player->v.ammo_nails = MH_ammo_nails;
	sv_player->v.ammo_cells = MH_ammo_cells;
	sv_player->v.armortype = MH_armortype;
	sv_player->v.armorvalue = MH_armorvalue;
	sv_player->v.weaponmodel = MH_weaponmodel;
	sv_player->v.currentammo = MH_currentammo;
	sv_player->v.weaponframe = MH_weaponframe;
	sv_player->v.frame = MH_frame;
	sv_player->v.weapon = MH_weapon;

	// restore modelindex properly otherwise we're fucked
	sv_player->v.modelindex = SV_ModelIndex ("progs/player.mdl");

	// remove powerups and keys - fixme - need to add rogue and hipnotic here
	sv_player->v.items = sv_player->v.items - ((int)sv_player->v.items & 
		(IT_KEY1|IT_KEY2|IT_INVISIBILITY|IT_INVULNERABILITY|IT_SUIT|IT_QUAD));

	player_state = EDICT_NUM (1);  // player is *always* edict 1...

	// reset it for both sv_player and edict 1
	player_state->v.ammo_shells = MH_ammo_shells;
	player_state->v.ammo_rockets = MH_ammo_rockets;
	player_state->v.ammo_nails = MH_ammo_nails;
	player_state->v.ammo_cells = MH_ammo_cells;
	player_state->v.armortype = MH_armortype;
	player_state->v.armorvalue = MH_armorvalue;
	player_state->v.weaponmodel = MH_weaponmodel;
	player_state->v.currentammo = MH_currentammo;
	player_state->v.weaponframe = MH_weaponframe;
	player_state->v.frame = MH_frame;
	player_state->v.weapon = MH_weapon;

	// restore modelindex properly otherwise we're fucked
	player_state->v.modelindex = SV_ModelIndex ("progs/player.mdl");

	// remove powerups and keys - fixme - need to add rogue and hipnotic here
	player_state->v.items = player_state->v.items - ((int)player_state->v.items & 
		(IT_KEY1|IT_KEY2|IT_INVISIBILITY|IT_INVULNERABILITY|IT_SUIT|IT_QUAD));
}


void Host_Changelevel_f (void)
{
	char level[MAX_QPATH];
	char return_point[MAX_QPATH];
	int same_game;

	if (Cmd_Argc () != 2)
	{
		Con_Printf ("changelevel <levelname> : continue game on a new level\n");
		return;
	}

	if (!sv.active || cls.demoplayback)
	{
		Con_Printf ("Only the server may changelevel\n");
		return;
	}

	GL_LoadTextureReplace (va ("mapshots/%s", Cmd_Argv (1)), "conback", false);

	worldtype_found = 0;

	// generate a new game id for start maps, otherwise keep the old one
	// this is because some people may use a "changelevel start" command while in a map to
	// start a new game (they shouldn't, but you never know) - since the start map is closed
	// off from previously visited maps in a normal quake sp game (and should never contain
	// any items or monsters anyway) this is not an issue.  some people may put items or
	// monsters in a start map - they shouldn't, but if they do these are normally meant to
	// be back there for each revisit anyway.
	if (!strcmp ("start", Cmd_Argv (1)))
		GenerateUniqueGameID ();

	// deathmatch doesn't use the hub system - co-op should
	if (deathmatch.value)
	{
		// use the normal changelevel functionality
		SV_SaveSpawnparms ();
		strcpy (level, Cmd_Argv (1));
		SV_SpawnServer (level);
		return;
	}

	// when leaving the start map, don't save a state
	if (!strcmp (sv.name, "start"))
	{
		// use the normal changelevel functionality
		SV_SaveSpawnparms ();
		strcpy (level, Cmd_Argv (1));
		SV_SpawnServer (level);
		return;
	}

	// save old map state here - set up a directory to put the state in
	sprintf (level, "%s/save/%s", com_gamedir, sv.name);
	Sys_mkdir (level);

	// we need to check for the same game here as well before saving the state cos
	// we need to get the file name correct...
	same_game = CheckSameGame (sv.name, return_point, true);
	sprintf (return_point, "%s/save/%s/s%i.hub", com_gamedir, sv.name, same_game);
	Con_DPrintf ("Saving state as %s\n", return_point);
	MH_SaveState (return_point, Cmd_Argv (1));

	// test for a possible return save
	same_game = CheckSameGame (Cmd_Argv (1), return_point, false);

	// if there's a previously saved state for this map and this game to restore, then restore
	// it (use -1 to test cos 0 is a valid return) - otherwise spawn normally.
	if (same_game != -1)
	{
		// before we load the new state, we wanna save the current player state.  we can't do
		// SV_SaveSpawnparms here cos loading edict 1 from the state file will overwrite the
		// necessary stuff.  this way is ugly, but ensures that the player returns with the
		// correct ammo, health, armour, active weapon, etc.  it may not work with mods,
		// though.  we'll see...
		MH_SavePlayerState ();
		Con_DPrintf ("Loading state from %s\n", return_point);
		MH_LoadState (return_point, sv.name);
		MH_RestorePlayerState ();
		return;
	}
	else
	{
		// just change the level normally
		Con_DPrintf ("Loading %s as normal\n", Cmd_Argv (1));
		SV_SaveSpawnparms ();
		strcpy (level, Cmd_Argv (1));
		SV_SpawnServer (level);
	}
}


/*
===============
MH_SaveState

saves current game state with enhancements for a player return...
===============
*/

dfunction_t *ED_FindFunction (char *);

void MH_SaveState (char *name, char *new_level)
{
	int		return_point_found;
	FILE	*f;
	dfunction_t	*func_mfh;
	int		i;
	char	comment[SAVEGAME_COMMENT_LENGTH+1];
	edict_t	*ent;
	vec3_t	ret_origin;
	int		changelevel_func;
	vec3_t	ret_angles;
		
	ddef_t		*def;
	char		*name2;
	int			type;

	Con_DPrintf ("Saving current game state to %s...\n", name);

	f = fopen (name, "w");

	if (!f)
	{
		Con_DPrintf ("ERROR: couldn't open.\n");
		return;
	}
	
	// the game id should be the first thing written to a state save so we can cleanly get out
	// when restoring if we're not in the same game
	fprintf (f, "%s\n", game_id);

	// use a SAVEGAME_VERSION of 0 to distinguish between state saves and game saves
	fprintf (f, "%i\n", 0);
	Host_SavegameComment (comment);
	fprintf (f, "%s\n", comment);

	for (i=0 ; i<NUM_SPAWN_PARMS ; i++)
		fprintf (f, "%f\n", svs.clients->spawn_parms[i]);

	fprintf (f, "%d\n", current_skill);
	fprintf (f, "%s\n", sv.name);
	fprintf (f, "%f\n",sv.time);

	// write the return info to the savegame header cos it's easier to get at
	// all we need is the origin & angles of the correct info_player_return,
	// so we can go through all the edicts until we find it, and then save it.
	return_point_found = 0;

	// look for an info_player_return for the correct map
	for (i = 0; i < sv.num_edicts; i++)
	{
		ent = EDICT_NUM (i);

		if (!strcmp (pr_strings + ent->v.classname, "info_player_return"))
		{
			// this should be the new map name, *not* sv.name!!!!!
			if (!strcmp (pr_strings + ent->v.targetname, new_level))
			{
				Con_DPrintf ("info_player_return from %s at: ", new_level);

				return_point_found = 1;
				goto return_done;
			}
		}
	}

	// if not found, look for an info_player_start2 (start map compatibility)
	if (!return_point_found)
	{
		// send back to info_player_start2
		for (i = 0; i < sv.num_edicts; i++)
		{
			ent = EDICT_NUM (i);

			if (!strcmp (pr_strings + ent->v.classname, "info_player_start2"))
			{
				Con_DPrintf ("info_player_start2 at: ");

				return_point_found = 1;
				goto return_done;
			}
		}
	}

	// otherwise, look for an info_player_start (returning to original start point)
	// this can't fail, cos *every* level *must* have an info_player_start!!!!!!!!!
	// send back to info_player_start
	for (i = 0; i < sv.num_edicts; i++)
	{
		ent = EDICT_NUM (i);

		if (!strcmp (pr_strings + ent->v.classname, "info_player_start"))
		{
			Con_DPrintf ("info_player_start at: ");

			return_point_found = 1;
			goto return_done;
		}
	}

	// goto is evil - what the fuck?
return_done:
	// set the correct origin and angles - these are common to all 3 types of return point
	// so we can do them here...
	Con_DPrintf ("%f ", ent->v.origin[0]);
	Con_DPrintf ("%f ", ent->v.origin[1]);
	Con_DPrintf ("%f\n", ent->v.origin[2]);

	ret_origin[0] = ent->v.origin[0];
	ret_origin[1] = ent->v.origin[1];
	ret_origin[2] = ent->v.origin[2];

	ret_angles[0] = ent->v.angles[0];
	ret_angles[1] = ent->v.angles[1];
	ret_angles[2] = ent->v.angles[2];

	fprintf (f, "%f\n", ret_origin[0]);
	fprintf (f, "%f\n", ret_origin[1]);
	fprintf (f, "%f\n", ret_origin[2]);

	fprintf (f, "%f\n", ret_angles[0]);
	fprintf (f, "%f\n", ret_angles[1]);
	fprintf (f, "%f\n", ret_angles[2]);

	// write the light styles
	for (i = 0; i < MAX_LIGHTSTYLES; i++)
	{
		if (sv.lightstyles[i])
			fprintf (f, "%s\n", sv.lightstyles[i]);
		else
			fprintf (f,"m\n");
	}

	// find the correct number of the trigger_changelevel function
	// (i know it's 223 in my own progs.dat - may be different in others!)
	func_mfh = ED_FindFunction ("changelevel_touch");

	for (i = 0; i < progs->numfunctions; i++)
		if (pr_functions[i].first_statement == func_mfh->first_statement) break;

	// because the progs.dat can change from level to level, we must set this
	// at each save for the correct level, rather than just once per game
	changelevel_func = i;

	// write the globals here instead of through ED_WriteGlobals,
	// cos I wanna hack the intermission stuff out of them...
	fprintf (f,"{\n");

	for (i = 0; i < progs->numglobaldefs; i++)
	{
		def = &pr_globaldefs[i];
		type = def->type;

		if ( !(def->type & DEF_SAVEGLOBAL) )
			continue;

		type &= ~DEF_SAVEGLOBAL;

		if (type != ev_string
			&& type != ev_float
			&& type != ev_entity)
			continue;

		name2 = pr_strings + def->s_name;	

		// velocity ain't in standard game saves...
		if (!strcmp (name2, "velocity")) continue;

		fprintf (f,"\"%s\" ", name2);

		// hack out the intermission info
		if (!strncmp (name2, "intermission", 12))
			fprintf (f, "\"0.000000\"\n");
		else if (!strcmp (name2, "nextmap"))
			fprintf (f, "\"\"\n");
		else fprintf (f,"\"%s\"\n", PR_UglyValueString (type, (eval_t *)&pr_globals[def->ofs]));
	}

	fprintf (f,"}\n");

	for (i = 0; i < sv.num_edicts; i++)
	{
		ent = EDICT_NUM(i);

		// edict 1 is always the player - in theory, we needn't write this, since it will
		// be restored properly during the load process, but we're going to anyway, cos
		// it's an easy way of filling up a lot of the fields that otherwise wouldn't be!

		// hack trigger_changelevel so it works properly on returning
		if (!strcmp (pr_strings + ent->v.classname, "trigger_changelevel"))
		{
			ent->v.touch = changelevel_func;	
			ent->v.think = 0;	
		}

		ED_Write (f, EDICT_NUM(i));
		fflush (f);
	}

	fclose (f);
	Con_DPrintf ("done.\n");
}


/*
===============
MH_LoadState

The returning to the proper point is done in here rather hacking the
proper point into the save function, because it's cleaner, neater and
more logical, as well as being easier, of course...
===============
*/
void MH_LoadState (char *name, char *old_level)
{
	FILE	*f;
	char	mapname[MAX_QPATH];
	float	time, tfloat;
	char	str[32768], *start;
	int		i, r;
	edict_t	*ent;
	int		entnum;
	int		version;
	float	spawn_parms[NUM_SPAWN_PARMS];
	vec3_t	ret_origin;
	vec3_t	ret_angles;

	char test_id[GAME_ID_LENGTH];

	Con_DPrintf ("Loading game from %s...\n", name);
	Con_DPrintf ("Returning from %s\n", old_level );

	// this is always gonna work cos we just tried it in the
	// function that calls this one, so we don't need to check it
	f = fopen (name, "r");

	// read the game id (it's already been checked)
	fscanf (f, "%s\n", test_id);

	// use a SAVEGAME_VERSION of 0 to distinguish between state saves and game saves
	fscanf (f, "%i\n", &version);

	if (version)
	{
		fclose (f);
		Con_Printf ("Savegame is version %i, not 0\n", version);
		return;
	}

	fscanf (f, "%s\n", str);

	for (i = 0; i < NUM_SPAWN_PARMS; i++)
		fscanf (f, "%f\n", &spawn_parms[i]);

	// this silliness is so we can load 1.06 save files, which have float skill values
	fscanf (f, "%f\n", &tfloat);
	current_skill = (int)(tfloat + 0.1);
	Cvar_SetValue ("skill", (float)current_skill);

	fscanf (f, "%s\n",mapname);
	fscanf (f, "%f\n",&time);

	CL_Disconnect_f ();
	
	SV_SpawnServer (mapname);

	if (!sv.active)
	{
		fclose (f);
		Con_Printf ("Couldn't load map\n");
		return;
	}

	sv.paused = true;		// pause until all clients connect
	sv.loadgame = true;

	// read the return info from the header
	fscanf (f, "%f\n",&ret_origin[0]);
	fscanf (f, "%f\n",&ret_origin[1]);
	fscanf (f, "%f\n",&ret_origin[2]);

	fscanf (f, "%f\n",&ret_angles[0]);
	fscanf (f, "%f\n",&ret_angles[1]);
	fscanf (f, "%f\n",&ret_angles[2]);

	// load the light styles
	for (i = 0; i < MAX_LIGHTSTYLES; i++)
	{
		fscanf (f, "%s\n", str);
		sv.lightstyles[i] = Hunk_Alloc (strlen(str)+1);
		strcpy (sv.lightstyles[i], str);
	}

	// load the edicts out of the savegame file
	entnum = -1;		// -1 is the globals

	while (!feof (f))
	{
		for (i = 0; i < sizeof (str) - 1; i++)
		{
			r = fgetc (f);

			if (r == EOF || !r)
				break;

			str[i] = r;

			if (r == '}')
			{
				i++;
				break;
			}
		}

		if (i == sizeof (str)-1)
		{
			// close the file and exit (slightly more) gracefully
			fclose (f);
			Host_Error ("Loadgame buffer overflow");
		}

		str[i] = 0;
		start = str;
		start = COM_Parse (str);

		if (!com_token[0])
			break;		// end of file

		if (strcmp (com_token,"{"))
		{
			// close the file and exit (slightly more) gracefully
			fclose (f);
			Host_Error ("Loadgame buffer overflow");
		}
			
		if (entnum == -1)
		{
			// parse the global vars
			ED_ParseGlobals (start);
		}
		else
		{
			// parse an edict
			ent = EDICT_NUM (entnum);
			memset (&ent->v, 0, progs->entityfields * 4);
			ent->free = false;
			ED_ParseEdict (start, ent);

			// link it into the bsp tree
			if (!ent->free) SV_LinkEdict (ent, false);
		}

		entnum++;
	}
	
	sv.num_edicts = entnum;
	sv.time = time;

	fclose (f);

	for (i = 0; i < NUM_SPAWN_PARMS; i++)
		svs.clients->spawn_parms[i] = spawn_parms[i];

	if (cls.state != ca_dedicated)
	{
		CL_EstablishConnection ("local");
		Host_Reconnect_f ();
	}

	// return at the correct position and facing in the correct direction
	sv_player->v.origin[0] = ret_origin[0];
	sv_player->v.origin[1] = ret_origin[1];
	sv_player->v.origin[2] = ret_origin[2];

	sv_player->v.oldorigin[0] = ret_origin[0];
	sv_player->v.oldorigin[1] = ret_origin[1];
	sv_player->v.oldorigin[2] = ret_origin[2];

	sv_player->v.angles[0] = ret_angles[0];
	sv_player->v.angles[1] = ret_angles[1];
	sv_player->v.angles[2] = ret_angles[2];

	sv_player->v.velocity[0] = 0;
	sv_player->v.velocity[1] = 0;
	sv_player->v.velocity[2] = 0;

	// oops!  nearly forgot this! - get out of intermission mode!
	sv_player->v.movetype = 3;
	sv_player->v.solid = 3;
	sv_player->v.view_ofs[0] = 0;
	sv_player->v.view_ofs[1] = 0;
	sv_player->v.view_ofs[2] = 22;
}


/*
==================
Host_Restart_f

Restarts the current server for a dead player
==================
*/
void Host_Restart_f (void)
{
	char	mapname[MAX_QPATH];

	if (cls.demoplayback || !sv.active)
		return;

	if (cmd_source != src_command)
		return;
	strcpy (mapname, sv.name);	// must copy out, because it gets cleared
								// in sv_spawnserver
	worldtype_found = 0;

	SV_SpawnServer (mapname);
}


/*
==================
Host_Reconnect_f

This command causes the client to wait for the signon messages again.
This is sent just before a server changes levels
==================
*/
void Host_Reconnect_f (void)
{
	SCR_BeginLoadingPlaque ();
	cls.signon = 0;		// need new connection messages
}


/*
=====================
Host_Connect_f

User command to connect to server
=====================
*/
void Host_Connect_f (void)
{
	char	name[MAX_QPATH];
	
	cls.demonum = -1;		// stop demo loop in case this fails

	if (cls.demoplayback)
	{
		CL_StopPlayback ();
		CL_Disconnect ();
	}

	strcpy (name, Cmd_Argv(1));
	CL_EstablishConnection (name);
	Host_Reconnect_f ();
}


/*
===============================================================================

LOAD / SAVE GAME

===============================================================================
*/

/*
===============
Host_SavegameComment

Writes a SAVEGAME_COMMENT_LENGTH character comment describing the current 
===============
*/
void Host_SavegameComment (char *text)
{
	int		i;
	char	kills[20];

	for (i=0 ; i<SAVEGAME_COMMENT_LENGTH ; i++)
		text[i] = ' ';

	memcpy (text, cl.levelname, strlen(cl.levelname));
	sprintf (kills,"kills:%3i/%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
	memcpy (text+22, kills, strlen(kills));

	// convert space to _ to make stdio happy
	for (i=0 ; i<SAVEGAME_COMMENT_LENGTH ; i++)
		if (text[i] == ' ')
			text[i] = '_';

	text[SAVEGAME_COMMENT_LENGTH] = '\0';
}


/*
===============
Host_Savegame_f
===============
*/
void Host_Savegame_f (void)
{
	char	name[256];
	FILE	*f;
	int		i;
	char	comment[SAVEGAME_COMMENT_LENGTH+1];

	char gid_file[MAX_OSPATH];
	FILE *g;

	if (cmd_source != src_command)
		return;

	if (!sv.active)
	{
		Con_Printf ("Not playing a local game.\n");
		return;
	}

	if (cl.intermission)
	{
		Con_Printf ("Can't save in intermission.\n");
		return;
	}

	if (svs.maxclients != 1)
	{
		Con_Printf ("Can't save multiplayer games.\n");
		return;
	}

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("save <savename> : save a game\n");
		return;
	}

	if (strstr(Cmd_Argv(1), ".."))
	{
		Con_Printf ("Relative pathnames are not allowed.\n");
		return;
	}
		
	for (i=0 ; i<svs.maxclients ; i++)
	{
		if (svs.clients[i].active && (svs.clients[i].edict->v.health <= 0) )
		{
			Con_Printf ("Can't savegame with a dead player\n");
			return;
		}
	}

	sprintf (name, "%s/save", com_gamedir);
	Sys_mkdir (name);

	sprintf (name, "%s/save/%s", com_gamedir, Cmd_Argv(1));
	COM_DefaultExtension (name, ".sav");
	
	Con_Printf ("Saving game to %s...\n", name);
	f = fopen (name, "w");

	if (!f)
	{
		Con_Printf ("ERROR: couldn't open %s\n", name);
		return;
	}

	// write the game id to the .GID file
	sprintf (gid_file, "%s/save/%s.gid", com_gamedir, Cmd_Argv(1));
	g = fopen (gid_file, "w");

	// write game id to the .GID file
	if (g)
	{
		fprintf (g, "%s\n\n", game_id);
		fclose (g);
	}

	fprintf (f, "%i\n", SAVEGAME_VERSION);
	Host_SavegameComment (comment);
	fprintf (f, "%s\n", comment);

	for (i=0 ; i<NUM_SPAWN_PARMS ; i++)
		fprintf (f, "%f\n", svs.clients->spawn_parms[i]);

	fprintf (f, "%d\n", current_skill);
	fprintf (f, "%s\n", sv.name);
	fprintf (f, "%f\n",sv.time);

	// write the light styles
	for (i=0 ; i<MAX_LIGHTSTYLES ; i++)
	{
		if (sv.lightstyles[i])
			fprintf (f, "%s\n", sv.lightstyles[i]);
		else
			fprintf (f,"m\n");
	}


	ED_WriteGlobals (f);

	for (i=0 ; i<sv.num_edicts ; i++)
	{
		ED_Write (f, EDICT_NUM(i));
		fflush (f);
	}

	fclose (f);
	Con_Printf ("done.\n");
}


/*
===============
Host_Loadgame_f
===============
*/
void Host_Loadgame_f (void)
{
	char name[MAX_OSPATH];
	FILE *f;
	char mapname[MAX_QPATH];
	float time, tfloat;
	char str[32768], *start;
	int i, r;
	edict_t *ent;
	int entnum;
	int version;
	float spawn_parms[NUM_SPAWN_PARMS];

	char gid_file[MAX_OSPATH];
	FILE *g;

	if (cmd_source != src_command)
		return;

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("load <savename> : load a game\n");
		return;
	}

	cls.demonum = -1;		// stop demo loop in case this fails

	sprintf (name, "%s/save", com_gamedir);
	Sys_mkdir (name);

	sprintf (name, "%s/save/%s", com_gamedir, Cmd_Argv(1));
	COM_DefaultExtension (name, ".sav");
	
	Con_Printf ("Loading game from %s...\n", name);

	// to maintain game save compatibility with all other engines, we have saved the game id in
	// a .gid file
	sprintf (gid_file, "%s/save/%s.gid", com_gamedir, Cmd_Argv(1));
	g = fopen (gid_file, "r");

	// load game id from the .GID file if present, otherwise generate a new game id
	if (g)
	{
		fscanf (g, "%s\n", game_id);

		// shouldn't be necessary, but what the hell...
		game_id[GAME_ID_LENGTH - 1] = '\0';
		fclose (g);
	}
	else GenerateUniqueGameID ();

	f = fopen (name, "r");

	if (!f)
	{
		Con_Printf ("ERROR: couldn't open %s\n", name);
		return;
	}

	fscanf (f, "%i\n", &version);

	if (version != SAVEGAME_VERSION)
	{
		fclose (f);
		Con_Printf ("Savegame is version %i, not %i\n", version, SAVEGAME_VERSION);
		return;
	}

	fscanf (f, "%s\n", str);

	for (i=0 ; i<NUM_SPAWN_PARMS ; i++)
		fscanf (f, "%f\n", &spawn_parms[i]);

	// this silliness is so we can load 1.06 save files, which have float skill values
	fscanf (f, "%f\n", &tfloat);
	current_skill = (int)(tfloat + 0.1);
	Cvar_SetValue ("skill", (float)current_skill);

	fscanf (f, "%s\n",mapname);
	fscanf (f, "%f\n",&time);

	CL_Disconnect_f ();
	
	worldtype_found = 0;

	SV_SpawnServer (mapname);

	if (!sv.active)
	{
		Con_Printf ("Couldn't load map\n");
		return;
	}

	sv.paused = true;		// pause until all clients connect
	sv.loadgame = true;

	// load the light styles
	for (i=0 ; i<MAX_LIGHTSTYLES ; i++)
	{
		fscanf (f, "%s\n", str);
		sv.lightstyles[i] = Hunk_Alloc (strlen(str)+1);
		strcpy (sv.lightstyles[i], str);
	}

	// load the edicts out of the savegame file
	entnum = -1;		// -1 is the globals

	while (!feof(f))
	{
		for (i=0 ; i<sizeof(str)-1 ; i++)
		{
			r = fgetc (f);

			if (r == EOF || !r)
				break;

			str[i] = r;

			if (r == '}')
			{
				i++;
				break;
			}
		}

		if (i == sizeof(str)-1)
		{
			// close the file and exit (slightly more) gracefully
			fclose (f);
			Host_Error ("Loadgame buffer overflow");
		}

		str[i] = 0;
		start = str;
		start = COM_Parse(str);

		if (!com_token[0])
			break;		// end of file

		if (strcmp(com_token,"{"))
		{
			// close the file and exit (slightly more) gracefully
			fclose (f);
			Host_Error ("Loadgame buffer overflow");
		}
			
		if (entnum == -1)
		{
			// parse the global vars
			ED_ParseGlobals (start);
		}
		else
		{
			// parse an edict
			ent = EDICT_NUM(entnum);
			memset (&ent->v, 0, progs->entityfields * 4);
			ent->free = false;
			ED_ParseEdict (start, ent);
	
			// link it into the bsp tree
			if (!ent->free)
				SV_LinkEdict (ent, false);
		}

		entnum++;
	}
	
	sv.num_edicts = entnum;
	sv.time = time;

	fclose (f);

	for (i=0 ; i<NUM_SPAWN_PARMS ; i++)
		svs.clients->spawn_parms[i] = spawn_parms[i];

	if (cls.state != ca_dedicated)
	{
		CL_EstablishConnection ("local");
		Host_Reconnect_f ();
	}
}



//============================================================================

/*
======================
Host_Name_f
======================
*/
void Host_Name_f (void)
{
	char	*newName;

	if (Cmd_Argc () == 1)
	{
		Con_Printf ("\"name\" is \"%s\"\n", cl_name.string);
		return;
	}
	if (Cmd_Argc () == 2)
		newName = Cmd_Argv(1);	
	else
		newName = Cmd_Args();
	newName[15] = 0;

	if (cmd_source == src_command)
	{
		if (strcmp(cl_name.string, newName) == 0)
			return;
		Cvar_Set ("_cl_name", newName);
		if (cls.state == ca_connected)
			Cmd_ForwardToServer ();
		return;
	}

	if (host_client->name[0] && strcmp(host_client->name, "unconnected") )
		if (strcmp(host_client->name, newName) != 0)
			Con_Printf ("%s renamed to %s\n", host_client->name, newName);
	strcpy (host_client->name, newName);
	host_client->edict->v.netname = host_client->name - pr_strings;
	
// send notification to all clients
	
	MSG_WriteByte (&sv.reliable_datagram, svc_updatename);
	MSG_WriteByte (&sv.reliable_datagram, host_client - svs.clients);
	MSG_WriteString (&sv.reliable_datagram, host_client->name);
}

	
void Host_Version_f (void)
{
	Con_Printf ("MHQuake Version 6\n");
}


void Host_Say (qboolean teamonly)
{
	client_t *client;
	client_t *save;
	int		j;
	char	*p;
	unsigned char	text[1024];	// was 64
	qboolean	fromServer = false;

	if (cmd_source == src_command)
	{
		if (cls.state == ca_dedicated)
		{
			fromServer = true;
			teamonly = false;
		}
		else
		{
			Cmd_ForwardToServer ();
			return;
		}
	}

	if (Cmd_Argc () < 2)
		return;

	save = host_client;

	p = Cmd_Args();

	// remove quotes if present
	if (*p == '"')
	{
		p++;
		p[strlen(p)-1] = 0;
	}

	// turn on color set 1
	if (!fromServer)
		sprintf (text, "%c%s: ", 1, save->name);
	else
		sprintf (text, "%c<%s> ", 1, hostname.string);

	//j = sizeof(text) - 2 - strlen(text);  // -2 for /n and null terminator
	j = sv_maxclientstring.value - 2 - strlen (text);  // -2 for /n and null terminator

	if (strlen(p) > j)
		p[j] = 0;

	strcat (text, p);
	strcat (text, "\n");

	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
	{
		if (!client || !client->active || !client->spawned) continue;
		if (teamplay.value && teamonly && client->edict->v.team != save->edict->v.team) continue;

		host_client = client;
		SV_ClientPrintf("%s", text);
	}

	host_client = save;

	Sys_Printf("%s", &text[1]);
}


void Host_Say_f(void)
{
	Host_Say(false);
}


void Host_Say_Team_f(void)
{
	Host_Say(true);
}


void Host_Tell_f(void)
{
	client_t *client;
	client_t *save;
	int		j;
	char	*p;
	char	text[1024];	// was 64

	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (Cmd_Argc () < 3)
		return;

	strcpy(text, host_client->name);
	strcat(text, ": ");

	p = Cmd_Args();

	// remove quotes if present
	if (*p == '"')
	{
		p++;
		p[strlen(p)-1] = 0;
	}

	// check length & truncate if necessary
	//j = sizeof(text) - 2 - strlen(text);  // -2 for /n and null terminator
	j = sv_maxclientstring.value - 2 - strlen (text);  // -2 for /n and null terminator

	if (strlen(p) > j)
		p[j] = 0;

	strcat (text, p);
	strcat (text, "\n");

	save = host_client;
	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
	{
		if (!client->active || !client->spawned) continue;
		if (stricmp(client->name, Cmd_Argv(1))) continue;

		host_client = client;
		SV_ClientPrintf("%s", text);
		break;
	}

	host_client = save;
}


/*
==================
Host_Color_f
==================
*/
void Host_Color_f(void)
{
	int		top, bottom;
	int		playercolor;
	
	if (Cmd_Argc() == 1)
	{
		Con_Printf ("\"color\" is \"%i %i\"\n", ((int)cl_color.value) >> 4, ((int)cl_color.value) & 0x0f);
		Con_Printf ("color <0-13> [0-13]\n");
		return;
	}

	if (Cmd_Argc() == 2)
		top = bottom = atoi(Cmd_Argv(1));
	else
	{
		top = atoi(Cmd_Argv(1));
		bottom = atoi(Cmd_Argv(2));
	}
	
	top &= 15;
	if (top > 13)
		top = 13;
	bottom &= 15;
	if (bottom > 13)
		bottom = 13;
	
	playercolor = top*16 + bottom;

	if (cmd_source == src_command)
	{
		Cvar_SetValue ("_cl_color", playercolor);
		if (cls.state == ca_connected)
			Cmd_ForwardToServer ();
		return;
	}

	host_client->colors = playercolor;
	host_client->edict->v.team = bottom + 1;

	// send notification to all clients
	MSG_WriteByte (&sv.reliable_datagram, svc_updatecolors);
	MSG_WriteByte (&sv.reliable_datagram, host_client - svs.clients);
	MSG_WriteByte (&sv.reliable_datagram, host_client->colors);
}

/*
==================
Host_Kill_f
==================
*/
void Host_Kill_f (void)
{
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (sv_player->v.health <= 0)
	{
		SV_ClientPrintf ("Can't suicide -- allready dead!\n");
		return;
	}
	
	pr_global_struct->time = sv.time;
	pr_global_struct->self = EDICT_TO_PROG(sv_player);
	PR_ExecuteProgram (pr_global_struct->ClientKill);
}


/*
==================
Host_Pause_f
==================
*/
void Host_Pause_f (void)
{
	
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}
	if (!pausable.value)
		SV_ClientPrintf ("Pause not allowed.\n");
	else
	{
		sv.paused ^= 1;

		if (sv.paused)
		{
			SV_BroadcastPrintf ("%s paused the game\n", pr_strings + sv_player->v.netname);
		}
		else
		{
			SV_BroadcastPrintf ("%s unpaused the game\n",pr_strings + sv_player->v.netname);
		}

	// send notification to all clients
		MSG_WriteByte (&sv.reliable_datagram, svc_setpause);
		MSG_WriteByte (&sv.reliable_datagram, sv.paused);
	}
}

//===========================================================================


/*
==================
Host_PreSpawn_f
==================
*/
void Host_PreSpawn_f (void)
{
	if (cmd_source == src_command)
	{
		Con_Printf ("prespawn is not valid from the console\n");
		return;
	}

	if (host_client->spawned)
	{
		Con_Printf ("prespawn not valid -- allready spawned\n");
		return;
	}
	
	SZ_Write (&host_client->message, sv.signon.data, sv.signon.cursize);
	MSG_WriteByte (&host_client->message, svc_signonnum);
	MSG_WriteByte (&host_client->message, 2);
	host_client->sendsignon = true;
}

/*
==================
Host_Spawn_f
==================
*/
void Host_Spawn_f (void)
{
	int		i;
	client_t	*client;
	edict_t	*ent;

	if (cmd_source == src_command)
	{
		Con_Printf ("spawn is not valid from the console\n");
		return;
	}

	if (host_client->spawned)
	{
		Con_Printf ("Spawn not valid -- allready spawned\n");
		return;
	}

	// run the entrance script
	if (sv.loadgame)
	{
		// loaded games are fully inited allready
		// if this is the last client to be connected, unpause
		sv.paused = false;
	}
	else
	{
		// set up the edict
		ent = host_client->edict;

		memset (&ent->v, 0, progs->entityfields * 4);
		ent->v.colormap = NUM_FOR_EDICT(ent);
		ent->v.team = (host_client->colors & 15) + 1;
		ent->v.netname = host_client->name - pr_strings;

		// copy spawn parms out of the client_t
		for (i=0 ; i< NUM_SPAWN_PARMS ; i++)
			(&pr_global_struct->parm1)[i] = host_client->spawn_parms[i];

		// call the spawn function
		pr_global_struct->time = sv.time;
		pr_global_struct->self = EDICT_TO_PROG(sv_player);
		PR_ExecuteProgram (pr_global_struct->ClientConnect);

		if ((Sys_FloatTime() - host_client->netconnection->connecttime) <= sv.time)
			Sys_Printf ("%s entered the game\n", host_client->name);

		PR_ExecuteProgram (pr_global_struct->PutClientInServer);	
	}


	// send all current names, colors, and frag counts
	SZ_Clear (&host_client->message);

	// send time of update
	MSG_WriteByte (&host_client->message, svc_time);
	MSG_WriteFloat (&host_client->message, sv.time);

	for (i=0, client = svs.clients ; i<svs.maxclients ; i++, client++)
	{
		Con_DPrintf ("Sending Client %i\n", i);
		MSG_WriteByte (&host_client->message, svc_updatename);
		MSG_WriteByte (&host_client->message, i);
		MSG_WriteString (&host_client->message, client->name);
		MSG_WriteByte (&host_client->message, svc_updatefrags);
		MSG_WriteByte (&host_client->message, i);
		MSG_WriteShort (&host_client->message, client->old_frags);
		MSG_WriteByte (&host_client->message, svc_updatecolors);
		MSG_WriteByte (&host_client->message, i);
		MSG_WriteByte (&host_client->message, client->colors);
	}
	
	// send all current light styles
	for (i=0 ; i<MAX_LIGHTSTYLES ; i++)
	{
		MSG_WriteByte (&host_client->message, svc_lightstyle);
		MSG_WriteByte (&host_client->message, (char)i);
		MSG_WriteString (&host_client->message, sv.lightstyles[i]);
	}

	// send some stats
	MSG_WriteByte (&host_client->message, svc_updatestat);
	MSG_WriteByte (&host_client->message, STAT_TOTALSECRETS);
	MSG_WriteLong (&host_client->message, pr_global_struct->total_secrets);

	MSG_WriteByte (&host_client->message, svc_updatestat);
	MSG_WriteByte (&host_client->message, STAT_TOTALMONSTERS);
	MSG_WriteLong (&host_client->message, pr_global_struct->total_monsters);

	MSG_WriteByte (&host_client->message, svc_updatestat);
	MSG_WriteByte (&host_client->message, STAT_SECRETS);
	MSG_WriteLong (&host_client->message, pr_global_struct->found_secrets);

	MSG_WriteByte (&host_client->message, svc_updatestat);
	MSG_WriteByte (&host_client->message, STAT_MONSTERS);
	MSG_WriteLong (&host_client->message, pr_global_struct->killed_monsters);

	
	// send a fixangle
	// Never send a roll angle, because savegames can catch the server
	// in a state where it is expecting the client to correct the angle
	// and it won't happen if the game was just loaded, so you wind up
	// with a permanent head tilt
	ent = EDICT_NUM( 1 + (host_client - svs.clients) );
	MSG_WriteByte (&host_client->message, svc_setangle);
	for (i=0 ; i < 2 ; i++)
		MSG_WriteAngle (&host_client->message, ent->v.angles[i] );
	MSG_WriteAngle (&host_client->message, 0 );

	SV_WriteClientdataToMessage (sv_player, &host_client->message);

	MSG_WriteByte (&host_client->message, svc_signonnum);
	MSG_WriteByte (&host_client->message, 3);
	host_client->sendsignon = true;
}

/*
==================
Host_Begin_f
==================
*/
void Host_Begin_f (void)
{
	if (cmd_source == src_command)
	{
		Con_Printf ("begin is not valid from the console\n");
		return;
	}

	host_client->spawned = true;
}

//===========================================================================


/*
==================
Host_Kick_f

Kicks a user off of the server
==================
*/
void Host_Kick_f (void)
{
	char		*who;
	char		*message = NULL;
	client_t	*save;
	int			i;
	qboolean	byNumber = false;

	if (cmd_source == src_command)
	{
		if (!sv.active)
		{
			Cmd_ForwardToServer ();
			return;
		}
	}
	else if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	save = host_client;

	if (Cmd_Argc() > 2 && strcmp(Cmd_Argv(1), "#") == 0)
	{
		i = atof(Cmd_Argv(2)) - 1;
		if (i < 0 || i >= svs.maxclients)
			return;
		if (!svs.clients[i].active)
			return;
		host_client = &svs.clients[i];
		byNumber = true;
	}
	else
	{
		for (i = 0, host_client = svs.clients; i < svs.maxclients; i++, host_client++)
		{
			if (!host_client->active)
				continue;
			if (stricmp(host_client->name, Cmd_Argv(1)) == 0)
				break;
		}
	}

	if (i < svs.maxclients)
	{
		if (cmd_source == src_command)
			if (cls.state == ca_dedicated)
				who = "Console";
			else
				who = cl_name.string;
		else
			who = save->name;

		// can't kick yourself!
		if (host_client == save)
			return;

		if (Cmd_Argc() > 2)
		{
			message = COM_Parse(Cmd_Args());
			if (byNumber)
			{
				message++;							// skip the #
				while (*message == ' ')				// skip white space
					message++;
				message += strlen(Cmd_Argv(2));	// skip the number
			}
			while (*message && *message == ' ')
				message++;
		}
		if (message)
			SV_ClientPrintf ("Kicked by %s: %s\n", who, message);
		else
			SV_ClientPrintf ("Kicked by %s\n", who);
		SV_DropClient (false);
	}

	host_client = save;
}

/*
===============================================================================

DEBUGGING TOOLS

===============================================================================
*/

/*
==================
Host_Give_f
==================
*/
void Host_Give_f (void)
{
	char	*t;
	int		v;
	eval_t	*val;

	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	t = Cmd_Argv(1);
	v = atoi (Cmd_Argv(2));
	
	switch (t[0])
	{
   case '0':
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      // MED 01/04/97 added hipnotic give stuff
      if (hipnotic)
      {
         if (t[0] == '6')
         {
            if (t[1] == 'a')
               sv_player->v.items = (int)sv_player->v.items | HIT_PROXIMITY_GUN;
            else
               sv_player->v.items = (int)sv_player->v.items | IT_GRENADE_LAUNCHER;
         }
         else if (t[0] == '9')
            sv_player->v.items = (int)sv_player->v.items | HIT_LASER_CANNON;
         else if (t[0] == '0')
            sv_player->v.items = (int)sv_player->v.items | HIT_MJOLNIR;
         else if (t[0] >= '2')
            sv_player->v.items = (int)sv_player->v.items | (IT_SHOTGUN << (t[0] - '2'));
      }
      else
      {
         if (t[0] >= '2')
            sv_player->v.items = (int)sv_player->v.items | (IT_SHOTGUN << (t[0] - '2'));
      }
		break;
	
    case 's':
		if (rogue)
		{
	        val = GetEdictFieldValue(sv_player, "ammo_shells1");
		    if (val)
			    val->_float = v;
		}

        sv_player->v.ammo_shells = v;
        break;		
    case 'n':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_nails1");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon <= IT_LIGHTNING)
					sv_player->v.ammo_nails = v;
			}
		}
		else
		{
			sv_player->v.ammo_nails = v;
		}
        break;		
    case 'l':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_lava_nails");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon > IT_LIGHTNING)
					sv_player->v.ammo_nails = v;
			}
		}
        break;
    case 'r':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_rockets1");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon <= IT_LIGHTNING)
					sv_player->v.ammo_rockets = v;
			}
		}
		else
		{
			sv_player->v.ammo_rockets = v;
		}
        break;		
    case 'm':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_multi_rockets");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon > IT_LIGHTNING)
					sv_player->v.ammo_rockets = v;
			}
		}
        break;		
    case 'h':
        sv_player->v.health = v;
        break;		
    case 'c':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_cells1");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon <= IT_LIGHTNING)
					sv_player->v.ammo_cells = v;
			}
		}
		else
		{
			sv_player->v.ammo_cells = v;
		}
        break;		
    case 'p':
		if (rogue)
		{
			val = GetEdictFieldValue(sv_player, "ammo_plasma");
			if (val)
			{
				val->_float = v;
				if (sv_player->v.weapon > IT_LIGHTNING)
					sv_player->v.ammo_cells = v;
			}
		}
        break;		
    }
}

edict_t	*FindViewthing (void)
{
	int		i;
	edict_t	*e;
	
	for (i=0 ; i<sv.num_edicts ; i++)
	{
		e = EDICT_NUM(i);
		if ( !strcmp (pr_strings + e->v.classname, "viewthing") )
			return e;
	}
	Con_Printf ("No viewthing on map\n");
	return NULL;
}

/*
==================
Host_Viewmodel_f
==================
*/
void Host_Viewmodel_f (void)
{
	edict_t	*e;
	model_t	*m;

	e = FindViewthing ();
	if (!e)
		return;

	m = Mod_ForName (Cmd_Argv(1), false);
	if (!m)
	{
		Con_Printf ("Can't load %s\n", Cmd_Argv(1));
		return;
	}
	
	e->v.frame = 0;
	cl.model_precache[(int)e->v.modelindex] = m;
}

/*
==================
Host_Viewframe_f
==================
*/
void Host_Viewframe_f (void)
{
	edict_t	*e;
	int		f;
	model_t	*m;

	e = FindViewthing ();
	if (!e)
		return;
	m = cl.model_precache[(int)e->v.modelindex];

	f = atoi(Cmd_Argv(1));
	if (f >= m->numframes)
		f = m->numframes-1;

	e->v.frame = f;		
}


void PrintFrameName (model_t *m, int frame)
{
	aliashdr_t 			*hdr;
	maliasframedesc_t	*pframedesc;

	hdr = (aliashdr_t *)Mod_Extradata (m);
	if (!hdr)
		return;
	pframedesc = &hdr->frames[frame];
	
	Con_Printf ("frame %i: %s\n", frame, pframedesc->name);
}

/*
==================
Host_Viewnext_f
==================
*/
void Host_Viewnext_f (void)
{
	edict_t	*e;
	model_t	*m;
	
	e = FindViewthing ();
	if (!e)
		return;
	m = cl.model_precache[(int)e->v.modelindex];

	e->v.frame = e->v.frame + 1;
	if (e->v.frame >= m->numframes)
		e->v.frame = m->numframes - 1;

	PrintFrameName (m, e->v.frame);		
}

/*
==================
Host_Viewprev_f
==================
*/
void Host_Viewprev_f (void)
{
	edict_t	*e;
	model_t	*m;

	e = FindViewthing ();
	if (!e)
		return;

	m = cl.model_precache[(int)e->v.modelindex];

	e->v.frame = e->v.frame - 1;
	if (e->v.frame < 0)
		e->v.frame = 0;

	PrintFrameName (m, e->v.frame);		
}

/*
===============================================================================

DEMO LOOP CONTROL

===============================================================================
*/


/*
==================
Host_Startdemos_f
==================
*/
void Host_Startdemos_f (void)
{
	int		i, c;

	if (cls.state == ca_dedicated)
	{
		if (!sv.active)
			Cbuf_AddText ("map start\n");
		return;
	}

	c = Cmd_Argc() - 1;
	if (c > MAX_DEMOS)
	{
		Con_Printf ("Max %i demos in demoloop\n", MAX_DEMOS);
		c = MAX_DEMOS;
	}

	//Con_Printf ("%i demo(s) in loop\n", c);

	for (i=1 ; i<c+1 ; i++)
		strncpy (cls.demos[i-1], Cmd_Argv(i), sizeof(cls.demos[0])-1);

	if (!sv.active && cls.demonum != -1 && !cls.demoplayback)
	{
		cls.demonum = 0;
		CL_NextDemo ();
	}
	else
		cls.demonum = -1;
}


/*
==================
Host_Demos_f

Return to looping demos
==================
*/
void Host_Demos_f (void)
{
	if (cls.state == ca_dedicated)
		return;
	if (cls.demonum == -1)
		cls.demonum = 1;
	CL_Disconnect_f ();
	CL_NextDemo ();
}

/*
==================
Host_Stopdemo_f

Return to looping demos
==================
*/
void Host_Stopdemo_f (void)
{
	if (cls.state == ca_dedicated)
		return;
	if (!cls.demoplayback)
		return;
	CL_StopPlayback ();
	CL_Disconnect ();
}


/*
==================
Host_Vidreset_f

at the present moment, this just rebuilds the lightmaps
==================
*/
void GL_BuildLightmaps (void);

void Host_Vidreset_f (void)
{
	GL_BuildLightmaps ();
}


/*
skybox control

(1) loading
(2) rotation

if the new skybox is the same as the old one, a skybox is not loaded again.  this can happen if
the map changes but the skybox does not.

if there is no skybox loaded, rotation cannot be set.
*/

int R_LoadSkys (char *path);
extern cvar_t _r_skybox;
extern int gl_useskybox;
char oldskybox[64];

void Host_LoadSkybox (char *sb_name, qboolean report)
{
	int valid_skybox;

#ifndef USE_EXTTEX
	// we need external textures to use a skybox, so don't use one if they're disabled
	Cvar_Set ("_r_skybox", "none");
	gl_useskybox = 0;
	return;
#else
	if (!strcmp (sb_name, "none"))
	{
		if (report) Con_Printf ("Skybox Disabled\n");
		Cvar_Set ("_r_skybox", "none");
		gl_useskybox = 0;
		return;
	}

	if (!strcmp (oldskybox, sb_name))
	{
		// it's the same skybox... these variables should already be set, but just in case...
		Cvar_Set ("_r_skybox", sb_name);
		gl_useskybox = 1;
		return;
	}

	if (report) Con_Printf ("Loading skybox: %s\n", sb_name);
	valid_skybox = R_LoadSkys (sb_name);

	if (!valid_skybox)
	{
		// an attempt to load an invalid skybox should leave it at it's current settings
		if (report) Con_Printf ("error : \"%s\" is not a valid skybox texture\n", sb_name);
		return;
	}

	if (report) Con_Printf ("\"%s\" loaded\n", sb_name);
	Cvar_Set ("_r_skybox", sb_name);
	sprintf (oldskybox, "%s", sb_name);
	gl_useskybox = 1;
#endif
}


void Host_Skybox_f (void)
{
	if (Cmd_Argc () != 2)
	{
		Con_Printf ("skybox <skybox name> : load a skybox texture\n");
		Con_Printf ("use a name of \"none\" for no skybox\n");
		return;
	}

	Host_LoadSkybox (Cmd_Argv (1), true);
}


// use this for any cvar/command/etc that is not supported in mhquake
void Fuck_Off_f (void)
{
	Con_Printf ("Not in this version of Quake, matey!\n");
}


void Host_Gameid_f (void)
{
	// just prints the current game id for testing and information purposes...
	Con_Printf ("Game ID is:\n%s\n", game_id);
}


//=============================================================================

void Host_Conback_f (void)
{
	// allow them to change the console background
	if (Cmd_Argc () != 2)
	{
		Con_Printf ("conback <texture path> : change the console background\n");
		return;
	}

	GL_LoadTextureReplace (Cmd_Argv (1), "conback", true);
}


void Host_TexReplace_f (void)
{
	// change any texture on the fly
	if (Cmd_Argc () != 3)
	{
		Con_Printf ("gl_texreplace <newtexturename> <oldtexturename>: change a texture\n");
		return;
	}

	GL_LoadTextureReplace (Cmd_Argv (1), Cmd_Argv (2), true);
}


/*
==================
Host_InitCommands
==================
*/
void Host_InitCommands (void)
{
	sprintf (oldskybox, "none");

	Cmd_AddCommand ("status", Host_Status_f);
	Cmd_AddCommand ("quit", Host_Quit_f);
	Cmd_AddCommand ("god", Host_God_f);
	Cmd_AddCommand ("notarget", Host_Notarget_f);
	Cmd_AddCommand ("fly", Host_Fly_f);
	Cmd_AddCommand ("map", Host_Map_f);
	Cmd_AddCommand ("restart", Host_Restart_f);
	Cmd_AddCommand ("changelevel", Host_Changelevel_f);
	Cmd_AddCommand ("connect", Host_Connect_f);
	Cmd_AddCommand ("reconnect", Host_Reconnect_f);
	Cmd_AddCommand ("name", Host_Name_f);
	Cmd_AddCommand ("noclip", Host_Noclip_f);
	Cmd_AddCommand ("version", Host_Version_f);
	Cmd_AddCommand ("say", Host_Say_f);
	Cmd_AddCommand ("say_team", Host_Say_Team_f);
	Cmd_AddCommand ("tell", Host_Tell_f);
	Cmd_AddCommand ("color", Host_Color_f);
	Cmd_AddCommand ("kill", Host_Kill_f);
	Cmd_AddCommand ("pause", Host_Pause_f);
	Cmd_AddCommand ("spawn", Host_Spawn_f);
	Cmd_AddCommand ("begin", Host_Begin_f);
	Cmd_AddCommand ("prespawn", Host_PreSpawn_f);
	Cmd_AddCommand ("kick", Host_Kick_f);
	Cmd_AddCommand ("ping", Host_Ping_f);
	Cmd_AddCommand ("load", Host_Loadgame_f);
	Cmd_AddCommand ("save", Host_Savegame_f);
	Cmd_AddCommand ("give", Host_Give_f);

	Cmd_AddCommand ("startdemos", Host_Startdemos_f);
	Cmd_AddCommand ("demos", Host_Demos_f);
	Cmd_AddCommand ("stopdemo", Host_Stopdemo_f);

	Cmd_AddCommand ("viewmodel", Host_Viewmodel_f);
	Cmd_AddCommand ("viewframe", Host_Viewframe_f);
	Cmd_AddCommand ("viewnext", Host_Viewnext_f);
	Cmd_AddCommand ("viewprev", Host_Viewprev_f);
	Cmd_AddCommand ("vid_reset", Host_Vidreset_f);
	Cmd_AddCommand ("loadsky", Host_Skybox_f);
	Cmd_AddCommand ("skybox", Host_Skybox_f);
	Cmd_AddCommand ("game_id", Host_Gameid_f);
	Cmd_AddCommand ("mcache", Mod_Print);
	Cmd_AddCommand ("conback", Host_Conback_f);
	Cmd_AddCommand ("gl_texreplace", Host_TexReplace_f);

	// generate a new game id at the start of each host
	GenerateUniqueGameID ();
}
