/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "quakedef.h"
#include "gl_rpart.h"

cvar_t r_particles = {"r_particles", "1", true};

/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int		i;

	i = COM_CheckParm ("-particles");

	if (i)
	{
		r_numparticles = (int)(atoi(com_argv[i+1]));
		if (r_numparticles < ABSOLUTE_MIN_PARTICLES) r_numparticles = ABSOLUTE_MIN_PARTICLES;
	}
	else r_numparticles = MAX_PARTICLES;

	Cvar_RegisterVariable (&r_particles);

	particles = (mh_particle_t *) Hunk_AllocName (r_numparticles * sizeof(mh_particle_t), "particles");
}


/*
===============
R_EntityParticles
===============
*/
void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	mh_particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;

	dist = 64;
	count = 50;

	if (!avelocities[0][0])
	{
		for (i=0 ; i<NUMVERTEXNORMALS*3 ; i++)
			avelocities[0][i] = (rand()&255) * 0.01;
	}

	if (r_particles.value)
	{
		for (i=0 ; i<NUMVERTEXNORMALS ; i++)
		{
			angle = cl.time * avelocities[i][0];
			sy = sin(angle);
			cy = cos(angle);
			angle = cl.time * avelocities[i][1];
			sp = sin(angle);
			cp = cos(angle);
			angle = cl.time * avelocities[i][2];
			sr = sin(angle);
			cr = cos(angle);
	
			forward[0] = cp*cy;
			forward[1] = cp*sy;
			forward[2] = -sp;

			if (!free_particles)
				return;

			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			p->die = cl.time + 0.01;
			p->color = 0x6f;
			p->type = pt_explode;
			p->radius = 15;
		
			p->org[0] = ent->origin[0] + r_avertexnormals[i][0]*dist + forward[0]*beamlength;			
			p->org[1] = ent->origin[1] + r_avertexnormals[i][1]*dist + forward[1]*beamlength;			
			p->org[2] = ent->origin[2] + r_avertexnormals[i][2]*dist + forward[2]*beamlength;			
		}
	}
	else
	{
		for (i=0 ; i<NUMVERTEXNORMALS ; i++)
		{
			angle = cl.time * avelocities[i][0];
			sy = sin(angle);
			cy = cos(angle);
			angle = cl.time * avelocities[i][1];
			sp = sin(angle);
			cp = cos(angle);
			angle = cl.time * avelocities[i][2];
			sr = sin(angle);
			cr = cos(angle);
	
			forward[0] = cp*cy;
			forward[1] = cp*sy;
			forward[2] = -sp;

			if (!free_particles)
				return;

			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			p->die = cl.time + 0.01;
			p->color = 0x6f;
			p->type = pt_explode_classic;
			p->radius = 0;
			p->particle_texture = particletexture;
			p->alpha_val = 0.9;
			p->vector_scale = 2;

			p->org[0] = ent->origin[0] + r_avertexnormals[i][0]*dist + forward[0]*beamlength;			
			p->org[1] = ent->origin[1] + r_avertexnormals[i][1]*dist + forward[1]*beamlength;			
			p->org[2] = ent->origin[2] + r_avertexnormals[i][2]*dist + forward[2]*beamlength;			
		}
	}
}


/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;
	
	free_particles = &particles[0];
	active_particles = NULL;

	for (i=0 ;i<r_numparticles ; i++)
		particles[i].next = &particles[i+1];
	particles[r_numparticles-1].next = NULL;
}


void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	mh_particle_t	*p;
	char	name[MAX_OSPATH];
	
	sprintf (name,"maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);

	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}
	
	Con_Printf ("Reading %s...\n", name);
	c = 0;

	for ( ;; )
	{
		r = fscanf (f,"%f %f %f\n", &org[0], &org[1], &org[2]);

		if (r != 3)
			break;

		c++;
		
		if (!free_particles)
		{
			Con_DPrintf ("Not enough free particles\n");
			break;
		}

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		
		p->die = 99999;
		p->color = (-c)&15;

		if (r_particles.value)
		{
			p->type = pt_static;
			p->radius = 2;
		}
		else
		{
			p->type = pt_static_classic;
			p->radius = 0;
			p->particle_texture = particletexture;
			p->alpha_val = 0.9;
			p->vector_scale = 2;
		}

		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, count, msgcount, color;

	for (i=0 ; i<3 ; i++)
		org[i] = MSG_ReadCoord ();
	for (i=0 ; i<3 ; i++)
		dir[i] = MSG_ReadChar () * (1.0/16);
	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

	if (msgcount == 255 && r_particles.value)
		count = NUM_EXPLODE_PARTICLES;	// was 1024
	else if (msgcount == 255 && !r_particles.value)
		count = NUM_EXP_PARTS_CLASSIC;	// was 1024
	else
		count = msgcount;
	
	R_RunParticleEffect (org, dir, color, count);
}
	

/*
===============
R_ParticleExplosionClassic

modified this to be the same as the original r_particleexplosion (classic particle system)
===============
*/
void R_ParticleExplosionClassic (vec3_t org)
{
	int			i, j;
	mh_particle_t	*p;

	for (i = 0; i < 1024; i++)
	{
		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = rand()&3;
		p->alpha_val = 0.9;
		p->vector_scale = 2;
		p->particle_texture = particletexture;

		if (i & 1)
		{
			p->type = pt_explode_classic;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
		else
		{
			p->type = pt_explode2_classic;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
	}
}


/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	mh_particle_t	*p;
	int			part_step;

	if (!r_particles.value)
	{
		R_ParticleExplosionClassic (org);
		return;
	}

	part_step = 0;

	for (i=0 ; i<NUM_EXPLODE_PARTICLES ; i++)	//was 1024
	{
		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 5;

		if (i & 1)
			p->type = pt_explode;
		else
			p->type = pt_explode2;

		if (part_step == 8)
		{
			p->die = p->die + 1;
			p->radius = 2;
			part_step = 0;
			p->ramp = 0;

			for (j=0 ; j<3 ; j++) p->vel[j] = (rand()%64)-32;
		}
		else
		{
			p->radius = 15;
			part_step++;
			p->ramp = rand()&3;

			for (j=0 ; j<3 ; j++) p->vel[j] = (rand()%64)-48;
		}

		for (j=0 ; j<3 ; j++) p->org[j] = org[j] + ((rand()%64)-32);

		p->color = ramp1[(int)p->ramp];
	}
}


void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	R_ParticleExplosion (org);
}


/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosionClassic (vec3_t org)
{
	int			i, j;
	mh_particle_t	*p;

	for (i = 0; i < NUM_EXP_PARTS_CLASSIC; i++)
	{
		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand()&8)*0.05;

		if (i & 1)
		{
			p->type = pt_blob_classic;
			p->color = 66 + rand()%6;

			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
		else
		{
			p->type = pt_blob2_classic;
			p->color = 150 + rand()%6;

			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
	}
}


void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	mh_particle_t	*p;
	int part_step;

	if (!r_particles.value)
	{
		R_BlobExplosionClassic (org);
		return;
	}

	part_step = 0;
	
	for (i=0 ; i<NUM_EXPLODE_PARTICLES ; i++)	// was 1024
	{
		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand()&8)*0.05;

		if (i & 1)
			p->type = pt_blob;
		else
			p->type = pt_blob2;

		if (part_step > 2)
		{
			p->radius = 2;
			p->ramp = 0;

			for (j=0 ; j<3 ; j++)
				p->vel[j] = (rand()%128)-64;
		}
		else
		{
			p->radius = 15;
			p->ramp = rand()&3;

			for (j=0 ; j<3 ; j++)
				p->vel[j] = 0;
		}

		if (part_step == 8)
			part_step = 0;
		else part_step++;

		for (j=0 ; j<3 ; j++)
			p->org[j] = org[j] + ((rand()%64)-32);

		p->color = ramp4[(int)p->ramp];
	}
}


/*
===============
R_RunParticleEffect

===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	mh_particle_t	*p;

	if (count == NUM_EXPLODE_PARTICLES)	// was 1024
	{
		R_ParticleExplosion (org);
		return;
	}
	else if (count == NUM_EXP_PARTS_CLASSIC)
	{
		R_ParticleExplosionClassic (org);
		return;
	}

	if (color == 73 || color == 225)
	{
		// reduce the number of blood particles - my new blood texture contains multiple different
		// sized splashes, so dividing the number of particles by 2, and adding 1 should be good.
		if (r_particles.value) count = (count / 2) + 1;

		for (i = 0; i < count; i++)
		{
			if (!free_particles)
				return;

			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			p->die = cl.time + 0.1*(rand()%5);

			// only the lightning gun uses 225 - so we'll change this sometime and do a different
			// effect for it... no need to modify these to classic particles here - the main
			// drawing function will do that for us.
			p->type = pt_blood;
			p->alpha_val = 0.9;
			p->vector_scale = 6;
			p->radius = 0;
			p->particle_texture = bloodtexture;
			p->color = 224 + (rand()&3);

			// use a lower velocity - the fact that we have a smaller number of blood particles
			// means that they seem to fly off faster - which is ugly.  reduce the origin too.
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()&15)-8);
				p->vel[j] = dir[j]*5;
			}
		}
	}
	else 
	{
		for (i = 0; i < count; i++)
		{
			if (!free_particles)
				return;

			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			p->die = cl.time + 0.1*(rand()%5);

			// retain the old particle texture here - glows look silly
			p->color = (color&~7) + (rand()&7);
			p->type = pt_slowgrav;
			p->radius = 0;
			p->particle_texture = particletexture;
			p->alpha_val = 0.9;
			p->vector_scale = 2;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand()&15)-8);
				p->vel[j] = dir[j]*15;// + (rand()%300)-150;
			}
		}
	}
}


void R_FlameSmokes (vec3_t org)
{
	mh_particle_t	*p;
	int rand_mod;

	// don't draw if paused or in the console, or otherwise
	if (key_dest != key_game) return;
	if (!r_torchsmoke.value) return;
	if (!free_particles) return;

	if (r_particles.value)
		rand_mod = 15;
	else rand_mod = 3;

	if (rand() % rand_mod) return;

	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	p->die = cl.time + (rand()%3);

	p->type = pt_antigrav;
	p->radius = 0;
	p->particle_texture = smoketexture;

	p->color = 3 + (rand() % 5);
	p->alpha_val = 0.3;
	p->vector_scale = 6;
	p->radius = 0;

	p->org[0] = org[0] - 3 + (rand () % 5);
	p->org[1] = org[1] - 3 + (rand () % 5);
	p->org[2] = org[2] + 40;

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 30;
}


/*
===============
R_LavaSplash

===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j;
	mh_particle_t	*p;
	float		vel;
	vec3_t		dir;
	int loop_inc;

	if (r_particles.value)
		loop_inc = 4;
	else loop_inc = 1;

	for (i = -16; i <= 16; i += loop_inc)
	{
		for (j = -16; j <= 16; j += loop_inc)
		{
			if (!free_particles)
				return;
			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;
		
			p->die = cl.time + 2 + (rand()&31) * 0.02;
			p->color = 224 + (rand()&7);
			p->type = pt_slowgrav;
			p->radius = 2;
				
			dir[0] = j*8 + (rand()&7);
			dir[1] = i*8 + (rand()&7);
			dir[2] = 256;
	
			p->org[0] = org[0] + dir[0];
			p->org[1] = org[1] + dir[1];
			p->org[2] = org[2] + (rand()&63);
	
			VectorNormalize (dir);						
			vel = 50 + (rand()&63);
			VectorScale (dir, vel, p->vel);
		}
	}
}

/*
===============
R_TeleportSplash
===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	mh_particle_t	*p;
	float		vel;
	vec3_t		dir;
	int loop_inc;

	if (r_particles.value)
		loop_inc = 8;
	else loop_inc = 4;

	for (i = -16; i <= 16; i += loop_inc)
	{
		for (j = -16; j <= 16; j += loop_inc)
		{
			for (k =- 24; k <= 32; k += loop_inc)
			{
				if (!free_particles)
					return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;
		
				p->die = cl.time + 0.2 + (rand()&7) * 0.02;
				p->color = 5 + (rand()&7);
				p->type = pt_slowgrav;
				p->radius = 1;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand()&3);
				p->org[1] = org[1] + j + (rand()&3);
				p->org[2] = org[2] + k + (rand()&3);

				VectorNormalize (dir);						
				vel = 50 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}
}


// use particles for zerstorer rain rather than sprites (faster)
void R_ZerstorerRain (vec3_t org)
{
	int i;
	mh_particle_t	*p;

	// don't draw if paused or in the console, or otherwise
	if (key_dest != key_game) return;

	for (i = 0; i < 16; i++)
	{
		if (!free_particles) return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1;

		p->type = pt_static;
		p->radius = 0;
		p->particle_texture = particletexture;

		p->color = 39 + (rand() % 5);
		p->alpha_val = 0.9;
		p->vector_scale = 1;
		p->radius = 0;

		p->org[0] = org[0] + (rand() % 256) - 128;
		p->org[1] = org[1] + (rand() % 256) - 128;
		p->org[2] = org[2] + (rand() % 256) - 128;

		p->vel[0] = 0;
		p->vel[0] = 0;
		p->vel[2] = -128;
	}
}


void MH_RocketTrail (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 2;
	p->ramp = (rand()&3);
	p->color = ramp3[(int)p->ramp];
	p->type = pt_trail;
	p->alpha_val = 0.3;
	p->vector_scale = 6;
	p->radius = 0;
	p->particle_texture = smoketexture;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_RocketTrailClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	p->ramp = (rand()&3);
	p->color = ramp3[(int)p->ramp];
	p->type = pt_trail_classic;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_SmokeSmoke (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 2;
	p->ramp = (rand()&3) + 2;
	p->color = ramp3[(int)p->ramp];
	p->type = pt_trail;
	p->alpha_val = 0.3;
	p->vector_scale = 6;
	p->radius = 0;
	p->particle_texture = smoketexture;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_SmokeSmokeClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	p->ramp = (rand()&3) + 2;
	p->color = ramp3[(int)p->ramp];
	p->type = pt_trail_classic;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_Blood (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 2;
	p->type = pt_blood;
	p->alpha_val = 0.9;
	p->vector_scale = 6;
	p->radius = 0;
	p->particle_texture = bloodtexture;
	p->color = 224 + (rand()&3);

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_BloodClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	p->type = pt_blood;
	p->color = 224 + (rand()&3);

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_WizardTrail (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int tracer;

	tracer = *tracercount;
	*len -= 3;
	p->die = cl.time + 0.5;
	p->type = pt_static;
	p->radius = 2;

	p->color = 52 + ((tracer&4)<<1);
	(*tracercount)++;

	VectorCopy (start, p->org);

	if (++tracer & 1)
	{
		p->vel[0] = 30*vec[1];
		p->vel[1] = 30*-vec[0];
	}
	else
	{
		p->vel[0] = 30*-vec[1];
		p->vel[1] = 30*vec[0];
	}
}


void MH_WizardTrailClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int tracer;

	tracer = *tracercount;
	p->die = cl.time + 0.5;
	p->type = pt_static_classic;

	p->color = 52 + ((tracer&4)<<1);
	(*tracercount)++;

	VectorCopy (start, p->org);

	if (++tracer & 1)
	{
		p->vel[0] = 30*vec[1];
		p->vel[1] = 30*-vec[0];
	}
	else
	{
		p->vel[0] = 30*-vec[1];
		p->vel[1] = 30*vec[0];
	}
}


void MH_SlightBlood (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 4;
	p->type = pt_blood;
	p->alpha_val = 0.9;
	p->vector_scale = 6;
	p->radius = 0;
	p->particle_texture = bloodtexture;
	p->color = 224 + (rand()&3);

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_SlightBloodClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 3;
	p->type = pt_blood_classic;
	p->color = 224 + (rand()&3);

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()%6)-3);
}


void MH_KnightTrail (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int tracer;

	tracer = *tracercount;
	*len -= 3;
	p->die = cl.time + 0.5;
	p->type = pt_static;
	p->radius = 2;

	p->color = 230 + ((tracer&4)<<1);
	(*tracercount)++;

	VectorCopy (start, p->org);

	if (++tracer & 1)
	{
		p->vel[0] = 30*vec[1];
		p->vel[1] = 30*-vec[0];
	}
	else
	{
		p->vel[0] = 30*-vec[1];
		p->vel[1] = 30*vec[0];
	}
}


void MH_KnightTrailClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int tracer;

	tracer = *tracercount;
	p->die = cl.time + 0.5;
	p->type = pt_static_classic;

	p->color = 230 + ((tracer&4)<<1);
	(*tracercount)++;

	VectorCopy (start, p->org);

	if (++tracer & 1)
	{
		p->vel[0] = 30*vec[1];
		p->vel[1] = 30*-vec[0];
	}
	else
	{
		p->vel[0] = 30*-vec[1];
		p->vel[1] = 30*vec[0];
	}
}


void MH_VoorTrail (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	*len -= 2;
	p->color = 9*16 + 8 + (rand()&3);
	p->type = pt_static;
	p->radius = 2;
	p->die = cl.time + 0.3;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()&15)-8);
}


void MH_VoorTrailClassic (mh_particle_t *p, float *len, int *tracercount, vec3_t start, vec3_t vec)
{
	int j;

	p->color = 9*16 + 8 + (rand()&3);
	p->type = pt_static_classic;
	p->die = cl.time + 0.3;

	for (j=0 ; j<3 ; j++)
		p->org[j] = start[j] + ((rand()&15)-8);
}


void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t vec;
	float len;
	int	j;
	mh_particle_t *p;
	int	dec;
	static int tracercount;
	qboolean sparky;
	int oldlen;
	vec3_t oldstart;

	void (*RocketTrailFunc) ();

	// could use VectorCopy, but wtf?
	oldstart[0] = start[0];
	oldstart[1] = start[1];
	oldstart[2] = start[2];

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 3;
	else
	{
		dec = 1;
		type -= 128;
	}

	if (!r_particles.value) type += 7;

	sparky = false;
	oldlen = len;

	switch (type)
	{
	case 0:		// rocket trail
		RocketTrailFunc = MH_RocketTrail;
		sparky = true;
		break;

	case 1:		// smoke smoke
		RocketTrailFunc = MH_SmokeSmoke;
		sparky = true;
		break;

	case 2:		// blood
		RocketTrailFunc = MH_Blood;
		break;

	case 3:		// wizard trail
		RocketTrailFunc = MH_WizardTrail;
		break;

	case 4:		// slight blood
		RocketTrailFunc = MH_SlightBlood;
		break;

	case 5:		// knight trail
		RocketTrailFunc = MH_KnightTrail;
		break;

	case 6:		// voor trail
		RocketTrailFunc = MH_VoorTrail;
		break;

	case 7:		// rocket trail
		RocketTrailFunc = MH_RocketTrailClassic;
		break;

	case 8:		// smoke smoke
		RocketTrailFunc = MH_SmokeSmokeClassic;
		break;

	case 9:		// blood
		RocketTrailFunc = MH_BloodClassic;
		break;

	case 10:	// wizard trail
		RocketTrailFunc = MH_WizardTrailClassic;
		break;

	case 11:	// slight blood
		RocketTrailFunc = MH_SlightBloodClassic;
		break;

	case 12:	// knight trail
		RocketTrailFunc = MH_KnightTrailClassic;
		break;

	case 13:	// voor trail
		RocketTrailFunc = MH_VoorTrailClassic;
		break;

	default:	// just in case...
		return;
	}

	while (len > 0)
	{
		// common
		len -= dec;

		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// different function for each type
		(*RocketTrailFunc) (p, &len, &tracercount, start, vec);

		VectorAdd (start, vec, start);
	}

	// speed up by allocating (and therefore rendering, cos particles are rendered in the order
	// they are allocated) trail sparks after the main trail allocation.
	if (sparky)
	{
		while (oldlen > 0)
		{
			oldlen -= 5;

			if (!free_particles)
				return;

			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			VectorCopy (vec3_origin, p->vel);
			p->die = cl.time + 1;	// die faster than the smoke

			p->ramp = (rand()&3);
			p->color = ramp3[(int)p->ramp];
			p->type = pt_trail_classic;

			p->vector_scale = 2;
			p->radius = 0;
			p->alpha_val = 0.9;
			p->particle_texture = particletexture;

			for (j=0 ; j<3 ; j++)
				p->org[j] = oldstart[j] + ((rand()%12)-6);

			VectorAdd (oldstart, vec, oldstart);
		}
	}
	else return;
}


/*
===============
R_DrawParticles
===============
*/
void R_DrawParticles (void)
{
	mh_particle_t		*p, *kill;
	float			grav;
	int				i;
	float			time2, time3;
	float			time1;
	float			dvel;
	
	vec3_t			up, right;
	float			scale;

	float rad;

	vec3_t pscalemody;
	vec3_t p_org;

	time3 = frametime * 15;
	time2 = frametime * 10;
	time1 = frametime * 5;
	grav = frametime * sv_gravity.value * 0.05;
	dvel = 4*frametime;

	glDepthMask (0);
	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// speed up computation of particle scale
	pscalemody[0] = r_origin[0] * vpn[0];
	pscalemody[1] = r_origin[1] * vpn[1];
	pscalemody[2] = r_origin[2] * vpn[2];

	glBindTexture (GL_TEXTURE_2D, particletexture);
	oldtexture = particletexture;
	glBegin (GL_QUADS);

	// find the first available particle
	for ( ;; ) 
	{
		kill = active_particles;
		if (kill && kill->die < cl.time)
		{
			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;
			continue;
		}

		break;
	}

	for (p=active_particles ; p ; p=p->next)
	{
		// find the next available particle
		for ( ;; )
		{
			kill = p->next;
			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				continue;
			}

			break;
		}

		// this is only really an issue for explosion particles (radius 3 or more)
		// particle checking in the next frame will kill off these particles and
		// compress the array for us.
		// should also apply to flame smoke
		if (p->radius > 2 || p->type == pt_antigrav)
		{
			// this was != contents_empty, meaning underwater explosions weren't being rendered
			if (Mod_PointInLeaf (p->org, cl.worldmodel)->contents == CONTENTS_SOLID)
			{
				p->die = -1;
				continue;
			}
		}

		// hack a scale up to keep particles from disapearing
		scale = ((p->org[0] * vpn[0]) - pscalemody[0]) +
			((p->org[1] * vpn[1]) - pscalemody[1]) +
			((p->org[2] * vpn[2]) - pscalemody[2]);

		// if we're using the classic particle system, set the radius, vector_scale, alpha_val
		// and particle_texture properly - this lets us switch between particle systems in the
		// middle of a frame in which particles are being drawn.
		if (!r_particles.value)
		{
			p->alpha_val = 0.9;
			p->particle_texture = particletexture;
			p->radius = 0;
			p->vector_scale = 2;
		}

		// don't bother drawing particles that are behind the viewpoint!  for glow effects, 
		// half the radius must be added to this, for smoke textures, use a minumum scale of 40, 
		// blood is OK to use 0 with.  note that glquake uses a depth of 4 as it's minimum.
		if (p->radius != 0 && scale > (p->radius / 2) + 4)
		{
			if (scale > p->radius / 2)
			{
				scale = 1.0 + scale * 0.0015;
				rad = p->radius * scale;

				R_AddGlowEffect (RTable[(int)p->color], GTable[(int)p->color], 
					BTable[(int)p->color], rad, p->org);
			}
		}
		else if ((scale > 40 && p->particle_texture == smoketexture) || 
			(scale > 0 && p->particle_texture == bloodtexture) ||
			scale > 0 && p->particle_texture == particletexture)
		{
			if (p->particle_texture != oldtexture)
			{
				// since particles are drawn in batches of all the same type, a state change
				// here won't really matter...
				glEnd ();
			    glBindTexture (GL_TEXTURE_2D, p->particle_texture);
				oldtexture = p->particle_texture;
				glBegin (GL_QUADS);
			}

			scale = 1.0 + scale * 0.004;

			VectorScale (vup, p->vector_scale, up);
			VectorScale (vright, p->vector_scale, right);

			up[0] = up[0] * scale;
			up[1] = up[1] * scale;
			up[2] = up[2] * scale;

			right[0] = right[0] * scale;
			right[1] = right[1] * scale;
			right[2] = right[2] * scale;

			// this makes the particle texture somewhat off center - so we must center the
			// texture exactly on the origin - up is not such an issue, but right matters a
			// hell of a lot.
			p_org[0] = p->org[0] - (right[0] / 2) - (up[0] / 2);
			p_org[1] = p->org[1] - (right[1] / 2) - (up[0] / 2);
			p_org[2] = p->org[2] - (right[2] / 2) - (up[0] / 2);

			glColor4f (RTable[(int)p->color], 
				GTable[(int)p->color], 
				BTable[(int)p->color], 
				p->alpha_val);

			glTexCoord2f (0, 0);
			glVertex3fv (p_org);

			glTexCoord2f (1, 0);
			glVertex3f (p_org[0] + up[0], p_org[1] + up[1], p_org[2] + up[2]);

			glTexCoord2f (1, 1);
			glVertex3f (p_org[0] + up[0] + right[0], p_org[1] + up[1] + right[1], p_org[2] + up[2] + right[2]);

			glTexCoord2f (0, 1);
			glVertex3f (p_org[0] + right[0], p_org[1] + right[1], p_org[2] + right[2]);
		}
		else
		{
			// kill the particle
			p->die = -1;
			continue;
		}

		// only modify these once per frame.
		if (mirror_particles) continue;

		if (p->vel[0] || p->vel[1] || p->vel[2])
		{
			p->org[0] += p->vel[0]*frametime;
			p->org[1] += p->vel[1]*frametime;
			p->org[2] += p->vel[2]*frametime;
		}

		switch (p->type)
		{
		case pt_static:
		case pt_static_classic:
			break;

		case pt_fire:
		case pt_trail:
		case pt_fire_classic:
		case pt_trail_classic:
			p->ramp += time1;

			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int)p->ramp];

			p->vel[2] += grav;
			break;

		case pt_explode_classic:
			p->ramp += time2;

			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp1[(int)p->ramp];

			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;

			p->vel[2] -= grav;
			break;

		case pt_explode:
			p->ramp += time2;

			// change the colour
			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp1[(int)p->ramp];

			if (p->radius > 2)
				p->radius = p->radius + time2;
			break;

		case pt_explode2_classic:
			p->ramp += time3;

			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp2[(int)p->ramp];

			for (i=0 ; i<3 ; i++)
				p->vel[i] -= p->vel[i]*frametime;

			p->vel[2] -= grav;
			break;

		case pt_explode2:
			p->ramp += time3;

			// change the colour
			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp2[(int)p->ramp];

			if (p->radius > 2)
				p->radius = p->radius + time3;
			break;

		case pt_blob_classic:
			for (i = 0; i < 3; i++)
				p->vel[i] += p->vel[i]*dvel;

			p->vel[2] -= grav;
			break;

		case pt_blob:
			p->ramp += time2;

			// change the colour
			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp4[(int)p->ramp];

			p->vel[2] += grav;
			break;

		case pt_blob2_classic:
			for (i = 0; i < 2; i++)
				p->vel[i] -= p->vel[i]*dvel;

			p->vel[2] -= grav;
			break;

		case pt_blob2:
			p->ramp += time3;

			// change the colour
			if (p->ramp >=8)
				p->die = -1;
			else p->color = ramp5[(int)p->ramp];

			p->vel[2] += grav;
			break;

		case pt_antigrav:
		case pt_antigrav_classic:
			// this isn't actually used in the game as yet, but can be...
			p->vel[2] += grav;
			break;

		case pt_blood:
		case pt_blood_classic:
		case pt_grav:
		case pt_slowgrav:
		case pt_tele:
		case pt_grav_classic:
		case pt_slowgrav_classic:
		case pt_tele_classic:
			p->vel[2] -= grav;
			break;

		default:
			// in case i forgot something...
			break;
		}
	}

	glEnd ();
	glDisable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDepthMask (1);
}

