/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// Quake is a trademark of Id Software, Inc., (c) 1996 Id Software, Inc. All
// rights reserved.

/*
MH - all ripped out in prep for MP3 playback
*/

#include <windows.h>
#include <vfw.h>
#include "quakedef.h"

// it's a dumb warning anyway...
#pragma warning(disable : 4761)

extern HINSTANCE global_hInstance;
extern	HWND	mainwindow;
extern	cvar_t	bgmvolume;
extern int sound_started;

HWND m_Video;
byte CurrentTrack;

qboolean Looping;
qboolean Paused;
qboolean Enabled;

static void CDAudio_Eject (void) {}
static void CDAudio_CloseDoor (void) {}
static int CDAudio_GetAudioDiskInfo (void) {return 0;}

void CDAudio_Stop (void)
{
	if (!sound_started) return;

	if (m_Video)
	{
		// prevent looping
		Looping = false;
		MCIWndStop (m_Video);
		MCIWndDestroy (m_Video);
		m_Video = NULL;
	}
}

void CDAudio_Pause (void)
{
	if (!sound_started) return;

	if (m_Video && !Paused)
	{
		MCIWndPause (m_Video);
		Paused = true;
	}
}

void CDAudio_Resume (void)
{
	if (!sound_started) return;

	if (m_Video && Paused)
	{
		MCIWndResume (m_Video);
		Paused = false;
	}
}

void CDAudio_Update (void) {}
static void CD_f (void);


int CDAudio_Init (void)
{
	m_Video = NULL;

	if (cls.state == ca_dedicated || COM_CheckParm("-nocdaudio"))
	{
		Enabled = false;
		return -1;
	}

	Looping = false;
	Paused = false;

	Enabled = true;

	Cmd_AddCommand ("cd", CD_f);
	Cmd_AddCommand ("mp3", CD_f);

	return 0;
}

void CDAudio_Shutdown (void)
{
	if (m_Video)
	{
		// prevent looping
		Looping = false;
		MCIWndStop (m_Video);
		MCIWndDestroy (m_Video);
		m_Video = NULL;
	}
}


qboolean MP3Touch (char *filePath, const char *track)
{
	FILE *f;

	if (!sound_started) return false;

	// first time we've tried to get this track.  otherwise we will already have it and
	// can just reuse the saved out path
	// first, 'touch' the file to see if it exists
	// try the current game dir first
	// do it this way cos va is buggy
	sprintf (filePath, "%s/music/%s.mp3", com_gamedir, track);
	f = fopen (filePath, "rb");

	Con_DPrintf ("Attempting %s\n", filePath);

	if (!f)
	{
		// try id1
		sprintf (filePath, "id1/music/%s.mp3", track);
		f = fopen (filePath, "rb");

		if (!f)
		{
			// standard quake has 10 tracks, numbered 01 to 10 with leading zeroes, track 05
			// is my personal fave so we'll have a look for that in id1
			sprintf (filePath, "id1/music/05.mp3", track);
			f = fopen (filePath, "rb");

			if (!f)
			{
				// she definitely ain't there, so git yo' ass out
				Con_DPrintf ("Could not find \"%s.mp3\"\n", track);
				return false;
			}
			else Con_DPrintf ("Could not find \"%s.mp3\"\nSubstituting with 05.mp3\n", track);
		}
	}

	// close the file, we don't need it any more
	fclose (f);

	// we've found *something* we can use, so use it
	Con_DPrintf ("Playing %s\n", filePath);

	return true;
}


void CDAudio_PlayMP3 (const char *track, qboolean looping)
{
	char filePath[256];

	// stop any previous instances
	CDAudio_Stop ();

	// find the MP3
	if (!MP3Touch (filePath, track))
		return;

	if (!sound_started) return;

	// play it without the toolbar
	// this creates an MCI device the lazy way...
	// a NULL HINSTANCE here gives better looping performance in the main game window
	m_Video = MCIWndCreate (mainwindow, NULL, MCIWNDF_NOPLAYBAR | MCIWNDF_NOTIFYMODE, filePath);

	// set the volume
	MCIWndSetVolume (m_Video, (int)(bgmvolume.value > 1.0f ? 1000.0f : bgmvolume.value * 999.0f));

	// and play it
	MCIWndPlay (m_Video);

	// OK, we're playing so we can go back to processing stop messages normally
	Looping = looping;
}


void CDAudio_Play (byte track, qboolean looping)
{
	if (!sound_started) return;

	// worldspawn tracks are one ahead of the game
	CurrentTrack = track - 1;

	// exit conditions
	if (CurrentTrack > 99) return;
	if (CurrentTrack < 1) return;
	if (!Enabled) return;

	CDAudio_PlayMP3 (va ("%02i", CurrentTrack), looping);
}


void CDAudioSetVolume (void)
{
	if (!sound_started) return;

	// set the volume
	MCIWndSetVolume (m_Video, (int)(bgmvolume.value > 1.0f ? 1000.0f : bgmvolume.value * 999.0f));
}


static void CD_f (void)
{
	char	*command;
	int		ret;
	int		n;
	int		startAddress;

	if (Cmd_Argc() < 2)
	{
		Con_Printf("commands:\n");
		Con_Printf("on, off, play, stop, loop, pause, resume\n");
		return;
	}

	command = Cmd_Argv (1);

	if (Q_strcasecmp(command, "on") == 0)
	{
		Enabled = true;
		return;
	}

	if (Q_strcasecmp(command, "off") == 0)
	{
		CDAudio_Stop ();

		Enabled = false;
		return;
	}

	if (Q_strcasecmp(command, "play") == 0)
	{
		int i;
		char MP3Name[256];

		// initialize so we have a valid first sprintf below
		memset (MP3Name, 0, sizeof (MP3Name));

		// get the full MP3 Name, including spaces
		for (i = 2; i < Cmd_Argc (); i++)
			sprintf (MP3Name, "%s %s", MP3Name, Cmd_Argv (i));

		// the first character will always be a space so skip it
		CDAudio_PlayMP3 (&MP3Name[1], false);

		return;
	}

	if (Q_strcasecmp(command, "loop") == 0)
	{
		int i;
		char MP3Name[256];

		// initialize so we have a valid first sprintf below
		MP3Name[0] = '\0';

		// get the full MP3 Name, including spaces
		for (i = 2; i < Cmd_Argc (); i++)
			sprintf (MP3Name, "%s %s", MP3Name, Cmd_Argv (i));

		// the first character will always be a space so skip it
		CDAudio_PlayMP3 (&MP3Name[1], true);

		return;
	}

	if (Q_strcasecmp(command, "stop") == 0)
	{
		CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "pause") == 0)
	{
		CDAudio_Pause();
		return;
	}

	if (Q_strcasecmp(command, "resume") == 0)
	{
		CDAudio_Resume();
		return;
	}

	Con_Printf("valid commands:\n");
	Con_Printf("on, off, play, stop, loop, pause, resume\n");
}


LONG CDAudio_MessageHandler (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (lParam)
	{
	case MCI_MODE_PLAY:
	case MCI_MODE_RECORD:
	case MCI_MODE_SEEK:
	case MCI_MODE_OPEN:
	case MCI_MODE_NOT_READY:
		break;

	case MCI_MODE_PAUSE:
		// pause or resume playback as required
		if (Paused)
			CDAudio_Resume ();
		else CDAudio_Pause ();
			break;

	case MCI_MODE_STOP:
		// this is the only one we're interested in, we have to handle a stop message
		// if we're looping playback - and only if we're looping playback...!
		if (!Looping) break;

		// just reuse the current window
		MCIWndPlay (m_Video);
		break;

	default:
		// bad mode, just quit uncomplainingly
		return 1;
	}

	return 0;
}
