// sbar widgets


#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


#define STAT_MINUS 10

// for the scorebar
qboolean sb_showscores;


// 2D sbar gfx
qpic_t		*sb_nums[2][11];
qpic_t		*sb_colon, *sb_slash;


// 3D sbar gfx
model_t *InvWeaps[10] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
model_t *HUD_BitModels[10] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
model_t *HUD_Ammo[7] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
model_t *HUD_Invl = NULL;
model_t *HUD_Quad = NULL;
model_t *HUD_Arm = NULL;
model_t *HUD_Face = NULL;


// names of the 3D stuff
char *Weaponses[] =
{
	// normal quake
	"progs/g_shot.mdl",
	"progs/g_shot.mdl",
	"progs/g_nail.mdl",
	"progs/g_nail2.mdl",
	"progs/g_rock.mdl",
	"progs/g_rock2.mdl",
	"progs/g_light.mdl",
	// hipnotic
	"progs/g_hammer.mdl",
	"progs/g_laserg.mdl",
	"progs/g_prox.mdl"
	// no g_*.mdl files for rogue
};

char *HUD_BitNames[] =
{
	"progs/end1.mdl",
	"progs/end2.mdl",
	"progs/end3.mdl",
	"progs/end4.mdl",
	"progs/b_s_key.mdl",
	"progs/b_g_key.mdl",
	"progs/m_s_key.mdl",
	"progs/m_g_key.mdl",
	"progs/w_s_key.mdl",
	"progs/w_g_key.mdl"
};

char *HUD_AmmoNames[] =
{
	// normal quake
	"maps/b_shell1.bsp",
	"maps/b_nail1.bsp",
	"maps/b_rock0.bsp",
	"maps/b_batt1.bsp",
	// rogue
	"maps/b_lnail1.bsp",
	"maps/b_mrock1.bsp",
	"maps/b_plas1.bsp"
	// hipnotic has no ammo box models
};


void Sbar_Preload3DModels (void)
{
	// preload all the models to prevent disk access at run time
	// done every map load to also prevent Quake from discarding the models or textures
	int i;

	// weapons
	for (i = 0; i < 10; i++)
		InvWeaps[i] = Mod_ForName (Weaponses[i], false);

	// bits (keys/sigils)
	for (i = 0; i < 10; i++)
		HUD_BitModels[i] = Mod_ForName (HUD_BitNames[i], false);

	// ammo - loaded as required (it messes with the world lighting)
	for (i = 0; i < 7; i++)
		HUD_Ammo[i] = Mod_ForName (HUD_AmmoNames[i], false);

	// powerups
	HUD_Invl = Mod_ForName ("progs/invulner.mdl", false);
	HUD_Quad = Mod_ForName ("progs/quaddama.mdl", false);

	// armour model
	HUD_Arm = Mod_ForName ("progs/armor.mdl", false);

	// face model
	HUD_Face = Mod_ForName ("maps/b_bh100.bsp", false);
}


void Sbar_ShowScores (void)
{
	sb_showscores = true;
}


void Sbar_DontShowScores (void)
{
	sb_showscores = false;
}


void Sbar_Init (void)
{
	int		i;

	// these are the only images we need now
	for (i = 0; i < 10; i++)
	{
		sb_nums[0][i] = Draw_PicFromWad (va ("num_%i", i));
		sb_nums[1][i] = Draw_PicFromWad (va ("anum_%i", i));
	}

	sb_nums[0][10] = Draw_PicFromWad ("num_minus");
	sb_nums[1][10] = Draw_PicFromWad ("anum_minus");

	sb_colon = Draw_PicFromWad ("num_colon");
	sb_slash = Draw_PicFromWad ("num_slash");

	// set up the item positions

	// preload the 3D stuff
	Sbar_Preload3DModels ();

	// scoreboard
	Cmd_AddCommand ("+showscores", Sbar_ShowScores);
	Cmd_AddCommand ("-showscores", Sbar_DontShowScores);
}


void Sbar_Changed (void)
{
	// do nothing!!!
}


void Sbar_DrawScoreboard (void)
{
}


int Sbar_itoa (int num, char *buf)
{
	char *str;
	int pow10;
	int dig;

	str = buf;

	if (num < 0)
	{
		*str++ = '-';
		num = -num;
	}

	for (pow10 = 10; num >= pow10; pow10 *= 10);

	do
	{
		pow10 /= 10;
		dig = num / pow10;
		*str++ = '0' + dig;
		num -= dig * pow10;
	} while (pow10 != 1);

	*str = 0;

	return str - buf;
}


void HUD_DrawNum (int x, int y, int num, int digits, int color)
{
	char str[12];
	char *ptr;
	int l, frame;

	l = Sbar_itoa (num, str);

	ptr = str;

	if (l > digits) ptr += (l - digits);
	if (l < digits) x += (digits - l) * 24;

	if (num < -99)
		num = -99;
	if (num < 0)
		;
	else if (num < 10) x -= 24;
	else if (num < 100) x -= 12;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		GL_Draw2DPic (x, y, sb_nums[color][frame]);

		x += 24;
		ptr++;
	}
}


void Draw_XHair (void)
{
	glDisable (GL_TEXTURE_2D);

	glBegin (GL_LINES);

	// make it green
	glColor3f (0, 1, 0);

	// vertical bar top
	glVertex2f (320, 253);
	glVertex2f (320, 256);

	// vertical bar bottom
	glVertex2f (320, 264);
	glVertex2f (320, 267);

	// horizontal bar left
	glVertex2f (313, 260);
	glVertex2f (316, 260);

	// horizontal bar right
	glVertex2f (324, 260);
	glVertex2f (327, 260);

	glColor3f (1, 1, 1);

	glEnd ();

	glEnable (GL_TEXTURE_2D);
}


void HUD_DrawAmmo (void)
{
	qboolean ammo_draw = false;

	// ammo icon
	if (rogue)
	{
		if (cl.items & (RIT_SHELLS | 
						RIT_NAILS | 
						RIT_ROCKETS | 
						RIT_CELLS | 
						RIT_LAVA_NAILS | 
						RIT_PLASMA_AMMO | 
						RIT_MULTI_ROCKETS))
			ammo_draw = true;
	}
	else
	{
		if (cl.items & (IT_SHELLS | IT_NAILS | IT_ROCKETS | IT_CELLS))
			ammo_draw = true;
	}

	// either we've got the axe, or some funky weapon we don't have a pic for, so just get out
	if (!ammo_draw) return;

	HUD_DrawNum (520, 447, cl.stats[STAT_AMMO], 3, cl.stats[STAT_AMMO] <= 10);

	Draw_XHair ();
}


void HUD_DrawArmor (void)
{
	int ArmourValue;
	int colour;

	if (cl.items & IT_INVULNERABILITY)
	{
		ArmourValue = 666;
		colour = 1;
	}
	else if (!cl.stats[STAT_ARMOR])
		return;
	else
	{
		ArmourValue = cl.stats[STAT_ARMOR];
		colour = (cl.stats[STAT_ARMOR] <= 25);
	}

	HUD_DrawNum (47, 
				 403, 
				 ArmourValue, 
				 3, 
				 colour);
}


void Sbar_Draw (void)
{
	// console is full screen
	if (scr_con_current == vid.height) return;

	// don't draw during an intermission
	if (cl.intermission) return;

	// don't draw the sbar for a QC cutscene
	if (scr_viewsize.value > 110) return;

	if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
		Sbar_DrawScoreboard ();

	// only draw if we're alive!!!
	if (cl.stats[STAT_HEALTH])
	{
		// health
		HUD_DrawNum (47, 
					 447, 
					 cl.stats[STAT_HEALTH], 
					 3,
					 cl.stats[STAT_HEALTH] <= 25);

		// current ammo
		HUD_DrawAmmo ();

		// armour
		HUD_DrawArmor ();
	}
}


void Sbar_FinaleOverlay (void)
{
	qpic_t	*pic;

	scr_copyeverything = 1;

	pic = Draw_CachePic ("gfx/finale.lmp");
	Draw_TransPic ((vid.width - pic->width) / 2, 16, pic);
}


void Sbar_IntermissionOverlay (void)
{
}


